#ifndef __NETMIST_H__

#define __NETMIST_H__


#if defined(WIN32)
#define PDSM_R_REGISTRY "c:\\tmp\\pdsm_registry_client"
#define PDSM_MM_REGISTRY "c:\\tmp\\tmp/pdsm_master_registry"
#define NETMIST_LIC_FILE "netmist_license_key"
#define WIN32_MAX_OPEN		60
#define WIN32_MAX_PROC		1000 /*This number limits number of process on each physical client*/
#else
#define PDSM_R_REGISTRY "/tmp/pdsm_registry_client"
#define PDSM_MM_REGISTRY "/tmp/pdsm_master_registry"
#define NETMIST_LIC_FILE "netmist_license_key"
#endif

#define MAXNAME 256
#define MAXHNAME 100
#define MAXFSNAME 20
#define OP_STR_LEN 35
#define MAX_DIST_ELEM 16

#define UNIFORM_RAND 0
#define NO_UNIFORM_DIST 0

/* Types of work objects */
#define METADATA 	0
#define BACKUP 		1
#define HOMEDIR 	2
#define HPC 		3
#define NUM_WORK_OBJ 	4

#define MAXCLIENTS 	10000
/*
 * If you don't have a licensed version...
 */
#define MAX_L_CLIENTS 50

/* 
 * Used in make_name 
 */
#define TYPE_FILE 0
#define TYPE_DIR 1
#define TYPE_SYMLINK 2
#define TYPE_TESTDIR 3

/*
 * Number of OP types
 */
#define NUM_OP_TYPES 24

/*
 * Define the operations to test 
 */
#define OP_WRITE 0
#define OP_WRITE_FILE 1
#define OP_READ 2
#define OP_READ_FILE 3
#define OP_WRITE_RAND 4
#define OP_READ_RAND 5
#define OP_RMW 6
#define OP_MKDIR 7
#define OP_UNLINK 8
#define OP_STAT 9
#define OP_APPEND 10
#define OP_CREATE 11
#define OP_LOCKING 12
#define OP_ACCESS 13
#define OP_CHMOD 14
#define OP_READDIR 15
#define OP_MM_WRITE 16
#define OP_MM_READ 17
#define OP_COPYFILE 18
#define OP_RENAME 19
#define OP_STATFS 20
#define OP_PATHCONF 21
#define OP_CUSTOM1 22
#define OP_CUSTOM2 23

/* 
 * Here are all of the default values for the workload objects 
 */

/*
 * This is the transfer size distribution table element.
 */
struct dist_entry {
        unsigned int size_min;
        unsigned int size_max;
        unsigned int percent;
};

/*
 * Define the mix. In the future this will be read in from a file,
 * and is transported to all of the client processes.
 * Important:
 * We implicitly maintain a mapping through "union", this mapping should *NOT* be changes. 
			int percent_read;      <==> percent_cmdline_file_delete
			int percent_mmap_read; <==> percent_cmdline_file_download
			int percent_read_rand; <==> percent_cmdline_file_upload
			int percent_write;     <==> percent_cmdline_navigation
			int percent_mmap_write;<==> percent_explorer_file_download
			int percent_write_rand;<==> percent_explorer_file_upload
			int percent_rmw;       <==> percent_explorer_file_delete
			int percent_mkdir;     <==> percent_explorer_navigation
			int percent_unlink;    <==> percent_explorer_select
			int percent_append;    <==> percent_word_edit_save
			int percent_locking;   <==> percent_word_file_open
			int percent_access;    <==> percent_word_file_close
 */
struct mix_table {
        char workload_name[MAXHNAME];
	union {
		struct {
		        int percent_read;
		        char read_string[OP_STR_LEN];
		        int percent_read_file;
		        char read_file_string[OP_STR_LEN];
		        int percent_mmap_read;
		        char mmap_read_string[OP_STR_LEN];
		        int percent_read_rand;
		        char read_rand_string[OP_STR_LEN];
		        int percent_write;
		        char write_string[OP_STR_LEN];
		        int percent_write_file;
		        char write_file_string[OP_STR_LEN];
		        int percent_mmap_write;
		        char mmap_write_string[OP_STR_LEN];
		        int percent_write_rand;
		        char write_rand_string[OP_STR_LEN];
		        int percent_rmw;
		        char rmw_string[OP_STR_LEN];
		        int percent_mkdir;
		        char mkdir_string[OP_STR_LEN];
		        int percent_unlink;
		        char unlink_string[OP_STR_LEN];
		        int percent_create;
		        char create_string[OP_STR_LEN];
		        int percent_append;
		        char append_string[OP_STR_LEN];
		        int percent_locking;
		        char locking_string[OP_STR_LEN];
		        int percent_access;
		        char access_string[OP_STR_LEN];
		        int percent_stat;
		        char stat_string[OP_STR_LEN];
		        int percent_chmod;
		        char chmod_string[OP_STR_LEN];
		        int percent_readdir;
		        char readdir_string[OP_STR_LEN];
		        int percent_copyfile;
		        char copyfile_string[OP_STR_LEN];
		        int percent_rename;
		        char rename_string[OP_STR_LEN];
		        int percent_statfs;
		        char statfs_string[OP_STR_LEN];
		        int percent_pathconf;
		        char pathconf_string[OP_STR_LEN];
		        int percent_custom1;
		        char custom1_string[OP_STR_LEN];
		        int percent_custom2;
		        char custom2_string[OP_STR_LEN];
		        struct dist_entry read_dist[MAX_DIST_ELEM];
		        struct dist_entry write_dist[MAX_DIST_ELEM];
		        int percent_commit;
		        int percent_direct;
		        int align;
		        int percent_osync;
		        int percent_geometric;
		        int percent_compress;
		        int background;
		        int sharemode;
		        int rand_dist_behavior;
		        int uniform_file_size_dist;
		};
		struct {
			int percent_cmdline_file_delete;
		        char cmdline_file_delete_string[OP_STR_LEN];
			int percent_cmdline_file_download;
		        char cmdline_file_download_string[OP_STR_LEN];
			int percent_cmdline_file_upload;
		        char cmdline_file_upload_string[OP_STR_LEN];
			int percent_cmdline_navigation;
		        char cmdline_navigation_string[OP_STR_LEN];
			int percent_explorer_file_download;
		        char explorer_file_download_string[OP_STR_LEN];
			int percent_explorer_file_upload;
		        char explorer_file_upload_string[OP_STR_LEN];
			int percent_explorer_file_delete;
		        char explorer_file_delete_string[OP_STR_LEN];
			int percent_explorer_navigation;
		        char explorer_navigation_string[OP_STR_LEN];
			int percent_explorer_select;
		        char explorer_select_string[OP_STR_LEN];
			int percent_word_edit_save;
		        char word_edit_save_string[OP_STR_LEN];
			int percent_word_file_open;
		        char word_file_open_string[OP_STR_LEN];
			int percent_word_file_close;
		        char word_file_close_string[OP_STR_LEN];
		};
	};
        char fs_type[MAXFSNAME];
};

#if defined(PDSM_R_REGISTRY)
char my_pdsm_file[256];
struct pdsm_remote_stats {
        double requested_op_rate;
        double achieved_op_rate;
        double run_time;
        double read_throughput;
        double write_throughput;
        int cur_state;
};

struct pdsm_remote_control {
        double set_op_rate;
        int set_op_rate_set;
        int set_workload;
        struct mix_table work_load_out[NUM_WORK_OBJ];
};
#endif
/*
 * This is the virtualized switch for all of the functions that
 * will get called in the OP mix.
 * We implicitly maintain a mapping through "union", this mapping should *NOT* be changes. 
			void (*read)(); <==> void (*cmdline_file_delete)();
			void (*mmap_read)(); <==> void (*cmdline_file_download)();
			void (*read_rand)(); <==> void (*cmdline_file_upload)();
			void (*write)(); <==> void (*cmdline_navigation)();
			void (*mmap_write)(); <==> void (*explorer_file_download)();
			void (*write_rand)(); <==> void (*explorer_file_upload)();
			void (*rmw)(); <==> void (*explorer_file_delete)();
			void (*mkdir)(); <==> void (*explorer_navigation)();
			void (*unlink)(); <==> void (*explorer_select)();
			void (*append)(); <==> void (*word_edit_save)();
			void (*locking)(); <==> void (*word_file_open)();
			void (*access)(); <==> void (*word_file_close)();
 */
struct method_table {
	char name[MAXNAME];
	union {
		struct {
			void (*read)();
			void (*read_file)();
			void (*mmap_read)();
			void (*read_rand)();
			void (*write)();
			void (*write_file)();
			void (*mmap_write)();
			void (*write_rand)();
			void (*rmw)();
			void (*mkdir)();
			void (*unlink)();
			void (*create)();
			void (*append)();
			void (*locking)();
			void (*access)();
			void (*stat)();
			void (*chmod)();
			void (*readdir)();
			void (*copyfile)();
			void (*rename)();
			void (*statfs)();
			void (*pathconf)();
			void (*custom1)();
			void (*custom2)();
		};
	};

	struct file_object * (*makepath)();
	void (*prefix_op)();
	void (*postfix_op)();
	void (*fill_buf)();
	int  (*init_dir)();
	void (*init_file)();
	void (*init_empty_file)();
	int  (*stat_workdir)();
	int  (*chdir)();
	int  (*remove_file)();
	int  (*remove_dir)();
	/*workload specific setup and cleanup*/
	void (*setup)();
	void (*teardown)();
	int (*impersonate)();
	void* (*make_impersonate_context)();
};

struct work_object {
	char work_obj_name[MAXNAME];
	union {
		struct {
			long run_time;
			long write_op_count;
			double write_op_time;
			long write_file_op_count;
			double write_file_op_time;
			long read_op_count;
			double read_op_time;
			long read_file_op_count;
			double read_file_op_time;
			long mkdir_op_count;
			double mkdir_op_time;
			long unlink_op_count;
			double unlink_op_time;
			long create_op_count;
			double create_op_time;
			long stat_op_count;
			double stat_op_time;
			long copyfile_op_count;
			double copyfile_op_time;
			long rename_op_count;
			double rename_op_time;
			long statfs_op_count;
			double statfs_op_time;
			long pathconf_op_count;
			double pathconf_op_time;
			long custom1_op_count;
			double custom1_op_time;
			long custom2_op_count;
			double custom2_op_time;
			long append_op_count;
			double append_op_time;
			long lock_op_count;
			double lock_op_time;
			long access_op_count;
			double access_op_time;
			long chmod_op_count;
			double chmod_op_time;
			long readdir_op_count;
			double readdir_op_time;
			long   write_rand_op_count;
			double write_rand_op_time;
			long   read_rand_op_count;
			double read_rand_op_time;
			long   rmw_op_count;
			double rmw_op_time;
			long mmap_write_op_count;
			double mmap_write_op_time;
			long mmap_read_op_count;
			double mmap_read_op_time;
			long open_op_count;
			long close_op_count;
		};
	};
	double total_file_op_time;
	double total_file_ops;
	/* Totals for all worker threads */
	double total_ops_per_sec;
	double total_latency;
	double total_count;
};

/*
 * mapping through "union", see netmist.h
 */
#define OP_WRITE 0
#define OP_WRITE_FILE 1
#define OP_READ 2
#define OP_READ_FILE 3
#define OP_WRITE_RAND 4
#define OP_READ_RAND 5
#define OP_RMW 6
#define OP_MKDIR 7
#define OP_UNLINK 8
#define OP_STAT 9
#define OP_APPEND 10
#define OP_CREATE 11
#define OP_LOCKING 12
#define OP_ACCESS 13
#define OP_CHMOD 14
#define OP_READDIR 15
#define OP_MM_WRITE 16
#define OP_MM_READ 17
#define OP_COPYFILE 18
#define OP_RENAME 19
#define OP_STATFS 20
#define OP_PATHCONF 21
#define OP_CUSTOM1 22
#define OP_CUSTOM2 23
#define OP_CMDLINE_FILE_DELETE		OP_READ			// 1
#define OP_CMDLINE_FILE_DOWNLOAD	OP_READ_FILE 		// 3
#define OP_CMDLINE_FILE_UPLOAD		OP_MM_READ		// 17
#define OP_CMDLINE_FILE_NAVIGATION	OP_READ_RAND		// 5
#define OP_EXPLORER_FILE_DOWNLOAD	OP_WRITE		// 0
#define OP_EXPLORER_FILE_UPLOAD		OP_WRITE_FILE		// 1
#define OP_EXPLORER_FILE_DELETE		OP_MM_WRITE		// 16
#define OP_EXPLORER_NAVIGATION		OP_WRITE_RAND		// 4
#define OP_EXPLORER_SELECT		OP_RMW			// 6
#define OP_WORD_EDIT_SAVE		OP_MKDIR		// 7
#define OP_WORD_FILE_OPEN		OP_UNLINK		// 8
#define OP_WORD_FILE_CLOSE		OP_CREATE		// 11


#endif 


/* 
 * Here are all of the default values for the workload objects 
 */
#define METADATA_NAME "METADATA"
#define METADATA_P_READ 10
#define METADATA_P_READ_STRING "Read"
#define METADATA_P_READ_FILE 10
#define METADATA_P_READ_FILE_STRING "Read_file"
#define METADATA_P_M_READ 0
#define METADATA_P_M_READ_STRING "Mmap_read"
#define METADATA_P_R_READ 8
#define METADATA_P_R_READ_STRING "Random_read"
#define METADATA_P_WRITE 7
#define METADATA_P_WRITE_STRING "Write"
#define METADATA_P_WRITE_FILE 10
#define METADATA_P_WRITE_FILE_STRING "Write_file"
#define METADATA_P_M_WRITE 0
#define METADATA_P_M_WRITE_STRING "Mmap_write"
#define METADATA_P_R_WRITE 2
#define METADATA_P_R_WRITE_STRING "Random_write"
#define METADATA_RMW 1
#define METADATA_RMW_STRING "Read_modify_write"
#define METADATA_MKDIR 2
#define METADATA_MKDIR_STRING "Mkdir"
#define METADATA_UNLINK 4
#define METADATA_UNLINK_STRING "Unlink"
#define METADATA_CREATE 1
#define METADATA_CREATE_STRING "Create"
#define METADATA_APPEND 3
#define METADATA_APPEND_STRING "Append"
#define METADATA_LOCK 5
#define METADATA_LOCK_STRING "Lock"
#define METADATA_ACCESS 14
#define METADATA_ACCESS_STRING "Access"
#define METADATA_STAT 14
#define METADATA_STAT_STRING "Stat"
#define METADATA_CHMOD 2
#define METADATA_CHMOD_STRING "Chmod"
#define METADATA_READDIR 4
#define METADATA_READDIR_STRING "Readdir"
#define METADATA_COPYFILE 0
#define METADATA_COPYFILE_STRING "Copyfile"
#define METADATA_RENAME 2
#define METADATA_RENAME_STRING "Rename"
#define METADATA_STATFS 1
#define METADATA_STATFS_STRING "Statfs"
#define METADATA_PATHCONF 0
#define METADATA_PATHCONF_STRING "Pathconf"
#define METADATA_CUSTOM1 0
#define METADATA_CUSTOM1_STRING "Custom1"
#define METADATA_CUSTOM2 0
#define METADATA_CUSTOM2_STRING "Custom2"
#define METADATA_IO_RS0_MIN 1
#define METADATA_IO_RS0_MAX 511
#define METADATA_IO_RS0_PCT 3
#define METADATA_IO_RS1_MIN 512
#define METADATA_IO_RS1_MAX 1023
#define METADATA_IO_RS1_PCT 1
#define METADATA_IO_RS2_MIN 1024
#define METADATA_IO_RS2_MAX 2047
#define METADATA_IO_RS2_PCT 2
#define METADATA_IO_RS3_MIN 2048
#define METADATA_IO_RS3_MAX 4095
#define METADATA_IO_RS3_PCT 1
#define METADATA_IO_RS4_MIN 4096
#define METADATA_IO_RS4_MAX 4096
#define METADATA_IO_RS4_PCT 16
#define METADATA_IO_RS5_MIN 4097
#define METADATA_IO_RS5_MAX 8191
#define METADATA_IO_RS5_PCT 6
#define METADATA_IO_RS6_MIN 8192
#define METADATA_IO_RS6_MAX 8192
#define METADATA_IO_RS6_PCT 36
#define METADATA_IO_RS7_MIN 8193
#define METADATA_IO_RS7_MAX 16383
#define METADATA_IO_RS7_PCT 7
#define METADATA_IO_RS8_MIN 16384
#define METADATA_IO_RS8_MAX 16384
#define METADATA_IO_RS8_PCT 7
#define METADATA_IO_RS9_MIN 16385
#define METADATA_IO_RS9_MAX 32767
#define METADATA_IO_RS9_PCT 2
#define METADATA_IO_RS10_MIN 32768
#define METADATA_IO_RS10_MAX 32768
#define METADATA_IO_RS10_PCT 9
#define METADATA_IO_RS11_MIN 65536
#define METADATA_IO_RS11_MAX 65536
#define METADATA_IO_RS11_PCT 4
#define METADATA_IO_RS12_MIN 98304
#define METADATA_IO_RS12_MAX 98304
#define METADATA_IO_RS12_PCT 3
#define METADATA_IO_RS13_MIN 131072
#define METADATA_IO_RS13_MAX 131072
#define METADATA_IO_RS13_PCT 2
#define METADATA_IO_RS14_MIN 262144
#define METADATA_IO_RS14_MAX 262144
#define METADATA_IO_RS14_PCT 1
#define METADATA_IO_RS15_MIN 524288
#define METADATA_IO_RS15_MAX 524288
#define METADATA_IO_RS15_PCT 0
#define METADATA_IO_WS0_MIN 1
#define METADATA_IO_WS0_MAX 511
#define METADATA_IO_WS0_PCT 13
#define METADATA_IO_WS1_MIN 512
#define METADATA_IO_WS1_MAX 1023
#define METADATA_IO_WS1_PCT 3
#define METADATA_IO_WS2_MIN 1024
#define METADATA_IO_WS2_MAX 2047
#define METADATA_IO_WS2_PCT 7
#define METADATA_IO_WS3_MIN 2048
#define METADATA_IO_WS3_MAX 4095
#define METADATA_IO_WS3_PCT 5
#define METADATA_IO_WS4_MIN 4096
#define METADATA_IO_WS4_MAX 4096
#define METADATA_IO_WS4_PCT 11
#define METADATA_IO_WS5_MIN 4097
#define METADATA_IO_WS5_MAX 8191
#define METADATA_IO_WS5_PCT 3
#define METADATA_IO_WS6_MIN 8192
#define METADATA_IO_WS6_MAX 8192
#define METADATA_IO_WS6_PCT 30
#define METADATA_IO_WS7_MIN 8193
#define METADATA_IO_WS7_MAX 16383
#define METADATA_IO_WS7_PCT 7
#define METADATA_IO_WS8_MIN 16384
#define METADATA_IO_WS8_MAX 16384
#define METADATA_IO_WS8_PCT 5
#define METADATA_IO_WS9_MIN 16385
#define METADATA_IO_WS9_MAX 32767
#define METADATA_IO_WS9_PCT 1
#define METADATA_IO_WS10_MIN 32768
#define METADATA_IO_WS10_MAX 32768
#define METADATA_IO_WS10_PCT 6
#define METADATA_IO_WS11_MIN 65536
#define METADATA_IO_WS11_MAX 65536
#define METADATA_IO_WS11_PCT 4
#define METADATA_IO_WS12_MIN 98304
#define METADATA_IO_WS12_MAX 98304
#define METADATA_IO_WS12_PCT 2
#define METADATA_IO_WS13_MIN 131072
#define METADATA_IO_WS13_MAX 131072
#define METADATA_IO_WS13_PCT 2
#define METADATA_IO_WS14_MIN 262144
#define METADATA_IO_WS14_MAX 262144
#define METADATA_IO_WS14_PCT 1
#define METADATA_IO_WS15_MIN 524288
#define METADATA_IO_WS15_MAX 524288
#define METADATA_IO_WS15_PCT 0
#define METADATA_COMMIT_PCT 33
#define METADATA_DIRECT_PCT 0
#define METADATA_ALIGN 0
#define METADATA_OSYNC_PCT 0
#define METADATA_POISSON_PCT 10
#define METADATA_COMPRESS_PCT 0
#define METADATA_BG 0
#define METADATA_SM 0
#define METADATA_RDBU 0
#define METADATA_UFSD 0
#define METADATA_FS "POSIX"

#define BACKUP_NAME "BACKUP"
#define BACKUP_P_READ 9
#define BACKUP_P_READ_STRING "Read"
#define BACKUP_P_READ_FILE 0
#define BACKUP_P_READ_FILE_STRING "Read_file"
#define BACKUP_P_M_READ 0
#define BACKUP_P_M_READ_STRING "Mmap_read"
#define BACKUP_P_R_READ 9
#define BACKUP_P_R_READ_STRING "Random_read"
#define BACKUP_P_WRITE 7
#define BACKUP_P_WRITE_STRING "Write"
#define BACKUP_P_WRITE_FILE 0
#define BACKUP_P_WRITE_FILE_STRING "Write_file"
#define BACKUP_P_M_WRITE 0
#define BACKUP_P_M_WRITE_STRING "Mmap_write"
#define BACKUP_P_R_WRITE 2
#define BACKUP_P_R_WRITE_STRING "Random_write"
#define BACKUP_RMW 1
#define BACKUP_RMW_STRING "Read_modify_write"
#define BACKUP_MKDIR 2
#define BACKUP_MKDIR_STRING "Mkdir"
#define BACKUP_UNLINK 3
#define BACKUP_UNLINK_STRING "Unlink"
#define BACKUP_CREATE 1
#define BACKUP_CREATE_STRING "Create"
#define BACKUP_APPEND 3
#define BACKUP_APPEND_STRING "Append"
#define BACKUP_LOCK 6
#define BACKUP_LOCK_STRING "Lock"
#define BACKUP_ACCESS 24
#define BACKUP_ACCESS_STRING "Access"
#define BACKUP_STAT 26
#define BACKUP_STAT_STRING "Stat"
#define BACKUP_CHMOD 2
#define BACKUP_CHMOD_STRING "Chmod"
#define BACKUP_READDIR 4
#define BACKUP_READDIR_STRING "Readdir"
#define BACKUP_COPYFILE 0
#define BACKUP_COPYFILE_STRING "Copyfile"
#define BACKUP_RENAME 0
#define BACKUP_RENAME_STRING "Rename"
#define BACKUP_STATFS 1
#define BACKUP_STATFS_STRING "Statfs"
#define BACKUP_PATHCONF 0
#define BACKUP_PATHCONF_STRING "Pathconf"
#define BACKUP_CUSTOM1 0
#define BACKUP_CUSTOM1_STRING "Custom1"
#define BACKUP_CUSTOM2 0
#define BACKUP_CUSTOM2_STRING "Custom2"
#define BACKUP_IO_RS0_MIN 1
#define BACKUP_IO_RS0_MAX 511
#define BACKUP_IO_RS0_PCT 3
#define BACKUP_IO_RS1_MIN 512
#define BACKUP_IO_RS1_MAX 1023
#define BACKUP_IO_RS1_PCT 1
#define BACKUP_IO_RS2_MIN 1024
#define BACKUP_IO_RS2_MAX 2047
#define BACKUP_IO_RS2_PCT 2
#define BACKUP_IO_RS3_MIN 2048
#define BACKUP_IO_RS3_MAX 4095
#define BACKUP_IO_RS3_PCT 1
#define BACKUP_IO_RS4_MIN 4096
#define BACKUP_IO_RS4_MAX 4096
#define BACKUP_IO_RS4_PCT 6
#define BACKUP_IO_RS5_MIN 4097
#define BACKUP_IO_RS5_MAX 8191
#define BACKUP_IO_RS5_PCT 6
#define BACKUP_IO_RS6_MIN 8192
#define BACKUP_IO_RS6_MAX 8192
#define BACKUP_IO_RS6_PCT 16
#define BACKUP_IO_RS7_MIN 8193
#define BACKUP_IO_RS7_MAX 16383
#define BACKUP_IO_RS7_PCT 7
#define BACKUP_IO_RS8_MIN 16384
#define BACKUP_IO_RS8_MAX 16384
#define BACKUP_IO_RS8_PCT 37
#define BACKUP_IO_RS9_MIN 16385
#define BACKUP_IO_RS9_MAX 32767
#define BACKUP_IO_RS9_PCT 2
#define BACKUP_IO_RS10_MIN 32768
#define BACKUP_IO_RS10_MAX 32768
#define BACKUP_IO_RS10_PCT 9
#define BACKUP_IO_RS11_MIN 65536
#define BACKUP_IO_RS11_MAX 65536
#define BACKUP_IO_RS11_PCT 4
#define BACKUP_IO_RS12_MIN 98304
#define BACKUP_IO_RS12_MAX 98304
#define BACKUP_IO_RS12_PCT 3
#define BACKUP_IO_RS13_MIN 131072
#define BACKUP_IO_RS13_MAX 131072
#define BACKUP_IO_RS13_PCT 2
#define BACKUP_IO_RS14_MIN 262144
#define BACKUP_IO_RS14_MAX 262144
#define BACKUP_IO_RS14_PCT 1
#define BACKUP_IO_RS15_MIN 524288
#define BACKUP_IO_RS15_MAX 524288
#define BACKUP_IO_RS15_PCT 0
#define BACKUP_IO_WS0_MIN 1
#define BACKUP_IO_WS0_MAX 511
#define BACKUP_IO_WS0_PCT 13
#define BACKUP_IO_WS1_MIN 512
#define BACKUP_IO_WS1_MAX 1023
#define BACKUP_IO_WS1_PCT 3
#define BACKUP_IO_WS2_MIN 1024
#define BACKUP_IO_WS2_MAX 2047
#define BACKUP_IO_WS2_PCT 7
#define BACKUP_IO_WS3_MIN 2048
#define BACKUP_IO_WS3_MAX 4095
#define BACKUP_IO_WS3_PCT 5
#define BACKUP_IO_WS4_MIN 4096
#define BACKUP_IO_WS4_MAX 4096
#define BACKUP_IO_WS4_PCT 11
#define BACKUP_IO_WS5_MIN 4097
#define BACKUP_IO_WS5_MAX 8191
#define BACKUP_IO_WS5_PCT 3
#define BACKUP_IO_WS6_MIN 8192
#define BACKUP_IO_WS6_MAX 8192
#define BACKUP_IO_WS6_PCT 0
#define BACKUP_IO_WS7_MIN 8193
#define BACKUP_IO_WS7_MAX 16383
#define BACKUP_IO_WS7_PCT 7
#define BACKUP_IO_WS8_MIN 16384
#define BACKUP_IO_WS8_MAX 16384
#define BACKUP_IO_WS8_PCT 35
#define BACKUP_IO_WS9_MIN 16385
#define BACKUP_IO_WS9_MAX 32767
#define BACKUP_IO_WS9_PCT 1
#define BACKUP_IO_WS10_MIN 32768
#define BACKUP_IO_WS10_MAX 32768
#define BACKUP_IO_WS10_PCT 6
#define BACKUP_IO_WS11_MIN 65536
#define BACKUP_IO_WS11_MAX 65536
#define BACKUP_IO_WS11_PCT 4
#define BACKUP_IO_WS12_MIN 98304
#define BACKUP_IO_WS12_MAX 98304
#define BACKUP_IO_WS12_PCT 2
#define BACKUP_IO_WS13_MIN 131072
#define BACKUP_IO_WS13_MAX 131072
#define BACKUP_IO_WS13_PCT 2
#define BACKUP_IO_WS14_MIN 262144
#define BACKUP_IO_WS14_MAX 262144
#define BACKUP_IO_WS14_PCT 1
#define BACKUP_IO_WS15_MIN 524288
#define BACKUP_IO_WS15_MAX 524288
#define BACKUP_IO_WS15_PCT 0
#define BACKUP_COMMIT_PCT 33
#define BACKUP_DIRECT_PCT 0
#define BACKUP_ALIGN 0
#define BACKUP_OSYNC_PCT 0
#define BACKUP_POISSON_PCT 0
#define BACKUP_COMPRESS_PCT 30
#define BACKUP_BG 0
#define BACKUP_SM 0
#define BACKUP_RDBU 0
#define BACKUP_UFSD 0
#define BACKUP_FS "POSIX"

#define HOMEDIR_NAME "HOMEDIR"
#define HOMEDIR_P_READ 8
#define HOMEDIR_P_READ_STRING "Read"
#define HOMEDIR_P_READ_FILE 0
#define HOMEDIR_P_READ_FILE_STRING "Read_file"
#define HOMEDIR_P_M_READ 0
#define HOMEDIR_P_M_READ_STRING "Mmap_read"
#define HOMEDIR_P_R_READ 10
#define HOMEDIR_P_R_READ_STRING "Random_read"
#define HOMEDIR_P_WRITE 7
#define HOMEDIR_P_WRITE_STRING "Write"
#define HOMEDIR_P_WRITE_FILE 0
#define HOMEDIR_P_WRITE_FILE_STRING "Write_file"
#define HOMEDIR_P_M_WRITE 0
#define HOMEDIR_P_M_WRITE_STRING "Mmap_write"
#define HOMEDIR_P_R_WRITE 2
#define HOMEDIR_P_R_WRITE_STRING "Random_write"
#define HOMEDIR_RMW 1
#define HOMEDIR_RMW_STRING "Read_modify_write"
#define HOMEDIR_MKDIR 2
#define HOMEDIR_MKDIR_STRING "Mkdir"
#define HOMEDIR_UNLINK 3
#define HOMEDIR_UNLINK_STRING "Unlink"
#define HOMEDIR_CREATE 1
#define HOMEDIR_CREATE_STRING "Create"
#define HOMEDIR_APPEND 3
#define HOMEDIR_APPEND_STRING "Append"
#define HOMEDIR_LOCK 6
#define HOMEDIR_LOCK_STRING "Lock"
#define HOMEDIR_ACCESS 18
#define HOMEDIR_ACCESS_STRING "Access"
#define HOMEDIR_STAT 20
#define HOMEDIR_STAT_STRING "Stat"
#define HOMEDIR_CHMOD 3
#define HOMEDIR_CHMOD_STRING "Chmod"
#define HOMEDIR_READDIR 5
#define HOMEDIR_READDIR_STRING "Readdir"
#define HOMEDIR_COPYFILE 10
#define HOMEDIR_COPYFILE_STRING "Copyfile"
#define HOMEDIR_RENAME 0
#define HOMEDIR_RENAME_STRING "Rename"
#define HOMEDIR_STATFS 1
#define HOMEDIR_STATFS_STRING "Statfs"
#define HOMEDIR_PATHCONF 0
#define HOMEDIR_PATHCONF_STRING "Pathconf"
#define HOMEDIR_CUSTOM1 0
#define HOMEDIR_CUSTOM1_STRING "Custom1"
#define HOMEDIR_CUSTOM2 0
#define HOMEDIR_CUSTOM2_STRING "Custom2"
#define HOMEDIR_IO_RS0_MIN 1
#define HOMEDIR_IO_RS0_MAX 511
#define HOMEDIR_IO_RS0_PCT 3
#define HOMEDIR_IO_RS1_MIN 512
#define HOMEDIR_IO_RS1_MAX 1023
#define HOMEDIR_IO_RS1_PCT 1
#define HOMEDIR_IO_RS2_MIN 1024
#define HOMEDIR_IO_RS2_MAX 2047
#define HOMEDIR_IO_RS2_PCT 2
#define HOMEDIR_IO_RS3_MIN 2048
#define HOMEDIR_IO_RS3_MAX 4095
#define HOMEDIR_IO_RS3_PCT 1
#define HOMEDIR_IO_RS4_MIN 4096
#define HOMEDIR_IO_RS4_MAX 4096
#define HOMEDIR_IO_RS4_PCT 16
#define HOMEDIR_IO_RS5_MIN 4097
#define HOMEDIR_IO_RS5_MAX 8191
#define HOMEDIR_IO_RS5_PCT 6
#define HOMEDIR_IO_RS6_MIN 8192
#define HOMEDIR_IO_RS6_MAX 8192
#define HOMEDIR_IO_RS6_PCT 36
#define HOMEDIR_IO_RS7_MIN 8193
#define HOMEDIR_IO_RS7_MAX 16383
#define HOMEDIR_IO_RS7_PCT 7
#define HOMEDIR_IO_RS8_MIN 16384
#define HOMEDIR_IO_RS8_MAX 16384
#define HOMEDIR_IO_RS8_PCT 7
#define HOMEDIR_IO_RS9_MIN 16385
#define HOMEDIR_IO_RS9_MAX 32767
#define HOMEDIR_IO_RS9_PCT 2
#define HOMEDIR_IO_RS10_MIN 32768
#define HOMEDIR_IO_RS10_MAX 32768
#define HOMEDIR_IO_RS10_PCT 9
#define HOMEDIR_IO_RS11_MIN 65536
#define HOMEDIR_IO_RS11_MAX 65536
#define HOMEDIR_IO_RS11_PCT 4
#define HOMEDIR_IO_RS12_MIN 98304
#define HOMEDIR_IO_RS12_MAX 98304
#define HOMEDIR_IO_RS12_PCT 3
#define HOMEDIR_IO_RS13_MIN 131072
#define HOMEDIR_IO_RS13_MAX 131072
#define HOMEDIR_IO_RS13_PCT 2
#define HOMEDIR_IO_RS14_MIN 262144
#define HOMEDIR_IO_RS14_MAX 262144
#define HOMEDIR_IO_RS14_PCT 1
#define HOMEDIR_IO_RS15_MIN 524288
#define HOMEDIR_IO_RS15_MAX 524288
#define HOMEDIR_IO_RS15_PCT 0
#define HOMEDIR_IO_WS0_MIN 1
#define HOMEDIR_IO_WS0_MAX 511
#define HOMEDIR_IO_WS0_PCT 13
#define HOMEDIR_IO_WS1_MIN 512
#define HOMEDIR_IO_WS1_MAX 1023
#define HOMEDIR_IO_WS1_PCT 3
#define HOMEDIR_IO_WS2_MIN 1024
#define HOMEDIR_IO_WS2_MAX 2047
#define HOMEDIR_IO_WS2_PCT 7
#define HOMEDIR_IO_WS3_MIN 2048
#define HOMEDIR_IO_WS3_MAX 4095
#define HOMEDIR_IO_WS3_PCT 5
#define HOMEDIR_IO_WS4_MIN 4096
#define HOMEDIR_IO_WS4_MAX 4096
#define HOMEDIR_IO_WS4_PCT 11
#define HOMEDIR_IO_WS5_MIN 4097
#define HOMEDIR_IO_WS5_MAX 8191
#define HOMEDIR_IO_WS5_PCT 3
#define HOMEDIR_IO_WS6_MIN 8192
#define HOMEDIR_IO_WS6_MAX 8192
#define HOMEDIR_IO_WS6_PCT 30
#define HOMEDIR_IO_WS7_MIN 8193
#define HOMEDIR_IO_WS7_MAX 16383
#define HOMEDIR_IO_WS7_PCT 7
#define HOMEDIR_IO_WS8_MIN 16384
#define HOMEDIR_IO_WS8_MAX 16384
#define HOMEDIR_IO_WS8_PCT 5
#define HOMEDIR_IO_WS9_MIN 16385
#define HOMEDIR_IO_WS9_MAX 32767
#define HOMEDIR_IO_WS9_PCT 1
#define HOMEDIR_IO_WS10_MIN 32768
#define HOMEDIR_IO_WS10_MAX 32768
#define HOMEDIR_IO_WS10_PCT 6
#define HOMEDIR_IO_WS11_MIN 65536
#define HOMEDIR_IO_WS11_MAX 65536
#define HOMEDIR_IO_WS11_PCT 4
#define HOMEDIR_IO_WS12_MIN 98304
#define HOMEDIR_IO_WS12_MAX 98304
#define HOMEDIR_IO_WS12_PCT 2
#define HOMEDIR_IO_WS13_MIN 131072
#define HOMEDIR_IO_WS13_MAX 131072
#define HOMEDIR_IO_WS13_PCT 2
#define HOMEDIR_IO_WS14_MIN 262144
#define HOMEDIR_IO_WS14_MAX 262144
#define HOMEDIR_IO_WS14_PCT 1
#define HOMEDIR_IO_WS15_MIN 524288
#define HOMEDIR_IO_WS15_MAX 524288
#define HOMEDIR_IO_WS15_PCT 0
#define HOMEDIR_COMMIT_PCT 33
#define HOMEDIR_DIRECT_PCT 0
#define HOMEDIR_ALIGN 0
#define HOMEDIR_OSYNC_PCT 0
#define HOMEDIR_POISSON_PCT 10
#define HOMEDIR_COMPRESS_PCT 30
#define HOMEDIR_BG 0
#define HOMEDIR_SM 0
#define HOMEDIR_RDBU 0
#define HOMEDIR_UFSD 0
#define HOMEDIR_FS "POSIX"

#define HPC_NAME "HPC"
#define HPC_P_READ 41
#define HPC_P_READ_STRING "Read"
#define HPC_P_READ_FILE 0
#define HPC_P_READ_FILE_STRING "Read_file"
#define HPC_P_M_READ 0
#define HPC_P_M_READ_STRING "Mmap_read"
#define HPC_P_R_READ 6
#define HPC_P_R_READ_STRING "Random_read"
#define HPC_P_WRITE 28
#define HPC_P_WRITE_STRING "Write"
#define HPC_P_WRITE_FILE 0
#define HPC_P_WRITE_FILE_STRING "Write_file"
#define HPC_P_M_WRITE 0
#define HPC_P_M_WRITE_STRING "Mmap_write"
#define HPC_P_R_WRITE 2
#define HPC_P_R_WRITE_STRING "Random_write"
#define HPC_RMW 1
#define HPC_RMW_STRING "Read_modify_write"
#define HPC_MKDIR 2
#define HPC_MKDIR_STRING "Mkdir"
#define HPC_UNLINK 1
#define HPC_UNLINK_STRING "Unlink"
#define HPC_CREATE 1
#define HPC_CREATE_STRING "Create"
#define HPC_APPEND 3
#define HPC_APPEND_STRING "Append"
#define HPC_LOCK 0
#define HPC_LOCK_STRING "Lock"
#define HPC_ACCESS 2
#define HPC_ACCESS_STRING "Access"
#define HPC_STAT 6
#define HPC_STAT_STRING "Stat"
#define HPC_CHMOD 2
#define HPC_CHMOD_STRING  "Chmod"
#define HPC_READDIR 4
#define HPC_READDIR_STRING "Readdir"
#define HPC_COPYFILE 0
#define HPC_COPYFILE_STRING "Copyfile"
#define HPC_RENAME 0
#define HPC_RENAME_STRING "Rename"
#define HPC_STATFS 1
#define HPC_STATFS_STRING "Statfs"
#define HPC_PATHCONF 0
#define HPC_PATHCONF_STRING "Pathconf"
#define HPC_CUSTOM1 0
#define HPC_CUSTOM1_STRING "Custom1"
#define HPC_CUSTOM2 0
#define HPC_CUSTOM2_STRING "Custom2"
#define HPC_IO_RS0_MIN 1
#define HPC_IO_RS0_MAX 511
#define HPC_IO_RS0_PCT 0
#define HPC_IO_RS1_MIN 512
#define HPC_IO_RS1_MAX 1023
#define HPC_IO_RS1_PCT 0
#define HPC_IO_RS2_MIN 1024
#define HPC_IO_RS2_MAX 2047
#define HPC_IO_RS2_PCT 0
#define HPC_IO_RS3_MIN 2048
#define HPC_IO_RS3_MAX 4095
#define HPC_IO_RS3_PCT 0
#define HPC_IO_RS4_MIN 4096
#define HPC_IO_RS4_MAX 4096
#define HPC_IO_RS4_PCT 0
#define HPC_IO_RS5_MIN 4096
#define HPC_IO_RS5_MAX 8191
#define HPC_IO_RS5_PCT 0
#define HPC_IO_RS6_MIN 8192
#define HPC_IO_RS6_MAX 8192
#define HPC_IO_RS6_PCT 0
#define HPC_IO_RS7_MIN 8192
#define HPC_IO_RS7_MAX 16383
#define HPC_IO_RS7_PCT 0
#define HPC_IO_RS8_MIN 16384
#define HPC_IO_RS8_MAX 16384
#define HPC_IO_RS8_PCT 0
#define HPC_IO_RS9_MIN 131072
#define HPC_IO_RS9_MAX 131072
#define HPC_IO_RS9_PCT 14
#define HPC_IO_RS10_MIN 262144
#define HPC_IO_RS10_MAX 262144
#define HPC_IO_RS10_PCT 14
#define HPC_IO_RS11_MIN 524288
#define HPC_IO_RS11_MAX 524288
#define HPC_IO_RS11_PCT 14
#define HPC_IO_RS12_MIN 1048576
#define HPC_IO_RS12_MAX 1048576
#define HPC_IO_RS12_PCT 14
#define HPC_IO_RS13_MIN 2097152
#define HPC_IO_RS13_MAX 2097152
#define HPC_IO_RS13_PCT 14
#define HPC_IO_RS14_MIN 4194304
#define HPC_IO_RS14_MAX 4194304
#define HPC_IO_RS14_PCT 15
#define HPC_IO_RS15_MIN 8388608
#define HPC_IO_RS15_MAX 8388608
#define HPC_IO_RS15_PCT 15
#define HPC_IO_WS0_MIN 1
#define HPC_IO_WS0_MAX 511
#define HPC_IO_WS0_PCT 0
#define HPC_IO_WS1_MIN 512
#define HPC_IO_WS1_MAX 1023
#define HPC_IO_WS1_PCT 0
#define HPC_IO_WS2_MIN 1024
#define HPC_IO_WS2_MAX 2047
#define HPC_IO_WS2_PCT 0
#define HPC_IO_WS3_MIN 2048
#define HPC_IO_WS3_MAX 4095
#define HPC_IO_WS3_PCT 0
#define HPC_IO_WS4_MIN 4096
#define HPC_IO_WS4_MAX 4096
#define HPC_IO_WS4_PCT 0
#define HPC_IO_WS5_MIN 4096
#define HPC_IO_WS5_MAX 8191
#define HPC_IO_WS5_PCT 0
#define HPC_IO_WS6_MIN 8192
#define HPC_IO_WS6_MAX 8192
#define HPC_IO_WS6_PCT 0
#define HPC_IO_WS7_MIN 8192
#define HPC_IO_WS7_MAX 16383
#define HPC_IO_WS7_PCT 0
#define HPC_IO_WS8_MIN 16384
#define HPC_IO_WS8_MAX 16384
#define HPC_IO_WS8_PCT 0
#define HPC_IO_WS9_MIN 131072
#define HPC_IO_WS9_MAX 131072
#define HPC_IO_WS9_PCT 14
#define HPC_IO_WS10_MIN 262144
#define HPC_IO_WS10_MAX 262144
#define HPC_IO_WS10_PCT 14
#define HPC_IO_WS11_MIN 524288
#define HPC_IO_WS11_MAX 524288
#define HPC_IO_WS11_PCT 14
#define HPC_IO_WS12_MIN 1048576
#define HPC_IO_WS12_MAX 1048576
#define HPC_IO_WS12_PCT 14
#define HPC_IO_WS13_MIN 2359296
#define HPC_IO_WS13_MAX 2359296
#define HPC_IO_WS13_PCT 14
#define HPC_IO_WS14_MIN 4194304
#define HPC_IO_WS14_MAX 4194304
#define HPC_IO_WS14_PCT 15
#define HPC_IO_WS15_MIN 8388608
#define HPC_IO_WS15_MAX 8388608
#define HPC_IO_WS15_PCT 15
#define HPC_COMMIT_PCT 0
#define HPC_DIRECT_PCT 0
#define HPC_ALIGN 0
#define HPC_OSYNC_PCT 0
#define HPC_POISSON_PCT 0
#define HPC_COMPRESS_PCT 0
#define HPC_BG 0
#define HPC_SM 0
#define HPC_RDBU 0
#define HPC_UFSD 0
#define HPC_FS "POSIX"

