/*****************************************************************************
 *  Copyright (c) 2002-2014 by Iozone.org
 *      All rights reserved.
 *              Iozone.org
 *              7417 Crenshaw Dr.
 *              Plano, TX 75025
 *
 *      This product contains benchmarks acquired from several sources who
 *      understand and agree with Iozone's goal of creating fair and objective
 *      benchmarks to measure computer performance.
 *
 *      This copyright notice is placed here only to protect Iozone.org in the
 *      event the source is misused in any manner that is contrary to the
 *      spirit, the goals and the intent of Iozone.org
 *****************************************************************************/

/*
 * This is a modify utility that can be used to alter the values in the
 * master shared memory segments on remote PDSM clients. It is a simple 
 * demonstration of the capabilty to modify variables used by a set 
 * of distributed clients, where variables are  transported via unified 
 * PDSM space
 *
 * In this case, Netmist is running on a bunch of clients. Netmist
 * detects if PDSM is present, and if so, it will pickup changes
 * and modify the internal variables to reflect these changes.
 */
#include "./copyright.txt" 
#include "./license.txt"

#ifdef WIN32
#pragma warning(disable: 4996)
#pragma warning(disable: 4101)
/*Windows specific headers*/
#include <Ws2tcpip.h>
#include <Windows.h>
#include <tchar.h>
#include <win32_getopt.h>
#include <win32_sub.h>
#endif

#if !defined(WIN32)
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <strings.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#else
#define MAXREMOTES 256
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include "netmist.h"

#define MAX_CLIENTS 256
#define THISVERSION "        $Revision: 739 $"

#ifdef WIN32
/*thread handles - threads are created to accept connections from remote client*/
HANDLE hThreadArray[MAXREMOTES];
SOCKET socks[MAXREMOTES]; //these are listening sockets, one per client.
/*
 * Define types and structures for Windows platform
 */
struct timespec
{
   time_t  tv_sec;         /* seconds */
   long    tv_nsec;        /* nanoseconds */
};
typedef unsigned int key_t;

/*
 * Implement function calls as macros for Windows platform
 */
#define bzero(addr, size) RtlZeroMemory((VOID UNALIGNED*)(addr), (SIZE_T)(size))
#define close(socket) closesocket((SOCKET)(socket))
/*
 * Windows "Sleep()/SleepEx()" can only achieve 1 milisecond resolution,
 * so the sleep time will be rounded up to milisecond,
 * and it always returns 0
 */

/* functions that have different signatures than their UNIX version*/
#endif


int getopt();
extern char *optarg;
#ifdef HAVE_ANSIC_C
void import_table(char *);
#else
import_table(char *);
#endif

void dump_table(void);

char version[]=THISVERSION;

char workloadfilename[MAXNAME];
/*
 * Define the mix.  Defaults come from netmist.h 
 * One can over-ride any of these by importing a new set 
 * of definistions.
 */
extern struct mix_table my_mix_table[NUM_WORK_OBJ];

/* #define PDSM_MM_REGISTRY "/tmp/pdsm_master_registry"*/

#if defined(PDSM_MM_REGISTRY)
char * get_master_master_zone(void);
char my_pdsm_file[MAXNAME];
struct pdsm_remote_control *pdsm_control;
#endif

int mo_key, mo_size, ro_key, ro_size, mo_seg_size, mo_num_segs;
time_t dummytime;
char time_string[30];
int pcdebug;
int cdebug;
int ro_seg_size, ro_num_segs;
int number_of_private_segs = 1;
int size_of_private_segs;
char *private_segs_mo_pointer;
int remove_flag = 0;
int total_size;
int cret;
char * get_sh_mem(key_t, size_t, int );
int shmid;
key_t rkey = 0x3141593;        /* Key for the unified ro segments  */
void usage(void);
char mystate[10];
int client_id;

double new_op_rate;

int set_workload,set_op_rate;

#include "./import.c"

int
main(int argc, char ** argv)
{
	
	char *myseg;

#ifdef WIN32
        WSADATA wsaData;
        if (0 != WSAStartup( MAKEWORD(2,2), &wsaData ))
        {
                /* If Windows Sockets does not initialize correctly
                 * we have trouble.
                 */
                printf("WSAStartup failed");
                exit(1);
        }
#endif

	if(argc == 1)
	{
		usage();
		exit(0);
	};

	while((cret = getopt(argc,argv,"Hvr:o:w:E ")) != EOF){
                switch(cret){
                case 'H':                 /* Help screen */
			usage();
			exit(0);
			break;
                case 'v':                 /* Version */
			printf("Version %s\n",version);
			exit(0);
			break;
                case 'r':                 /* Remote client ID */
			client_id = atoi(optarg);
			break;
                case 'o':                 /* set_op rate */
			sscanf(optarg,"%lf",&new_op_rate);
			/* new_op_rate = atoi(optarg); */
			set_op_rate = 1;
			break;
		case 'w':		/* set workload */
			strcpy(workloadfilename, optarg);
			set_workload=1;
			break;
		case 'E':		/* output workload */
			dump_table();
			exit(0);
			break;
		};
	}	
	private_segs_mo_pointer=get_master_master_zone();
	if(private_segs_mo_pointer==NULL)
	{
		printf("Get_master_master_zone failed\n");
		exit(1);
	}

	total_size = mo_size;
	number_of_private_segs = mo_num_segs;
	size_of_private_segs = mo_seg_size;

	myseg = (char *)private_segs_mo_pointer +
		   (client_id*size_of_private_segs);
	pdsm_control = (struct pdsm_remote_control *) myseg;
	if(set_op_rate)
	{
	   pdsm_control->set_op_rate = new_op_rate;
	   pdsm_control->set_op_rate_set = 1;
	   printf("Setting op_rate for client %d to %lf\n",client_id, new_op_rate);
	}
	pdsm_control->set_workload=0;
	if(set_workload)
	{
		import_table(workloadfilename);
	        printf("Setting workload parameters for client %d\n",client_id);
		pdsm_control->set_workload=1;
		memcpy(pdsm_control->work_load_out,my_mix_table,sizeof(struct mix_table)*NUM_WORK_OBJ);
	}
	exit(0);
}

#ifdef WIN32

#define MAX_SHARE_MEMORY 32 //max number of shared memory one process can request, regardless of their size
typedef struct _SHARED_MEMORY
{
        HANDLE hFile; //associated file handle
        HANDLE hMapFile; //associated mapping object handle
        key_t  key; //shared memory key
        BOOL   valid;
} SHARED_MEMORY;

static SHARED_MEMORY sharedMem[MAX_SHARE_MEMORY] = {0};
/*
 *      Win32 implementation using memory mapped file
 *  key - share memory id, it is used to generate mapping object
 *  size - the size of shared memory requested
 *  flag - delete shared memory if its value is non-zero
 */
char *
get_sh_mem(key_t key, size_t size, int flag)
{
        TCHAR szObjName[255];
        TCHAR szFileName[255];
        char* pMem = NULL;
        int position=-1, i=0; //index in shared memory array where we can save shared_memory information
        HANDLE hFile;

        _stprintf(szObjName, TEXT("Global\\pdsm-%x"), key);
        _stprintf(szFileName, TEXT("SharedMemory.%x"), key);

        /*First we search the shared memory array, try to get existing one*/
        for (i=0; i<MAX_SHARE_MEMORY; i++)
        {
                if (sharedMem[i].valid && sharedMem[i].key == key)
                {
                        position = i;
                        break;
                }
                else if (position < 0  && !sharedMem[i].valid)
                        position = i; //find a unused slot
        }
        /* if we reach here, we did not find exising shared memory*/
        if (flag && sharedMem[position].valid)
        {
                /*do we need to delete memory mapped file here???*/
                hFile = CreateFile(szFileName,
                                                        GENERIC_WRITE | GENERIC_READ,
                                                        FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,
                                                        NULL,
                                                        OPEN_EXISTING,
                                                        FILE_FLAG_DELETE_ON_CLOSE, //delete file on close
                                                        NULL);
                if (hFile != INVALID_HANDLE_VALUE)
                        CloseHandle(hFile);
        }

        if (position < 0)
                /*the process has already allocated max-allowed shared memory, we have no choice but fail the call*/
                return NULL;
        /* Now we start to create new shared memory*/
        if (!sharedMem[position].valid)
        {
                sharedMem[position].hMapFile = NULL;
                /* We open or create a file to map to memory*/
                if ((sharedMem[position].hFile = CreateFile(szFileName,

                        GENERIC_WRITE | GENERIC_READ,

                        FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,

                        NULL,

                        OPEN_ALWAYS,     // create a new file if it does not exist, otherwise open the old file

                        (flag?FILE_FLAG_DELETE_ON_CLOSE:FILE_ATTRIBUTE_NORMAL),

                        NULL)) == (HANDLE)INVALID_HANDLE_VALUE)
                {
                        if(cdebug >= 2)
                                printf("Unable to open or create file for memory mapping:\"SharedMemory.%x\"\n",key);
                }
                else
                {
                        /*Create mapping object*/
                        sharedMem[position].hMapFile = CreateFileMapping(

                                        sharedMem[position].hFile,    // use paging file
                                                                                                                        NULL,                    // default security

                                        PAGE_READWRITE,          // read/write access
                                                                                                                        0,                       // maximum object size (high-order DWORD)

                                        (DWORD)size,                // maximum object size (low-order DWORD)

                                        szObjName);                 // name of mapping object
                        if (sharedMem[position].hMapFile == NULL)
                        {
                                if (cdebug >= 2)
                                        printf("Unable to create mapping object: \"Global\\pdsm-%x\", error code: %d\n",key, GetLastError());
                                CloseHandle(sharedMem[position].hFile);
                        }
                        else
                        {
                                if (cdebug >= 2 && GetLastError() == ERROR_ALREADY_EXISTS)
                                        printf("CreateFileMapping returned a handle of existing mapping object\n");
                                sharedMem[position].key = key;
                                sharedMem[position].valid = TRUE;
                        }
                }
        }

        if (sharedMem[position].hMapFile != NULL)
        {
                //
                //now we need to mapped the view.
                //
                pMem = (char*) MapViewOfFile(sharedMem[position].hMapFile,   // handle to map object

FILE_MAP_ALL_ACCESS, // read/write permission

0,

0,

(DWORD)size);
                if (pMem == NULL && cdebug >= 2)
                        printf("Unable to map file to memory, error code: %d\n", GetLastError());
        }
        return pMem;
}
#else
char *
get_sh_mem(key_t key, size_t size, int flag)
{
	struct stats *statsp;

	if(cdebug == 1)
	{
	  printf("Trying to get shared memory segment with key %x\n",(int)key);
	  printf("Size %d\n",(int)size);
	}
        shmid=shmget((key_t)key,(size_t)size,(int)(0666));
        if(shmid < 0)
        {
                printf("Shmid %d Errno %d\n",shmid,errno);
        	return((char *)0);
        } 

	if(cdebug == 2)
	  printf("Got shared memory with key %x\n",(int)key);

        statsp=(struct stats *)shmat(shmid,0,0);
        if(statsp == (struct stats *)-1)
        {
                printf("Attach to shared memory failed\n");
        	return((char *)0);
        }
	/* Attach ref count is now all that is holding on to this segment */
	if(flag)
		shmctl(shmid, IPC_RMID, 0);  /* cleanup */

	if(cdebug == 2)
	  printf("Attached to shared memory with key %x\n",(int)key);

	if(cdebug == 1)
	  printf("Shared memory key %x Size = %d\n",(int)key, (int)size);
	/*
	 * Don't need to call quit, as any termination will cleanup
 	 * automatically.
	 */

        /*quit();*/
        return((char *)statsp);
}
#endif

char * 
get_master_master_zone(void)
{
	FILE *reg;
	char my_reg_name[256];

	/*
 	 * Pick the file that has this client's shared memory key.
 	 */
	sprintf(my_reg_name,"%s",PDSM_MM_REGISTRY);
	reg = fopen(my_reg_name,"r");
	if(reg == NULL)
	{
		if(pcdebug)
		   printf("Open of registry %s failed\n",my_reg_name);
             	return((char *)0);
	}
	/*
 	 * Read in the shared memory key and size.
 	 */
/*
*/
	fscanf(reg,"PDSM_MO_key %x PDSM_MO_size %d Number_of_segs %d Seg_size %d\n",&mo_key,&mo_size,&mo_num_segs, &mo_seg_size);
	fscanf(reg,"PDSM_RO_key %x PDSM_RO_size %d Number_of_segs %d Seg_size %d\n",&ro_key,&ro_size,&ro_num_segs, &ro_seg_size);

	printf("%d Active clients\n",mo_num_segs);
        if(pcdebug)
		printf("Using RO key %x \n",mo_key);
	if(mo_key==0)
		return(0);
	if(mo_size==0)
		return(0);
        private_segs_mo_pointer = get_sh_mem(mo_key, mo_size,
             remove_flag);
        if(private_segs_mo_pointer == NULL)
        {
             return((char *)0);
        }
	fclose(reg);
	return(private_segs_mo_pointer);
}

void
usage()
{
	printf("\t-H............  Help screen\n");
	printf("\t-r #..........  Client ID\n");
	printf("\t-o #..........  Set Op_rate\n");
	printf("\t-E ...........  Export workload description\n");
	printf("\t-w filename ..  Import workload description\n");
	printf("\n");
}

#ifdef HAVE_ANSIC_C
void
dump_table(void)
#else
void
dump_table()
#endif
{
	int i,j;
	for(j=0;j<NUM_WORK_OBJ;j++)
	{
	   printf("Workload name %s\n",my_mix_table[j].workload_name);
	   printf("Percent %s %d\n",my_mix_table[j].read_string, my_mix_table[j].percent_read);
	   printf("Percent %s %d\n",my_mix_table[j].read_file_string,my_mix_table[j].percent_read_file);
	   printf("Percent %s %d\n",my_mix_table[j].mmap_read_string,my_mix_table[j].percent_mmap_read);
	   printf("Percent %s %d\n",my_mix_table[j].read_rand_string,my_mix_table[j].percent_read_rand);
	   printf("Percent %s %d\n",my_mix_table[j].write_string,my_mix_table[j].percent_write);
	   printf("Percent %s %d\n",my_mix_table[j].write_file_string,my_mix_table[j].percent_write_file);
	   printf("Percent %s %d\n",my_mix_table[j].mmap_write_string,my_mix_table[j].percent_mmap_write);
	   printf("Percent %s %d\n",my_mix_table[j].write_rand_string,my_mix_table[j].percent_write_rand);
	   printf("Percent %s %d\n",my_mix_table[j].rmw_string,my_mix_table[j].percent_rmw);
	   printf("Percent %s %d\n",my_mix_table[j].mkdir_string,my_mix_table[j].percent_mkdir);
	   printf("Percent %s %d\n",my_mix_table[j].unlink_string,my_mix_table[j].percent_unlink);
	   printf("Percent %s %d\n",my_mix_table[j].append_string,my_mix_table[j].percent_append);
	   printf("Percent %s %d\n",my_mix_table[j].locking_string,my_mix_table[j].percent_locking);
	   printf("Percent %s %d\n",my_mix_table[j].access_string,my_mix_table[j].percent_access);
	   printf("Percent %s %d\n",my_mix_table[j].stat_string,my_mix_table[j].percent_stat);
	   printf("Percent %s %d\n",my_mix_table[j].chmod_string,my_mix_table[j].percent_chmod);
	   printf("Percent %s %d\n",my_mix_table[j].readdir_string,my_mix_table[j].percent_readdir);
	   printf("Percent %s %d\n",my_mix_table[j].copyfile_string,my_mix_table[j].percent_copyfile);
	   printf("Percent %s %d\n",my_mix_table[j].rename_string,my_mix_table[j].percent_rename);
	   printf("Percent %s %d\n",my_mix_table[j].pathconf_string,my_mix_table[j].percent_pathconf);
	   printf("Percent %s %d\n",my_mix_table[j].statfs_string,my_mix_table[j].percent_statfs);
	   printf("Percent %s %d\n",my_mix_table[j].custom1_string,my_mix_table[j].percent_custom1);
	   printf("Percent %s %d\n",my_mix_table[j].custom2_string,my_mix_table[j].percent_custom2);
	   for(i=0;i<MAX_DIST_ELEM;i++)
	   {
	     printf("Read elem %d xfer min size %d\n",i,my_mix_table[j].read_dist[i].size_min);
	     printf("Read elem %d xfer max size %d\n",i,my_mix_table[j].read_dist[i].size_max);
	     printf("Read elem %d xfer percent %d\n",i,my_mix_table[j].read_dist[i].percent);
	   }
	   for(i=0;i<MAX_DIST_ELEM;i++)
	   {
	     printf("Write elem %d xfer min size %d\n",i,my_mix_table[j].write_dist[i].size_min);
	     printf("Write elem %d xfer max size %d\n",i,my_mix_table[j].write_dist[i].size_max);
	     printf("Write elem %d xfer percent %d\n",i,my_mix_table[j].write_dist[i].percent);
	   }
	   printf("Percent write commit %d\n",my_mix_table[j].percent_commit);
	   printf("Percent direct %d\n",my_mix_table[j].percent_direct);
	   printf("Align %d\n",my_mix_table[j].align);
	   printf("Percent osync %d\n",my_mix_table[j].percent_osync);
	   printf("Percent geometric %d\n",my_mix_table[j].percent_geometric);
	}
}

