/*****************************************************************************
 *  Copyright (c) 2002-2014 by Iozone.org
 *      All rights reserved.
 *              Iozone.org
 *              7417 Crenshaw Dr.
 *              Plano, TX 75025
 *
 *      This product contains benchmarks acquired from several sources who
 *      understand and agree with Iozone's goal of creating fair and objective
 *      benchmarks to measure computer performance.
 *
 *      This copyright notice is placed here only to protect Iozone.org in the
 *      event the source is misused in any manner that is contrary to the
 *      spirit, the goals and the intent of Iozone.org
 *****************************************************************************/

/*
 * This is a monitor that can be used to watch values in the
 * remote shared memory segments. It is a simple demonstration
 * of the capabilty to monitor variables that are modified
 * by a set of distributed clients, and transported into
 * a unified space for monitoring. 
 *
 * In this case, Netmist is running on a bunch of clients. Netmist
 * detects if PDSM is present, and if so, it will deposit the 
 * current requested ops/sec and the currently achieved ops/sec
 * from every client into the PDSM space. PDSM then transports
 * all of the state to a unified shared memory space on the 
 * master, where this tool is connecting and monitoring the 
 * unified shared memory space.
 */
#include "./copyright.txt" 
#include "./license.txt"

#if defined(WIN32)
#pragma warning(disable:4996)
#pragma warning(disable:4267)
#pragma warning(disable:4133)
#pragma warning(disable:4244)
#pragma warning(disable:4102)
#pragma warning(disable:4018)

#include <win32_sub.h>
#include <win32_getopt.h>
#include <time.h>
#endif

#include <stdio.h>
#include <stdlib.h>

#if !defined (WIN32)
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <strings.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#endif

#include <errno.h>
#include <time.h>
#include "netmist.h"

#define MAX_CLIENTS 256
#define GEN_SZ (4096)
#define INBUFSZ (GEN_SZ)
#define UPDATE_TIME 1
#define THISVERSION "        $Revision: 610 $"
#define INIT_BEAT 1
#define WARM_BEAT 2
#define RUN_BEAT 3
#define PRE_WARM_BEAT 4
#define CLEAN_BEAT 5

char version[]=THISVERSION;

#define PDSM_M_REGISTRY "/tmp/pdsm_master_registry"

#if defined(PDSM_M_REGISTRY)
char * get_master_zone(void);
struct pdsm_remote_stats *pdsm_stats;
#endif
char *get_timestamp(char *);
int getopt();

int mo_key, mo_size, ro_key, ro_size, mo_seg_size, mo_num_segs;
time_t dummytime;
char time_string[30];
int pcdebug;
int cdebug;
int mdebug;
int ro_seg_size, ro_num_segs;
int number_of_private_segs = 1;
int size_of_private_segs = INBUFSZ;
char *private_segs_ro_pointer;
int remove_flag = 0;
int total_size;
int cret;
double save_requested_op_rate[MAX_CLIENTS];
double save_read_throughput[MAX_CLIENTS];
double save_write_throughput[MAX_CLIENTS];
double save_achieved_op_rate[MAX_CLIENTS];
double save_run_time[MAX_CLIENTS];
#ifdef WIN32
/*thread handles - threads are created to accept connections from remote client*/
HANDLE hThreadArray[MAX_CLIENTS];
SOCKET socks[MAX_CLIENTS]; //these are listening sockets, one per client.
/*
 * Define types and structures for Windows platform
 */
struct timespec
{
   time_t  tv_sec;         /* seconds */
   long    tv_nsec;        /* nanoseconds */
};
typedef unsigned int key_t;

/*
 * Implement function calls as macros for Windows platform
 */
#define bzero(addr, size) RtlZeroMemory((VOID UNALIGNED*)(addr), (SIZE_T)(size))
#define close(socket) closesocket((SOCKET)(socket))
/*
 * Windows "Sleep()/SleepEx()" can only achieve 1 milisecond resolution,
 * so the sleep time will be rounded up to milisecond,
 * and it always returns 0
 */

/* functions that have different signatures than their UNIX version*/
#endif

char * get_sh_mem(key_t, size_t, int );
int shmid;
void usage(void);
char mystate[10];

int
main(int argc, char ** argv)
{
	
	int j,output=0;
	char *myseg;

#ifdef WIN32
        WSADATA wsaData;
        if (0 != WSAStartup( MAKEWORD(2,2), &wsaData ))
        {
                /* If Windows Sockets does not initialize correctly
                 * we have trouble.
                 */
                printf("WSAStartup failed");
                exit(1);
        }
#endif

	while((cret = getopt(argc,argv,"hv")) != EOF){
                switch(cret){
                case 'h':                 /* Help screen */
			usage();
			exit(0);
			break;
                case 'v':                 /* Help screen */
			printf("Version %s\n",version);
			exit(0);
			break;
		};
	}	
	private_segs_ro_pointer=get_master_zone();
	if(private_segs_ro_pointer==NULL)
	{
		printf("Get_master_zone failed\n");
		exit(1);
	}

	total_size = ro_size;
	number_of_private_segs = ro_num_segs;
	size_of_private_segs = ro_seg_size;
	/* 
	 * Display the RO slots
	 */
	while(1)
	{
	   for(j=0;j<number_of_private_segs;j++)
	   {
		output=0;
		myseg = (char *)private_segs_ro_pointer +
		   (j*size_of_private_segs);
		pdsm_stats = (struct pdsm_remote_stats *) myseg;
		switch (pdsm_stats->cur_state ) {
			case INIT_BEAT: 
				strcpy(mystate,"INIT");
				break;
			case WARM_BEAT: 
				strcpy(mystate,"WARM");
				break;
			case PRE_WARM_BEAT: 
				strcpy(mystate,"PRE-WARM");
				break;
			case RUN_BEAT: 
				strcpy(mystate,"RUN ");
				break;
			case CLEAN_BEAT: 
				strcpy(mystate,"CLEANUP ");
				break;
		};
		if((pdsm_stats->achieved_op_rate != save_achieved_op_rate[j]) ||
		   (pdsm_stats->run_time != save_run_time[j]) || 
		    (pdsm_stats->read_throughput != save_read_throughput[j]) ||
		    (pdsm_stats->write_throughput != save_write_throughput[j]))
			
		{
		   output=1;
		   if(pdsm_stats->requested_op_rate == (double)0.0)
		   {

		     printf("%s CL %d: S:%s Cur op_rate %10.2f Ops/sec. Run_time %3.0f secs RT: %.2f WT: %.2f\n",
			get_timestamp(&time_string[0]),
			j,
			mystate,
			pdsm_stats->achieved_op_rate,
			pdsm_stats->run_time,
			pdsm_stats->read_throughput,
			pdsm_stats->write_throughput);
		   }
		   else
		   {

		     printf("%s CL %d: S:%s Req op_rate %10.2f Cur op_rate %10.2f Ops/sec. Run_time %3.0f secs RT %.2f WT: %.2f\n",
			get_timestamp(&time_string[0]),
			j,
			mystate,
			pdsm_stats->requested_op_rate,
			pdsm_stats->achieved_op_rate,
			pdsm_stats->run_time,
			pdsm_stats->read_throughput,
			pdsm_stats->write_throughput);
		   }
		}
		save_requested_op_rate[j] = pdsm_stats->requested_op_rate;
		save_achieved_op_rate[j] = pdsm_stats->achieved_op_rate;
		save_read_throughput[j] = pdsm_stats->read_throughput;
		save_write_throughput[j] = pdsm_stats->write_throughput;
		save_run_time[j] = pdsm_stats->run_time;
	   }
	   if(output)
	      printf("\n");
	   sleep(UPDATE_TIME);
	}
}

#ifdef WIN32

#define MAX_SHARE_MEMORY 32 //max number of shared memory one process can request, regardless of their size
typedef struct _SHARED_MEMORY
{
        HANDLE hFile; //associated file handle
        HANDLE hMapFile; //associated mapping object handle
        key_t  key; //shared memory key
        BOOL   valid;
} SHARED_MEMORY;

static SHARED_MEMORY sharedMem[MAX_SHARE_MEMORY] = {0};
/*
 *      Win32 implementation using memory mapped file
 *  key - share memory id, it is used to generate mapping object
 *  size - the size of shared memory requested
 *  flag - delete shared memory if its value is non-zero
 */
char *
get_sh_mem(key_t key, size_t size, int flag)
{
        TCHAR szObjName[255];
        TCHAR szFileName[255];
        char* pMem = NULL;
        int position=-1, i=0; //index in shared memory array where we can save shared_memory information
        HANDLE hFile;
        BOOL foundExisting = FALSE;

        /* generate mapping object name */
        _stprintf(szObjName, TEXT("Global\\pdsm-%x"), key);
        _stprintf(szFileName, TEXT("SharedMemory.%x"), key);
        /*First we search the shared memory array, try to get existing one*/
        for (i=0; i<MAX_SHARE_MEMORY; i++)
        {
                if (sharedMem[i].valid && sharedMem[i].key == key)
                {
                        position = i;
                        break;
                }
                else if (position < 0  && !sharedMem[i].valid)
                        position = i; //find a unused slot
        }
        /* if we reach here, we did not find exising shared memory*/
        if (flag && sharedMem[position].valid)
        {
                /*do we need to delete memory mapped file here???*/
                hFile = CreateFile(szFileName,
                                                        GENERIC_WRITE | GENERIC_READ,
                                                        FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,
                                                        NULL,
                                                        OPEN_EXISTING,
                                                        FILE_FLAG_DELETE_ON_CLOSE, //delete file on close
                                                        NULL);
                if (hFile != INVALID_HANDLE_VALUE)
                        CloseHandle(hFile);
        }

        if (position < 0)
                /*the process has already allocated max-allowed shared memory, we have no choice but fail the call*/
                return NULL;
        /* Now we start to create new shared memory*/
        if (!sharedMem[position].valid)
        {
                sharedMem[position].hMapFile = NULL;
                /* We open or create a file to map to memory*/
                if ((sharedMem[position].hFile = CreateFile(szFileName,

                        GENERIC_WRITE | GENERIC_READ,

                        FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,

                        NULL,

                        OPEN_ALWAYS,     // create a new file if it does not exist, otherwise open the old file

                        (flag?FILE_FLAG_DELETE_ON_CLOSE:FILE_ATTRIBUTE_NORMAL),

                        NULL)) == (HANDLE)INVALID_HANDLE_VALUE)
                {
                        if(mdebug >= 2)
                                printf("Unable to open or create file for memory mapping:\"SharedMemory.%x\"\n",key);
                }
                else
                {
                        /*Create mapping object*/
                        sharedMem[position].hMapFile = CreateFileMapping(

                                        sharedMem[position].hFile,    // use paging file
                                                                                                                        NULL,                    // default security

                                        PAGE_READWRITE,          // read/write access
                                                                                                                        0,                       // maximum object size (high-order DWORD)

                                        (DWORD)size,                // maximum object size (low-order DWORD)

                                        szObjName);                 // name of mapping object
                        if (sharedMem[position].hMapFile == NULL)
                        {
                                if (mdebug >= 2)
                                        printf("Unable to create mapping object: \"Global\\pdsm-%x\", error code: %d\n",key, GetLastError());
                                CloseHandle(sharedMem[position].hFile);
                        }
                        else
                        {
                                if (mdebug >= 2 && GetLastError() == ERROR_ALREADY_EXISTS)
                                        printf("CreateFileMapping returned a handle of existing mapping object\n");
                                sharedMem[position].key = key;
                                sharedMem[position].valid = TRUE;
                        }
                }
        }

        if (sharedMem[position].hMapFile != NULL)
        {
                //
                //now we need to mapped the view.
                //
                pMem = (char*) MapViewOfFile(sharedMem[position].hMapFile,   // handle to map object

FILE_MAP_ALL_ACCESS, // read/write permission

0,

0,

(DWORD)size);
                if (pMem == NULL && mdebug >= 2)
                        printf("Unable to map file to memory, error code: %d\n", GetLastError());
        }
        return pMem;
}

#else

char *
get_sh_mem(key_t key, size_t size, int flag)
{
	struct stats *statsp;

	if(cdebug == 1)
	{
	  printf("Trying to get shared memory segment with key %x\n",(int)key);
	  printf("Size %d\n",(int)size);
	}
        shmid=shmget((key_t)key,(size_t)size,(int)(0666));
        if(shmid < 0)
        {
                printf("Shmid %d Errno %d\n",shmid,errno);
        	return((char *)0);
        } 

	if(cdebug == 2)
	  printf("Got shared memory with key %x\n",(int)key);

        statsp=(struct stats *)shmat(shmid,0,0);
        if(statsp == (struct stats *)-1)
        {
                printf("Attach to shared memory failed\n");
        	return((char *)0);
        }
	/* Attach ref count is now all that is holding on to this segment */
	if(flag)
		shmctl(shmid, IPC_RMID, 0);  /* cleanup */

	if(cdebug == 2)
	  printf("Attached to shared memory with key %x\n",(int)key);

	if(cdebug == 1)
	  printf("Shared memory key %x Size = %d\n",(int)key, (int)size);
	/*
	 * Don't need to call quit, as any termination will cleanup
 	 * automatically.
	 */

        /*quit();*/
        return((char *)statsp);
}
#endif

char * 
get_master_zone(void)
{
	FILE *reg;
	char my_reg_name[256];

	/*
 	 * Pick the file that has this client's shared memory key.
 	 */
	sprintf(my_reg_name,"%s",PDSM_M_REGISTRY);
	reg = fopen(my_reg_name,"r");
	if(reg == NULL)
	{
		if(pcdebug)
		   printf("Open of registry %s failed\n",my_reg_name);
             	return((char *)0);
	}
	/*
 	 * Read in the shared memory key and size.
 	 */
/*
*/
	fscanf(reg,"PDSM_MO_key %x PDSM_MO_size %d Number_of_segs %d Seg_size %d\n",&mo_key,&mo_size,&mo_num_segs, &mo_seg_size);
	fscanf(reg,"PDSM_RO_key %x PDSM_RO_size %d Number_of_segs %d Seg_size %d\n",&ro_key,&ro_size,&ro_num_segs, &ro_seg_size);

	printf("Monitoring %d clients\n",ro_num_segs);
        if(pcdebug)
		printf("Using RO key %x \n",ro_key);
	if(ro_key==0)
		return(0);
	if(ro_size==0)
		return(0);
        private_segs_ro_pointer = get_sh_mem(ro_key, ro_size,
             remove_flag);
        if(private_segs_ro_pointer == NULL)
        {
             return((char *)0);
        }
	fclose(reg);
	return(private_segs_ro_pointer);
}
void
usage()
{
	printf("\t-h............  Help screen\n");
	printf("\n");
}

char *
get_timestamp(char *string)
{
    time_t      run_str;

    run_str = time((time_t *)&dummytime);
    (void) strncpy((char *) string, (char *) ctime((time_t *)&run_str), 24);
    return(string);
}

