/*                                                                           */
/* Copyright 2011                                                            */
/* Created: May 2011                                                         */
/* Author: Don Capps (capps@iozone.org)                                      */
/* Location: Iozone.org                                                      */
/*                                                                           */
/* --------------------------------------------------------------------------*/
/*                                                                           */
/*       Author: Don Capps (capps@iozone.org)                                */
/*               7417 Crenshaw                                               */
/*              Plano, TX 75025                                              */
/*                                                                           */
/*  Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011 Don Capps    */
/*                                                                           */
/*  License to freely use and distribute this software is hereby granted     */
/*  by the author, subject to the condition that this copyright notice       */
/*  remains intact.  The author retains the exclusive right to publish       */
/*  derivative works based on this work, including, but not limited to,      */
/*  revised versions of this work.                                           */
/*                                                                           */
/*  Universities, and research communities are hereby granted by the         */
/*  author, subject to the condition that this copyright notice remains      */
/*  intact, the right to use and distribute modified versions of this        */
/*  software providing that all modifications are clearly stated in the      */
/*  NOTICE file that will be provided with all derivative works.  Any        */
/*  derivative works must also remain available for free.  Attribution of    */
/*  the original authors and contributors must also remain intact.           */
/*                                                                           */
/*  Trademarks. This License does not grant permission to use the trade      */
/*  names, trademarks, service marks, or product names of the Licensor,      */
/*  except as required for reasonable and customary use in describing the    */
/*  origin of the Work and reproducing the content of the NOTICE file.       */
/*                                                                           */
/*  Disclaimer of Warranty. Unless required by applicable law or             */
/*  agreed to in writing, Licensor provides the Work (and each               */
/*  Contributor provides its Contributions) on an "AS IS" BASIS,             */
/*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or          */
/*  implied, including, without limitation, any warranties or conditions     */
/*  of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A            */
/*  PARTICULAR PURPOSE. You are solely responsible for determining the       */
/*  appropriateness of using or redistributing the Work and assume any       */
/*  risks associated with Your exercise of permissions under this License.   */
/*                                                                           */
/*  Limitation of Liability. In no event and under no legal theory,          */
/*  whether in tort (including negligence), contract, or otherwise,          */
/*  unless required by applicable law (such as deliberate and grossly        */
/*  negligent acts) or agreed to in writing, shall any Contributor be        */
/*  liable to You for damages, including any direct, indirect, special,      */
/*  incidental, or consequential damages of any character arising as a       */
/*  result of this License or out of the use or inability to use the         */
/*  Work (including but not limited to damages for loss of goodwill,         */
/*  work stoppage, computer failure or malfunction, or any and all           */
/*  other commercial damages or losses), even if such Contributor            */
/*  has been advised of the possibility of such damages.                     */
/*---------------------------------------------------------------------------*/
/*
 * --------------------------------------------------------------------------
 * pdsm_master  ( The master instance )
 *
 * Persistent shared memory with *some* globally distributed 
 * attributes.
 *
 * "ro" (remote owned) Private segments are for children to send 
 * data to the master.
 *
 * "mo" (master owned ) Shared segment is for master to send data to 
 * remote children.
 *
 *
 *   ---------------------    ---------------------  ->
 *   | Private seg       |    | Private seg       |  ->
 *   | Remote Owned (ro) |    | Master owned (mo) |  ->
 *   ---------------------    ---------------------  ->
 *   *This* remote writes     Master writes here
 *   here, and sends          and sends update to
 *   updates to master.       *this* remote.
 *
 *
 */

#ifdef WIN32
#pragma warning(disable: 4996)
#pragma warning(disable: 4101)
/*Windows specific headers*/
#include <Ws2tcpip.h>
#include <Windows.h>
#include <tchar.h>
#include <win32_getopt.h>
#include <win32_sub.h>
#endif

#if defined(_macos_) || defined(_bsd_)
#include <sys/wait.h>
#include <ctype.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#ifndef WIN32
/* UNIX specific headers*/
#include <unistd.h>
/* Yes, HPUX is this BROKEN !!! */
#if defined(_hpux)
#  ifndef _CMPTID_T
#    define _CMPTID_T
     typedef int32_t   cid_t;           /* For compartment IDs */
     typedef cid_t     cmpt_t;
#  endif /* _CMPTID_T */
#else
#include <sys/ipc.h>
#endif
#include <sys/shm.h>
#include <strings.h>
/* 
 * This __USE_GNU gets the strcasestr to return a correct char * (pointer) 
 * and is needed on 64 bit platforms.
 */
#define __USE_GNU
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#endif

#ifdef WIN32
/*
 * Create file on C: drive
 */
#define PDSM_MM_REGISTRY "c:\\tmp\\pdsm_master_registry"
#else
#define PDSM_MM_REGISTRY "/tmp/pdsm_master_registry"
#endif

#define MAXREMOTES 255
/* The TCP ports where things connect for the coherency traffic */
#define MASTER_ALIST_PORT  22000
#define MASTER_ASEND_PORT  24000
#define REMOTE_ASEND_PORT  26000
#define REMOTE_ALIST_PORT  28000

#define THISVERSION "        $Revision: 793 $"
char version[]=THISVERSION;
/* 
 * Used to debug the master 
 * 
 * 1 ... every day stuff.
 * 2 ... very verbose stuff.
 */

int mdebug = 2;
int ipv6_enable;
extern char *optarg;
/*
 * Format of segments:
 *
 *   struct internal {
 *      long Which_seg;
 *      char data[GEN_SZ]
 *   }
 *
 */

#define GEN_SZ (4096)
#define INBUFSZ (GEN_SZ)
/* How often to sync the memory segments */
#define UPDATE_TIME 1

/* Temporary buffers */
char remote_hostname[MAXREMOTES][255];   /* Remote hostnames */
char *extbuf;                   /* string output buffer */
char dummybuf[1024*1024];                   /* string output buffer */
char *intbuf;                    	/* Binary input buffer */
char *oldseg[MAXREMOTES];                /* Used to detect changes */
int mo_changed[MAXREMOTES];              /* If a segment was changed */
char dummybuf1[1024*1024];                   /* string output buffer */
int master_asend_socket_val[MAXREMOTES]; /* sockets to async remote listeners */
int size_of_extbuf;

/* 
 * Prototypes & types
 */
#ifdef WIN32
/*thread handles - threads are created to accept connections from remote client*/
HANDLE hThreadArray[MAXREMOTES];
SOCKET socks[MAXREMOTES]; //these are listening sockets, one per client.
/*
 * Define types and structures for Windows platform
 */
struct timespec
{
   time_t  tv_sec;         /* seconds */
   long    tv_nsec;        /* nanoseconds */
};
typedef unsigned int key_t;

/*
 * Implement function calls as macros for Windows platform
 */
#define bzero(addr, size) RtlZeroMemory((VOID UNALIGNED*)(addr), (SIZE_T)(size))
#define close(socket) closesocket((SOCKET)(socket))
/*
 * Windows "Sleep()/SleepEx()" can only achieve 1 milisecond resolution, 
 * so the sleep time will be rounded up to milisecond, 
 * and it always returns 0 
 */
int nanosleep(const struct timespec *req, struct timespec *rem);
/* functions that have different signatures than their UNIX version*/
DWORD WINAPI start_m_listen_proc(__in  LPVOID lpParameter);
int start_master_listen_async(int size_of_message,int which, void* inout_addr);
int master_attach(int s, void* addr);
void master_listen_async(int sock, int size_of_message, char* cnc);
#else
/*Function prototype on UNIX*/
typedef int SOCKET;
void start_m_listen_proc(int who);
int start_master_listen_async(int size_of_message,int which);
int master_attach(int s);
void master_listen_async(int, int);
#endif
int num_clients;
int gflag;
char client_tok_filename[256];

/* 
 * Prototypes 
 */
void nap( int );
char * get_sh_mem(key_t, size_t, int);
void master_parent(void);
void master_child(void);
void remote_parent(void);
void remote_child(void);
void master_send(int, char *, char *, int, int);
int nlongs_to_ulongs(size_t, unsigned int *, unsigned int *);
void ulongs_to_nlongs(size_t, unsigned int *, unsigned int *, int );
unsigned int * remote_ipaddr(char *);
unsigned int * my_ipaddr(void);
int mo_change_check(int);
void usage(void);
void get_names(void);
int start_master_send_async(char *child_host_name, int remote_port, unsigned int * my_addr_buf);
void read_client_tok_file(void);
int validate_all_tokens(int);
int parse_token(char *, char *, char *);
char * my_strcasestr(char *, char *);
int get_host_address(char *, unsigned int *);
int parse_all_tokens(char *input );
/*
 * Allocate shared pools of memory
 */
#define PRIVATE_SEG_SIZE (INBUFSZ)

/* 
 * The pools of memory 
 */
int number_of_private_segs;    /* Number of segs & remotes */
int size_of_private_segs = PRIVATE_SEG_SIZE; /* page size + long */

char *private_segs_mo_pointer; /* shared memory pointer for mo */
char *private_segs_ro_pointer; /* shared memory pointer for ro */

int shmid;
key_t mkey = 0x3141592;        /* Key for all mo segments  */
key_t rkey = 0x3141593;        /* Key for all ro segments  */
size_t total_size;

/* Temp variables */
int cret;
int remove_flag;
int chid;                       /* my unique ID */
int child_pid;                  /* just used for forking */
int once[MAXREMOTES];
FILE *fd;
int total_clients;
int k = 0;

/*
 * Main: Start it up :-)
 */
int 
main(int argc, char **argv)
{
	int page_size;
#ifdef WIN32	
	WSADATA wsaData;
	if (0 != WSAStartup( MAKEWORD(2,2), &wsaData ))
	{
		/* If Windows Sockets does not initialize correctly
		 * we have trouble.
		 */
		printf("WSAStartup failed");
		exit(1);
	}
#endif
	if(argc == 1)
	{
		usage();
		exit(0);
	};
	page_size = getpagesize();
	while((cret = getopt(argc,argv,"hrvs:g:y:")) != EOF){
                switch(cret){
                case 'h':                 /* Help screen */
			usage();
			exit(0);
			break;
                case 'r':		  /* Remove segments */
			remove_flag = 1;
			break;
                case 'g':
			strcpy(client_tok_filename,optarg);
			gflag=1;
			break;
                case 's':
			size_of_private_segs = atoi(optarg);
			/*
			 * Round up to page size.
			 */
        		size_of_private_segs= ((size_of_private_segs + 
				(page_size-1)) & ~(page_size-1));
			break;
#ifdef OLD
                case 'p':                 /* Total count of remotes */
			number_of_private_segs = atoi(optarg);
			if(number_of_private_segs  < 0)
				number_of_private_segs = 0;
			break;
#endif
                case 'y':
			ipv6_enable = atoi(optarg);
			break;
                case 'v':		  /* Remove segments */
			printf("pdsm_master: Version %s\n",version);
			exit(0);
			break;
		};
	}	
	if(!gflag)
	{
		usage();
		exit(1);
	}
 	/* 
  	 * Get the names of all of the remote clients.
	 */
	read_client_tok_file();
	number_of_private_segs = k;

	intbuf=(char *)malloc(size_of_private_segs);
	size_of_extbuf = sizeof(int) + size_of_private_segs;
	extbuf = (char *)malloc(size_of_extbuf);
 	/* 
  	 * Allocate, or attach, to shared memory segments 
	 */
	if((number_of_private_segs != 0) && (size_of_private_segs != 0))
	{
		total_size = (number_of_private_segs * size_of_private_segs);	
		/* Remember... sizes are actually rounded up to page size */
		private_segs_mo_pointer = get_sh_mem(mkey, total_size,
			remove_flag);
		if(private_segs_mo_pointer<=0)
		{
			printf("Unable to attach to MO segment\n");
			exit(1);
		}
		if(mdebug>=1 && !remove_flag)
		  printf("Master allocated/attached %d bytes of (MO) pdsm. Key = %lx\n",
			(int)total_size, (unsigned long)mkey);
		bzero((char *)private_segs_mo_pointer,total_size);

		private_segs_ro_pointer = get_sh_mem(rkey, total_size,
			remove_flag);
		if(private_segs_ro_pointer<=0)
		{
			printf("Unable to attach to RO segment\n");
			exit(1);
		}
		if(mdebug>=1 && !remove_flag)
		  printf("Remote allocated/attached %d bytes of (RO) pdsm. Key = %lx\n",
			(int)total_size, (unsigned long)rkey);
		bzero((char *)private_segs_ro_pointer,total_size);
	}

	if(remove_flag)	 /* this will trigger the auto removal of segs */
	{
		  printf("Master released %d bytes of (MO) pdsm. Key = %lx\n",
			(int)total_size, (unsigned long)mkey);
		  printf("Remote released %d bytes of (RO) pdsm. Key = %lx\n",
			(int)total_size, (unsigned long)rkey);
		unlink(PDSM_MM_REGISTRY);
		exit(0);
	}

	fd = fopen(PDSM_MM_REGISTRY,"w");
	if(fd < 0)
	{
		printf("Unable to create %s.\n",PDSM_MM_REGISTRY);
		exit(1);	
	}
	fprintf(fd,"PDSM_MO_key 0x%lx PDSM_MO_size %d Number_of_segs %d Seg_size %d\n",
		(unsigned long)mkey,(int)total_size,number_of_private_segs,size_of_private_segs);

	fprintf(fd,"PDSM_RO_key 0x%lx PDSM_RO_size %d Number_of_segs %d Seg_size %d\n",
		(unsigned long)rkey,(int)total_size,number_of_private_segs,size_of_private_segs);
	fclose(fd);
/*
	get_names();
*/
#ifdef WIN32
	master_child(); //start thread to listen and receive from remote client
	master_parent(); //main thread to send the update to remote clients
	 WSACleanup();
#else
	child_pid=fork();
	if(child_pid==0)
		master_child();      /* Listens for updates (ro) */
	else
		master_parent();     /* Sends updates (mo)       */
	wait(0);
#endif
}


/*
 * Parent loops, and sends messages to children when things 
 * have changed in the shared seg. 
 */
void
master_parent(void)
{
	int sock,s,i;
	struct in_addr r_saddr;
	unsigned int *myseg;
	int dest_who;
	unsigned int *my_addr_buf;

	if(mdebug>=2)
	{
		printf("Master parent starting\n");
	}

	for(i=0; i<number_of_private_segs; i++)
	{
		my_addr_buf = remote_ipaddr(remote_hostname[i]);
		s=start_master_send_async(remote_hostname[i], 
			REMOTE_ALIST_PORT+i,my_addr_buf);
		master_asend_socket_val[i]=s;
	}

	/* 
	 * One send to sync every client remote mo 
 	 */
	for(i=0; i<number_of_private_segs; i++)
		 once[i]=1;
	/* 
	 * Loop, sending mo updates when needed.
 	 */
	while(1)
	{
		for(i=0; i<number_of_private_segs; i++)
		{
		  mo_changed[i]=mo_change_check(i);
		  if(mo_changed[i] || once[i])
		  {
			if(mdebug>=2)
			  printf("Sync for master owned segment %d for %s\n",i,&remote_hostname[i][0]);
			mo_changed[i] = 0;
			if(master_asend_socket_val[i])
			{
			  /* Copy from mo_private to inbuf */
			  myseg = (unsigned int *)(private_segs_mo_pointer + 
				(size_of_private_segs * i));
			  dest_who=i;
			  /* Send to remote client */
			  master_send(master_asend_socket_val[i], &remote_hostname[i][0],(char *)myseg,
				size_of_private_segs,dest_who);
			}
			once[i]=0;
		  }

		}
		sleep(UPDATE_TIME);
	}
}

/*
 * Check to see if the mo segment has changed 
 */
int
mo_change_check(int i)
{
	int j;
	int flag;
	char *myseg;
	int *p1,*p2;

	if(oldseg[i]==NULL)
	{
		oldseg[i]=(char *)malloc(size_of_private_segs);
		bzero(oldseg[i],size_of_private_segs);
	}

	flag = 0;
	p1= (int *)(private_segs_mo_pointer + (size_of_private_segs * i));
	p2= (int *) oldseg[i];
	for(j=0;j< (int)(size_of_private_segs/sizeof(int));j++)
	{
		if(*p2 != *p1)
		{
		   flag=1;
/*
		   if(mdebug>=2)
		      printf("Master %llx changed was %d now  %d\n",(unsigned long long)p1,(int)(*p2), (int)(*p1));
*/
		}
		*p2=*p1;
		p2++;
		p1++;
	}
	return(flag);
}

/*
 * The master's children become listeners for remote's 
 * messages to update the "ro" segs.
*/
void
master_child(void)
{
	int i;

	if(mdebug >= 2)
	{
		printf("Master children starting\n");
	}
	for(i=0;i < number_of_private_segs; i++)
	{
#ifdef WIN32
		/*We create threads to process the incoming message from remote clients*/
		hThreadArray[i] = CreateThread( 
										NULL,                   // default security attributes
										0,                      // use default stack size  
										start_m_listen_proc,       // thread function name
										(LPVOID)i,          // argument to thread function 
										0,                      // use default creation flags 
										NULL);   // returns the thread identifier 

#else
		start_m_listen_proc(i);
#endif
	}
#ifndef WIN32
	wait(0); /* Wait for all children to terminate */
#endif
}

/*
 * Start a child proc to listen for ro updates.
 */
#ifdef WIN32
DWORD WINAPI start_m_listen_proc(__in  LPVOID lpParameter)
#else
void start_m_listen_proc(int who)
#endif
{
	int sock,p;
	char *which_ro_pointer;
	int dest_who;

#ifdef WIN32
	char* extbuf_local = (char *)malloc(size_of_private_segs);
	char* intbuf_local = (char *)malloc(size_of_extbuf);
	int who = (int)lpParameter;
	struct sockaddr_in async_sock; //use local variable for thread safety
	struct sockaddr_in6 async_sock6; //use local variable for thread safety
	void* addr = NULL; //use local variable for thread safety
#else
	p=fork();
	if(p==0)	
	{
#endif

#ifdef WIN32
	   //save the listening socket to the array
	   //the rountine is modified to return addr structure	
	   if (ipv6_enable)
		   addr = (void*)&async_sock6;
	   else 
		   addr = (void*)&async_sock;
	   socks[who]=start_master_listen_async(size_of_extbuf, who, addr);
	   sock=master_attach((int)socks[who], addr);
#else
	   sock=start_master_listen_async(size_of_extbuf, who);
	   sock=master_attach(sock);
#endif
	   /*
	    * Loop forever, getting ro updates 
	    */
	   if(mdebug >= 2 )
	      printf("Master looping in receive \n");
	   while(1)
	   {

#ifdef WIN32
		master_listen_async(sock,size_of_extbuf, extbuf_local);
		dest_who=nlongs_to_ulongs(size_of_private_segs,
			(unsigned int *)extbuf_local,(unsigned int *)intbuf_local);
#else
		master_listen_async(sock,size_of_extbuf);
		dest_who=nlongs_to_ulongs(size_of_private_segs,
			(unsigned int *)extbuf,(unsigned int *)intbuf);
#endif
		printf("Master received update from remote id %d\n",dest_who);
		which_ro_pointer = (char *)private_segs_ro_pointer + 
			(dest_who*size_of_private_segs);
#ifdef WIN32
		memcpy(which_ro_pointer,intbuf_local,size_of_private_segs);
	   } //end of while
	   free(extbuf_local);
	   free(intbuf_local);
	   return 0;
#else
		memcpy(which_ro_pointer,intbuf,size_of_private_segs);
	   }
	}
	else
	{
		return; /* let caller go to the next */
	}
#endif
}

/* 
 * This is like an xdr layer, but chunky style and portable.
 * These two subroutines convert binary format data into string data, 
 * and converts string data back into the binary format.
 */

void
ulongs_to_nlongs(size_t size, unsigned int *inbuf, unsigned int *outbuf,
		int who )
{
	int i;
	unsigned int *outp;
	/*
	 * The first external buffer long is "who" 
	 */
	outp = (unsigned int *)outbuf;
	*outp = htonl(who);
	outp++;
	for(i = 0; i < (int)(size/sizeof(int)); i++)
	{
	   *outp = htonl(inbuf[i]);
	   outp ++;
	}
}

int
nlongs_to_ulongs(size_t size, unsigned int *inbuf, unsigned int *outbuf)
{
	int i;
	int who;
	unsigned int *pointer;

	/*
	 * The first external buffer int is "who" 
	 */
	pointer= (unsigned int *)inbuf;

	who=ntohl(*pointer);
	pointer++;
	if(mdebug>=2)
		printf("Master got dest_who of %d\n",who);
	for(i = 0; i < (int)(size/sizeof(int)); i++)
	{
	   outbuf[i]=ntohl(*pointer);
	   pointer++;
	}
	return(who);
}

/*
 * Creates shared memory, or if it exists... just attaches and bumps
 * the ref count. 
 * 
 * @param key 
 *        Used to identify the unique public shared memory seg.
 * @param Size 
 *   	  Needs to be big enough to handle all results 
 *
 * This mechanism creates, if needed, and then bumps a ref attach 
 * count and then attempts to remove the shared memory segement. 
 * The "flag" will permit an automatic delete of the shared 
 * memory segment whenever the process exits. (normal, or abnormal exit)
 */
#ifdef WIN32

#define MAX_SHARE_MEMORY 32 //max number of shared memory one process can request, regardless of their size
typedef struct _SHARED_MEMORY
{
	HANDLE hFile; //associated file handle
	HANDLE hMapFile; //associated mapping object handle
	key_t  key; //shared memory key
	BOOL   valid;
} SHARED_MEMORY;

static SHARED_MEMORY sharedMem[MAX_SHARE_MEMORY] = {0};
/*
 *	Win32 implementation using memory mapped file
 *  key - share memory id, it is used to generate mapping object
 *  size - the size of shared memory requested
 *  flag - delete shared memory if its value is non-zero
 */
char *
get_sh_mem(key_t key, size_t size, int flag)
{
	TCHAR szObjName[255];
	TCHAR szFileName[255];
	char* pMem = NULL;
	int position=-1, i=0; //index in shared memory array where we can save shared_memory information 
	HANDLE hFile;
	BOOL foundExisting = FALSE;
	
	/* generate mapping object name */
	_stprintf(szObjName, TEXT("Global\\pdsm-%lx"), (unsigned long)key);
	_stprintf(szFileName, TEXT("SharedMemory.%lx"), (unsigned long)key);
	/*First we search the shared memory array, try to get existing one*/
	for (i=0; i<MAX_SHARE_MEMORY; i++)
	{
		if (sharedMem[i].valid && sharedMem[i].key == key)
		{
			position = i;
			break;
		}
		else if (position < 0  && !sharedMem[i].valid)
			position = i; //find a unused slot
	}
	/* if we reach here, we did not find exising shared memory*/
	if (flag && sharedMem[position].valid)
	{
		/*do we need to delete memory mapped file here???*/
		hFile = CreateFile(szFileName,
							GENERIC_WRITE | GENERIC_READ,
							FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,
							NULL,
							OPEN_EXISTING,	 
							FILE_FLAG_DELETE_ON_CLOSE, //delete file on close
							NULL);
		if (hFile != INVALID_HANDLE_VALUE)
			CloseHandle(hFile);
	}

	if (position < 0)
		/*the process has already allocated max-allowed shared memory, we have no choice but fail the call*/
		return NULL; 
	/* Now we start to create new shared memory*/
	if (!sharedMem[position].valid)
	{
		sharedMem[position].hMapFile = NULL;
		/* We open or create a file to map to memory*/
		if ((sharedMem[position].hFile = CreateFile(szFileName,
													GENERIC_WRITE | GENERIC_READ,
													FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,
													NULL,
													OPEN_ALWAYS,	 // create a new file if it does not exist, otherwise open the old file
													(flag?FILE_FLAG_DELETE_ON_CLOSE:FILE_ATTRIBUTE_NORMAL), 
													NULL)) == (HANDLE)INVALID_HANDLE_VALUE)
		{
			if(mdebug >= 2)
				printf("Unable to open or create file for memory mapping:\"SharedMemory.%lx\"\n",(unsigned long)key);
		} 
		else
		{
			/*Create mapping object*/
			sharedMem[position].hMapFile = CreateFileMapping(
															sharedMem[position].hFile,    // use paging file
															NULL,                    // default security
															PAGE_READWRITE,          // read/write access
															0,                       // maximum object size (high-order DWORD)
															(DWORD)size,                // maximum object size (low-order DWORD)
															szObjName);                 // name of mapping object
			if (sharedMem[position].hMapFile == NULL)
			{
				if (mdebug >= 2)
					printf("Unable to create mapping object: \"Global\\pdsm-%lx\", error code: %d\n",(unsigned long)key, GetLastError());
				CloseHandle(sharedMem[position].hFile);
			}
			else 
			{
				if (mdebug >= 2 && GetLastError() == ERROR_ALREADY_EXISTS)
					printf("CreateFileMapping returned a handle of existing mapping object\n");
				sharedMem[position].key = key;
				sharedMem[position].valid = TRUE;
			}
		}
	}

	if (sharedMem[position].hMapFile != NULL)
	{
		//
		//now we need to mapped the view. 
		//
		pMem = (char*) MapViewOfFile(sharedMem[position].hMapFile,   // handle to map object
										FILE_MAP_ALL_ACCESS, // read/write permission
										0,
										0,
										(DWORD)size);
		if (pMem == NULL && mdebug >= 2)
			printf("Unable to map file to memory, error code: %d\n", GetLastError());
	}
	return pMem;
}


/*The sleep cannot be interrpted, so field in "rem" will be always 0*/
int nanosleep(const struct timespec *req, struct timespec *rem)
{
	unsigned int milisec = 0;

	if (!req || req->tv_nsec == 0 && req->tv_sec == 0)
		return -1;
	milisec = (unsigned int)req->tv_sec*1000; //convert to milisecond
	milisec += (req->tv_nsec/1000000); //convert to milisecond
	if (req->tv_nsec % 1000000)
		milisec++; //round the remainder up.
	Sleep(milisec);
	
	if (rem)
	{
		rem->tv_nsec = 0;
		rem->tv_sec = 0;
	}
	return 0;
}


#else
char *
get_sh_mem(key_t key, size_t size, int flag)
{
	struct stats *statsp;

	if(mdebug >= 2)
	  printf("Trying to get shared memory segment with key %lx\n",(unsigned long)key);
        shmid=shmget((key_t)key,(size_t)size,(int)(0666)|IPC_CREAT);
        if(shmid < 0)
        {
                printf("Shmid %d Errno %d\n",shmid,errno);
        	return((char *)0);
        } 
	if(mdebug >= 2)
	  printf("Got shared memory with key %lx\n",(unsigned long)key);
        statsp=(struct stats *)shmat(shmid,0,0);
        if(statsp == (struct stats *)-1)
        {
                printf("Attach to shared memory failed\n");
        	return((char *)0);
        }
	/* Attach ref count is now all that is holding on to this segment   */
	if(flag)
		shmctl(shmid, IPC_RMID, 0);  /* Remove from the namespace.  */
					     /* This arms a auto cleanup on */
					     /* exit 			    */
	if(mdebug >= 2)
	  printf("Attached to shared memory with key %lx\n",(unsigned long)key);

	if(mdebug >= 2)
	  printf("Shared memory key %lx Size = %d\n",(unsigned long)key, (int)size);
	/*
	 * Don't need to call quit, as any termination will cleanup
 	 * automatically.
	 */
        /*quit();*/
        return((char *)statsp);
}
#endif
struct sockaddr_in master_async_sock; /* Temp variable */
struct sockaddr_in6 master_async_sock6; /* Temp variable */

/*
 * Start the master async listening service for messages from the master.
 */
#ifdef WIN32
/*The addr passed in will be used later by other function*/
int
start_master_listen_async(int size_of_message,int which, void* inout_addr)
#else
int
start_master_listen_async(int size_of_message,int which)
#endif
{
	int tsize;
	SOCKET s;
	int rc;
	int xx;
	int tmp_port;
	int sockerr;
	int recv_buf_size=65536;
	int optval=1;
#ifndef WIN32
	struct sockaddr_in* addr = &master_async_sock;
	struct sockaddr_in6* addr6 = &master_async_sock6;
#else
	struct sockaddr_in* addr = NULL;
	struct sockaddr_in6* addr6 = NULL;
	if(ipv6_enable)
		addr6 = (struct sockaddr_in6*)inout_addr;
	else
		addr = (struct sockaddr_in*)inout_addr;
#endif
	xx = 0;
	tsize=size_of_message; /* Number of messages to receive */
	if(ipv6_enable)
           s = socket(AF_INET6, SOCK_STREAM, 0);
	else
           s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0)
    {
            perror("socket failed:");
            exit(19);
    }
	sockerr = setsockopt (s, SOL_SOCKET, SO_RCVBUF, (char *)
		&recv_buf_size, sizeof(int));
	if ( sockerr == -1 ) {
		perror("Error in setsockopt 5\n");
	}
	sockerr = setsockopt (s, SOL_SOCKET, SO_REUSEADDR, (char *)
		&optval, sizeof(int));
	if ( sockerr == -1 ) {
		perror("Error in setsockopt 6\n");
	}
	tmp_port=MASTER_ALIST_PORT + which;
	if(ipv6_enable)
	{
           bzero(addr6, sizeof(struct sockaddr_in6));
           addr6->sin6_port = htons(tmp_port);
           addr6->sin6_family = AF_INET6;
			/* This line needs work */
		   memset((void*)&addr6->sin6_addr.s6_addr[0],0,16);
	}
	else
	{
           bzero(&master_async_sock, sizeof(struct sockaddr_in));
           addr->sin_port = htons(tmp_port);
           addr->sin_family = AF_INET;
           addr->sin_addr.s_addr = INADDR_ANY;
	}
        rc = -1;
        while (rc < 0)
        {
		if(ipv6_enable)
                   rc = bind(s, (struct sockaddr *)addr6,
                                sizeof(struct sockaddr_in6));
		else
                   rc = bind(s, (struct sockaddr *)addr,
                                sizeof(struct sockaddr_in));
		if(rc < 0)
		{
			tmp_port++;
			if(ipv6_enable)
                	  addr6->sin6_port=htons(tmp_port);
			else
                	  addr->sin_port=htons(tmp_port);
			continue;
		}
        }
	if(mdebug >= 2)
	{
		printf("Master's child: Async Listen: Bound at port %d\n",
			tmp_port);
	}
	if(rc < 0)
	{
		printf("bind failed. Errno %d \n",errno);
		exit(20);
	}
	return((int)s);
}

/*
 * Start the channel for the master to send a message to 
 * a particular child on a particular port that the child
 * has created for the parent to use to communicate.
 */
int
start_master_send_async(char *child_host_name, int remote_port, 
		unsigned int * my_addr_buf)
{
	int rc;
	SOCKET master_socket_val;
	struct sockaddr_in addr,raddr;
	struct sockaddr_in6 addr6,raddr6;
	int port,tmp_port;
	int ecount = 0;
/*
	struct timespec req,rem;
	req.tv_sec = 0;
	req.tv_nsec = 10000000;
	rem.tv_sec = 0;
	rem.tv_nsec = 10000000;
*/


	port=remote_port;
	nap(1);
/*
	nanosleep(&req,&rem);
*/

over:
	if(ipv6_enable)
	{
           raddr6.sin6_family = AF_INET6;
           raddr6.sin6_port = htons(port);
           memcpy(&raddr6.sin6_addr.s6_addr,my_addr_buf,16);
           master_socket_val = socket(AF_INET6, SOCK_STREAM, 0);
	}
	else
	{
        raddr.sin_family = AF_INET;
        raddr.sin_port = htons(port);
           memcpy(&raddr.sin_addr.s_addr,my_addr_buf,4);
           master_socket_val = socket(AF_INET, SOCK_STREAM, 0);
	}
    if (master_socket_val < 0)
    {
            perror("Master: async socket failed:");
            exit(23);
    }

	tmp_port=MASTER_ASEND_PORT; 
	if(ipv6_enable)
	{
           bzero(&addr6, sizeof(struct sockaddr_in6));
           addr6.sin6_port = htons(tmp_port);
           addr6.sin6_family = AF_INET6;
		   memset(&addr6.sin6_addr.s6_addr[0],0,16);
	}
	else
	{
           bzero(&addr, sizeof(struct sockaddr_in));
           addr.sin_port = htons(tmp_port);
           addr.sin_family = AF_INET;
           addr.sin_addr.s_addr = INADDR_ANY;
	}
	//tmp_port=MASTER_ASEND_PORT; /*move this line up, otherwise we will use uninitilized tmp_port*/
        rc = -1;
        while (rc < 0)
        {
		if(ipv6_enable)
                   rc = bind(master_socket_val, (struct sockaddr *)&addr6,
                                                sizeof(struct sockaddr_in6));
		else
                   rc = bind(master_socket_val, (struct sockaddr *)&addr,
                                                sizeof(struct sockaddr_in));
		if(rc < 0)
		{
			tmp_port++;
			if(ipv6_enable)
                	  addr6.sin6_port=htons(tmp_port);
			else
                	  addr.sin_port=htons(tmp_port);
			continue;
		}
        }
	if(mdebug >= 2)
	{
		if(ipv6_enable)
		  printf("Master: Bound async send port %d\n",
			ntohs(addr6.sin6_port));
		else
		  printf("Master: Bound async send port %d\n",
			ntohs(addr.sin_port));
	}
        if (rc < 0)
        {
                perror("Master: bind async failed\n");
                exit(24);
        }
again:

	if(mdebug >= 2)
           printf("Master: calling connect..\n");
	if(ipv6_enable)
           rc = connect(master_socket_val, (struct sockaddr *)&raddr6, 
			sizeof(struct sockaddr_in6));
	else
           rc = connect(master_socket_val, (struct sockaddr *)&raddr, 
			sizeof(struct sockaddr_in));
	if (rc < 0)
        {
		if(ecount++ < 300)
		{
		/* Really need this sleep for Windows */
			sleep(1);
			goto again;
		}
                perror("Master: async connect failed\n");
                close(master_socket_val);

		sleep(1);

		/* nanosleep(&req,&rem);*/
                ecount=0;
                goto over;
        }
	if(mdebug >= 2)
	{
		printf("Master async send Connected\n");
		fflush(stdout);
	}
	sleep(1);
	return ((int)master_socket_val);
}

	
/*
master_send(master_send_async_sockets[i],child_idents[i].child_name, 
		&cc,sizeof(struct client_command));
*/

/*
 * Master sending message to a child
 * There should be a unique child_socket_val for each
 * child.
 */
void
master_send(int remote_socket_val, char *host_name, char *inbuf, 
	int send_size, int who)
{
	int rc;
	unsigned int *myseg;

	bzero(extbuf,size_of_extbuf);
	if(mdebug>=2)
	{
		printf("Neutral size = %d\n",(unsigned int)size_of_extbuf);
	}
	/* 
	 * Convert internal commands to string format for neutral 
	 * format/portability
	 */
	myseg = (unsigned int *)inbuf;

	ulongs_to_nlongs((size_t)size_of_private_segs,(unsigned int *)inbuf,
		(unsigned int *)extbuf,who);

	if(mdebug >= 1)
		printf("Master sending update message to remote \"%s\" size %d ID: %d\n",
			host_name,size_of_extbuf,who);
        rc = write(remote_socket_val, (char *)extbuf, size_of_extbuf);
        if (rc < 0)
        {
                perror("write failed\n");
                exit(26);
        }
}

unsigned int addr_buf[16];

unsigned int * 
remote_ipaddr(char *hostname)
{

	get_host_address(hostname, addr_buf);
	return(addr_buf);

}

unsigned int *
my_ipaddr(void)
{

	char myhostname[200];

	gethostname(myhostname,200);
	get_host_address(myhostname, addr_buf);
	return(addr_buf);
}

/*
 * Master async listener
 */
#ifdef WIN32
void
master_listen_async(int sock, int size_of_message, char* cnc)
#else
void
master_listen_async(int sock, int size_of_message)
#endif
{
	int tsize;
	int rcvd;
	int s;
	int rc;
#ifndef WIN32
	/*use global buffer for UNIX as it is single threaded*/
	char *cnc;
	cnc = extbuf;
#endif
	s = sock;
	tsize=size_of_message; /* Number of messages to receive */
	rcvd = 0;
	while(rcvd < tsize)
	{
		if(mdebug >= 2)
		{
			printf("Master %d In async recieve \n",(int)chid);
		}
		rc=read(s,cnc,size_of_message);
		if(rc < 0)
		{
			printf("Master failed. Errno %d \n",errno);
			exit(21);
		}
		/* Special case. If master gets final results, it can 
                   exit, and close the connection to the async child
                   too quickly. When this happens the child gets a 
		   read() that returns 0. It just needs to exit here.
		*/
		if(rc==0)
			exit(0);
		if(mdebug >= 2)
		{
			printf("Master: Got %d bytes (async) \n",rc);
		}
		rcvd+=rc;
		cnc+=rc;
	}
}
#ifdef WIN32
int
master_attach(int s, void *addr)
#else
int
master_attach(int s)
#endif
{
	unsigned int me,me6;
	SOCKET ns;
	struct sockaddr_in6 *addr6;
#ifndef WIN32	
	
	struct sockaddr_in *addr;
	addr6=&master_async_sock6;
	addr=&master_async_sock;
#else
	addr6 = (struct sockaddr_in6 *)addr;
#endif
	if(mdebug>=2)
	{
		printf("Master %d attach async\n",(int)chid);
	}
	if(ipv6_enable)
	   me6=sizeof(struct sockaddr_in6);
	else
	   me=sizeof(struct sockaddr_in);
	if(mdebug>=2)
	{
		printf("Master %d enters listen\n",(int)chid);
	}
	listen(s,10);
	if(mdebug>=2)
	{
		printf("Master %d enters accept\n",(int)chid);
	}
	if(ipv6_enable)
	   ns=accept(s,(void *)addr6,(socklen_t *)&me6);
	else
	   ns=accept(s,(void *)addr,(socklen_t *)&me);
	if(mdebug>=2)
	{
		printf("Master %d attached for receive. Sock %d  %d\n",
		    (int)chid, ns,errno);
	}
	return((int)ns);
}

void get_names(void)
{
	char buffer[200];
	FILE *fd;
	int count = 0;
	char *ret1;
	fd=fopen("pdsm.conf","r");
	if(fd == (FILE *)NULL)
	{
		printf("Unable to open client file \"%s\"\n",
			"pdsm.conf");
		exit(1);
	}
	while(1)
	{
          	if (count > MAXREMOTES) {                                                                           
            	  printf("Too many lines in config - max of %d supported\n",
			MAXREMOTES);
            	  exit(2);
          	}
		if(count >= number_of_private_segs)
			break;
		ret1=fgets(buffer,200,fd);
		if(ret1== (char *)NULL)
			break;
		sscanf(buffer,"%s",&remote_hostname[count][0]);
		if(mdebug>=2)
		   printf("Client %s\n",&remote_hostname[count][0]);
		count++;
	}
	fclose(fd);
}	
/*
 * Section for the parser. This parser pulls out the configuration
 * information using a token based scheme.
 */

#define NUM_INPUT_TOKENS 2
#define MAX_INPUT_TOKENLEN 255
#define T_CLIENT_NAME 0
#define T_INSTANCES 1

struct token {
        int optional;              /* Set to 1 to ignore if token is not present */
        char tname[MAX_INPUT_TOKENLEN];  /* String to search for */
        char tvalue[MAX_INPUT_TOKENLEN]; /* Resulting value */
} token_objects[NUM_INPUT_TOKENS] =
           { {0,"Clientname=",""},
            {1,"Instances=",""}};

#ifdef HAVE_ANSIC_C
void
read_client_tok_file(void)
#else
void
read_client_tok_file()
#endif
{
        int client_count;
        char buf[600];
        FILE *fd;
        int ret;
        char *cret;

        client_count=0;
        fd=fopen(client_tok_filename,"r");
        if(fd==0)
        {
                printf("Bad client filename > %s <\n",client_tok_filename);
                exit(0);
        }
        while(1)
        {
                bzero(buf,600);
                cret=fgets(buf,600,fd);
                if(cret == (char *)0)
                {
                        break;
                }
                if(buf[0]=='#')
                        continue;
                if(buf[strlen(buf)-1]=='\n')
                        buf[strlen(buf)-1]=0;;
                ret = parse_all_tokens(buf);
                if(ret)
                {
                     printf("Error parsing config\n");
                     exit(1);
                }
                ret = validate_all_tokens(client_count);
                if(ret)
                {
                        printf("Error validating config\n");
                     exit(1);
                }
		if(total_clients)
	 	{
                   client_count+=total_clients;
		   total_clients=0;
		}
		else
                   client_count++;

                if(client_count > MAXREMOTES)
                {
                        printf("Master: Too many clients. Increase MAXREMOTES and recompile.\n");
                        exit(1);
                }
        }
        num_clients=client_count;
}

/*
 * Parse an individual token and put result in destination
 */
int
parse_token(char *inputlinep, char *searchtokenp, char *destinationp)
{
        char *tokenp,*cp,*cd;
        int count = 0;
        tokenp = ( char *)my_strcasestr(inputlinep,searchtokenp);
        if(tokenp)
        {
                cp = tokenp+strlen(searchtokenp);
                cd = destinationp;
                while((*cp != ' ') && (cp != inputlinep+strlen(inputlinep)) && count < MAX_INPUT_TOKENLEN )
                {
                   count++;
                   *cd++=*cp++;
                }
                *cd=0;
                return(0);
        }
        return(1);
}

/*
 * Parse all known tokens from the line.
 */
int
parse_all_tokens(char *input )
{
        int i,ret;
        for(i=0;i<NUM_INPUT_TOKENS;i++)
        {
           ret = parse_token(input,token_objects[i].tname,token_objects[i].tvalue);
           if(ret && !token_objects[i].optional)
           {
              printf("Parsing error while looking for %s\n",token_objects[i].tname);
              return(1);
           }
        }
        return(0);
}
/*
 * Validate all tokens are there, and place their values in the
 * client_object table.
 */

int
validate_all_tokens(int client_num)
{
	int j;
        /* Check for Clientname */
        if(strlen(token_objects[T_CLIENT_NAME].tvalue)==0)
        {
                printf("Clientname is missing\n");
                return(1);
        }

        if(strlen(token_objects[T_INSTANCES].tvalue) != 0)
        {

	/* printf("Instances %d\n", atoi(token_objects[T_INSTANCES].tvalue));*/
                total_clients+=atoi(token_objects[T_INSTANCES].tvalue);
		for(j=0;j<total_clients;j++)
		{
        	   strcpy(remote_hostname[k],token_objects[T_CLIENT_NAME].tvalue);
        	   printf("Clientname is %s\n",remote_hostname[k++]);
		}
        }
	else
	{
        	printf("Clientname is %s\n",token_objects[T_CLIENT_NAME].tvalue);
        	strcpy(remote_hostname[client_num],token_objects[T_CLIENT_NAME].tvalue);
		k++;
	}

        return(0);
}

void usage(void)
{
	printf("pdsm_master: [-h] [-r] [-s size ] [-g filename] [-v ]\n");
	printf("\t[-h] ............... help screen.\n");
	printf("\t[-r] ............... Release of shared memory.\n");
	printf("\t[-g config_file ]... Configuration filename.\n");
	printf("\t[-s size ].......... Size of shared memory in bytes.\n");
	printf("\t[-y 6] ............. Use IPv6.\n");
	printf("\t[-v] ............... Print version number and exit.\n");
	printf("\n");
}
void
nap( int select_msecs )
{
#ifndef WIN32
	struct timeval	sleeptime;
	sleeptime.tv_sec = select_msecs / 1000;
	sleeptime.tv_usec = (select_msecs % 1000) * 1000;
	select(0, NULL, NULL , NULL, &sleeptime);
#else
	Sleep(select_msecs);
#endif
}

char *
my_strcasestr(char *s, char *find)
{
	char c, sc;
	size_t len;

	if ((c = *find++) != 0) {
		c = (char)tolower((int)c);
		len = strlen(find);
		do {
			do {
				if ((sc = *s++) == 0)
					return (NULL);
			} while ((char)tolower((int)sc) != c);
		} while (strncasecmp(s, find, len) != 0);
		s--;
	}
	return ((char *)s);
}

int
get_host_address(char *host, unsigned int *buffer)
{
	struct addrinfo     *server_addr;
	struct addrinfo     ai_hints;
	struct sockaddr_in *sain;
	struct sockaddr_in6 *sain6;
	int error;

    	memset((char *)&ai_hints, '\0', sizeof (ai_hints));
    	/* Get host's address */
    	if (ipv6_enable) {
            ai_hints.ai_family = AF_INET6;
            error = getaddrinfo(host, NULL, &ai_hints, &server_addr);
            if (error != 0) {
                    (void) fprintf(stderr,
                        "%s could not be located by getaddrinfo\n", host);
                    return (1);
            }
    	}
    	else
    	{
            ai_hints.ai_family = AF_INET;
            error = getaddrinfo(host, NULL, &ai_hints, &server_addr);
            if (error != 0) {
                    (void) fprintf(stderr,
                        "%s could not be located by getaddrinfo\n",host);
                    return (1);
            }
    	}
	if(ipv6_enable)
	{
		sain6 = (struct sockaddr_in6 *)(server_addr->ai_addr);
        	memcpy(buffer,&sain6->sin6_addr,16);
	}
	else
	{
		sain = (struct sockaddr_in *)(server_addr->ai_addr);
        	memcpy(buffer,&sain->sin_addr,4);
	}
        return (0);
}	
