/*                                                                           */
/* Copyright 2011                                                            */
/* Created: May 2011                                                         */
/* Author: Don Capps (capps@iozone.org)                                      */
/* Location: Iozone.org                                                      */
/*                                                                           */
/* --------------------------------------------------------------------------*/
/*                                                                           */
/*       Author: Don Capps (capps@iozone.org)                                */
/*               7417 Crenshaw                                               */
/*              Plano, TX 75025                                              */
/*                                                                           */
/*  Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011 Don Capps    */
/*                                                                           */
/*  License to freely use and distribute this software is hereby granted     */
/*  by the author, subject to the condition that this copyright notice       */
/*  remains intact.  The author retains the exclusive right to publish       */
/*  derivative works based on this work, including, but not limited to,      */
/*  revised versions of this work.                                           */
/*                                                                           */
/*  Universities, and research communities are hereby granted by the         */
/*  author, subject to the condition that this copyright notice remains      */
/*  intact, the right to use and distribute modified versions of this        */
/*  software providing that all modifications are clearly stated in the      */
/*  NOTICE file that will be provided with all derivative works.  Any        */
/*  derivative works must also remain available for free.  Attribution of    */
/*  the original authors and contributors must also remain intact.           */
/*                                                                           */
/*  Trademarks. This License does not grant permission to use the trade      */
/*  names, trademarks, service marks, or product names of the Licensor,      */
/*  except as required for reasonable and customary use in describing the    */
/*  origin of the Work and reproducing the content of the NOTICE file.       */
/*                                                                           */
/*  Disclaimer of Warranty. Unless required by applicable law or             */
/*  agreed to in writing, Licensor provides the Work (and each               */
/*  Contributor provides its Contributions) on an "AS IS" BASIS,             */
/*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or          */
/*  implied, including, without limitation, any warranties or conditions     */
/*  of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A            */
/*  PARTICULAR PURPOSE. You are solely responsible for determining the       */
/*  appropriateness of using or redistributing the Work and assume any       */
/*  risks associated with Your exercise of permissions under this License.   */
/*                                                                           */
/*  Limitation of Liability. In no event and under no legal theory,          */
/*  whether in tort (including negligence), contract, or otherwise,          */
/*  unless required by applicable law (such as deliberate and grossly        */
/*  negligent acts) or agreed to in writing, shall any Contributor be        */
/*  liable to You for damages, including any direct, indirect, special,      */
/*  incidental, or consequential damages of any character arising as a       */
/*  result of this License or out of the use or inability to use the         */
/*  Work (including but not limited to damages for loss of goodwill,         */
/*  work stoppage, computer failure or malfunction, or any and all           */
/*  other commercial damages or losses), even if such Contributor            */
/*  has been advised of the possibility of such damages.                     */
/*---------------------------------------------------------------------------*/
/*
 * ----------------------------------------------------------------------
 *
 * pdsm_remote 	( the remote nodes run this )
 *
 * Persistent shared memory with some globally distributed 
 * attributes.
 * Private segments are for children to send data to the master.
 * Shared segemt is for master to broadcast to children.
 * 
 *
 *   ---------------------    ---------------------  ->
 *   | Private seg       |    | Private seg       |  ->
 *   | Remote Owned (ro) |    | Master owned (mo) |  ->
 *   ---------------------    ---------------------  ->
 *   *This* remote writes     Master writes here
 *   here, and sends          and sends update to
 *   updates to master.       *this* remote.
 *
 *
 */
  
#ifdef WIN32
#pragma warning(disable: 4996)
#pragma warning(disable: 4101)
/*Windows specific headers*/
#include <Ws2tcpip.h>
#include <Windows.h>
#include <tchar.h>
#include <win32_getopt.h>
#include <win32_sub.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#ifndef WIN32
/* UNIX specific headers*/
#include <unistd.h>
#include <stdlib.h>
#if defined(_hpux_)
#  ifndef _CMPTID_T
#    define _CMPTID_T
     typedef int32_t   cid_t;           /* For compartment IDs */
     typedef cid_t     cmpt_t;
#  endif /* _CMPTID_T */
#else
#include <sys/ipc.h>
#endif
#include <sys/shm.h>
#include <strings.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#endif
#ifdef _bsd_
#include <sys/wait.h>
#endif

#ifdef WIN32
/*
 * Create file on C: drive
 */
#define PDSM_R_REGISTRY "c:\\tmp\\pdsm_registry_client"
#define ERRNO GetLastError()
#else
#define PDSM_R_REGISTRY "/tmp/pdsm_registry_client"
#endif
#define MAXREMOTES 255

/* 
 * Number of bytes to set aside for encode and decode of ulongs
 * to external format.
 */

#define MASTER_ALIST_PORT  22000
#define MASTER_ASEND_PORT  24000
#define REMOTE_ASEND_PORT  26000
#define REMOTE_ALIST_PORT  28000

#define THISVERSION "        $Revision: 793 $"
char version[]=THISVERSION;

int cdebug = 1;
#define UPDATE_TIME 1

/*
 * Format of segments:
 *
 * struct format {
 *    long Which_seg;
 *    char data[INBUFSZ]
 * }
 *
 */
#define INBUFSZ (4096)

/* Temporary buffers */
unsigned int *oldseg;
unsigned int *intbuf;
unsigned int *extbuf;
int ro_changed;
char master_hostname[255];
int remote_asend_socket_val[MAXREMOTES]; /* socket to master async listeners */
extern char *optarg;

#ifdef WIN32
/*thread handles - thread that handles start_r_listen_proc */
HANDLE hChildThread;
/*
 * Define types and structures for Windows platform
 */
struct timespec
{
   time_t  tv_sec;         /* seconds */
   long    tv_nsec;        /* nanoseconds */
};
typedef unsigned int key_t;

/*
 * Implement function calls as macros for Windows platform
 */
#define bzero(addr, size) RtlZeroMemory((VOID UNALIGNED*)(addr), (SIZE_T)(size))
#define close(socket) closesocket((SOCKET)(socket))
/*
 * Windows "Sleep()/SleepEx()" can only achieve 1 milisecond resolution, 
 * so the sleep time will be rounded up to milisecond, 
 * and it always returns 0 
 */
int nanosleep(const struct timespec *req, struct timespec *rem);
/* functions that have different signatures than their UNIX version*/
DWORD WINAPI start_r_listen_proc(__in  LPVOID lpParameter);
void remote_listen_async(int, int, char*);
#else
/*To compatible with Windows SOCKET type*/
typedef int SOCKET;
/*Function prototype on UNIX*/
void start_r_listen_proc(int who);
void remote_listen_async(int, int);
#endif

int ipv6_enable;
/* 
 * Prototypes 
 */
void nap(int);
char * get_sh_mem(key_t, size_t, int);
void remote_parent(void);
void remote_child(void);
void remote_send(int, char *, char *, int, int);
unsigned int * remote_ipaddr(char *);
unsigned int * my_ipaddr(void);
void ulongs_to_nlongs(size_t, unsigned int *, unsigned int *, int);
int nlongs_to_ulongs(size_t, unsigned int *, unsigned int *);
int ro_change_check(void);
void usage(void);
int start_remote_send_async(char *master_host_name, int remote_port, unsigned int* my_s_addr);
int start_remote_listen_async(int size_of_message);
int remote_attach(int s);
int get_host_address(char *, unsigned int *);
/*
 * Allocate shared pools of memory 
 */

/* The pools of memory */
int size_of_private_segs = INBUFSZ; /* Default INBUFSZ of 8k */
int number_of_private_segs = 1;
int size_of_extbuf;

char *private_segs_mo_pointer;  /* Master's mo segments */
char *private_segs_ro_pointer;  /* Remote's ro segments */

int shmid;
/*
 *     REMOTE KEYS... ONE PER NODE 
 *     REMOTE KEYS... ONE PER NODE 
 *     REMOTE KEYS... ONE PER NODE 
*/
key_t mkey = 0x2718281; /* Key for all inbound Private segs  */
key_t rkey = 0x2718282; /* Key for all outbound private segs */
size_t total_size;

/* Temp variables */
int cret;
int remove_flag;
int master_flag;
unsigned long chid;
int child_pid;
int hflag, nflag;
FILE *fd;
char my_reg_name[256];

int
main(int argc, char **argv)
{

	int page_size;

#ifdef WIN32	
	WSADATA wsaData;
	if (0 != WSAStartup( MAKEWORD(2,2), &wsaData ))
	{
		/* If Windows Sockets does not initialize correctly
		 * we have trouble.
		 */
		printf("WSAStartup failed");
		exit(1);
	}
#endif

	if(argc == 1)
	{
		usage();
		exit(0);
	};

	page_size = getpagesize();
	while((cret = getopt(argc,argv,"vhrn:H:s:y:")) != EOF){
                switch(cret){
                case 'r':
			remove_flag = 1;
			break;
                case 'n':
			nflag=1;
			chid = (unsigned long)atoi(optarg);
			if((long)chid  < 0)
				chid = 0;
			break;
                case 'H':
			hflag=1;
			strcpy(master_hostname,optarg);
			break;
                case 's':
			size_of_private_segs = atoi(optarg);
			/*
			 * Round up to page size 
			 */
        		size_of_private_segs = ((size_of_private_segs + 
				(page_size-1)) & ~(page_size-1));
			break;
                case 'h':
			usage();
			exit(0);
			break;
                case 'y':
			ipv6_enable=atoi(optarg);
			break;
                case 'v':
			printf("pdsm_remote: Version %s\n",version);
			exit(0);
			break;
		};
	}	
	if(!hflag || !nflag )
	{
		usage();
		exit(0);
	}

	intbuf=(unsigned int *)malloc(size_of_private_segs);
	bzero((char *)intbuf,size_of_private_segs);

	size_of_extbuf = sizeof(int) + size_of_private_segs;

	extbuf = (unsigned int *)malloc(size_of_extbuf);
	bzero((char *)extbuf,size_of_extbuf);

	oldseg=(unsigned int *)malloc(size_of_private_segs);
	bzero((char *)oldseg,size_of_private_segs);


	if((number_of_private_segs != 0) && (size_of_private_segs != 0))
	{
	   total_size = (number_of_private_segs * size_of_private_segs);	
	   /* Remember... sizes are actually rounded up to page size */
	   private_segs_mo_pointer = get_sh_mem(mkey+(2*chid), total_size,
			remove_flag);
	   if(private_segs_mo_pointer<=0)
	   {
		printf("Unable to attach to MO segment \n");
		exit(1);	
	   }
	   if(cdebug>=1 && !remove_flag)
	   {
	      printf("Remote allocated/attached %d bytes of MO pdsm memory. Key is 0x%lx\n",
		(int)total_size, (unsigned long)(mkey+(2*chid)));
	   }

	   private_segs_ro_pointer = get_sh_mem(rkey+(2*chid), total_size,
		remove_flag);
	   if(private_segs_ro_pointer<=0)
	   {
		printf("Unable to attach to RO segment \n");
		exit(1);	
	   }
	   if(cdebug>=1 && !remove_flag)
	   {
	     printf("Remote allocated/attached %d bytes of RO pdsm memory. Key is 0x%lx\n",
		(int)total_size,(unsigned long)(rkey+(2*chid)));
	   }
	   sprintf(my_reg_name,"%s_%u",PDSM_R_REGISTRY,(unsigned int)chid);
	   fd = fopen(my_reg_name,"w");
	   if(fd < 0)
	   {
		printf("Unable to create %s.\n",my_reg_name);
		exit(1);	
	   }
	   fprintf(fd,"PDSM_MO_key 0x%lx PDSM_MO_size %d \nPDSM_RO_key 0x%lx PDSM_RO_size %d\n",
		(unsigned long)(mkey+(2*chid)),(int)total_size, (unsigned long)(rkey+(2*chid)),(int)total_size);
	   fclose(fd);
	   
	}
	if(remove_flag)
	{
	     printf("Remote released %d bytes of MO pdsm memory. Key is 0x%lx\n",
		(int)total_size, (unsigned long)(mkey+(2*chid)));
	     printf("Remote released %d bytes of RO pdsm memory. Key is 0x%lx\n",
		(int)total_size,(unsigned long)(rkey+(2*chid)));
	     unlink(my_reg_name);
	     exit(0);
	}
#ifdef WIN32
	/*start child thread*/
	remote_child();
	/*main thread continue to run parent loop*/
	remote_parent();
	WSACleanup();
#else
	child_pid=fork();
	if(child_pid==0)
		remote_child();
	else
	{
		sleep(1);
		remote_parent();
	    	wait(0);
	}
#endif
}

void
remote_parent(void)
{
	/*
	 * Parent loops, and sends messages to master when things 
	 * have changed in the remote private seg. 
	*/
	int m_changed = 0;
	//int s;
	int sock,s,i;
	char *myseg ;
	int dest_who,once;
	unsigned int *p2;
	unsigned int * my_buf_addr;

	myseg = (char *)(private_segs_ro_pointer);

	if(cdebug>=2)
	{
		printf("Remote parent starting\n");
	}

	my_buf_addr = remote_ipaddr(master_hostname);
	s=start_remote_send_async(master_hostname, MASTER_ALIST_PORT+chid,my_buf_addr);
	remote_asend_socket_val[chid]=s;

	once=1; /* Do once for sure */
	bzero((char *)oldseg,size_of_private_segs);
	while(1)
	{
		ro_changed = ro_change_check();
		if(ro_changed || once)
		{
			dest_who=chid;
			remote_send(remote_asend_socket_val[chid],
				master_hostname,myseg,size_of_private_segs,
				dest_who);
			ro_changed = 0;
			once=0;
		}
		sleep(UPDATE_TIME);
	}
}

void
remote_child(void)
{
	/*
	The remote's child becomes listener for master's messages to 
	update the mo private segs.
	*/
	int i;
	if(cdebug>=2)
	{
		printf("Remote children starting\n");
	}
#ifdef WIN32
	/*We create threads to process the incoming message from remote clients*/
	hChildThread = CreateThread( 
								NULL,                   // default security attributes
								0,                      // use default stack size  
								start_r_listen_proc,       // thread function name
								(LPVOID)NULL,          // argument to thread function 
								0,                      // use default creation flags 
								NULL);   // returns the thread identifier 

#else
	start_r_listen_proc(chid);
#endif
}

#ifdef WIN32
DWORD WINAPI start_r_listen_proc(__in  LPVOID lpParameter)
#else
void start_r_listen_proc(int who)
#endif
{
	int sock,p;
	char *which_mo_pointer;
	int dest_who;

#ifdef WIN32
	char* extbuf_local = (char *)malloc(size_of_private_segs);
	char* intbuf_local = (char *)malloc(size_of_extbuf);
	//int who = (int)lpParameter;
	//struct sockaddr_in async_sock; //use local variable for thread safety
	sock=start_remote_listen_async(size_of_extbuf);
#else
	sock=start_remote_listen_async(size_of_extbuf);
#endif

	sock=remote_attach(sock);
	while(1)
	{
#ifdef WIN32
		remote_listen_async(sock,size_of_extbuf, extbuf_local);
		dest_who = nlongs_to_ulongs(size_of_private_segs,
			(unsigned int *)extbuf_local,(unsigned int *)intbuf_local);
#else
		remote_listen_async(sock,size_of_extbuf);
		dest_who = nlongs_to_ulongs(size_of_private_segs,
			(unsigned int *)extbuf,(unsigned int *)intbuf);
#endif
		/* Find this child's mo (master owned) segment */
		if(cdebug>=1)
		  printf("Remote[%d] got update for mo segment #%d\n",(int)chid,(int)dest_who);

		which_mo_pointer = (char *)private_segs_mo_pointer;
/*
		which_mo_pointer = (char *)private_segs_mo_pointer + 
			(dest_who*size_of_private_segs);
*/
#ifdef WIN32
		memcpy(which_mo_pointer,intbuf_local,size_of_private_segs);
#else
		memcpy(which_mo_pointer,intbuf,size_of_private_segs);
#endif
	}
#ifdef WIN32
	free(extbuf_local);
	free(intbuf_local);
	return 0;
#endif
}

/* 
 * Messages from master to children through shared segment.
 *
 * This is like an xdr layer, but chunky style and portable.
 */

void
ulongs_to_nlongs(size_t size, unsigned int *inbuf, unsigned int *outbuf,
		int dest_who)
{
	int i;
	unsigned int *outp;
	/* 
	 * First int in the external buffer is the WHO 
	 */
	outp=(unsigned int *)outbuf;
	*outp = htonl(dest_who);
	outp++;
	for(i = 0; i < (int)(size/sizeof(int)); i++)
	{
	   *outp = (unsigned int )htonl(inbuf[i]);
	   outp++;
	}
}

int
nlongs_to_ulongs(size_t size, unsigned int *inbuf, unsigned int *outbuf)
{
	int i;
	int dest_who;
	unsigned int *inp;
	/* 
	 * First int in the external buffer is the WHO 
	 */
	inp=(unsigned int *)inbuf;	
	dest_who = ntohl(*inp);
	inp++;
	for(i = 0; i < (int)(size/sizeof(int)); i++)
	{
	   outbuf[i]=ntohl(*inp);
	   inp++;
	}
	return(dest_who);
}

/*
 * Creates shared memory, or if it exists... just attaches and bumps
 * the ref count. 
 * 
 * @param key Used to identify the unique public shared memory seg.
 *
 * @param Size needs to be big enough to handle all results 
 *
 * This mechanism creates, if needed, and then bumps a ref attach 
 * count and then attempts to remove the shared memory segement. 
 * The "flag" will permit an automatic delete of the shared 
 * memory segment whenever the process exits. (normal, or abnormal exit)
 */

#ifdef WIN32

#define MAX_SHARE_MEMORY 32 //max number of shared memory one process can request, regardless of their size
typedef struct _SHARED_MEMORY
{
	HANDLE hFile; //associated file handle
	HANDLE hMapFile; //associated mapping object handle
	key_t  key; //shared memory key
	BOOL   valid;
} SHARED_MEMORY;

static SHARED_MEMORY sharedMem[MAX_SHARE_MEMORY] = {0};
/*
 *	Win32 implementation using memory mapped file
 *  key - share memory id, it is used to generate mapping object
 *  size - the size of shared memory requested
 *  flag - delete shared memory if its value is non-zero
 */
char *
get_sh_mem(key_t key, size_t size, int flag)
{
	TCHAR szObjName[255];
	TCHAR szFileName[255];
	char* pMem = NULL;
	int position=-1, i=0; //index in shared memory array where we can save shared_memory information 
	HANDLE hFile;

	_stprintf(szObjName, TEXT("Global\\pdsm-%lx"), (unsigned long)key);
	_stprintf(szFileName, TEXT("SharedMemory.%lx"), (unsigned long)key);
	
	/*First we search the shared memory array, try to get existing one*/
	for (i=0; i<MAX_SHARE_MEMORY; i++)
	{
		if (sharedMem[i].valid && sharedMem[i].key == key)
		{
			position = i;
			break;
		}
		else if (position < 0  && !sharedMem[i].valid)
			position = i; //find a unused slot
	}
	/* if we reach here, we did not find exising shared memory*/
	if (flag && sharedMem[position].valid)
	{
		/*do we need to delete memory mapped file here???*/
		hFile = CreateFile(szFileName,
							GENERIC_WRITE | GENERIC_READ,
							FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,
							NULL,
							OPEN_EXISTING,	 
							FILE_FLAG_DELETE_ON_CLOSE, //delete file on close
							NULL);
		if (hFile != INVALID_HANDLE_VALUE)
			CloseHandle(hFile);
	}

	if (position < 0)
		/*the process has already allocated max-allowed shared memory, we have no choice but fail the call*/
		return NULL; 
	/* Now we start to create new shared memory*/
	if (!sharedMem[position].valid)
	{
		sharedMem[position].hMapFile = NULL;
		/* We open or create a file to map to memory*/
		if ((sharedMem[position].hFile = CreateFile(szFileName,
													GENERIC_WRITE | GENERIC_READ,
													FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,
													NULL,
													OPEN_ALWAYS,	 // create a new file if it does not exist, otherwise open the old file
													(flag?FILE_FLAG_DELETE_ON_CLOSE:FILE_ATTRIBUTE_NORMAL), 
													NULL)) == (HANDLE)INVALID_HANDLE_VALUE)
		{
			if(cdebug >= 2)
				printf("Unable to open or create file for memory mapping:\"SharedMemory.%lx\"\n",(unsigned long)key);
		} 
		else
		{
			/*Create mapping object*/
			sharedMem[position].hMapFile = CreateFileMapping(
															sharedMem[position].hFile,    // use paging file
															NULL,                    // default security
															PAGE_READWRITE,          // read/write access
															0,                       // maximum object size (high-order DWORD)
															(DWORD)size,                // maximum object size (low-order DWORD)
															szObjName);                 // name of mapping object
			if (sharedMem[position].hMapFile == NULL)
			{
				if (cdebug >= 2)
					printf("Unable to create mapping object: \"Global\\pdsm-%lx\", error code: %d\n",(unsigned long)key, GetLastError());
				CloseHandle(sharedMem[position].hFile);
			}
			else 
			{
				if (cdebug >= 2 && GetLastError() == ERROR_ALREADY_EXISTS)
					printf("CreateFileMapping returned a handle of existing mapping object\n");
				sharedMem[position].key = key;
				sharedMem[position].valid = TRUE;
			}
		}
	}

	if (sharedMem[position].hMapFile != NULL)
	{
		//
		//now we need to mapped the view. 
		//
		pMem = (char*) MapViewOfFile(sharedMem[position].hMapFile,   // handle to map object
										FILE_MAP_ALL_ACCESS, // read/write permission
										0,
										0,
										(DWORD)size);
		if (pMem == NULL && cdebug >= 2)
			printf("Unable to map file to memory, error code: %d\n", GetLastError());
	}
	return pMem;
}

/*The sleep cannot be interrpted, so field in "rem" will be always 0*/
int nanosleep(const struct timespec *req, struct timespec *rem)
{
	unsigned int milisec = 0;

	if (!req || req->tv_nsec == 0 && req->tv_sec == 0)
		return -1;
	milisec = (unsigned int)req->tv_sec*1000; //convert to milisecond
	milisec += (req->tv_nsec/1000000); //convert to milisecond
	if (req->tv_nsec % 1000000)
		milisec++; //round the remainder up.
	Sleep(milisec);
	
	if (rem)
	{
		rem->tv_nsec = 0;
		rem->tv_sec = 0;
	}
	return 0;
}


#else
char *
get_sh_mem(key_t key, size_t size, int flag)
{
	struct stats *statsp;

	if(cdebug >= 2)
	  printf("Trying to get shared memory segment with key %lx\n",(unsigned long)key);

        shmid=shmget((key_t)key,(size_t)size,(int)(0666)|IPC_CREAT);
        if(shmid < 0)
        {
                printf("Shmid %d Errno %d\n",shmid,errno);
        	return((char *)0);
        } 

	if(cdebug >= 2)
	  printf("Got shared memory with key %lx\n",(unsigned long)key);

        statsp=(struct stats *)shmat(shmid,0,0);
        if(statsp == (struct stats *)-1)
        {
                printf("Attach to shared memory failed\n");
        	return((char *)0);
        }
	/* Attach ref count is now all that is holding on to this segment */
	if(flag)
		shmctl(shmid, IPC_RMID, 0);  /* cleanup */

	if(cdebug >= 2)
	  printf("Shared memory key %lx Size = %d\n",(unsigned long)key, (int)size);
	/*
	 * Don't need to call quit, as any termination will cleanup
 	 * automatically.
	 */

        /*quit();*/
        return((char *)statsp);
}
#endif

int remote_async_port; /* Virtualized due to fork */
int child_async_port; /* Virtualized due to fork */
struct sockaddr_in child_async_sock, remote_async_sock;
struct sockaddr_in6 child_async_sock6, remote_async_sock6;

/*
 * Start the remote async listening service for messages from the master.
 */
#ifdef HAVE_ANSIC_C
int
start_remote_listen_async(int size_of_message)
#else
int
start_remote_listen_async(size_of_message)
int size_of_message;
#endif
{
	int tsize;
	SOCKET s;
	int rc;
	int xx;
	int tmp_port;
	int sockerr;
	int recv_buf_size=size_of_message;
	int optval=1;
	xx = 0;
	tsize=size_of_message; /* Number of messages to receive */
	if(ipv6_enable)
           s = socket(AF_INET6, SOCK_STREAM, 0);
	else
           s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0)
    {
            perror("socket failed:");
            exit(19);
    }
	sockerr = setsockopt (s, SOL_SOCKET, SO_RCVBUF, (char *)
		&recv_buf_size, sizeof(int));
	if ( sockerr == -1 ) {
		perror("Error in setsockopt 5\n");
	}
	sockerr = setsockopt (s, SOL_SOCKET, SO_REUSEADDR, (char *)
		&optval, sizeof(int));
	if ( sockerr == -1 ) {
		perror("Error in setsockopt 6\n");
	}
	if(ipv6_enable)
           bzero(&remote_async_sock6, sizeof(struct sockaddr_in6));
	else
           bzero(&remote_async_sock, sizeof(struct sockaddr_in));

	tmp_port=REMOTE_ALIST_PORT+chid;

	if(ipv6_enable)
	{
           remote_async_sock6.sin6_port = htons(tmp_port);
           remote_async_sock6.sin6_family = AF_INET6;
	   memset(&remote_async_sock6.sin6_addr.s6_addr[0],0,16);
	}
	else
	{
           remote_async_sock.sin_port = htons(tmp_port);
           remote_async_sock.sin_family = AF_INET;
           remote_async_sock.sin_addr.s_addr = INADDR_ANY;
	}
    rc = -1;
    while (rc < 0)
    {
		if(ipv6_enable)
                   rc = bind(s, (struct sockaddr *)&remote_async_sock6,
                                sizeof(struct sockaddr_in6));
		else
                   rc = bind(s, (struct sockaddr *)&remote_async_sock,
                                sizeof(struct sockaddr_in));
		if(rc < 0)
		{
			tmp_port++;
			if(ipv6_enable)
                	  remote_async_sock6.sin6_port=htons(tmp_port);
			else
                	  remote_async_sock.sin_port=htons(tmp_port);
			continue;
		}
    }
	if(ipv6_enable)
	  remote_async_port = ntohs(remote_async_sock6.sin6_port);
	else
	  remote_async_port = ntohs(remote_async_sock.sin_port);
	if(cdebug >=2)
	{
		printf("Remote %d: Async Listen: Bound at port %d\n",
			(int)chid,(int)tmp_port);
	}
	if(rc < 0)
	{
		printf("bind failed. Errno %d \n",errno);
		exit(20);
	}
	if(cdebug >=2)
	   printf("Remote async listen socket %d \n",s);
	return((int)s);
}

/*
 * Start the channel for the remote to send a message to 
 * the master.
 */
#ifdef HAVE_ANSIC_C
int
start_remote_send_async(char *master_host_name, int remote_port, 
		unsigned int * my_buf_addr)
#else
int
start_remote_send_async(child_host_name, remote_port, my_buf_addr)
char *child_host_name; 
int remote_port;
unsigned int * my_buf_addr;
#endif
{
	int rc;
	SOCKET remote_socket_val;
	struct sockaddr_in addr,raddr;
	struct sockaddr_in6 addr6,raddr6;
	int port,tmp_port;
	int ecount = 0;
/*
	struct timespec req,rem;
	req.tv_sec = 0;
	req.tv_nsec = 10000000;
	rem.tv_sec = 0;
	rem.tv_nsec = 10000000;
*/


	port=remote_port;
	nap(1);
/*
	nanosleep(&req,&rem);
*/

over:
	if(ipv6_enable)
	{
           raddr6.sin6_family = AF_INET6;
           raddr6.sin6_port = htons(port);
	   memcpy(&raddr6.sin6_addr.s6_addr[0],my_buf_addr,16);
           remote_socket_val = socket(AF_INET6, SOCK_STREAM, 0);
	}
	else
	{
           raddr.sin_family = AF_INET;
           raddr.sin_port = htons(port);
	   memcpy(&raddr.sin_addr.s_addr,my_buf_addr,4);
           remote_socket_val = socket(AF_INET, SOCK_STREAM, 0);
	}
        if (remote_socket_val < 0)
        {
                perror("Remote: async socket failed:");
                exit(23);
        }
	tmp_port=REMOTE_ASEND_PORT;
	if(ipv6_enable)
	{
           bzero(&addr6, sizeof(struct sockaddr_in6));
           addr6.sin6_port = htons(tmp_port);
           addr6.sin6_family = AF_INET6;
	   memset(&addr6.sin6_addr.s6_addr[0],0,16);
	}
	else
	{
           bzero(&addr, sizeof(struct sockaddr_in));
           addr.sin_port = htons(tmp_port);
           addr.sin_family = AF_INET;
           addr.sin_addr.s_addr = INADDR_ANY;
	}
        rc = -1;
        while (rc < 0)
        {
		if(ipv6_enable)
                   rc = bind(remote_socket_val, (struct sockaddr *)&addr6,
                                                sizeof(struct sockaddr_in6));
		else
                   rc = bind(remote_socket_val, (struct sockaddr *)&addr,
                                                sizeof(struct sockaddr_in));
		if(rc < 0)
		{
			tmp_port++;
			if(ipv6_enable)
                	   addr6.sin6_port=htons(tmp_port);
			else
                	   addr.sin_port=htons(tmp_port);
			continue;
		}
        }
	if(cdebug >= 2)
	{
		printf("Remote: Bound async port\n");
	}
        if (rc < 0)
        {
                perror("Remote: bind async failed\n");
                exit(24);
        }
again:

	if(ipv6_enable)
           rc = connect(remote_socket_val, (struct sockaddr *)&raddr6, 
			sizeof(struct sockaddr_in6));
	else
           rc = connect(remote_socket_val, (struct sockaddr *)&raddr, 
			sizeof(struct sockaddr_in));
	if (rc < 0)
        {
		if(ecount++ < 300)
		{
		/* Really need this sleep for Windows */
			sleep(1);
			/*nanosleep(&req,&rem);*/
			goto again;
		}
                perror("Remote: async connect failed\n");
                close(remote_socket_val);

		sleep(1);
		/*nanosleep(&req,&rem);*/
                ecount=0;
                goto over;
        }
	sleep(1);
	if(cdebug >= 2)
	{
		printf("Remote async Connected\n");
	}
	return ((int)remote_socket_val);
}


/*
 * Remote sending message to the master
 * There should be a unique remote_socket_val for each
 * master listener.
 */
void
remote_send(int remote_socket_val, char *host_name, char *inbuf, int send_size,
	int dest_who)
{
	int rc;
	unsigned int *myseg;

	bzero((char *)extbuf,size_of_extbuf);
	/* 
	 * Convert internal commands to string format for neutral 
	 * format/portability
	 */
	myseg = (unsigned int *)inbuf;

	if(cdebug>=2)
		printf("Child sending to %d\n",dest_who);

	ulongs_to_nlongs((size_t)send_size, (unsigned int *)inbuf, 
		(unsigned int *)extbuf,dest_who);

	if(cdebug >= 1)
		printf("Remote[%d] sending update to master \"%s\" \n",
		   (int)chid, host_name);
        rc = write(remote_socket_val, (char *)extbuf, size_of_extbuf);
        if (rc < 0)
        {
                perror("write failed\n");
                exit(26);
        }
}

unsigned int buf_addr[16];

unsigned int * 
remote_ipaddr(char *hostname)
{

/*
	struct in_addr *ip;
	struct hostent *he;

        he = gethostbyname(hostname);
        if (he == NULL)
        {
                printf("Bad hostname 1 >%s<\n",hostname);
		fflush(stdout);
                exit(22);
        }
        ip = (struct in_addr *)he->h_addr_list[0];
	return(*ip);
*/
	get_host_address(hostname,buf_addr);
	return(buf_addr);
}

unsigned int * 
my_ipaddr(void)
{

/*
	struct in_addr *ip;
	struct hostent *he;
*/
	char myhostname[200];

	gethostname(myhostname,200);
	get_host_address(myhostname,buf_addr);
	return(buf_addr);
/*
        he = gethostbyname(myhostname);
        if (he == NULL)
        {
                printf("Bad hostname 2 >%s<\n",myhostname);
		fflush(stdout);
                exit(22);
        }
        ip = (struct in_addr *)he->h_addr_list[0];
	return(*ip);
*/
}

/*
 * Remote async listener
 */
#ifdef WIN32
void
remote_listen_async(int sock, int size_of_message, char* cnc)
#else
#ifdef HAVE_ANSIC_C
void
remote_listen_async(int sock, int size_of_message)
#else
void
remote_listen_async(sock, size_of_message)
int sock, size_of_message;
#endif
#endif
{
	int tsize;
	int rcvd;
	int s;
	int rc;
#ifndef WIN32
	char *cnc;
	cnc = (char *)extbuf;
#endif

	s = sock;
	tsize=size_of_message; /* Number of messages to receive */
	if(cdebug>=2)
	   printf("Remote child %d Sock %d Entering async recieve %d\n",
				(int)chid,(int)sock,(int)tsize);
	rcvd = 0;
	while(rcvd < tsize)
	{
		if(cdebug >= 2)
		{
			printf("Remote %d Sock %d In async recieve %d\n",
				(int)chid,(int)sock,(int)tsize);
		}
		rc=read(s,cnc,size_of_message);
		if(rc < 0)
		{
			printf("Remote failed. Errno %d \n",errno);
#ifdef WIN32
			if (WSAECONNRESET == ERRNO || 
				WSAENETDOWN == ERRNO ||
				WSAECONNABORTED == ERROR ||
				WSAENETRESET == ERROR)
			{
				/*connection is closed by remote end or network failure, so we terminate the programe here*/
				printf("Connection is closed by master or network error is encountered, program terminated.");
				exit(21);
			}
#endif
			/*exit(21);*/
		}
		/*sleep(1); XXX */
		if(rc > 0)
		{
			if(cdebug >= 2 )
			   printf("Remote read %d bytes \n",rc);
		}

		/* Special case. If remote gets final results, it can 
                   exit, and close the connection to the async child
                   too quickly. When this happens the child gets a 
		   read() that returns 0. It just needs to exit here.
		*/
		if(rc==0)
		{
			printf("Remote socket closed. Exiting... \n");
			exit(0);
		}
		if(cdebug >= 2)
		{
			printf("Remote %d: Got %d bytes (async) \n",(int)chid,
				(int)rc);
		}
		rcvd+=rc;
		cnc+=rc;
	}
	if(cdebug >= 2)
	{
		printf("Remote %d: return from async listen\n",(int)chid);
	}
}
#ifdef HAVE_ANSIC_C
int
remote_attach(int s)
#else
int
remote_attach(s)
int s;
#endif
{
	unsigned int me;
	unsigned int me6;
	SOCKET ns;
	struct sockaddr_in6 *addr6;
	struct sockaddr_in *addr;

	if(ipv6_enable)
	  addr6=&remote_async_sock6;
	else
	  addr=&remote_async_sock;
	if(cdebug >= 2)
	{
		printf("Remote %d attach async. Sock %d\n",(int)chid,(int)s);
	}
	if(ipv6_enable)
	  me6=sizeof(struct sockaddr_in6);
	else
	  me=sizeof(struct sockaddr_in);
	
	if(cdebug >= 2)
	{
		printf("Remote %d enters listen\n",(int)chid);
	}
	listen(s,10);
	if(cdebug >= 2)
	{
		printf("Remote %d enters accept\n",(int)chid);
	}
	if(ipv6_enable)
	  ns=accept(s,(void *)addr6,(socklen_t *)&me6);
	else
	  ns=accept(s,(void *)addr,(socklen_t *)&me);
	if(cdebug >= 2)
	{
		printf("Remote %d attached for receive. Sock %d\n",
		    (int)chid, (int)ns);
	}
	if(ns<=0)
		printf("Remote %d attached for receive. Sock %d errno %d\n",
		    (int)chid, (int)ns,(int)errno);
	return((int)ns);
}

int
ro_change_check(void)
{
        int j;
        int flag;
        char *myseg;
	unsigned int *p1, *p2;

    flag = 0;
	if(cdebug>=2)
      printf("Sync time for remote segments \n");
	p1 = (unsigned int *)(private_segs_ro_pointer );
	p2 = (unsigned int *)oldseg;
    for(j=0;j< (int)(size_of_private_segs/sizeof(int));j++)
    {
            if(*p1 != *p2)
            {
                *p2=*p1;
                flag=1;
            }
            *p2=*p1;
            p2++;
	p1++;
    }
    if(cdebug >= 2 && flag)
	printf("Sync needed.\n");
    return(flag);
}

void usage(void)
{
	printf("pdsm_remote: [-h] [-H Master_hostname] [-n client_id] [-r] [-v]\n");
	printf("\t[-h].................... Help screen\n");
	printf("\t[-H Master_hostname].... Hostname of the master.\n");
	printf("\t[-n client_id].......... Unique client identifier.\n");
	printf("\t[-s].................... Size of shared memory in bytes.\n");
	printf("\t[-r].................... Release shared memory.\n");
	printf("\t[-y 6].................. Use IPv6.\n");
	printf("\t[-v].................... Print version number and exit.\n");
	printf("\n");
}

void
nap( int select_msecs )
{
#ifndef WIN32
	struct timeval	sleeptime;
	sleeptime.tv_sec = select_msecs / 1000;
	sleeptime.tv_usec = (select_msecs % 1000) * 1000;
	select(0, NULL, NULL , NULL, &sleeptime);
#else
	Sleep(select_msecs);
#endif
}

int
get_host_address(char *host, unsigned int *buffer)
{
	struct addrinfo     *server_addr;
	struct addrinfo     ai_hints;
	struct sockaddr_in *sain;
	struct sockaddr_in6 *sain6;
	int error;

    	memset((char *)&ai_hints, '\0', sizeof (ai_hints));
    	/* Get host's address */
    	if (ipv6_enable) {
            ai_hints.ai_family = AF_INET6;
            error = getaddrinfo(host, NULL, &ai_hints, &server_addr);
            if (error != 0) {
                    (void) fprintf(stderr,
                        "%s could not be located by getaddrinfo\n", host);
                    return (1);
            }
    	}
    	else
    	{
            ai_hints.ai_family = AF_INET;
            error = getaddrinfo(host, NULL, &ai_hints, &server_addr);
            if (error != 0) {
                    (void) fprintf(stderr,
                        "%s could not be located by getaddrinfo\n",host);
                    return (1);
            }
    	}
	if(ipv6_enable)
	{
		sain6 = (struct sockaddr_in6 *)(server_addr->ai_addr);
        	memcpy(buffer,&sain6->sin6_addr,16);
	}
	else
	{
		sain = (struct sockaddr_in *)(server_addr->ai_addr);
        	memcpy(buffer,&sain->sin_addr,4);
	}
        return (0);
}	
