/*                                                                           */
/* Copyright 2011                                                            */
/* Created: May 2011                                                         */
/* Author: Don Capps (capps@iozone.org)                                      */
/* Location: Iozone.org                                                      */
/*                                                                           */
/* --------------------------------------------------------------------------*/
/*                                                                           */
/*       Author: Don Capps (capps@iozone.org)                                */
/*               7417 Crenshaw                                               */
/*              Plano, TX 75025                                              */
/*                                                                           */
/*  Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011 Don Capps    */
/*                                                                           */
/*  License to freely use and distribute this software is hereby granted     */
/*  by the author, subject to the condition that this copyright notice       */
/*  remains intact.  The author retains the exclusive right to publish       */
/*  derivative works based on this work, including, but not limited to,      */
/*  revised versions of this work.                                           */
/*                                                                           */
/*  Universities, and research communities are hereby granted by the         */
/*  author, subject to the condition that this copyright notice remains      */
/*  intact, the right to use and distribute modified versions of this        */
/*  software providing that all modifications are clearly stated in the      */
/*  NOTICE file that will be provided with all derivative works.  Any        */
/*  derivative works must also remain available for free.  Attribution of    */
/*  the original authors and contributors must also remain intact.           */
/*                                                                           */
/*  Trademarks. This License does not grant permission to use the trade      */
/*  names, trademarks, service marks, or product names of the Licensor,      */
/*  except as required for reasonable and customary use in describing the    */
/*  origin of the Work and reproducing the content of the NOTICE file.       */
/*                                                                           */
/*  Disclaimer of Warranty. Unless required by applicable law or             */
/*  agreed to in writing, Licensor provides the Work (and each               */
/*  Contributor provides its Contributions) on an "AS IS" BASIS,             */
/*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or          */
/*  implied, including, without limitation, any warranties or conditions     */
/*  of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A            */
/*  PARTICULAR PURPOSE. You are solely responsible for determining the       */
/*  appropriateness of using or redistributing the Work and assume any       */
/*  risks associated with Your exercise of permissions under this License.   */
/*                                                                           */
/*  Limitation of Liability. In no event and under no legal theory,          */
/*  whether in tort (including negligence), contract, or otherwise,          */
/*  unless required by applicable law (such as deliberate and grossly        */
/*  negligent acts) or agreed to in writing, shall any Contributor be        */
/*  liable to You for damages, including any direct, indirect, special,      */
/*  incidental, or consequential damages of any character arising as a       */
/*  result of this License or out of the use or inability to use the         */
/*  Work (including but not limited to damages for loss of goodwill,         */
/*  work stoppage, computer failure or malfunction, or any and all           */
/*  other commercial damages or losses), even if such Contributor            */
/*  has been advised of the possibility of such damages.                     */
/*---------------------------------------------------------------------------*/
/*
 * ----------------------------------------------------------------------
 * tickle:
 *
 * This is a tickler that can be used to set values in the
 * master or remote shared memory segments. One can
 * then monitor the coherency traffic and make sure that
 * it is working correctly. It's only use is for debug and test.
 *
 */

#ifdef WIN32
#pragma warning(disable: 4996)
/*Windows specific headers*/
#include <ws2tcpip.h>
#include <Windows.h>
#include <tchar.h>
#include <time.h>
#include <win32_getopt.h>
#include <win32_sub.h>
#endif

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#ifndef WIN32
/* UNIX specific headers*/
#include <unistd.h>
#if defined(_hpux_)
#  ifndef _CMPTID_T
#    define _CMPTID_T
     typedef int32_t   cid_t;           /* For compartment IDs */
     typedef cid_t     cmpt_t;
#  endif /* _CMPTID_T */
#else
#include <sys/ipc.h>
#endif
#include <sys/shm.h>
#include <strings.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#define GEN_SZ (4096)
#define INBUFSZ (GEN_SZ)
#define UPDATE_TIME 2

extern char *optarg;
int number_of_private_segs = 1;
int size_of_private_segs = INBUFSZ;
char *private_segs_mo_pointer;
char *private_segs_ro_pointer;
int remove_flag = 0;
int total_size,single_size;
int chid = -1;
int cret;
int remote_only;

#ifdef WIN32
typedef unsigned int key_t;
#endif

char * get_sh_mem(key_t, size_t, int );
void usage(void);

int cdebug = 1;

int shmid;
key_t mkey = 0x3141592; /* Key for all mo segments on the master */
key_t rkey = 0x2718282; /* Key for the outbound private seg on the remote */

int
main(int argc, char ** argv)
{
	
	int j,page_size;
	char *myseg;

	if(argc == 1)
	{
		usage();
		exit(0);
	};
	page_size = getpagesize();
	while((cret = getopt(argc,argv,"hcrRp:s:")) != EOF){
                switch(cret){
                case 'h':                 /* Help screen */
			usage();
			exit(0);
			break;
                case 'r':		  /* Remove segments */
			remove_flag = 1;
			break;
                case 'R':		  /* Remote segments only */
			remote_only = 1;
			break;
                case 'c':                 /* Child to tickle */
			chid = 0;
			break;
                case 'p':                 /* Total count of remotes */
			number_of_private_segs = atoi(optarg);
			break;
                case 's':                 /* Total count of remotes */
			size_of_private_segs = atoi(optarg);
			/*
			 * Round up to page size.
			 */
        		size_of_private_segs= ((size_of_private_segs + 
				(page_size-1)) & ~(page_size-1));
			break;
		};
	}	
	total_size = (number_of_private_segs * size_of_private_segs);	
	single_size = size_of_private_segs;	
	/* Remember... sizes are actually rounded up to page size */

	if(!remote_only)
	{
		printf("Using MO key %lx \n",(unsigned long)mkey);
		private_segs_mo_pointer = get_sh_mem(mkey, total_size,
			remove_flag);
		if(private_segs_mo_pointer<=0)
		{
		   printf("Unable to attach to MO segment\n");
		   exit(1);
		}
/*
		if(cdebug==1)
	  		printf("Tickle allocated %d bytes of psm MO at %llx\n",
				(int)total_size, (long long)private_segs_mo_pointer);
*/
	}

	srand((unsigned int)time(0));

	/*
	 * Tickle all of the MO slots 
	 */
	if(!remote_only)
	{
		for(j=0;j<number_of_private_segs;j++)
		{
	   		myseg = (char *)private_segs_mo_pointer +(size_of_private_segs*j);
	   		myseg[0] = (char)rand();
		}
	}

	/* 
	 * Tickle my RO slot
	 */
	if(chid !=-1)
	{
		for(j=0;j<number_of_private_segs;j++)
		{
	   	   printf("Using RO key %lx \n",(unsigned long)rkey + (2*j));
	   	   private_segs_ro_pointer = get_sh_mem(rkey+(2*(j)), 
			single_size, remove_flag);
	   	   if(private_segs_ro_pointer<=0)
	   	   {
			   printf("Unable to attach to RO segment\n");
			   exit(1);
	   	   }
	   	   myseg = (char *)private_segs_ro_pointer;
	   	   myseg[0] = (char)rand();
		}
/*
	   myseg = (char *)private_segs_ro_pointer;
	   myseg[0] = (char)rand();
*/
	}
}
#ifdef WIN32

#define MAX_SHARE_MEMORY 32 //max number of shared memory one process can request, regardless of their size
typedef struct _SHARED_MEMORY
{
	HANDLE hFile; //associated file handle, unused here
	HANDLE hMapFile; //associated mapping object handle
	key_t  key; //shared memory key
	BOOL   valid;
} SHARED_MEMORY;

static SHARED_MEMORY sharedMem[MAX_SHARE_MEMORY] = {0};

/*
 *	Win32 implementation using memory mapped file
 *  key - share memory id, it is used to generate mapping object
 *  size - the size of shared memory requested
 *  flag - delete shared memory if its value is non-zero
 */
char *
get_sh_mem(key_t key, size_t size, int flag)
{
	TCHAR szObjName[255];
	TCHAR szFileName[255];
	char* pMem = NULL;
	int position=-1, i=0; //index in shared memory array where we can save shared_memory information 
	HANDLE hFile;

	_stprintf(szObjName, TEXT("Global\\pdsm-%lx"), (unsigned long)key);
	_stprintf(szFileName, TEXT("SharedMemory.%lx"), (unsigned long)key);
	
	/*First we search the shared memory array, try to get existing one*/
	for (i=0; i<MAX_SHARE_MEMORY; i++)
	{
		if (sharedMem[i].valid && sharedMem[i].key == key)
		{
			position = i;
			break;
		}
		else if (position < 0  && !sharedMem[i].valid)
			position = i; //find a unused slot
	}
	/* if we reach here, we did not find exising shared memory*/
	if (flag && sharedMem[position].valid)
	{
		/*do we need to delete memory mapped file here???*/
		hFile = CreateFile(szFileName,
							GENERIC_WRITE | GENERIC_READ,
							FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,
							NULL,
							OPEN_EXISTING,	 
							FILE_FLAG_DELETE_ON_CLOSE, //delete file on close
							NULL);
		if (hFile != INVALID_HANDLE_VALUE)
			CloseHandle(hFile);
	}

	if (position < 0)
		/*the process has already allocated max-allowed shared memory, we have no choice but fail the call*/
		return NULL; 
	/* Now we start to create new shared memory*/
	if (!sharedMem[position].valid)
	{
		sharedMem[position].hMapFile = NULL;
		/* We open or create a file to map to memory*/
		if ((sharedMem[position].hFile = CreateFile(szFileName,
													GENERIC_WRITE | GENERIC_READ,
													FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,
													NULL,
													OPEN_ALWAYS,	 // create a new file if it does not exist, otherwise open the old file
													(flag?FILE_FLAG_DELETE_ON_CLOSE:FILE_ATTRIBUTE_NORMAL), 
													NULL)) == (HANDLE)INVALID_HANDLE_VALUE)
		{
			if(cdebug >= 2)
				printf("Unable to open or create file for memory mapping:\"SharedMemory.%lx\"\n",(unsigned long)key);
		} 
		else
		{
			/*Create mapping object*/
			sharedMem[position].hMapFile = CreateFileMapping(
															sharedMem[position].hFile,    // use paging file
															NULL,                    // default security
															PAGE_READWRITE,          // read/write access
															0,                       // maximum object size (high-order DWORD)
															(DWORD)size,                // maximum object size (low-order DWORD)
															szObjName);                 // name of mapping object
			if (sharedMem[position].hMapFile == NULL)
			{
				if (cdebug >= 2)
					printf("Unable to create mapping object: \"Global\\pdsm-%lx\", error code: %d\n",(unsigned long)key, GetLastError());
				CloseHandle(sharedMem[position].hFile);
			}
			else 
			{
				if (cdebug >= 2 && GetLastError() == ERROR_ALREADY_EXISTS)
					printf("CreateFileMapping returned a handle of existing mapping object\n");
				sharedMem[position].key = key;
				sharedMem[position].valid = TRUE;
			}
		}
	}

	if (sharedMem[position].hMapFile != NULL)
	{
		//
		//now we need to mapped the view. 
		//
		pMem = (char*) MapViewOfFile(sharedMem[position].hMapFile,   // handle to map object
										FILE_MAP_ALL_ACCESS, // read/write permission
										0,
										0,
										(DWORD)size);
		if (pMem == NULL && cdebug >= 2)
			printf("Unable to map file to memory, error code: %d\n", GetLastError());
	}
	return pMem;
}

#else
char *
get_sh_mem(key_t key, size_t size, int flag)
{
	struct stats *statsp;

	if(cdebug == 1)
	  printf("Trying to get shared memory segment with key %lx\n",(unsigned long)key);
	printf("Size %d\n",(int)size);
        shmid=shmget((key_t)key,(size_t)size,(int)(0666));
        if(shmid < 0)
        {
                printf("Shmid %d Errno %d\n",shmid,errno);
        	return((char *)0);
        } 

	if(cdebug == 2)
	  printf("Got shared memory with key %lx\n",(unsigned long)key);

        statsp=(struct stats *)shmat(shmid,0,0);
        if(statsp == (struct stats *)-1)
        {
                printf("Attach to shared memory failed\n");
        	return((char *)0);
        }
	/* Attach ref count is now all that is holding on to this segment */
	if(flag)
		shmctl(shmid, IPC_RMID, 0);  /* cleanup */

	if(cdebug == 2)
	  printf("Attached to shared memory with key %lx\n",(unsigned long)key);

	if(cdebug == 1)
	  printf("Shared memory key %lx Size = %d\n",(unsigned long)key, (int)size);
	/*
	 * Don't need to call quit, as any termination will cleanup
 	 * automatically.
	 */

        /*quit();*/
        return((char *)statsp);
}
#endif
void usage(void)
{
	printf("\t-h............  Help screen\n");
	printf("\t-r............  Remove segements.\n");
	printf("\t-R............  Remote only segements\n");
	printf("\t-p............  Number of nodes\n");
	printf("\t-c............  Tickle child RO segment\n");
	printf("\t-s............  Size of shared memory in bytes\n");
	printf("\n");
}
