/*****************************************************************************
 *  Copyright (c) 2002-2014 by Iozone.org
 *      All rights reserved.
 *              Iozone.org
 *              7417 Crenshaw Dr.
 *              Plano, TX 75025
 *
 *      This product contains benchmarks acquired from several sources who
 *      understand and agree with Iozone's goal of creating fair and objective
 *      benchmarks to measure computer performance.
 *
 *      This copyright notice is placed here only to protect Iozone.org in the
 *      event the source is misused in any manner that is contrary to the
 *      spirit, the goals and the intent of Iozone.org
 *****************************************************************************/

#include "./copyright.txt" 
#include "./license.txt"

/* 
 * Turn this on to enable insanly fast run configuration mode. 
 * The run_time period (see -t) can go down to 60 seconds.
 */
/*
#define FAST
*/

#if defined(WIN32)
#pragma warning(disable:4996)
#pragma warning(disable:4267)
#pragma warning(disable:4133)
#pragma warning(disable:4244)
#pragma warning(disable:4102)
#pragma warning(disable:4018)

#define _CRT_RAND_S

#include <win32_sub.h>
#include <win32_getopt.h>
#include <time.h>
#endif

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#if defined(_linux_)
#include <fcntl.h>
#endif

#if defined(_solaris_)
#include <fcntl.h>
#endif

#if !defined(WIN32)
#include <sys/file.h>
#endif

#if defined(_solaris_)
  #include <time.h>
  #include <sys/time.h>
  #include <sys/fcntl.h>
  #define O_DIRECT (0)
  #include <signal.h>
  #include <unistd.h>
  #include <dirent.h>

#elif !defined(WIN32)
  #include <sys/time.h>
  #include <sys/signal.h>
  #include <unistd.h>
  #include <dirent.h>
  #include <arpa/inet.h>
#endif

#if defined(_macos_)
  #include <signal.h>
  #include <sys/mount.h>
  #include <sys/param.h>
  #define O_DIRECT (0)
#endif
#if defined(_linux_)
  #include <time.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <ctype.h>
#include "netmist.h"
#if !defined(WIN32)
#include <sys/resource.h>
#endif


#if defined(_vmware_)
#define O_DIRECT (0)
#define MALLOC(x) pool_alloc(x)
#else
#define MALLOC(x) my_malloc(x)
#endif
#define THISVERSION "        $Revision: 955 $"
char version[]=THISVERSION;

#if defined(Windows) || defined(_linux_) || defined (_bsd_)||defined(WIN32) ||defined(_cygwin_) || defined(_solaris_)
#define _SHUTDOWN_
#endif

#if defined(_bsd_)
  #define off64_t off_t
#endif

#if defined(_macos_)

  #define off64_t off_t
  #ifdef _LARGEFILE64_SOURCE
    #define I_LSEEK(x,y,z)  lseek(x,(off64_t)(y),z)
    #define I_OPEN(x,y,z)   open(x,(int)(y),(int)(z))
    #define I_CREAT(x,y)    creat(x,(int)(y))
    #define I_FOPEN(x,y)    fopen(x,y)
  /* 
   * Heavy magic happens with the size of the stat and statfs structures
   * about the time that 10.6 came out.
   */
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  > 1050
      #define I_STAT(x,y)     stat(x,y)
      #define I_STATFS(x,y)     statfs(x,y)
    #endif
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  < 1050
      #define I_STAT(x,y)     stat(x,y)
      #define I_STATFS(x,y)     statfs(x,y)
    #endif
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  == 1050
      #define I_STAT(x,y)     stat64(x,y)
      #define I_STATFS(x,y)     statfs64(x,y)
    #endif

    #define I_MMAP(a,b,c,d,e,f)     mmap((void *)(a),(size_t)(b),(int)(c),(int)(d),(int)(e),(off64_t)(f))
    #define I_TRUNC(x,y)  truncate(x,(off64_t)(y))
  #else
    #define I_LSEEK(x,y,z)  lseek(x,(off_t)(y),z)
    #define I_OPEN(x,y,z)   open(x,(int)(y),(int)(z))
    #define I_CREAT(x,y)    creat(x,(int)(y))
    #define I_FOPEN(x,y)    fopen(x,y)
    #define I_STAT(x,y)     stat(x,y)
    #define I_STATFS(x,y)     statfs(x,y)
    #define I_STATVFS(x,y)     statvfs(x,y)
    #define I_MMAP(a,b,c,d,e,f)     mmap((void *)(a),(size_t)(b),(int)(c),(int)(d),(int)(e),(off_t)(f))
    #define I_TRUNC(x,y)  truncate(x,(off_t)(y))
  #endif
#else
  #ifdef _LARGEFILE64_SOURCE
    #define I_LSEEK(x,y,z)  lseek64(x,(off64_t)(y),z)
    #define I_OPEN(x,y,z)   open64(x,(int)(y),(int)(z))
    #define I_CREAT(x,y)    creat64(x,(int)(y))
    #define I_FOPEN(x,y)    fopen64(x,y)
    #define I_STAT(x,y)     stat64(x,y)
    #define I_STATFS(x,y)     statfs64(x,y)
    #define I_STATVFS(x,y)     statvfs64(x,y)
    #if defined(_linux_)
      #define I_MMAP(a,b,c,d,e,f)     mmap64((void *)(a),(size_t)(b),(int)(c),(int)(d),(int)(e),(off_t)(f/4096))
    #else
      #define I_MMAP(a,b,c,d,e,f)     mmap64((void *)(a),(size_t)(b),(int)(c),(int)(d),(int)(e),(off64_t)(f))
    #endif
    #define I_TRUNC(x,y)  truncate64(x,(off64_t)(y))
  #else
    #define I_LSEEK(x,y,z)  lseek(x,(off_t)(y),z)
    #define I_OPEN(x,y,z)   open(x,(int)(y),(int)(z))
    #define I_CREAT(x,y)    creat(x,(int)(y))
    #define I_FOPEN(x,y)    fopen(x,y)
    #if defined(WIN32)
      #define I_STAT(x,y)     _stati64(x,y)
    #else
      #define I_STAT(x,y)     stat(x,y)
    #endif
    #define I_STATFS(x,y)     statfs(x,y)
    #define I_STATVFS(x,y)     statvfs(x,y)
    #define I_MMAP(a,b,c,d,e,f)     mmap((void *)(a),(size_t)(b),(int)(c),(int)(d),(int)(e),(off_t)(f))
    #define I_TRUNC(x,y)  truncate(x,(off_t)(y))
  #endif
#endif

#if defined(WIN32)
/* 
 * There are two ways to get here in Windows. 
 *    #define I_FSYNC(x) fsync(x)
 *    #define I_FSYNC(x) FlushFileBuffers(x)
 * For now, it looks like sticking with vanilla fsync() is ok.
 */
#define I_FSYNC(x) fsync(x)
#else
#define I_FSYNC(x) fsync(x)
#endif

/*
 * If you don't have a licensed version...
 */
#define MAX_L_CLIENTS 50

int max_licensed_clients = MAX_L_CLIENTS;

/* Seconds between heartbeats */

#define HEARTBEAT_TICK 60

/* 
 * Maximum latency during INIT before delay is introduced to slow things down.
 */
#define MAX_LATENCY 5

/* Meta-data constant */
#define META_B 8192

/* 
 * Tell subroutine to run the external script 
 */
#define SRUN 1
/* 
 * Tell subroutine to check to see if script exists and is
 * executable 
 */
#define SCHECK 2

/*
 * Max command size for external scripts 
 */
#define MAXECOMMAND 255

/*
 * Minimum sizes to satisfy users of O_DIRECT
 * Linux 2.6 and later support a minimum transfer size of 512 bytes
 * for O_DIRECT. 
 * BUT.... it jumps to 4k if the disk drive has 4096 byte sectors.
 */
#define MIN_DIRECT_SIZE 512
#define MIN2_DIRECT_SIZE 4096

#define RANDOMIZE_ACROSS_LOAD_POINTS 1

/*
 * We default to the 512 byte sector size, and if this fails, then
 * we try the 4k sector size. If that fails, then we log error and
 * terminate.
 */
int min_direct_size = MIN_DIRECT_SIZE;

/*
 * Number of tokens in the client configuration file.
 */
#define NUM_TOKENS 5

#define POWER_CHECK_LEVEL 70
int power_once;

#if defined(WIN32)
/*
 * Name of the power monitor script 
 */
char power_script[] = "netmist_power.cmd";
/*
 * Name of the external monitor script 
 */
char mon_script[]   = "netmist_mon.cmd";

/*
 * Name of the external stat script 
 */
char stat_script[]   = "netmist_stat.cmd";
#else
/*
 * Name of the power monitor script 
 */
char power_script[] = "./netmist_power.sh";
/*
 * Name of the external monitor script 
 */
char mon_script[]   = "./netmist_mon.sh";

/*
 * Name of the external stat script 
 */
char stat_script[]   = "./netmist_stat.sh";
#endif
/*
 * Name of the external client startup script 
 * 
 * Everyone else gets a standard Korn shell type
 * script file.
 *
 * REMEMBER !!! This thing gets started from the default
 * location of a remote login. 
 * So... be sure you put this thing where it belongs.
 *
 * Note: This script can map shares in Unix, 
 * but things may get weird in SUA or in an Active Directory
 * Domain, where logins may land you in strange lands.
 * 
 */
#if defined (WIN32)
char pretest_script[]   = "netmist_pretest.cmd";
#else 
char pretest_script[]   = "./netmist_pretest.sh";
#endif

#define MERSENNE

#ifdef MERSENNE
unsigned long long init[4]={0x12345ULL, 0x23456ULL, 0x34567ULL, 0x45678ULL};
unsigned long long length=4;
void init_genrand64(unsigned long long );
void init_by_array64(unsigned long long *, unsigned long long );
unsigned long long genrand64_int64(void);
long long genrand64_int63(void);
double genrand64_real1(void);
double genrand64_real2(void);
double genrand64_real3(void);
#endif
#if defined(WIN32)
void n_spec_srand(int);
int n_spec_rand(void);
int n_spec_gseed(void);
void n_ran(void);
#endif
void name_srand(int);
int name_rand(void);
int name_gseed(void);
void name_ran(void);
unsigned int rotateR(unsigned int, int);

/****************************** Plug-ins ***************************/
/* A new plugin will need these functions implemented.             */
/* There are generic implementations for all of these, and they    */
/* are what is used for the current workloads.                     */
/****************************** Plug-ins ***************************/
/*
	read, read_rand, read_file, mmap_read, write, write_rand, 
	write_file, mmap_write, rmw, init_dir, init_file, init_empty_file,
	init_dir, mkdir, unlink, append, locking, access, stat, chmod, 
	remove_file, remove_dir, readdir, rename, copyfile, statfs,
	stat_workdir, make_stat, chdir, fill_buf, prefix_op, postfix_op,
	pathconf.
*/
void generic_read();
void generic_read_file();
void generic_mmap_read();
void generic_read_rand();
void generic_write();
void generic_write_file();
void generic_write_rand();
void generic_mmap_write();
void generic_rmw();
void generic_mkdir();
int  generic_init_dir();
void generic_init_file();
void generic_init_empty_file();
void generic_unlink();
void generic_create();
void generic_append();
void generic_locking();
void generic_access();
void generic_stat();
int  generic_stat_workdir();
int  generic_chdir();
int  generic_remove_file();
int  generic_remove_dir();
void generic_chmod();
void generic_readdir();
void generic_make_stat();
void generic_copyfile();
void generic_rename();
void generic_statfs();
void generic_pathconf();
void generic_custom1();
void generic_custom2();
void generic_prefix_op();
void generic_postfix_op();
void generic_fill_buf();
void generic_setup();
void generic_teardown();
int generic_impersonate();
void* generic_make_impersonate_context();
void donotuse();
void donotuse2();
void donotuse3();

struct file_object *generic_make_path();

int geometric_rand(int);
int netmist_rand( void );
unsigned long long netmist_llrand( void );
int picker(int,int);
int check_dir_w(char *);
void close_cached_file_desc(void);

/*
 * Buckets used for histograms 
*/
#define BUCKETS 40
long long buckets[BUCKETS];
long long global_buckets[BUCKETS];
long long bucket_val[BUCKETS] =
        { 20,40,60,80,100,
        200,400,600,800,1000,
        2000,4000,6000,8000,10000,
        12000,14000,16000,18000,20000,
        40000,60000,80000,100000,
        200000,400000,600000,800000,1000000,
        2000000,4000000,6000000,8000000,10000000,
        20000000,30000000,60000000,90000000,120000000,120000001};

char option_buf[4];
void hist_insert(double);
void child_dump_hist(FILE *);
void dump_hist(void);

int validate_all_tokens(int);
int parse_all_tokens(char * );
int parse_token(char *, char *, char *);
unsigned short check_sum(void);
int license_check(void);
void set_pdsm_attributes(void);
int check_op_total(void);
int modified_run;
long long cache_file_space;
void purge_file(struct file_object *, char *);
unsigned long long get_next_fsize( long long );
void push_name(char *);
void drain_list(void);

int client_number;
/************************** Plug-ins ************************************/
/* This is where your default plug-in goes so that the gen_op( ) will   */
/* go through your plug-in implementation of the functions.             */
/* 								        */
/* Please remember, these will be over-ridden by the workload fs        */
/* type. The default is POSIX, but others may exist 	                */
/* The client_startup will look at the mix_table.fs_type, and will      */
/* replace all of these pointers, with the ones for the specific 	*/
/* workload.								*/
/* 								        */
/************************** Plug-ins ************************************/
struct method_table Netmist_Ops[NUM_WORK_OBJ];

/*
 * Initialize the random number generator, for reproducable
 * results.
 */

#define INIT_SEED 0x456

/*
 * For everyone else use ssh and scp.
 */
#if !defined(REMSH)
#define REMSH "ssh"
#define REMCP "scp"
#endif

char remsh[]=REMSH;
char remcp[]=REMCP;

long avg_read_size,avg_write_size;

/*
 * Define the mix. The defaults are imported from netmist.h
 * All of these can be set to new values by importing a new 
 * set of workload definitions from a file.
 */
struct mix_table my_mix_table[NUM_WORK_OBJ];

int ipv6_enable = 0;
int my_workload;
int first_time;
int generation;
int init_phase = 0;
int licensed = 0;
time_t delivered_heartbeat, current_time;
time_t delivered_log_heartbeat;

#define RW_ARRAY 100

int read_dist_array[RW_ARRAY];
int write_dist_array[RW_ARRAY];

#define FSIZE_MIN_SLOTS 20

double read_throughput, write_throughput;

#if defined(PDSM_R_REGISTRY)
char * get_remote_zone(int,int);
char * get_my_master_zone(int,int);
char my_pdsm_file[MAXNAME];

struct pdsm_remote_stats *pdsm_stats;
struct pdsm_remote_control *pdsm_control;

#endif

#if defined(WIN32)
char wmiUser[MAXNAME] = "administrator";
char wmiDomain[MAXNAME] = "fsperf";
char wmiPassword[MAXNAME] = "Lab$Man01";
#endif 
/*
 * Initialize the size distribution mechanism
 * Note: my_workload is already set by now !!
 */
void
init_dist_tables(void)
{
	int nval,val,minv,maxv;
	int i,j,k;
	long long total;
	
	total = 0LL;
	k=0;
	for(i=0;i<MAX_DIST_ELEM;i++)
	{
	  for(j = my_mix_table[my_workload].read_dist[i].percent; j > 0; j--)
	  {
	    minv = my_mix_table[my_workload].read_dist[i].size_min;
	    maxv = my_mix_table[my_workload].read_dist[i].size_max;
	    val = (maxv-minv +1);
	    nval = (netmist_rand()%val) + minv;
	    read_dist_array[k]=nval;
	    k++;
	  }
	}

	for(i=0;i<RW_ARRAY;i++)
	{
	  total+=read_dist_array[i];
	}
        avg_read_size = (long) (total/(long long)RW_ARRAY)/1024;
	printf("Average read size = %ld Kbytes\n",avg_read_size);
   	fflush(stdout);
	
	k=0;
	for(i=0;i<MAX_DIST_ELEM;i++)
	{
	  for(j = my_mix_table[my_workload].write_dist[i].percent; j > 0; j--)
	  {
	    minv = my_mix_table[my_workload].write_dist[i].size_min;
	    maxv = my_mix_table[my_workload].write_dist[i].size_max;
	    val = (maxv-minv)+1;
	    nval = (netmist_rand()%val) + minv;
	    write_dist_array[k]=nval;
	    k++;
	  }
	}
	total= 0LL;
	for(i=0;i<RW_ARRAY;i++)
	{
	  total+=write_dist_array[i];
	}
        avg_write_size = (long) (total/(long long)RW_ARRAY)/1024;
	printf("Average write size = %ld Kbytes\n",avg_write_size);
   	fflush(stdout);
}
	
/* 
 * Defines for the mixture of types of ops.
 * "my_workload" must be set before using these !!
 */

#define PERCENT_READ	     (my_mix_table[my_workload].percent_read)
#define PERCENT_READ_FILE    (my_mix_table[my_workload].percent_read_file)
#define PERCENT_MM_READ      (my_mix_table[my_workload].percent_mmap_read)
#define PERCENT_READ_RAND    (my_mix_table[my_workload].percent_read_rand)
#define PERCENT_WRITE        (my_mix_table[my_workload].percent_write)
#define PERCENT_WRITE_FILE   (my_mix_table[my_workload].percent_write_file)
#define PERCENT_MM_WRITE     (my_mix_table[my_workload].percent_mmap_write)
#define PERCENT_WRITE_RAND   (my_mix_table[my_workload].percent_write_rand)
#define PERCENT_RMW          (my_mix_table[my_workload].percent_rmw)
#define PERCENT_MKDIR        (my_mix_table[my_workload].percent_mkdir)
#define PERCENT_UNLINK	     (my_mix_table[my_workload].percent_unlink)
#define PERCENT_CREATE	     (my_mix_table[my_workload].percent_create)
#define PERCENT_APPEND       (my_mix_table[my_workload].percent_append)
#define PERCENT_LOCKING      (my_mix_table[my_workload].percent_locking)
#define PERCENT_ACCESS       (my_mix_table[my_workload].percent_access)
#define PERCENT_STAT	     (my_mix_table[my_workload].percent_stat)
#define PERCENT_CHMOD	     (my_mix_table[my_workload].percent_chmod)
#define PERCENT_READDIR	     (my_mix_table[my_workload].percent_readdir)
#define PERCENT_COPYFILE     (my_mix_table[my_workload].percent_copyfile)
#define PERCENT_RENAME       (my_mix_table[my_workload].percent_rename)
#define PERCENT_STATFS       (my_mix_table[my_workload].percent_statfs)
#define PERCENT_PATHCONF     (my_mix_table[my_workload].percent_pathconf)
#define PERCENT_CUSTOM1      (my_mix_table[my_workload].percent_custom1)
#define PERCENT_CUSTOM2      (my_mix_table[my_workload].percent_custom2)


/****************************************************************************/
/* COMM SECTION */
/****************************************************************************/
/*
 * Communication library for controlling many hosts over the net.
 */
#if !defined(WIN32)

#include <sys/signal.h>
#include <sys/socket.h>
#include <netdb.h>
#else
#include <signal.h>
#endif

#if defined(_bsd_) ||defined(_cygwin_)
#include <signal.h>
#endif

#if defined(_solaris_) || defined(_bsd_) ||defined(_cygwin_)
#include <arpa/inet.h>
#include <netinet/in.h>
#endif

/*
 * Master:
 * Call start_master_init() very early before any children start.
 * After all the children are running then call master_start_comm(number of 
 * children);
 * If you wish to wait for all children the call master_wait(number of children)
 * If you wish to tell all children to go then call master_send_go
 * (number of children);
 * After all children are gone then call master_stop_comm(number of children);
 * 
 * Children:
 * Call child_start_comm(controller's name) early
 * Call child_wait() to block for message from master.
 * If you wish to tell the master you are ready then call 
 * tell_master_ready(child_id);
 *
 */

char c_master_name[MAXHNAME];
char child_name[MAXHNAME];

int *child_pid; /* Master only, malloc space */
int *don_master_send_socket_port;
char *don_master_send_socket_name;
int master_listen_socket;
int child_listen_socket;
int child_socket_val;
int child_port;
int master_listen_port;
long *mem_buf;
char *main_buf;
char *copy_file_buf;
char pit_hostname[40];
char pit_service[8];
char time_string[30];
int flush_flag = 1;
int p_ext_mon_flag;
int p_ext_mon_arg_flag;
int log_beats;
int fpd_flag = 0;
int max_open;
int max_proc;
int rot;
int report_name;
int cached_open_count = 0;
int cleanup_flag = 1;
int dump_files_flag = 0;
int sharing_flag = 0;
int heartbeat_flag = 0;
int fd_caching_limit = 0;
int Op_lat_flag = 0;
int Multi_load_disp = 0;
int junk;
int localflag=0;
int gg_flag;
int gg_offset;
char *big_buf;
typedef struct sockaddr_in       sockaddr_in_t;
typedef struct sockaddr_in6      sockaddr_in6_t;
struct addrinfo *ai;
int              aiErr;
struct addrinfo  hints;
struct addrinfo *aiHead;
struct sockaddr_in6  *pSadrIn6;

#define SPOT_PCT 5
#define NSPOTS (100/SPOT_PCT)
long long hit_in_spot = 1000; /* default . Is calculated later */
long long spot_size;
long long offset_in_spot;
/*
sockaddr_in6_t  *pSadrIn6;
*/
int once;
int ecount;

int map_index;
int xdebug = 0;
int cdebug = 0;
int sockdebug = 0;
int mdebug = 0;
int ldebug = 0;
FILE *newstdin;
FILE *newstdout;
FILE *newstderr;
FILE *tracefile;
int LogonUser_res=0;
int ImpersonateLoggedOnUser_res=0;
int retry1_counter, retry2_counter;

int log_flag;
FILE *log_out,*log_beat_fd;
char log_filename[MAXNAME];
char log_beat_filename[MAXNAME];

/* Note: with trace debug, O_CREAT=1 and O_RDWR=1 are assumed and not reported */
int tracedebug = 0;
/* enabling tracing during init makes very large trace logs */
/*
#define INIT_TRACE_DEBUG
*/
unsigned long trace_seq = 0; /* trace log sequence number */
#define TRACE_STATE_SETUP 0
#define TRACE_STATE_VALIDATION 1
#define TRACE_STATE_CREATE 2
#define TRACE_STATE_RUN 3
#define TRACE_STATE_CLEANUP 4
int tracestate = TRACE_STATE_SETUP; /* tracks the state of client for trace log */
const char *trace_states[] = { "Setup","Validation","Create","Run","Cleanup" };

/*
 * Holds name of the file that contains the import op dist
 */
char itable[MAXNAME];

/*
 * Internal command structure that is sent from the clients to the master.
 */
struct master_command {
	int m_command;
	int m_generation;
	int m_client_number;
	int m_work_obj_index;
	int m_child_flag;
	char m_host_name[MAXHNAME];
	char m_client_name[MAXHNAME];
	int m_client_eval;
	int m_child_port;
	int m_child_pid;
	int m_percent_complete;
	int m_background;
	int m_sharemode;
	int m_rand_dist_behavior;
	int m_uniform_file_size_dist;
	int m_fs_type[MAXFSNAME];
	char m_workload_name[MAXHNAME];
	long m_run_time;
	long m_min_direct_size;
	double m_total_file_op_time;
	double m_total_file_ops;
	double m_ops_per_second;
	double m_average_latency;
	double m_read_throughput;
	double m_write_throughput;
	double m_current_op_rate;
	int  m_read_count;
	char m_read_string[OP_STR_LEN];
	int  m_read_file_count;
	char m_read_file_string[OP_STR_LEN];
	int  m_mmap_read_count;
	char m_mmap_read_string[OP_STR_LEN];
	int m_read_rand_count;
	char m_read_rand_string[OP_STR_LEN];
	int m_write_count;
	char m_write_string[OP_STR_LEN];
	int m_write_file_count;
	char m_write_file_string[OP_STR_LEN];
	int m_open_count;
	int m_close_count;
	int m_mmap_write_count;
	char m_mmap_write_string[OP_STR_LEN];
	int m_write_rand_count;
	char m_write_rand_string[OP_STR_LEN];
	int m_rmw_count;
	char m_rmw_string[OP_STR_LEN];
	int m_mkdir_count;
	char m_mkdir_string[OP_STR_LEN];
	int m_unlink_count;
	char m_unlink_string[OP_STR_LEN];
	int m_create_count;
	char m_create_string[OP_STR_LEN];
	int m_append_count;
	char m_append_string[OP_STR_LEN];
	int m_locking_count;
	char m_locking_string[OP_STR_LEN];
	int m_access_count;
	char m_access_string[OP_STR_LEN];
	int m_chmod_count;
	char m_chmod_string[OP_STR_LEN];
	int m_readdir_count;
	char m_readdir_string[OP_STR_LEN];
	int m_stat_count;
	char m_stat_string[OP_STR_LEN];
	int m_copyfile_count;
	char m_copyfile_string[OP_STR_LEN];
	int m_rename_count;
	char m_rename_string[OP_STR_LEN];
	int m_statfs_count;
	char m_statfs_string[OP_STR_LEN];
	int m_pathconf_count;
	char m_pathconf_string[OP_STR_LEN];
	int m_custom1_count;
	char m_custom1_string[OP_STR_LEN];
	int m_custom2_count;
	char m_custom2_string[OP_STR_LEN];

	double m_read_time;
	double m_read_file_time;
	double m_mmap_read_time;
	double m_read_rand_time;
	double m_write_time;
	double m_write_file_time;
	double m_mmap_write_time;
	double m_write_rand_time;
	double m_rmw_time;
	double m_mkdir_time;
	double m_unlink_time;
	double m_create_time;
	double m_append_time;
	double m_locking_time;
	double m_access_time;
	double m_chmod_time;
	double m_readdir_time;
	double m_stat_time;
	double m_copyfile_time;
	double m_rename_time;
	double m_statfs_time;
	double m_pathconf_time;
	double m_custom1_time;
	double m_custom2_time;
	long long m_bands[BUCKETS];
	int m_heartbeat_type;
	int m_modified_run;
}master_rcv_buf;

/*
 * External command structure that is sent from the clients to the master.
 */
struct ext_master_command {
	char m_command[10];
	char m_generation[10];
	char m_client_number[10];
	char m_work_obj_index[10];
	char m_child_flag[10];
	char m_host_name[MAXHNAME];
	char m_client_name[MAXHNAME];
	char m_client_eval[10];
	char m_child_port[10];
	char m_child_pid[10];
	char m_percent_complete[10]; /* update msgs */
	char m_background[4]; 
	char m_sharemode[4]; 
	char m_rand_dist_behavior[4]; 
	char m_uniform_file_size_dist[4]; 
	char m_fs_type[MAXFSNAME]; 
	char m_workload_name[MAXHNAME]; /* result msgs */
        char m_run_time[20];
        char m_min_direct_size[20];
        char m_total_file_op_time[20];
        char m_total_file_ops[20];
        char m_ops_per_second[20];
        char m_average_latency[20];
        char m_read_throughput[20];
        char m_write_throughput[20];
        char m_current_op_rate[20];
/* HERE */
        char m_read_count[10];
        char m_read_string[OP_STR_LEN];
        char m_read_file_count[10];
        char m_read_file_string[OP_STR_LEN];
        char m_mmap_read_count[10];
        char m_mmap_read_string[OP_STR_LEN];
        char m_read_rand_count[10];
        char m_read_rand_string[OP_STR_LEN];
        char m_write_count[10];
        char m_write_string[OP_STR_LEN];
        char m_write_file_count[10];
        char m_write_file_string[OP_STR_LEN];
        char m_open_count[10];
        char m_close_count[10];
        char m_mmap_write_count[10];
        char m_mmap_write_string[OP_STR_LEN];
        char m_write_rand_count[10];
        char m_write_rand_string[OP_STR_LEN];
        char m_rmw_count[10];
        char m_rmw_string[OP_STR_LEN];
        char m_mkdir_count[10];
        char m_mkdir_string[OP_STR_LEN];
        char m_unlink_count[10];
        char m_unlink_string[OP_STR_LEN];
        char m_create_count[10];
        char m_create_string[OP_STR_LEN];
        char m_append_count[10];
        char m_append_string[OP_STR_LEN];
        char m_locking_count[10];
        char m_locking_string[OP_STR_LEN];
        char m_access_count[10];
        char m_access_string[OP_STR_LEN];
        char m_chmod_count[10];
        char m_chmod_string[OP_STR_LEN];
        char m_readdir_count[10];
        char m_readdir_string[OP_STR_LEN];
        char m_stat_count[10];
        char m_stat_string[OP_STR_LEN];
        char m_copyfile_count[10];
        char m_copyfile_string[OP_STR_LEN];
        char m_rename_count[10];
        char m_rename_string[OP_STR_LEN];
        char m_statfs_count[10];
        char m_statfs_string[OP_STR_LEN];
        char m_pathconf_count[10];
        char m_pathconf_string[OP_STR_LEN];
        char m_custom1_count[10];
        char m_custom1_string[OP_STR_LEN];
        char m_custom2_count[10];
        char m_custom2_string[OP_STR_LEN];

        char m_read_time[20];
        char m_read_file_time[20];
        char m_mmap_read_time[20];
        char m_read_rand_time[20];
        char m_write_time[20];
        char m_write_file_time[20];
        char m_mmap_write_time[20];
        char m_write_rand_time[20];
        char m_rmw_time[20];
        char m_mkdir_time[20];
        char m_unlink_time[20];
        char m_create_time[20];
        char m_append_time[20];
        char m_locking_time[20];
        char m_access_time[20];
        char m_chmod_time[20];
        char m_readdir_time[20];
        char m_stat_time[20];
        char m_copyfile_time[20];
        char m_rename_time[20];
        char m_statfs_time[20];
        char m_pathconf_time[20];
        char m_custom1_time[20];
        char m_custom2_time[20];
        char m_bands[BUCKETS][20];
        char m_heartbeat_type[10];
        char m_modified_run[10];
}ext_master_rcv_buf;

/*
 * Internal command structure that is sent from the master to the clients.
 */
struct work_desc {
	int c_op_array[NUM_OP_TYPES];
	char c_op_array_string[NUM_OP_TYPES][OP_STR_LEN];
	struct dist_entry c_rdist_array[MAX_DIST_ELEM];
	struct dist_entry c_wdist_array[MAX_DIST_ELEM];
	int c_percent_commit;
	int c_percent_direct;
	int c_percent_osync;
	int c_percent_geometric;
	int c_percent_compress;
	int c_background;
	int c_sharemode;
	int c_rand_dist_behavior;
	int c_uniform_file_size_dist;
	int c_align;
	char c_fs_type[MAXFSNAME];
	char c_workload_name[MAXHNAME];
};

/*
 * Internal command structure that is sent from the master to the clients.
 */
struct client_command {
	int c_generation;
	int c_command;
	int c_cleanup;
	int c_dump_files_flag;
	int c_tracedebug;
	int c_ipv6_enable;
	int c_licensed;
	int c_heartbeat;
	int c_fd_caching_limit;
	int c_op_lat_flag;
	int c_gg_offset;
	int c_gg_flag;
	int c_client_number;
	int c_child_flag;
	int c_flush_flag;
	int c_sockdebug;
	char c_host_name[MAXHNAME];
	char c_pit_hostname[40];
	char c_pit_service[8];
	char c_client_name[MAXHNAME];
	char c_client_version[100];
	struct work_desc client_worker[NUM_WORK_OBJ];
}child_rcv_buf;

/*
 * External command structure that is sent from the master to the clients.
 */
struct ext_work {
	char c_op_array[NUM_OP_TYPES][4];
	char c_op_array_string[NUM_OP_TYPES][OP_STR_LEN];
	char c_rdist_array_size_min[MAX_DIST_ELEM][8];
	char c_rdist_array_size_max[MAX_DIST_ELEM][8];
	char c_rdist_array_percent[MAX_DIST_ELEM][8];
	char c_wdist_array_size_min[MAX_DIST_ELEM][8];
	char c_wdist_array_size_max[MAX_DIST_ELEM][8];
	char c_wdist_array_percent[MAX_DIST_ELEM][8];
	char c_percent_commit[4];
	char c_percent_direct[4];
	char c_percent_osync[4];
	char c_percent_geometric[4];
	char c_percent_compress[4];
	char c_background[4];
	char c_sharemode[4];
	char c_rand_dist_behavior[4];
	char c_uniform_file_size_dist[4];
	char c_align[8];
	char c_fs_type[MAXFSNAME];
	char c_workload_name[MAXHNAME];
};
/*
 * External command structure that is sent from the master to the clients.
 */
struct ext_client_command {
	char c_generation[20];
	char c_command[20];
	char c_cleanup[4];
	char c_dump_files_flag[4];
	char c_tracedebug[4];
	char c_ipv6_enable[4];
	char c_licensed[4];
	char c_heartbeat[4];
	char c_fd_caching_limit[8];
	char c_op_lat_flag[4];
	char c_gg_offset[20];
	char c_gg_flag[4];
	char c_client_number[20];
	char c_child_flag[20];
	char c_flush_flag[4];
	char c_sockdebug[4];
	char c_host_name[MAXHNAME];
	char c_client_name[MAXHNAME];
	char c_pit_hostname[48];
	char c_pit_service[8];
	char c_client_version[100];
	struct ext_work ext_worker[NUM_WORK_OBJ];
	
}ext_child_rcv_buf;
int master_p; /* Port the master is listening on */

/*
 * The prototypes used for the comm mechanism.
 */
#if defined(HAVE_ANSIC_C)
void master_int_to_ext(struct master_command *, struct ext_master_command *);
void master_ext_to_int(struct master_command *, struct ext_master_command *);
void client_int_to_ext(struct client_command *, struct ext_client_command *);
void client_ext_to_int(struct client_command *, struct ext_client_command *);
char *my_strcasestr(char *, char *);
int my_strcasecmp(char *, char *);
void *my_malloc(size_t);
void init_vars(void);
void re_init_vars(void);
void init_space(void);
void dump_file_objects(FILE *);
int get_host_address(char *, unsigned int *, char*);
int get_host_ip_address(char *, char *);

#if defined(_solaris_)
int pit_gettimeofday( struct timeval *, long, char *, char *);
void pit( int, struct timeval *);
#endif

#if !defined(_solaris_)
int pit_gettimeofday( struct timeval *, struct timezone *, char *, char *);
void pit( int, struct timeval *);
#endif

#if defined(_solaris_)
double m_gettimeofday(struct timeval *, long );
#else
double m_gettimeofday(struct timeval *, struct timezone *);
#endif

int openSckt( const char *, const char *, unsigned int );
int start_master_listen(void);
void stop_master_listen(int);
int Master_send(char *, int, char *);
void master_listen(int);
int start_child_listen(void);
void stop_child_listen(int);
int Child_send(char *, char *);
void child_listen(int);
void start_master_init(void);
int power_mon(int, char *);
int extern_mon(int, char *);
int u_extern_mon(int, char *);
int extern_stat(int, char *);
void master_start_comm(int);
void master_stop_comm(int);
void child_start_comm(char *);
void child_wait(void);
void tell_master_ready(int);
void print_results(void);
void tell_master_pct_complete(int,int,int);
void tell_master_heartbeat(int,int,double);
void tell_master_phase(int,int);
void tell_master_results(int);
void tell_master_error(int,int);
void master_send_go(int);
int master_wait_barrier(int);
int master_wait_results(int);
char * get_timestamp(char *);
void make_name(char *, int, int, int);
void make_short_name(char *, int, int);
static int prename_len(void);
static int postname_len(void);
static void init_dist(void);
static char get_ch1(void);
void netmist_srand( int );
int netmist_gseed(void);
#else
char *my_strcasestr();
int my_strcasecmp();
int get_host_address();
int pit_gettimeofday( );
double m_gettimeofday();
void master_int_to_ext();
void init_vars();
void re_init_vars();
void init_space();
void dump_file_objects();
void master_ext_to_int();
void client_int_to_ext();
void client_ext_to_int();
int start_master_listen();
void stop_master_listen();
int Master_send();
void stop_master_send();
void master_listen();
int start_child_listen();
void stop_child_listen();
int Child_send();
void child_listen();
int power_mon();
int extern_mon();
int u_extern_mon();
int extern_stat();
void start_master_init();
void master_start_comm();
void master_stop_comm();
void child_start_comm();
void child_wait();
void print_results();
void tell_master_pct_complete();
void tell_master_heartbeat();
void tell_master_phase();
void tell_master_results();
void tell_master_ready();
void tell_master_error();
void master_send_go();
int master_wait_barrier();
int master_wait_results();
char * get_timestamp();
void make_name();
void make_short_name();
static int prename_len();
static int postname_len();
static void init_dist();
static char get_ch1();
void netmist_srand();
int netmist_rand();
unsigned long long netmist_llrand();
int netmist_gseed();
#endif 
/*
 * The operation types that are sent back and forth.
 */
#define R_CHILD_JOIN 1
#define R_FLAG_DATA 2
#define R_FAILED 3
#define R_TERMINATE 4
#define R_GO 5
#define R_PERCENT_COMPLETE 6
#define R_SENDING_RESULTS 7
#define R_IM_DONE 8
#define R_HEARTBEAT 9
#define R_PERCENT_W_COMPLETE 10
#define R_PERCENT_I_COMPLETE 11
#define INIT_PHASE 12
#define RUN_PHASE 13
#define WARM_PHASE 14


#define INIT_BEAT 1
#define WARM_BEAT 2
#define RUN_BEAT 3
#define PRE_WARM_BEAT 4
#define CLEAN_BEAT 5
/* 
 * These are the ports that will be used for comm.
 */
#define HOST_LIST_PORT 5000
#define HOST_ESEND_PORT (HOST_LIST_PORT+MAXCLIENTS)
#define CHILD_ESEND_PORT (HOST_ESEND_PORT+MAXCLIENTS)
#define CHILD_LIST_PORT (CHILD_ESEND_PORT+MAXCLIENTS)
#define PORT_RANGE 10000

#define CHILD_STATE_READY 1
#define CHILD_STATE_BEGIN 2

/*
 * Each client is assigned an ID when the comm is set up.
 */
int c_chid;



/****************************************************************************/
/* END COMM SECTION */
/****************************************************************************/

/*
 * All of the various names of things. Hostnames, filenames....
 */

char master_results_dir[MAXNAME];         /* Path to master's results dir */
char client_log_dir[MAXNAME];             /* Path to client's log dir */
char master_name[MAXHNAME];	          /* The master's name */
long long client_memsize_mb=(long long)1024; 	  /* memsize of the clients */
long long agg_set_memsize_mb=(long long)1024; 	  /* memsize of the agg_set */

/* password for users for "home folder" workload*/
char password[MAXNAME]="abcd_1234"; /*this is default password*/
char usr_dom[MAXNAME];	          	  /* User & domain name */
char userdomain[MAXNAME]=".";
char mypassword[MAXNAME];
//char user[MAXNAME] = "nobody";

struct result_object {
	int  client_id;
	int  work_obj_index;
	char work_obj_name[MAXNAME];
	char client_name[MAXNAME];
	long run_time;
	long min_direct_size;
	long write_count;
	char write_string[OP_STR_LEN];
	long write_file_count;
	char write_file_string[OP_STR_LEN];
	long open_count;
	long close_count;
	long mmap_write_count;
	char mmap_write_string[OP_STR_LEN];
	long read_count;
	char read_string[OP_STR_LEN];
	long read_file_count;
	char read_file_string[OP_STR_LEN];
	long mmap_read_count;
	char mmap_read_string[OP_STR_LEN];
	long mkdir_count;
	char mkdir_string[OP_STR_LEN];
	long unlink_count;
	char unlink_string[OP_STR_LEN];
	long create_count;
	char create_string[OP_STR_LEN];
	long stat_count;
	char stat_string[OP_STR_LEN];
	long copyfile_count;
	char copyfile_string[OP_STR_LEN];
	long rename_count;
	char rename_string[OP_STR_LEN];
	long statfs_count;
	char statfs_string[OP_STR_LEN];
	long pathconf_count;
	char pathconf_string[OP_STR_LEN];
	long custom1_count;
	char custom1_string[OP_STR_LEN];
	long custom2_count;
	char custom2_string[OP_STR_LEN];
	long append_count;
	char append_string[OP_STR_LEN];
	long lock_count;
	char lock_string[OP_STR_LEN];
	long access_count;
	char access_string[OP_STR_LEN];
	long chmod_count;
	char chmod_string[OP_STR_LEN];
	long readdir_count;
	char readdir_string[OP_STR_LEN];
	long write_rand_count;
	char write_rand_string[OP_STR_LEN];
	long read_rand_count;
	char read_rand_string[OP_STR_LEN];
	long rmw_count;
	char rmw_string[OP_STR_LEN];
	double write_time;
	double write_file_time;
	double mmap_write_time;
	double read_time;
	double read_file_time;
	double mmap_read_time;
	double mkdir_time;
	double unlink_time;
	double create_time;
	double stat_time;
	double copyfile_time;
	double rename_time;
	double statfs_time;
	double pathconf_time;
	double custom1_time;
	double custom2_time;
	double append_time;
	double lock_time;
	double access_time;
	double chmod_time;
	double readdir_time;
	double write_rand_time;
	double read_rand_time;
	double rmw_time;
	double ops_per_second;
	double total_file_op_time;
	double total_file_ops;
	double average_latency;
	double read_throughput;
	double write_throughput;
	long long bands[BUCKETS];
	int modified_run;
	int background;
	int sharemode;
	int uniform_file_size_dist;
	int rand_dist_behavior;
	/*char fs_type[MAXFSNAME];*/
};

void master_dump_hist(struct result_object *, FILE *);
/*
 * The client object contains all of the properties and attributes of 
 * a given client worker thread/proc.
 */
struct client_object {
	int client_id;				/* Client number            */
	char client_name[MAXHNAME];		/* Client hostname          */
	char client_workdir[MAXNAME];		/* Client workdir	    */
	char client_dom_user[MAXNAME];		/* Client login userid      */
	char client_password[MAXNAME];		/* Client login password    */
	char client_execdir[MAXNAME];		/* Client exec path	    */
	char client_workload_name[MAXHNAME];    /* Client name of workload  */
	long long memsize;			/* Client's memory size     */
	double op_rate;				/* Requested op rate 	    */
	int instances;				/* Number of copies  	    */
	unsigned long open_flags;		/* Flags to open files      */
	int instant_ops_per_sec;		/* Realtime rate            */
	int instant_status;			/* Realtime status	    */
	struct result_object results;		/* This child's result	    */
} ;
/* ptr_client_obj[MAXCLIENTS];*/

struct client_object *client_obj;
/* 
 * Defaults for file and record size 
 */
#define FSIZE (1) /* in Kbytes */
#define RSIZE (1) /* Minimum xfer size */
int rsize = RSIZE; /* in kBytes */
int max_xfer_size ;  /* Scope: Global:  In bytes */
long long fsize = FSIZE; /* In Kbytes */
long long fsize1,fsize2;

/* 
 * Used to manually set the average file size in Kbytes 
 */
int setfsize = 0; /* In Kbytes */
int setfsflag;
/* 
 * Just in case there is no workdir provided 
 */
#define HOME "/tmp"

/*
 * Directory,on the client,  where work will be conducted.
 */
char workdir[MAXNAME];

/* 
 * Location of executable for test on client 
 */
char execdir[MAXNAME]; 


/*
 * Other op types not yet implemented.
 * OP_LINK
 * OP_SYMLINK
 */


/*
 * Define the directory structure
 */
int client_dirs;
int client_files;
int client_files_pop;
int client_dirs_pop;
int client_files_rpop;
int client_files_wpop;
int max_client_dirs_pop;


/* 
 * Status flag values .. Currently not used by test.
 */
#define NONE 0
#define INIT 1
#define INUSE 2

/* 
 * Error codes returned by children.
 */
#define TOO_SHORT 199
#define WRONG_VERS 197
#define BAD_MIX 196
#define NO_CREATE 195
#define BAD_PIT 194
#define BAD_PIT_CONN 193
#define BAD_WORKLOAD 192
#define BAD_DNS 191
#define NO_WORKDIR 190
#define LOGON_FAILURE 189
#define LOAD_GEN_ERR 187

int bad_name_option; /* Used by client to kill bad workload names */
/*
 * Create a table of 2000 test op entries. Then
 * initialize according to the distribution described
 * by the mix.
 */
#define TESTOPS 3000

int test_ops = TESTOPS;

/* 
 * These are the objects associated with each unique
 * workload in the mix 
 */
struct work_object workloads[NUM_WORK_OBJ];


/* struct result_object *ro_list_head, *ro_list_tail; */

/*
 * There are one of these work_objects for each unique
 * workload 
 */
struct work_object work_obj[NUM_WORK_OBJ];

struct file_object {
	char filename[MAXNAME];
	char dirname[MAXNAME];
	long op_count;
	long access_count;
	long long db_access_count;
	long long which_spot;
	unsigned long long Current_fsize;
	unsigned long long Original_fsize;
#if defined(WIN32)
	HANDLE  file_desc; //we cache opened file handle for Windows
#else
	int  file_desc;
#endif 
	/* Concurency for random ops */
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t offset;
#else
	off_t offset;
#endif
	/* Concurency for sequential read ops */
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t seqread_offset;
#else
	off_t seqread_offset;
#endif
	/* Concurency for sequential write ops */
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t seqwrite_offset;
#else
	off_t seqwrite_offset;
#endif
	/* Concurency for whole file read ops */
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t wholefileread_offset;
#else
	off_t wholefileread_offset;
#endif
	/* Concurency for whole file write ops */
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t wholefilewrite_offset;
#else
	off_t wholefilewrite_offset;
#endif
	/* Concurency for copy file ops */
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t copyfile_offset;
#else
	off_t copyfile_offset;
#endif
	/* Concurency for rmw ops */
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t rmw_offset;
#else
	off_t rmw_offset;
#endif
	/* Concurency for mmapread ops */
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t mmapread_offset;
#else
	off_t mmapread_offset;
#endif
	/* Concurency for mmapwrite ops */
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t mmapwrite_offset;
#else
	off_t mmapwrite_offset;
#endif
#if defined(_solaris_) || defined(_macos_)
	unsigned long directio_on;
#endif
};
	
struct file_object *op_X_array[NUM_OP_TYPES];


/*
 * Some global counters used for statistics.
 */
long long total_space,total_max_space;

/*
 * Total number of clients running the test.
 */
int num_clients,p_num_clients;

int debug = 0;

/*
 * The test array of possible ops to be performed.
 */
struct test{
	int op_type;
	int status_flags; /* Future */
	int writer_id; /* Future */
};

/*
 * Global counters for statistics. Needs to be in 
 * shared memory if the client forks.
 */

double total_rmw_ops;
double total_meta_ops;
double total_read_bytes;
double total_read_rand_bytes;
double total_write_bytes;
double total_copyfile_bytes;
double total_write_rand_bytes;
double total_rmw_bytes;
double total_file_ops;
double prev_file_ops;
double ops_this_tick;
double total_file_op_time;
double my_gen_op_time;
double total_nap_time;

/*
 * Pointer off to the test entries array. Allocated dynamically.
 */
struct test *tests;

/*
 * Prototypes
 */
#if defined(HAVE_ANSIC_C)
double gettime(void);
void clean_exit(int);
void pipe_elog(int);
void R_exit(int);
#else
double gettime();
void clean_exit();
void pipe_elog();
void R_exit();
#endif

/*
 * Needed externs
 */
#if !defined(_cygwin_)
extern int optind;
extern char *optarg;
#endif

#if !defined(WIN32)
extern int errno;
#endif

/*
 * The total number of operations to perform. In the future it
 * may be infinite, as the test may terminate based on run
 * time.
 */
int number_ops;

/*
 * Used to create unique directory structure for each client or 
 * child process of a client.
 */
int client_id,impersonate,real_client_id;

/* 
 * Name of benchmark. "Generic Remote All-purpose Filesystem Test" 
 */
#if !defined(LITE)
char testname[] = "SPECsfs2014";
char short_testname[] = "SPECsfs2014";
#else
char testname[] = "Netmist_2014";
char short_testname[] = "Netmist_2014";
#endif
char Author[] = "Don Capps. Email: capps@iozone.org";
char Maintainer[] = "Don Capps. Email: capps@iozone.org";
char Contributors[] = "Don Capps, Carol Capps, Darren Sawyer, Jerry Lohr,";
char Contributors2[] = "George Dowding, Gary Little, Terry Capps,";
char Contributors3[] = "Robin Miller, Sorin Faibish, Raymond Wang,";
char Contributors4[] = "Tanmay Waghmare, Yansheng Zhang, Vernon Miller,";
char Contributors5[] = "Nick Principe, Zach Jones";
char bin_testname[] = "netmist";

/*
 * Flags used for input validation on getopt() post processing.
 */

int orflag, cflag, nflag, rflag, sflag, pflag, hflag,kflag,dflag,im_flag;
int Cflag, eflag,fflag,gflag,Rflag, Mflag,Uflag,bflag,iflag,Bflag,vflag;
int lflag,c_eflag,c_eval,Dflag,Pflag,wflag,Q_flag,T_flag,Vflag;
int Password_flag;
int Usr_dom_flag;
int my_child_pid; /* used by child processes to kill procs */
int ppc;
double op_rate;
int mflag = 1;
int gdir_count;
int gfile_count;
int gempty_count;
int my_ops, my_warmup_ops;
char var_opt[256];

/* 
 * Used for latency statistics
 */
double my_time_res = 9999999.99;
double cur_time,my_run_time;
double my_start_time, post_time;
double my_start_time2;
int prev_heartbeat;


/* ELAP_T Collect this many seconds worth of samples */
#define ELAP_T 2.0

/* ISCALE helps with the clock resolution issues. */
/* Every ISCALE ops, inject delay to maintain requested op_rate */
#define ISCALE 50	/* always overridden */

/* Keep track of time spent napping */
double seconds_to_nap, total_seconds_napped;
int naps_taken;

int tick;
int twarn, wwarn;

char my_tmp_workload_name[MAXNAME];

/*
 * Time for the test to run on the client.
 */
int run_time = 300;  /* Default run time of 5 mins */
int warm_time =300; /* Default warmup time of 5 mins */

time_t dummytime;

/*
 * Name of the client where the controlling process can use
 * remsh/rsh to start the client test running.
 */
char client_name[MAXHNAME];
char client_localname[MAXHNAME];

/*
 * Name of the file that contains the list of clients.
 */
char client_filename[MAXNAME];
/*
 * Name of the token based file that contains the clients info.
 */
char client_tok_filename[MAXNAME];
/*
 * Name of the external monitor script, and its arguments.
 */
char u_external_script_name[MAXNAME];
char u_external_script_args[MAXNAME];

/*
 * Prototypes and forward declarations.
 */

#if defined(HAVE_ANSIC_C)
void generic_null(void);
void clear_stats(void);
void client_startup(void);
void launch_clients(void);
void read_clientfile(void);
void read_client_tok_file(void);
void usage(void);
void create_kill_file(void);
void remove_kill_file(void);
void use_kill_file(void);
void kill_clients(void);
void init_tests(int,int);
void report_stats(void);
void create_tests(int,int);
int gen_op(int,int);
void client_init_files(int);
void client_validate_ops(void);
int my_gethostname(char *, int);
void client_remove_files(int);
void generic_write(struct file_object *);
void generic_write_file(struct file_object *);
void generic_write_rand(struct file_object *);
void generic_mmap_write(struct file_object *);
void generic_read(struct file_object *);
void generic_read_file(struct file_object *);
void generic_mmap_read(struct file_object *);
void generic_read_rand(struct file_object *);
void generic_rmw(struct file_object *);
void generic_stat(struct file_object *);
void generic_copyfile(struct file_object *);
void generic_rename(struct file_object *);
void generic_statfs(struct file_object *);
void generic_pathconf(struct file_object *);
void generic_custom1(struct file_object *);
void generic_custom2(struct file_object *);
void generic_readdir(struct file_object *);
void generic_chmod(struct file_object *);
void generic_access(struct file_object *);
void generic_mkdir(struct file_object *);
void generic_append(struct file_object *);
void generic_unlink(struct file_object *, int);
void generic_create(struct file_object *, int);
void generic_locking(struct file_object *);
void generic_prefix_op(int,struct file_object *);
void generic_postfix_op(int,struct file_object *);
struct file_object *generic_make_path(int,int);
int lock_file(char *, struct file_object *);
#ifdef WIN32
int try_lock_file_hold(HANDLE);
#else
int try_lock_file_hold(int);
void bzero(void *, size_t );
#endif
int master_wait(int);
void distribute(void);
void purge_cache(void);
void dump_table(void);
void import_table(char *);
int import_dist_table(struct client_command *);
void export_dist_table(struct client_command *);
void nap(int);
void make_non_dedup(long *, int);
#else
void nap();
void bzero();
void clear_stats();
void client_startup();
void launch_clients();
void read_clientfile();
void read_client_tok_file();
void usage();
void create_kill_file();
void remove_kill_file();
void use_kill_file();
void kill_clients();
void init_tests();
void report_stats();
void create_tests();
int gen_op();
void client_init_files();
void client_validate_ops();
int my_gethostname();
void client_remove_files();
void generic_write();
void generic_write_file();
void generic_mmap_write();
void generic_read();
void generic_read_file();
void generic_mmap_read();
void generic_rmw();
void generic_stat();
void generic_copyfile();
void generic_rename();
void generic_statfs();
void generic_pathconf();
void generic_custom1();
void generic_custom2();
void generic_readdir();
void generic_chmod();
void generic_access();
void generic_mkdir();
void generic_append();
void generic_unlink();
void generic_create();
void generic_locking();
struct file_object *generic_make_path();
void generic_prefix_op();
void generic_postfix_op();
int lock_file();
int try_lock_file_hold();
int master_wait_barrier();
int master_wait_results();
void distribute();
void purge_cache();
void dump_table();
void import_table();
void export_dist_table();
int import_dist_table();
void make_non_dedup(long *, int);
#endif

/* This is different for FULL versus "lite" versions. */

#include "./import.c"

/*
 * Entry point where the controlling process starts the
 * universe. The clients also enter here but take a 
 * different path to their destiny.
 */
int
main(int argc, char **argv)
{
	int mret;
	long long fsize_per_client;
	long long total_set_size;
	int picked;
#if defined(RLIMIT_NPROC)
	struct rlimit rlimit_struct;
#endif
	int cret;
	/*
         * Handy signal handler used to cleanup when things go wrong.
	 */
	signal(SIGINT,clean_exit);
	strcpy(workdir,HOME);
	if(argc==1)
	{
		usage();
		exit(0);
	}
#if !defined(WIN32)
	/*
 	 * If any process gets a broken pipe (on its sockets), then log and terminate.
	 * This should **NEVER** happen, but it can if the operating system
 	 * on the clients starts closing sockets with software resets....
	 */
	signal(SIGPIPE,pipe_elog);
#endif

	/* Setup default */
#if !defined(WIN32)
  #if defined(RLIMIT_NPROC)
    mret=getrlimit(RLIMIT_NPROC,&rlimit_struct);
    max_proc = rlimit_struct.rlim_max;
    if(max_proc == (int) RLIM_INFINITY)
    {
        max_proc = MAXCLIENTS;
    }
  #endif
    if(max_proc <= 0)
	max_proc=sysconf(_SC_CHILD_MAX);

    if(max_proc <= 0)
    {
  #if defined(CHILD_MAX)
       max_proc = CHILD_MAX;
  #endif
    }
    if(max_proc <= 0)
    {
        printf("Unable to obtain MAX_CHILD.. Max_child = %d Fatal\n",max_proc);
        fflush(stdout);
        exit(1);
    }
    max_open = sysconf(_SC_OPEN_MAX);

#else
	/* WINDOWS HERE: 
	 * There is no hard-set limit for max. open and max. processes 
	 * on Windows
	 */
	max_open = WIN32_MAX_OPEN;
	max_proc = WIN32_MAX_PROC;
#endif

#ifdef MERSENNE
	/* Initialize the 64 bit random number generator */
	init_by_array64(init, length);
#endif

#if defined(WIN32)
	strcpy(client_log_dir,"C:\\tmp");
	win32_init();
#else
	strcpy(client_log_dir,"/tmp");
#endif

	/* Init file state variables */
	init_vars();

	strcpy(pit_service,"PIT"); /* Default */

	while((cret = getopt(argc,argv,"zNKYXFGDLvkimhux2c:o:O:Q:T:H:l:S:n:r:s:p:C:e:g:f:j:t:M:R:U:b:B:d:P:w:W:EI:a:y:J:V:A:Z:q: ")) != EOF){
		switch(cret){
		case 'F': /* Disable flushes */
			flush_flag = 0;
			break;
		case 'x': /* Enable sockdebug */
			sockdebug = 1;
			break;
		case 'G': /* enable heartbeat messages */
			heartbeat_flag = 1;
			break;
		case 'N': /* enable op_latency collection */
			Op_lat_flag = 1;
			break;
		case 'J': /* limit file descriptor caching to this many */
			fd_caching_limit=atoi(optarg);
			break;
		case 'Q': /* Files per dir */
			Q_flag = 1;
			client_files=atoi(optarg);
			if(client_files < 1)
			   client_files = 1;
			re_init_vars();
			fpd_flag++;
			break;
		case 'T': /* Number of dirs */
			T_flag=1;
			client_dirs=atoi(optarg);
			if(client_dirs < 1)
			   client_dirs = 1;
			re_init_vars();
			break;
		case 'c': /* Total number of clients */
			num_clients=atoi(optarg);
			cflag++;
			break;
		case 'n': /* Unique client id */
			client_id=atoi(optarg);
			real_client_id = client_id;
			nflag++;
			break;
		case 'a': /* Gary & George offset */
			gg_offset=atoi(optarg);
			gg_flag++;
			break;
		case 'r': /* Manual over-ride of calculated fsize */
			picked = 0;
                        setfsize  = (atoi(optarg));
                        if(optarg[strlen(optarg)-1]=='k' ||
                                optarg[strlen(optarg)-1]=='K'){
                                setfsize  = (atoi(optarg));
				picked = 1;
                        }
                        if(optarg[strlen(optarg)-1]=='m' ||
                                optarg[strlen(optarg)-1]=='M'){
                                setfsize  = (atoi(optarg))*1024;
				picked = 1;
                        }
                        if(optarg[strlen(optarg)-1]=='g' ||
                                optarg[strlen(optarg)-1]=='G'){
                                setfsize  = (atoi(optarg))*1024*1024;
				picked = 1;
                        }
			if(mflag && !picked) /* If master, then has suffix, else in KB */
			{
				printf("-r # must have a suffix kKmMgG\n");
				exit(1);
			}
			if(setfsize != 0)
			   setfsflag++;
			break;
		case 's': /* File size to test */
			sscanf(optarg,"%lld",&fsize);
			re_init_vars();
			sflag++;
			break;
		case 'd': /* Procs per client */
			ppc=atoi(optarg);
			if(ppc==0)
			{
				printf("Proc count can not equal zero\n");
   				fflush(stdout);
				usage();
				exit(1);
			}
			dflag++;
			break;
		case 'u': /* Use a different remote shell */
			strcpy(remsh,"rsh");
			strcpy(remcp,"rcp");
			break;
		case 'p': /* Directory to do work */
			pflag++;
			strcpy(workdir,optarg);
			break;
		case 'q': /* Heartbeat log file name */
			log_beats++;
			strcpy(log_beat_filename,optarg);
			log_beat_fd = fopen(log_beat_filename,"a+");
			if(log_beat_fd == 0)
			{
				printf("Unable to open the heartbeat log %s\n",log_beat_filename);
				exit(1);
			}
			break;
		case 'z': /* Use sh to run locaally */
			localflag=1;
			break;
		case 'h': /* Help menu */
			usage();
			hflag++;
			exit(0);
		case 'C': /* Name of the client to run test on */
			Cflag++;
			strcpy(client_name,optarg);
			break;
		case 'e': /* Path to executable for the test */
			eflag++;
			strcpy(execdir,optarg);
			break;
		case 'm': /* Master/controlling process flag */
			mflag=0;
			break;
		case 'i': /* Show but do not run */
			iflag++;
			break;
		case 'g': /* Name of the token based file that contains the client config info */
			gflag++;
			strcpy(client_tok_filename,optarg);
			break;
		case 'f': /* Name of the external monitor script */
			p_ext_mon_flag++;
			strcpy(u_external_script_name,optarg);
			break;
		case 'j': /* Arguments to external monitor script */
			p_ext_mon_arg_flag++;
			strcpy(u_external_script_args,optarg);
			break;
		case 't': /* Timed test for this many seconds */
			run_time=atoi(optarg);
#if defined(FAST)
			if(run_time < 60)
			{
			   run_time = 60;
			   twarn=1;
			}
#else
			if(run_time < 300)
			{
			   run_time = 300;
			   twarn=1;
			}
#endif
			break;
		case 'M': /* Name of the master */
			Mflag++;
			strcpy(master_name,optarg);
			break;
		case 'A': /* Password */
			Password_flag++;
			strcpy(password,optarg);
			if(strcmp(password,"_")==0)
				Password_flag=0; /* Invalid password */
			strcpy(mypassword,password);
			break;
		case 'Z': /* User/domain */
			Usr_dom_flag++;
			strcpy(usr_dom,optarg);
			break;
		case 'y': /* Enable ipv6 */
			ipv6_enable = atoi(optarg);
			break;
		case 'U': /* Name of the child log directory */
			Uflag++;
			strcpy(client_log_dir,optarg);
			break;
		case 'R': /* Name of the masters results directory */
			Rflag++;
			strcpy(master_results_dir,optarg);
			break;
		case 'V': /* PID of child to kill */
			Vflag++;
			my_child_pid=atoi(optarg);
			break;
		case 'b': /* Memory size of clients in MiBytes */
			picked = 0;
			client_memsize_mb=atoi(optarg);
                        if(optarg[strlen(optarg)-1]=='m' ||
                                optarg[strlen(optarg)-1]=='M'){
                                client_memsize_mb = (long long)(atoi(optarg));
				picked = 1;
                        }
                        if(optarg[strlen(optarg)-1]=='g' ||
                                optarg[strlen(optarg)-1]=='G'){
                                client_memsize_mb = (long long)(1024 *(long long)atoi(optarg));
				picked = 1;
                        }
                        if(optarg[strlen(optarg)-1]=='t' ||
                                optarg[strlen(optarg)-1]=='T'){
                                client_memsize_mb = (long long)(1024 * 1024 * (long long)atoi(optarg));
				picked = 1;
                        }
			if(mflag && !picked)
			{
				printf("-b # must have a suffix mMbGtT\n");
				exit(1);
			}
			bflag++;
			break;
		case 'B': /* Aggregate data set size */
			picked = 0;
			agg_set_memsize_mb = atoi(optarg);
                        if(optarg[strlen(optarg)-1]=='m' ||
                                optarg[strlen(optarg)-1]=='M'){
                                agg_set_memsize_mb = (long long)(atoi(optarg));
				picked = 1;
                        }
                        if(optarg[strlen(optarg)-1]=='g' ||
                                optarg[strlen(optarg)-1]=='G'){
                                agg_set_memsize_mb = (long long)(1024 *(long long)atoi(optarg));
				picked = 1;
                        }
                        if(optarg[strlen(optarg)-1]=='t' ||
                                optarg[strlen(optarg)-1]=='T'){
                                agg_set_memsize_mb = (long long)(1024 * 1024 * (long long)atoi(optarg));
				picked = 1;
                        }
			if(mflag && !picked)
			{
				printf("-B # must have a suffix mMbGtT\n");
				exit(1);
			}
			Bflag++;
			break;
		case 'k': /* kill the tests on the clients and exit */
			kflag++;
			break;
		case 'v': /* Version info only */
			vflag++;
			break;
		case 'L': /* Local file system flag for debug */
			lflag++;
			strcpy(var_opt," -L ");
			break;
		case 'l': /* Log file */
			log_flag++;
			strcpy(log_filename,optarg);
			break;
		case 'D': /* Distribute the benchmark to the clients */
			Dflag++;
			break;
		case 'P': /* Port the master is listening on */
			Pflag++;
			master_p = atoi(optarg);
			break;
		case 'w': /* Warmup time */
			wflag++;
			warm_time = atoi(optarg);
			if(warm_time < 60)
			{
			   warm_time = 60;
			   wwarn=1;
			}
			break;
		case 'E': /* export current table */
			dump_table();
			exit(0);
			break;
		case 'I': /* Import current table */
			im_flag++;
			strcpy(itable,optarg);
			import_table(itable);
			break;
		case 'O': /* Op rate to test */
			sscanf(optarg,"%lf",&op_rate);
			orflag++;
			if(op_rate < 0.0)
			{
		 	  printf("Negative op rate not supported.\n");
   			  fflush(stdout);
			  exit(0);
			}
			break;
		case 'K': /* Cleanup flag */
			cleanup_flag=0;
			break;
		case 'Y': /* Dump files flag */
			dump_files_flag=1;
			break;
		case '2': /* Dump files flag */
			tracedebug=1;
			break;
		case 'X': /* Sharing flag */
			sharing_flag=1;
			break;
                case 'H':  /* Argument is hostname of the PIT */
                        strcpy(pit_hostname,optarg);
                        break;
                case 'W':  /* Argument is workload name */
			if (strlen(optarg) > MAXNAME)
				strncpy(my_tmp_workload_name,optarg,MAXNAME);
			else
			  	strcpy(my_tmp_workload_name,optarg);
                        break;

                case 'S':  /* Argument is port of the PIT */
                        strcpy(pit_service,optarg);
                        break;
		}
	}

	if(mflag) /* Only the master needs the client_obj array */
	{
		client_obj = (struct client_object *)MALLOC( sizeof(struct client_object) * (MAXCLIENTS+1));
		bzero(client_obj, sizeof(struct client_object) * (MAXCLIENTS+1));

		child_pid = (int *)MALLOC( sizeof(int) * (MAXCLIENTS+1));
		bzero(child_pid, sizeof(int) * (MAXCLIENTS+1));

		don_master_send_socket_name = (char *)MALLOC( sizeof(char) * (MAXCLIENTS+1) * MAXHNAME);
		bzero(don_master_send_socket_name, sizeof(char) * (MAXCLIENTS+1) * MAXHNAME);

		don_master_send_socket_port = (int *) MALLOC( sizeof(int) *  (MAXCLIENTS+1));
		bzero(don_master_send_socket_port, sizeof(int) * (MAXCLIENTS+1));
	}

	if(sflag && setfsflag) /* Use the manual over-ride. */
		fsize = setfsize;

	if(Dflag && (fflag||gflag)) /* Distribute the benchmark to the clients */
	{
		if(fflag)
			read_clientfile();
		else
			read_client_tok_file(); /*it should also allow token file*/
		distribute();
		exit(0);
	}
	/*
	 * Print version and exit
	 */
	if(vflag)
	{
		printf("%s %s\n",short_testname,version);
   		fflush(stdout);
		exit(0);
	}

	/* 
	 * Require ops, or client_memsize and  client count or 
	 * client_filename and Aggregate data_set size size.
 	 */
	if(!(cflag||fflag||gflag))
	{
		printf("Missing configuration file name. (-g)\n");
   		fflush(stdout);
		usage();
		exit(0);
	}
	/*
	 * Master and no aggregate data set size.. not permitted 
	 */
	if( (mflag && (!Bflag)) )
	{
		printf("Missing minimum aggregate data set size. (-B) \n");
   		fflush(stdout);
		usage();
		exit(0);
	}
	/*
	 * Must have procs per client.
	 */
	if( mflag && !dflag )
	{
		printf("Missing procs per client. (-d)\n");
   		fflush(stdout);
		usage();
		exit(0);
	}
	/*
	 * Must have memory size per client.
	 */
	if( (mflag && (!bflag)) )
	{
		printf("Missing client memory size. (-b) \n");
   		fflush(stdout);
		usage();
		exit(0);
	}
	/*
	 * Master reads the client file to get count.
	 */
	if(mflag)
	{
		if(fflag)
			read_clientfile();
		if(gflag)
			read_client_tok_file();
	}
	if(num_clients <= 0)
	{
		printf("Something is wrong with the configuration file. When parsed, it\n");
		printf("appears that there are zero total client entries.\n");
   		fflush(stdout);
		exit(1);
	}
	if(!licensed && im_flag)
	{
		printf("Import of workload objects not permitted in unlicensed version\n");
   		fflush(stdout);
		exit(1);
	}

	/* Init file object space */
	init_space();

	/* 
	 * If more than one proc per client then adjust memory size 
	 */
	if(mflag && dflag) 
	{
		client_memsize_mb=((client_memsize_mb)*(num_clients/ppc));
	}

	/*
 	 * Re-initialize the state variables, after any size adjustments.
	 */
	re_init_vars();	

	/*
	 * p_num_clients is used to have all files be shared by 
 	 * all procs, on all clients. By using p_num_clients all 
 	 * of the files appear to be created, and accessed, by
	 * a single client ID. (client id = 0) and no scaling 
	 * is done because all files are shared by everyone. 
	 */
	if(sharing_flag)
		p_num_clients = 1;
	else
		p_num_clients = num_clients;

	/* 
	 * Scale the file size to fit the client memory 
	 */
	if((bflag && mflag)  && !setfsflag)
	{
		fsize1=(long long) (((long long)(client_memsize_mb*(long long)1024 ))/ (client_files_rpop * p_num_clients))*
			(long long)1;
		fsize2=(long long) (((long long)(agg_set_memsize_mb*(long long)1024 ))/ (client_files_rpop * p_num_clients))*
			(long long)1;
		if(fsize1 > fsize2)
			fsize=fsize1;
		else
			fsize=fsize2;
		fsize=(fsize +((rsize)-1)) & ~((rsize)-1);
	}

	if(setfsflag && (setfsize!=0))
		fsize=(setfsize +((rsize)-1)) & ~((rsize)-1);
	/*
	 * Round up to at least one transfer
	 */
	if(mflag && (fsize < rsize))
		fsize=rsize;

	/*
	 * If killing test and child then use the kill file to send a signal
	 * to the test running on this client. This signal tells the client
	 * to do a clean_exit and exit.
	 *
	 * If I'm the master then I need to send the kill command to the 
	 * clients.
	 */
	if(kflag)
	{
		if(!mflag)
			use_kill_file();
		else
			kill_clients();
		exit(0);
	}
	
	/*
	 * If I am a client then I need to start the test.
	 */
	if(!mflag)
	{
		client_startup();
		exit(0);
	}

	/*
	 * If I am here, then I am the master/controlling instance.
	 */
	if(log_flag)
	{
	   log_out=fopen(log_filename,"a+");
	   if(log_out==0)
	   {
		printf("%s Unable to open log file %s\n",get_timestamp(&time_string[0]),log_filename);
   		fflush(stdout);
		exit(0);
	   }
	}

	if(flush_flag == 0)
	{
	    printf("**** Flushes are disabled. ****\n");
   	    fflush(stdout);
	    if(log_flag)
	    {
	       fprintf(log_out,"**** Flushes are disabled. ****\n");
   	       fflush(log_out);
	    }

	}
	
	if(twarn)
	{
	    printf("**** Minimum run time of 300 enforced. ****\n");
   	    fflush(stdout);
	    if(log_flag)
	    {
	       fprintf(log_out,"**** Minimum run time of 300 enforced. ****\n");
   	       fflush(log_out);
	    }
	}
	if(wwarn)
	{
	    printf("**** Minimum warm-up time of 300 enforced. ****\n");
   	    fflush(stdout);
	    if(log_flag)
	    {
	       fprintf(log_out,"**** Minimum warm-up time of 300 enforced. ****\n");
   	       fflush(log_out);
	    }
	}
	if(log_flag)
	   fflush(log_out);
	
	/*
 	 * Can't have both. Pick file with space delimited info, or token based.
	 */
	if( fflag && gflag )
	{
		usage();
		exit(0);
	}
	/*
	 * At this point, I am the controlling process for the benchmark and 
	 * will coordinate the testing.
  	 */
	if(!(fflag||gflag))
	{
		usage();
		exit(0);
	}

	if(power_mon(SCHECK,"Baseline")==0)
	   power_mon(SRUN,"Baseline");

	/*
	 * Save the master's current working directory. This is where the 
	 * clients will deposit their results.
	 */
	if( !Rflag )
	   getcwd(master_results_dir,MAXNAME);

	/*
	 * Check to see if master's results directory exists, and is writable.
	 * If not, then try to create it now...
	 */
	if(mflag && (check_dir_w(master_results_dir) != 0) )
	{
		/* If it doesn't exist, try creating it */
		cret = mkdir(master_results_dir,0777);
		if(cret < 0)
		{
 		   printf("%s: ",get_timestamp(&time_string[0]));
		   printf("Unable to create master results directory > %s <  %s\n",
			master_results_dir,strerror(errno));
   	           fflush(stdout);
		   exit(1);
		}
	}

	/*
	 * Setup the generation number. Generation numbers are used to protect
	 * the messaging system. It prevents previous runs from corruptin the
	 * the messages from this run.
	 */
	netmist_srand((int)time(0));
	generation = netmist_rand();

	/*
	 * Save the master's name. The clients use this to  deposit their 
	 * results.
	 */
	while(1)
	{
	   mret = gethostname(master_name,MAXHNAME);
	   if(mret == 0)
		break;
	   perror("gethostname failure");
	}
		

	/* Now converto to IP address, just in case the client's resolve
 	 * is not setup correctly, or the client is not in the same DNS
	 * domain.
	 *
	
	 *** THIS CAUSES PROBLEMS IN ONE OF THE LABS !!! Revert back
	     to using hostnames and not IP addresses. In one lab
	     the IP is coming back in the 169.*** range. Bad news.

	get_host_ip_address(master_name, tmp_buf);
	strcpy(master_name,tmp_buf);
	*/

	/*
	 * Calculate the current aggregate data set size.  At this point 
	 * it should be large enough to use all client memory but may 
 	 * not be big enough to consume the specified aggregate data set size.
	 */
	if( !sharing_flag)
	{
	   total_set_size= 
		(long long)(fsize*(client_dirs_pop)*(((double)num_clients/(double)ppc)*(double)ppc));

	   /*
	    * Calculate the total file system space that can be used.
	    */
	   total_space= 
		(long long)(fsize*(client_dirs_pop)* (((double)num_clients/(double)ppc) * (double)ppc));

	   total_max_space= 
		(long long)(fsize*(max_client_dirs_pop)* (((double)num_clients/(double)ppc) * (double)ppc));
	}
	else
	{
	   total_set_size= fsize*client_dirs_pop;

	   /*
	    * Calculate the total file system space that can be used.
	    */
	   total_space= fsize*client_dirs_pop;

	   total_max_space= fsize*max_client_dirs_pop;
	}

	/*
	 * If the aggregate data set size is not big enough to consume
	 * all of the specified agg data set then make another adjustment on the
	 * file size.
	 */
	if( !setfsflag && (total_set_size < (agg_set_memsize_mb*(long long)1024)))
	{
		fsize=(((long long)((agg_set_memsize_mb*(long long)1024))/(client_files_rpop*p_num_clients))*
			(long long)1);
		fsize=(fsize +((rsize)-1)) & ~((rsize)-1);
		if(xdebug) 
			printf("Data set size = %lld kbytes, Adjustable aggregate Data set value = %lld MiBytes\n", total_set_size, agg_set_memsize_mb/(long long)1024);
		if(xdebug) 
			printf("Resetting fsize to %lld Kbytes\n",fsize);
		if( !sharing_flag)
		{
		   total_space= (long long)(fsize*(client_dirs_pop)* (((double)num_clients/(double)ppc) * (double)ppc));
		   total_max_space= (long long)(fsize*(max_client_dirs_pop)* (((double)num_clients/(double)ppc) * (double)ppc));
		}
		else
		{
		   total_space= fsize*client_dirs_pop;
		   total_max_space= fsize*max_client_dirs_pop;
		}
	}

#if defined(LITE)
	printf("\n");
	printf("    *************************************************************************\n");
	printf("    |     %-15s Benchmark         %24s        |\n",testname,version);
	printf("    |                                    by                                 |\n");
	printf("    |     Author:       %-50s  |\n",Author);
	printf("    |     Maintainer:   %-50s  |\n",Maintainer);
	printf("    |     Contributors: %-50s  |\n",Contributors);
	printf("    |                   %-50s  |\n",Contributors2);
	printf("    |                   %-50s  |\n",Contributors3);
	printf("    |                   %-50s  |\n",Contributors4);
	printf("    |                   %-50s  |\n",Contributors5);
	printf("    *************************************************************************\n");
	printf("\n");
	printf("     %s %s\n\n",short_testname,version);
   	fflush(stdout);
#endif
#if defined(PRO)
	printf("\n");
        printf("     SPECsfs2014 Release %24s\n",version);
	printf("\n");
	printf("     This product contains benchmarks acquired from several sources who\n");
	printf("     understand and agree with SPEC's goal of creating fair and objective\n");
	printf("     benchmarks to measure computer performance.\n");
	printf("\n");
	printf("     This copyright notice is placed here only to protect SPEC in the\n");
	printf("     event the source is misused in any manner that is contrary to the\n");
	printf("     spirit, the goals and the intent of SPEC.\n");
	printf("\n");
	printf("     The source code is provided to the user or company under the license\n");
	printf("     agreement for the SPEC Benchmark Suite for this product.\n");
	printf("\n");
   	fflush(stdout);
#endif

	if(log_flag)
	{
#if defined(LITE)
	  fprintf(log_out,"\n");
	  fprintf(log_out,"    *************************************************************************\n");
	  fprintf(log_out,"    |     %-15s Benchmark         %24s        |\n",testname,version);
	  fprintf(log_out,"    |                                    by                                 |\n");
	  fprintf(log_out,"    |     Author:       %-50s  |\n",Author);
	  fprintf(log_out,"    |     Maintainer:   %-50s  |\n",Maintainer);
	  fprintf(log_out,"    |     Contributors: %-50s  |\n",Contributors);
	  fprintf(log_out,"    |                   %-50s  |\n",Contributors2);
	  fprintf(log_out,"    |                   %-50s  |\n",Contributors3);
	  fprintf(log_out,"    |                   %-50s  |\n",Contributors4);
	  fprintf(log_out,"    |                   %-50s  |\n",Contributors5);
	  fprintf(log_out,"    *************************************************************************\n");
	  fprintf(log_out,"\n");
	  fprintf(log_out,"     %s %s\n\n",short_testname,version);
	  fflush(log_out);
#endif
#if defined(PRO)
	  fprintf(log_out,"\n");
          fprintf(log_out,"     SPECsfs2014 Release %24s\n",version);
	  fprintf(log_out,"\n");
	  fprintf(log_out,"     This product contains benchmarks acquired from several sources who\n");
	  fprintf(log_out,"     understand and agree with SPEC's goal of creating fair and objective\n");
	  fprintf(log_out,"     benchmarks to measure computer performance.\n");
	  fprintf(log_out,"\n");
	  fprintf(log_out,"     This copyright notice is placed here only to protect SPEC in the\n");
	  fprintf(log_out,"     event the source is misused in any manner that is contrary to the\n");
	  fprintf(log_out,"     spirit, the goals and the intent of SPEC.\n");
	  fprintf(log_out,"\n");
	  fprintf(log_out,"     The source code is provided to the user or company under the license\n");
	  fprintf(log_out,"     agreement for the SPEC Benchmark Suite for this product.\n");
	  fprintf(log_out,"\n");
	  fflush(log_out);
#endif
	}
	
	printf("     Test run time = %d seconds, Warmup = %d seconds.\n",run_time,warm_time);
   	fflush(stdout);
	if(log_flag)
	  fprintf(log_out,"     Test run time = %d seconds, Warmup = %d seconds.\n",run_time,warm_time);
	if(orflag)
	{
		printf("     Request Op_rate = %.2f ops/second\n",op_rate);
		if(log_flag)
		  fprintf(log_out,"     Request Op_rate = %.2f ops/second\n",op_rate);
	}

	printf("     Running %d copies of the test on %d clients\n",num_clients,num_clients/ppc);
	printf("     Results directory: %s\n",master_results_dir);
   	fflush(stdout);
	if(Op_lat_flag)
	{
	   printf("     Op latency reporting activated\n");
	   if(log_flag)
	   	fprintf(log_out,"     Op latency reporting activated\n");
	}
	if(Q_flag)
	{
	   printf("     Files per directory set to %d\n",client_files);
	   if(log_flag)
	   	fprintf(log_out,"     Files per directory set to %d\n",client_files);
	}
	if(T_flag)
	{
	   printf("     Directories per proc set to %d\n",client_dirs);
	   if(log_flag)
	   	fprintf(log_out,"     Directories per proc set to %d\n",client_dirs);
	}
	if(gg_flag)
	{
	   printf("     %d byte offset added to all xfers\n",gg_offset);
	   if(log_flag)
	   	fprintf(log_out,"     %d byte offset added to all xfers\n",
			gg_offset);
	}
	if(im_flag)
	{
	   printf("     Using custom workload\n");
	   if(log_flag)
	   	fprintf(log_out,"     Using custom workload\n");
	}
	if(setfsflag)
	{
	   printf("     Using custom file size of %d Kbytes\n",setfsize);
	   if(log_flag)
	   	fprintf(log_out,"     Using custom file size %d Kbytes\n",setfsize);
	}
	if(log_flag)
	{
	  fprintf(log_out,"     Running %d copies of the test on %d clients\n",num_clients,num_clients/ppc);
	  fprintf(log_out,"     Results directory: %s\n",master_results_dir);
	}
	if(bflag)
	{
		if(dflag)
		{
		   printf("     Clients have a total of %lld MiBytes of memory\n",client_memsize_mb);
		   printf("     Clients have %lld MiBytes of memory size per process\n",client_memsize_mb/num_clients);
		   if(log_flag)
		   {
		     fprintf(log_out,"     Clients have a total of %lld MiBytes of memory\n",client_memsize_mb);
		     fprintf(log_out,"     Clients have %lld MiBytes of memory size per process\n",client_memsize_mb/num_clients);
		   }
		}
		else
		{
		   printf("     Clients have %lld MiBytes of memory\n",client_memsize_mb);
		   if(log_flag)
		     fprintf(log_out,"     Clients have %lld MiBytes of memory\n",client_memsize_mb);
		}
	}
	if(dflag)
	{
		printf("     Clients each have %d processes\n",ppc);
		if(log_flag)
		  fprintf(log_out,"     Clients each have %d processes\n",ppc);
	}
	if(Bflag)
	{
		printf("     Adjustable aggregate data set value set to %lld MiBytes\n",agg_set_memsize_mb);
		if(log_flag)
		{
		  fprintf(log_out,"     Adjustable aggregate data set value set to %lld MiBytes\n",agg_set_memsize_mb);
		}
	}
	printf("     Each process file size = %lld kbytes\n",fsize);
   	fflush(stdout);
	if(log_flag)
	   fprintf(log_out,"     Each process file size = %lld kbytes\n",fsize);
	if(sharing_flag)
	   fsize_per_client = (fsize*client_dirs_pop);
	else
	   fsize_per_client = (fsize*client_dirs_pop*ppc);
	printf("     Client data set size       = %lld MiBytes\n", fsize_per_client/(long long)1024);
	if(sharing_flag)
	   printf("     Total starting data set size = %lld MiBytes\n",
		(fsize_per_client)/(long long)1024);
	else
	   printf("     Total starting data set size = %lld MiBytes\n",
		((num_clients/ppc) * fsize_per_client)/(long long)1024);
	printf("     Total initial file space   = %lld MiBytes\n",total_space/(long long)1024);
	printf("     Total max file space       = %lld MiBytes\n",total_max_space/(long long)1024);
	
	printf("\n");
   	fflush(stdout);
	if(log_flag)
	{
	   fprintf(log_out,"     Client data set size       = %lld MiBytes\n", fsize_per_client/(long long)1024);
	   fprintf(log_out,"     Total starting data set size = %lld MiBytes\n",
		(long long)((num_clients/ppc) * fsize_per_client)/(long long)1024);
	   fprintf(log_out,"     Total initial file space   = %lld MiBytes\n",total_space/(long long)1024);
	   fprintf(log_out,"     Total max file space       = %lld MiBytes\n",total_max_space/(long long)1024);
	
	   fprintf(log_out,"\n");
	   fflush(log_out);
	}
	if(iflag)
		exit(0);
	printf("     Starting tests: %s \n\n",get_timestamp(&time_string[0]));
	fflush(stdout);
	if(log_flag)
	{
	   fprintf(log_out,"     Starting tests: %s \n\n",get_timestamp(&time_string[0]));
	   fflush(log_out);
	}
	start_master_init(); 		/* Start comm channel  */
	launch_clients();		/* Start the clients   */
	if(mflag)
	   master_start_comm(num_clients); /* Finish comm linkage */

	/* 
 	 * Let the external monitor know that we are entering the INIT phase.
	 */
	if(extern_mon(SCHECK,"INIT")==0)
	   extern_mon(SRUN,"INIT");

	/* 
 	 * Let the user defined external monitor know that we are 
	 * entering the INIT phase.
	 */
	if(u_extern_mon(SCHECK,"INIT")==0)
	   u_extern_mon(SRUN,"INIT");

	printf("\tWaiting to finish initialization. %s\n", get_timestamp(&time_string[0]));
	fflush(stdout);
	if(log_flag)
	{
	   fprintf(log_out,"\tWaiting to finish initialization. %s\n", get_timestamp(&time_string[0]));
	   fflush(log_out);
	}

	mret=master_wait_barrier(num_clients); /* Wait for ready or error */
	if(mret)
	{
		printf("Client initialization failed.\n");
		fflush(stdout);
		if(log_flag)
		{
		  fprintf(log_out,"Client initialization failed.\n");
		  fflush(log_out);
		}
		clean_exit(0);
		exit(100);
	}
	sleep(1); 	/* Let children get into their child_wait() */
	printf("\tInitialization finished: %s\n", get_timestamp(&time_string[0]));
	printf("\tTesting begins: %s\n",get_timestamp(&time_string[0]));
	fflush(stdout);
	if(log_flag)
	{
	   fprintf(log_out,"\tInitialization finished: %s\n", get_timestamp(&time_string[0]));
	   fprintf(log_out,"\tTesting begins: %s\n",get_timestamp(&time_string[0]));
	   fflush(log_out);
	}
	/* Let the consumer defined monitor script know we're entering RUN */
	if(extern_mon(SCHECK,"WARMUP")==0)
	   extern_mon(SRUN,"WARMUP");

	/* Let the user defined monitor script know we're entering RUN */
	if(u_extern_mon(SCHECK,"WARMUP")==0)
	   u_extern_mon(SRUN,"WARMUP");

	/* Consumer defined stat script. This is so vendor can zero counters */
	if(extern_stat(SCHECK,"ZERO")==0)
	   extern_stat(SRUN,"ZERO");
	/*
	 * Tell the clients to GO
	 */
	master_send_go(num_clients); /* Start testing */
	printf("\tWaiting for tests to finish. %s\n",get_timestamp(&time_string[0]));
	fflush(stdout);
	if(log_flag)
	{
	   fprintf(log_out,"\tWaiting for tests to finish. %s\n",get_timestamp(&time_string[0]));
	   fflush(log_out);
	}
	/*  sleep(2);  */
	mret=master_wait_results(num_clients); /* Wait for done (their done) */
	if(mret)
	{
		printf("Client testing failed: %s\n",get_timestamp(&time_string[0]));
		fflush(stdout);
		if(log_flag)
		{
		   fprintf(log_out,"Client testing failed: %s\n",get_timestamp(&time_string[0]));
	           fflush(log_out);
		}
		clean_exit(101);
		exit(101);
	}
	printf("\tTests finished: %s %c\n",get_timestamp(&time_string[0]),(char)7);
	fflush(stdout);
	if(log_flag)
	{
	  fprintf(log_out,"\tTests finished: %s \n",get_timestamp(&time_string[0]));
	  fflush(log_out);
	}
	/* Consumer defined monitor script */
	if(extern_mon(SCHECK,"STOP")==0)
	   extern_mon(SRUN,"STOP");

	/* User defined monitor script */
	if(u_extern_mon(SCHECK,"STOP")==0)
	   u_extern_mon(SRUN,"STOP");

	/* Consumer defined stat script. This is so vendor can collect stats */
	if(extern_stat(SCHECK,"STOP")==0)
	   extern_stat(SRUN,"STOP");
	master_stop_comm(num_clients);

#if defined(WIN32)
	/* stop network */
	win32_close();
#endif
	print_results();
	sleep(1);
	exit(0);		
}

void
print_results(void)
{
	struct result_object *here;
	double total_latency = 0.0;
	double total_ops_per_second = 0.0;
	double average_latency,ops_per_second;
	double total_read_throughput = 0.0;
	double total_write_throughput = 0.0;
	double total_throughput = 0.0;
	double count = 0.0;
	unsigned short sum_val = 0;
	int u,i;
	double skip = 0.0;
	char ch_filename[400];
	FILE *ch_out;

	for(i=0;i<num_clients;i++)
	{
	   here=&client_obj[i].results;

#if defined(WIN32)
	   sprintf(ch_filename,"%s\\%s%d_results",master_results_dir,"Client_",here->client_id);
#else
	   sprintf(ch_filename,"%s/%s%d_results",master_results_dir,"Client_",here->client_id);
#endif
	   ch_out=fopen(ch_filename,"a+");
	   if(ch_out < 0)
	   {
 		printf("%s: ",get_timestamp(&time_string[0]));
		printf("Error opening children logs: %s\n",strerror(errno));
   	        fflush(stdout);
		exit(1);
	   }
	   
	   fprintf(ch_out,"--------------------------------------------------------------------\n");
	   fprintf(ch_out,"Client %s    ID:%d\n",here->client_name, here->client_id);
	   fprintf(ch_out,"Workload Name    : %s\n",here->work_obj_name);
	   fprintf(ch_out,"Run time           %ld\n",here->run_time);
	   fprintf(ch_out,"Ops/sec            %10.2f\n",here->ops_per_second);
	   fprintf(ch_out,"Avg Latency        %7.3f\n",here->average_latency);
	   fprintf(ch_out,"Total file ops     %10.0f\n",here->total_file_ops);
	   fprintf(ch_out,"Read throughput    %10.3f Kbytes/sec\n",here->read_throughput/1024);
	   fprintf(ch_out,"Write throughput   %10.3f Kbytes/sec\n",here->write_throughput/1024);
	   fprintf(ch_out,"Min Direct I/O size = %ld\n",here->min_direct_size);

	   /* Write */
	   fprintf(ch_out,"\nDetails\n");
	   /*
            * Here are the outputs for all of the standard workloads. 
	    */

	   fprintf(ch_out,"\t%-35s ops = %10ld",here->write_string, here->write_count);
	   if(Op_lat_flag && (here->write_count != 0) && (here->write_time != 0.0))
	           fprintf(ch_out," \tAvg Latency: %10.6f\n",here->write_time/(double)here->write_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* Write file*/
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->write_file_string,here->write_file_count);
	   if(Op_lat_flag && (here->write_file_count != 0) && (here->write_file_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->write_file_time/(double)here->write_file_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* mmap write */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->mmap_write_string,here->mmap_write_count);
	   if(Op_lat_flag && (here->mmap_write_count != 0) && (here->mmap_write_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->mmap_write_time/(double)here->mmap_write_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* mmap read */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->mmap_read_string,here->mmap_read_count);
	   if(Op_lat_flag && (here->mmap_read_count != 0) && (here->mmap_read_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->mmap_read_time/(double)here->mmap_read_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* read */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->read_string,here->read_count);
	   if(Op_lat_flag && (here->read_count != 0) && (here->read_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->read_time/(double)here->read_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* read file */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->read_file_string,here->read_file_count);
	   if(Op_lat_flag && (here->read_file_count != 0) && (here->read_file_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->read_file_time/(double)here->read_file_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* mkdir */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->mkdir_string,here->mkdir_count);
	   if(Op_lat_flag && (here->mkdir_count != 0) && (here->mkdir_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->mkdir_time/(double)here->mkdir_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* unlink */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->unlink_string,here->unlink_count);
	   if(Op_lat_flag && (here->unlink_count != 0) && (here->unlink_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->unlink_time/(double)here->unlink_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* create */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->create_string,here->create_count);
	   if(Op_lat_flag && (here->create_count != 0) && (here->create_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->create_time/(double)here->create_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* stat */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->stat_string,here->stat_count);
	   if(Op_lat_flag && (here->stat_count != 0) && (here->stat_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->stat_time/(double)here->stat_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* append */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->append_string, here->append_count);
	   if(Op_lat_flag && (here->append_count != 0) && (here->append_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->append_time/(double)here->append_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* Lock */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->lock_string,here->lock_count);
	   if(Op_lat_flag && (here->lock_count != 0) && (here->lock_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->lock_time/(double)here->lock_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* append */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->access_string,here->access_count);
	   if(Op_lat_flag && (here->access_count != 0) && (here->access_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->access_time/(double)here->access_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* chmod */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->chmod_string,here->chmod_count);
	   if(Op_lat_flag && (here->chmod_count != 0) && (here->chmod_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->chmod_time/(double)here->chmod_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* readdir */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->readdir_string,here->readdir_count);
	   if(Op_lat_flag && (here->readdir_count != 0) && (here->readdir_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->readdir_time/(double)here->readdir_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* write rand */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->write_rand_string,here->write_rand_count);
	   if(Op_lat_flag && (here->write_rand_count != 0) && (here->write_rand_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->write_rand_time/(double)here->write_rand_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* Read rand */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->read_rand_string,here->read_rand_count);
	   if(Op_lat_flag && (here->read_rand_count != 0) && (here->read_rand_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->read_rand_time/(double)here->read_rand_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* rmw */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->rmw_string,here->rmw_count);
	   if(Op_lat_flag && (here->rmw_count != 0) && (here->rmw_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->rmw_time/(double)here->rmw_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   fprintf(ch_out,"\t%-35s ops = %10ld","Open file", here->open_count);
	   fprintf(ch_out,"\tAvg Latency:   Not collected\n");

	   fprintf(ch_out,"\t%-35s ops = %10ld","Close file", here->close_count);
	   fprintf(ch_out,"\tAvg Latency:   Not collected\n");

	   /* Copyfile */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->copyfile_string,here->copyfile_count);
	   if(Op_lat_flag && (here->copyfile_count != 0) && (here->copyfile_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->copyfile_time/(double)here->copyfile_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* Rename */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->rename_string, here->rename_count);
	   if(Op_lat_flag && (here->rename_count != 0) && (here->rename_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->rename_time/(double)here->rename_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* Statfs */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->statfs_string, here->statfs_count);
	   if(Op_lat_flag && (here->statfs_count != 0) && (here->statfs_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->statfs_time/(double)here->statfs_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* Pathconf */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->pathconf_string, here->pathconf_count);
	   if(Op_lat_flag && (here->pathconf_count != 0) && (here->pathconf_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->pathconf_time/(double)here->pathconf_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* custom1 */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->custom1_string, here->custom1_count);
	   if(Op_lat_flag && (here->custom1_count != 0) && (here->custom1_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->custom1_time/(double)here->custom1_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");

	   /* custom2 */
	   fprintf(ch_out,"\t%-35s ops = %10ld",here->custom2_string, here->custom2_count);
	   if(Op_lat_flag && (here->custom2_count != 0) && (here->custom2_time != 0.0))
	   	   fprintf(ch_out," \tAvg Latency: %10.6f\n",here->custom2_time/(double)here->custom2_count);
	   else
	   	   fprintf(ch_out," \tAvg Latency:   Not collected\n");


	   fprintf(ch_out,"\n");
	   if(Op_lat_flag)
	   {
	      master_dump_hist(here,ch_out);
	      fprintf(ch_out,"\n");
	   }

	   fflush(ch_out);
	   fclose(ch_out);

	   if(Op_lat_flag)
	   {
	      for(u=0;u<BUCKETS;u++)
		global_buckets[u]+=here->bands[u];
	   }
	   if(here->background !=0)
	   { 
		skip+=1.0;
	   } 
	   else
	   {
	      total_ops_per_second += here->ops_per_second;
	      total_latency  += here->average_latency;
	      total_read_throughput  += here->read_throughput;
	      total_write_throughput += here->write_throughput;
	      total_throughput  += here->read_throughput + here->write_throughput;
	      /* Totals for each workload type */	
	      strcpy(workloads[here->work_obj_index].work_obj_name,here->work_obj_name);
	      workloads[here->work_obj_index].total_ops_per_sec += here->ops_per_second;
	      workloads[here->work_obj_index].total_latency += here->average_latency;
	      workloads[here->work_obj_index].total_count += 1.0;
	   }
	   if(here->modified_run) /* set this if any client was modified */
		modified_run = 1;
	   count+=1.0;
	}
	count = count - skip;
	average_latency = total_latency/count;
	ops_per_second = total_ops_per_second;
	printf("\t------------------------------------------------------------\n");
	printf("\t%-30s  %10.2f Milli-seconds\n",
		"Overall average latency",average_latency);
	printf("\t%-7s %-23s %10.2f Ops/sec\n",
		"Overall",short_testname,ops_per_second);
	printf("\t%-30s ~%10.2f Kbytes/sec\n",
		"Overall Read_throughput",total_read_throughput/1024);
	printf("\t%-30s ~%10.2f Kbytes/sec\n",
		"Overall Write_throughput",total_write_throughput/1024);
	printf("\t%-30s ~%10.2f Kbytes/sec\n",
		"Overall throughput",total_throughput/1024);
	sum_val = check_sum();
	/* 
	 * Licensed means that we users will obey the rules and not 
	 * modify things.
	 */ 
	if(!licensed && modified_run)
	    sum_val = 0;
	if(licensed)
	   printf("\t%-30s  %5u\n",
		"Registered Finger Print",(unsigned int)sum_val);
	else
	   printf("\t%-30s  %5u\n",
		"Public Finger Print",(unsigned int)sum_val);
	printf("\t------------------------------------------------------------\n");
   	fflush(stdout);
	if(log_flag)
	{
	   fprintf(log_out,"\t------------------------------------------------------------\n");
	   fprintf(log_out,"\t%-30s  %10.2f Milli-seconds\n",
		"Overall average latency",average_latency);
	   fprintf(log_out,"\t%-7s %-23s %10.2f Ops/sec\n",
		"Overall", short_testname,ops_per_second);
	   fprintf(log_out,"\t%-30s ~%10.2f Kbytes/sec\n",
		"Overall Read_throughput",total_read_throughput/1024);
	   fprintf(log_out,"\t%-30s ~%10.2f Kbytes/sec\n",
		"Overall Write_throughput",total_write_throughput/1024);
	   fprintf(log_out,"\t%-30s ~%10.2f Kbytes/sec\n",
		"Overall throughput",total_throughput/1024);
	   if(licensed)
	      fprintf(log_out,"\t%-30s  %5u\n",
		"Registered Finger Print",(unsigned int)sum_val);
	   else
	      fprintf(log_out,"\t%-30s  %5u\n",
		"Public Finger Print",(unsigned int)sum_val);
	   fprintf(log_out,"\t------------------------------------------------------------\n");
	   fflush(log_out);
	}
	if(Op_lat_flag)
	{
		dump_hist();
	        printf("\t------------------------------------------------------------\n");
   		fflush(stdout);
		if(log_flag)
		{
	          fprintf(log_out,"\t------------------------------------------------------------\n");
	   	  fflush(log_out);
		}
	}
	/* Scan... looking to see if there was more than one workload contributing */
	for(i=0;i<NUM_WORK_OBJ;i++)
	{
		if(workloads[i].total_count != 0.0)
		   Multi_load_disp++;
	}
	if(Multi_load_disp > 1)
	{
	    printf("\t%-30s\n\n","Contributing workloads:");
	    if(log_flag)
	    {
	        fprintf(log_out,"\t%-15s\n\n","Contributing workloads:");
	   	fflush(log_out);
	    }
	    for(i=0;i<NUM_WORK_OBJ;i++)
	    {
		if(workloads[i].total_count != 0.0)
		{
		      printf("\t%-12s Procs %4.0f   Ops/sec %7.1f  Avg Latency %5.1f\n",
			workloads[i].work_obj_name, workloads[i].total_count, 
			workloads[i].total_ops_per_sec, 
			workloads[i].total_latency/workloads[i].total_count);
	           if(log_flag)
		   {
		      fprintf(log_out,"\t%-12s Procs %4.0f   Ops/sec %7.1f  Avg Latency %5.1f\n",
			workloads[i].work_obj_name, workloads[i].total_count, 
			workloads[i].total_ops_per_sec, 
			workloads[i].total_latency/workloads[i].total_count);
		   }
		}
	    }
	    printf("\t------------------------------------------------------------\n");
   	    fflush(stdout);
	    if(log_flag)
	    {
	          fprintf(log_out,"\t------------------------------------------------------------\n");
	   	  fflush(log_out);
	    }
	}
}

/* 
 * Prepare the distribution tables for messaging to the clients.
 */
#ifdef HAVE_ANSIC_C
void
export_dist_table(struct client_command *cc)
#else
void
export_dist_table(cc)
struct client_command *cc;
#endif
{
	int i,j;
	for(j=0;j<NUM_WORK_OBJ;j++)
	{
	   strcpy(cc->client_worker[j].c_workload_name,my_mix_table[j].workload_name);
	   strcpy(cc->client_worker[j].c_fs_type,my_mix_table[j].fs_type);

	   strcpy(cc->client_worker[j].c_op_array_string[OP_READ], my_mix_table[j].read_string);
	   cc->client_worker[j].c_op_array[OP_READ]      = my_mix_table[j].percent_read;

	   cc->client_worker[j].c_op_array[OP_READ_FILE] = my_mix_table[j].percent_read_file;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_READ_FILE],my_mix_table[j].read_file_string);

	   cc->client_worker[j].c_op_array[OP_MM_READ]   = my_mix_table[j].percent_mmap_read;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_MM_READ],my_mix_table[j].mmap_read_string);

	   cc->client_worker[j].c_op_array[OP_READ_RAND] = my_mix_table[j].percent_read_rand;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_READ_RAND],my_mix_table[j].read_rand_string);

	   cc->client_worker[j].c_op_array[OP_WRITE]     = my_mix_table[j].percent_write;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_WRITE],my_mix_table[j].write_string);

	   cc->client_worker[j].c_op_array[OP_WRITE_FILE]= my_mix_table[j].percent_write_file;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_WRITE_FILE],my_mix_table[j].write_file_string);

	   cc->client_worker[j].c_op_array[OP_MM_WRITE]  = my_mix_table[j].percent_mmap_write;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_MM_WRITE],my_mix_table[j].mmap_write_string);

	   cc->client_worker[j].c_op_array[OP_WRITE_RAND]= my_mix_table[j].percent_write_rand;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_WRITE_RAND],my_mix_table[j].write_rand_string);

	   cc->client_worker[j].c_op_array[OP_RMW]       = my_mix_table[j].percent_rmw;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_RMW],my_mix_table[j].rmw_string);

	   cc->client_worker[j].c_op_array[OP_MKDIR]     = my_mix_table[j].percent_mkdir;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_MKDIR],my_mix_table[j].mkdir_string);

	   cc->client_worker[j].c_op_array[OP_UNLINK]    = my_mix_table[j].percent_unlink;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_UNLINK],my_mix_table[j].unlink_string);

	   cc->client_worker[j].c_op_array[OP_CREATE]    = my_mix_table[j].percent_create;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_CREATE],my_mix_table[j].create_string);

	   cc->client_worker[j].c_op_array[OP_APPEND]    = my_mix_table[j].percent_append;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_APPEND],my_mix_table[j].append_string);

	   cc->client_worker[j].c_op_array[OP_LOCKING]   = my_mix_table[j].percent_locking;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_LOCKING],my_mix_table[j].locking_string);

	   cc->client_worker[j].c_op_array[OP_ACCESS]    = my_mix_table[j].percent_access;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_ACCESS],my_mix_table[j].access_string);

	   cc->client_worker[j].c_op_array[OP_STAT]      = my_mix_table[j].percent_stat;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_STAT],my_mix_table[j].stat_string);

	   cc->client_worker[j].c_op_array[OP_CHMOD]     = my_mix_table[j].percent_chmod;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_CHMOD],my_mix_table[j].chmod_string);

	   cc->client_worker[j].c_op_array[OP_READDIR]   = my_mix_table[j].percent_readdir;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_READDIR],my_mix_table[j].readdir_string);

	   cc->client_worker[j].c_op_array[OP_COPYFILE]  = my_mix_table[j].percent_copyfile;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_COPYFILE],my_mix_table[j].copyfile_string);

	   cc->client_worker[j].c_op_array[OP_RENAME]    = my_mix_table[j].percent_rename;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_RENAME],my_mix_table[j].rename_string);

	   cc->client_worker[j].c_op_array[OP_STATFS]    = my_mix_table[j].percent_statfs;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_STATFS],my_mix_table[j].statfs_string);

	   cc->client_worker[j].c_op_array[OP_PATHCONF]  = my_mix_table[j].percent_pathconf;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_PATHCONF],my_mix_table[j].pathconf_string);

	   cc->client_worker[j].c_op_array[OP_CUSTOM1]   = my_mix_table[j].percent_custom1;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_CUSTOM1],my_mix_table[j].custom1_string);

	   cc->client_worker[j].c_op_array[OP_CUSTOM2]   = my_mix_table[j].percent_custom2;
	   strcpy(cc->client_worker[j].c_op_array_string[OP_CUSTOM2],my_mix_table[j].custom2_string);

	   cc->client_worker[j].c_percent_commit = my_mix_table[j].percent_commit;
	   cc->client_worker[j].c_percent_direct = my_mix_table[j].percent_direct;
	   cc->client_worker[j].c_align = my_mix_table[j].align;
	   cc->client_worker[j].c_percent_osync = my_mix_table[j].percent_osync;
	   cc->client_worker[j].c_percent_geometric = my_mix_table[j].percent_geometric;
	   cc->client_worker[j].c_percent_compress = my_mix_table[j].percent_compress;
	   cc->client_worker[j].c_background = my_mix_table[j].background;
	   cc->client_worker[j].c_sharemode = my_mix_table[j].sharemode;
	   cc->client_worker[j].c_uniform_file_size_dist = my_mix_table[j].uniform_file_size_dist;
	   cc->client_worker[j].c_rand_dist_behavior = my_mix_table[j].rand_dist_behavior;

	   for(i=0;i<MAX_DIST_ELEM;i++)
	   {
	       cc->client_worker[j].c_rdist_array[i].size_min = my_mix_table[j].read_dist[i].size_min;
	       cc->client_worker[j].c_rdist_array[i].size_max = my_mix_table[j].read_dist[i].size_max;
	       cc->client_worker[j].c_rdist_array[i].percent = my_mix_table[j].read_dist[i].percent;
	       cc->client_worker[j].c_wdist_array[i].size_min = my_mix_table[j].write_dist[i].size_min;
	       cc->client_worker[j].c_wdist_array[i].size_max = my_mix_table[j].write_dist[i].size_max;
	       cc->client_worker[j].c_wdist_array[i].percent = my_mix_table[j].write_dist[i].percent;
	   }
	}
}

/*
 * If using an external Programmable Interval Timer, here is 
 * the routine to import the PIT service information.
 */
#ifdef HAVE_ANSIC_C
int
import_pit(struct client_command *cc)
#else
int
import_pit(cc)
struct client_command *cc;
#endif
{
	pit_hostname[0]=0;
	pit_service[0]=0;
        if(cc->c_pit_hostname[0]!=0)
	  strcpy(pit_hostname,&cc->c_pit_hostname[0]);
        if(cc->c_pit_service[0]!=0)
	  strcpy(pit_service,&cc->c_pit_service[0]);
	else
	  strcpy(pit_service,"PIT"); /* default */
	return(0);
}

/*
 * If using an external Programmable Interval Timer, here is 
 * the routine to exports the PIT service information.
 */
#ifdef HAVE_ANSIC_C
void
export_pit(struct client_command *cc)
#else
void
export_pit(cc)
struct client_command *cc;
#endif
{
	cc->c_pit_hostname[0]=0;
	cc->c_pit_service[0]=0;
	if(pit_hostname[0]!=0)
	  strcpy(&cc->c_pit_hostname[0],pit_hostname);
	if(pit_service[0]!=0)
	  strcpy(&cc->c_pit_service[0],pit_service);
}

/*
 * The routine to import (from the messaging system) the
 * distribution tables.
 */
#ifdef HAVE_ANSIC_C
int
import_dist_table(struct client_command *cc)
#else
int
import_dist_table(cc)
struct client_command *cc;
#endif
{
	int i,j;
	int total = 0;
	int rtotal,wtotal;

	for(j=0;j<NUM_WORK_OBJ;j++)
	{
	   fflush(newstdout);
	   strcpy(my_mix_table[j].workload_name, cc->client_worker[j].c_workload_name);
	   strcpy(my_mix_table[j].fs_type, cc->client_worker[j].c_fs_type);
	   my_mix_table[j].percent_read       = cc->client_worker[j].c_op_array[OP_READ];
	   strcpy(my_mix_table[j].read_string, cc->client_worker[j].c_op_array_string[OP_READ]);

	   my_mix_table[j].percent_read_file  = cc->client_worker[j].c_op_array[OP_READ_FILE];
	   strcpy(my_mix_table[j].read_file_string, cc->client_worker[j].c_op_array_string[OP_READ_FILE]);

	   my_mix_table[j].percent_mmap_read  = cc->client_worker[j].c_op_array[OP_MM_READ]; 
	   strcpy(my_mix_table[j].mmap_read_string, cc->client_worker[j].c_op_array_string[OP_MM_READ]);

	   my_mix_table[j].percent_read_rand  = cc->client_worker[j].c_op_array[OP_READ_RAND];
	   strcpy(my_mix_table[j].read_rand_string, cc->client_worker[j].c_op_array_string[OP_READ_RAND]);

	   my_mix_table[j].percent_write      = cc->client_worker[j].c_op_array[OP_WRITE]; 
	   strcpy(my_mix_table[j].write_string, cc->client_worker[j].c_op_array_string[OP_WRITE]);

	   my_mix_table[j].percent_write_file = cc->client_worker[j].c_op_array[OP_WRITE_FILE]; 
	   strcpy(my_mix_table[j].write_file_string, cc->client_worker[j].c_op_array_string[OP_WRITE_FILE]);

	   my_mix_table[j].percent_mmap_write = cc->client_worker[j].c_op_array[OP_MM_WRITE]; 
	   strcpy(my_mix_table[j].mmap_write_string, cc->client_worker[j].c_op_array_string[OP_MM_WRITE]);

	   my_mix_table[j].percent_write_rand = cc->client_worker[j].c_op_array[OP_WRITE_RAND];
	   strcpy(my_mix_table[j].write_rand_string, cc->client_worker[j].c_op_array_string[OP_WRITE_RAND]);

	   my_mix_table[j].percent_rmw        = cc->client_worker[j].c_op_array[OP_RMW];
	   strcpy(my_mix_table[j].rmw_string, cc->client_worker[j].c_op_array_string[OP_RMW]);

	   my_mix_table[j].percent_mkdir      = cc->client_worker[j].c_op_array[OP_MKDIR];
	   strcpy(my_mix_table[j].mkdir_string, cc->client_worker[j].c_op_array_string[OP_MKDIR]);

	   my_mix_table[j].percent_unlink     = cc->client_worker[j].c_op_array[OP_UNLINK];
	   strcpy(my_mix_table[j].unlink_string, cc->client_worker[j].c_op_array_string[OP_UNLINK]);

	   my_mix_table[j].percent_create     = cc->client_worker[j].c_op_array[OP_CREATE];
	   strcpy(my_mix_table[j].create_string, cc->client_worker[j].c_op_array_string[OP_CREATE]);

	   my_mix_table[j].percent_append     = cc->client_worker[j].c_op_array[OP_APPEND];
	   strcpy(my_mix_table[j].append_string, cc->client_worker[j].c_op_array_string[OP_APPEND]);

	   my_mix_table[j].percent_locking    = cc->client_worker[j].c_op_array[OP_LOCKING];
	   strcpy(my_mix_table[j].locking_string, cc->client_worker[j].c_op_array_string[OP_LOCKING]);

	   my_mix_table[j].percent_access     = cc->client_worker[j].c_op_array[OP_ACCESS];
	   strcpy(my_mix_table[j].access_string, cc->client_worker[j].c_op_array_string[OP_ACCESS]);

	   my_mix_table[j].percent_stat       = cc->client_worker[j].c_op_array[OP_STAT];
	   strcpy(my_mix_table[j].stat_string, cc->client_worker[j].c_op_array_string[OP_STAT]);

	   my_mix_table[j].percent_chmod      = cc->client_worker[j].c_op_array[OP_CHMOD];
	   strcpy(my_mix_table[j].chmod_string, cc->client_worker[j].c_op_array_string[OP_CHMOD]);

	   my_mix_table[j].percent_readdir    = cc->client_worker[j].c_op_array[OP_READDIR];
	   strcpy(my_mix_table[j].readdir_string, cc->client_worker[j].c_op_array_string[OP_READDIR]);

	   my_mix_table[j].percent_copyfile   = cc->client_worker[j].c_op_array[OP_COPYFILE];
	   strcpy(my_mix_table[j].copyfile_string, cc->client_worker[j].c_op_array_string[OP_COPYFILE]);

	   my_mix_table[j].percent_rename     = cc->client_worker[j].c_op_array[OP_RENAME];
	   strcpy(my_mix_table[j].rename_string, cc->client_worker[j].c_op_array_string[OP_RENAME]);

	   my_mix_table[j].percent_statfs     = cc->client_worker[j].c_op_array[OP_STATFS];
	   strcpy(my_mix_table[j].statfs_string, cc->client_worker[j].c_op_array_string[OP_STATFS]);

	   my_mix_table[j].percent_pathconf   = cc->client_worker[j].c_op_array[OP_PATHCONF];
	   strcpy(my_mix_table[j].pathconf_string, cc->client_worker[j].c_op_array_string[OP_PATHCONF]);

	   my_mix_table[j].percent_custom1    = cc->client_worker[j].c_op_array[OP_CUSTOM1];
	   strcpy(my_mix_table[j].custom1_string, cc->client_worker[j].c_op_array_string[OP_CUSTOM1]);

	   my_mix_table[j].percent_custom2    = cc->client_worker[j].c_op_array[OP_CUSTOM2];
	   strcpy(my_mix_table[j].custom2_string, cc->client_worker[j].c_op_array_string[OP_CUSTOM2]);

	   my_mix_table[j].percent_commit = cc->client_worker[j].c_percent_commit;
	   my_mix_table[j].percent_direct = cc->client_worker[j].c_percent_direct;
	   my_mix_table[j].align = cc->client_worker[j].c_align;
	   my_mix_table[j].percent_osync = cc->client_worker[j].c_percent_osync;
	   my_mix_table[j].percent_geometric = cc->client_worker[j].c_percent_geometric;
	   my_mix_table[j].percent_compress = cc->client_worker[j].c_percent_compress;
	   my_mix_table[j].background = cc->client_worker[j].c_background;
	   my_mix_table[j].sharemode = cc->client_worker[j].c_sharemode;
	   my_mix_table[j].uniform_file_size_dist = cc->client_worker[j].c_uniform_file_size_dist;
	   my_mix_table[j].rand_dist_behavior = cc->client_worker[j].c_rand_dist_behavior;
	
	   total=0;
	   for(i=0;i< NUM_OP_TYPES; i++)
	   {
		total += cc->client_worker[j].c_op_array[i];
	   }
	   if (total != 100)
	   {
	     if(cdebug || 1)
	     {
 	        fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
 	        fprintf(newstdout,"Invalid OP distribution. Total = %d \n",total);
		fflush(newstdout);
	     }
	     return(1);
	   }
	   wtotal = rtotal =0;
	   for(i=0;i< MAX_DIST_ELEM; i++)
	   {

	      my_mix_table[j].read_dist[i].size_min = cc->client_worker[j].c_rdist_array[i].size_min;
	      my_mix_table[j].read_dist[i].size_max = cc->client_worker[j].c_rdist_array[i].size_max;
	      my_mix_table[j].read_dist[i].percent = cc->client_worker[j].c_rdist_array[i].percent;
	      my_mix_table[j].write_dist[i].size_min = cc->client_worker[j].c_wdist_array[i].size_min;
	      my_mix_table[j].write_dist[i].size_max = cc->client_worker[j].c_wdist_array[i].size_max;
	      my_mix_table[j].write_dist[i].percent = cc->client_worker[j].c_wdist_array[i].percent;
	      rtotal  += cc->client_worker[j].c_rdist_array[i].percent;
	      wtotal += cc->client_worker[j].c_wdist_array[i].percent;
	   }
	   if ((rtotal != 100) || (wtotal != 100))
	   {
	     if(cdebug || 1)
	     {
 	        fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
 	        fprintf(newstdout,"Invalid xfer distribution. \n");
	        fprintf(newstdout,"Read total percent = %d \n",rtotal);
	        fprintf(newstdout,"Write total percent = %d \n",wtotal);
		fflush(newstdout);
	     }
	     return(1);
	   }
	   if(ldebug)
	   {
 	     fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
	     fprintf(newstdout,"-------------------\n");
	     fprintf(newstdout,"Distribution table:\n");
	     fprintf(newstdout,"-------------------\n");
	     fprintf(newstdout,"Workload name        = %s\n",my_mix_table[j].workload_name);
	     fprintf(newstdout,"Percent read         = %d\n",my_mix_table[j].percent_read);
	     fprintf(newstdout,"Percent read file    = %d\n",my_mix_table[j].percent_read_file);
	     fprintf(newstdout,"Percent mmap read    = %d\n",my_mix_table[j].percent_mmap_read);
	     fprintf(newstdout,"Percent rand read    = %d\n",my_mix_table[j].percent_read_rand);
	     fprintf(newstdout,"Percent write        = %d\n",my_mix_table[j].percent_write);
	     fprintf(newstdout,"Percent write file   = %d\n",my_mix_table[j].percent_write_file);
	     fprintf(newstdout,"Percent mmap write   = %d\n",my_mix_table[j].percent_mmap_write);
	     fprintf(newstdout,"Percent rand write   = %d\n",my_mix_table[j].percent_write_rand);
	     fprintf(newstdout,"Percent rmw          = %d\n",my_mix_table[j].percent_rmw);
	     fprintf(newstdout,"Percent mkdir        = %d\n",my_mix_table[j].percent_mkdir);
	     fprintf(newstdout,"Percent unlink       = %d\n",my_mix_table[j].percent_unlink);
	     fprintf(newstdout,"Percent create       = %d\n",my_mix_table[j].percent_create);
	     fprintf(newstdout,"Percent append       = %d\n",my_mix_table[j].percent_append);
	     fprintf(newstdout,"Percent locking      = %d\n",my_mix_table[j].percent_locking);
	     fprintf(newstdout,"Percent access       = %d\n",my_mix_table[j].percent_access);
	     fprintf(newstdout,"Percent stat         = %d\n",my_mix_table[j].percent_stat);
	     fprintf(newstdout,"Percent chmod        = %d\n",my_mix_table[j].percent_chmod);
	     fprintf(newstdout,"Percent readdir      = %d\n",my_mix_table[j].percent_readdir);
	     fprintf(newstdout,"Percent copyfile     = %d\n",my_mix_table[j].percent_copyfile);
	     fprintf(newstdout,"Percent rename       = %d\n",my_mix_table[j].percent_rename);
	     fprintf(newstdout,"Percent statfs       = %d\n",my_mix_table[j].percent_statfs);
	     fprintf(newstdout,"Percent pathconf     = %d\n",my_mix_table[j].percent_pathconf);
	     fprintf(newstdout,"Percent custom1      = %d\n",my_mix_table[j].percent_custom1);
	     fprintf(newstdout,"Percent custom2      = %d\n",my_mix_table[j].percent_custom2);
	     fprintf(newstdout,"Percent write commit = %d\n",my_mix_table[j].percent_commit);
	     fprintf(newstdout,"Percent osync        = %d\n",my_mix_table[j].percent_osync);
	     fprintf(newstdout,"Percent direct       = %d\n",my_mix_table[j].percent_direct);
	     fprintf(newstdout,"Align                = %d\n",my_mix_table[j].align);
	     fprintf(newstdout,"Percent geometric    = %d\n",my_mix_table[j].percent_geometric);
	     fprintf(newstdout,"Percent compress     = %d\n",my_mix_table[j].percent_compress);
	     fprintf(newstdout,"Background           = %d\n",my_mix_table[j].background);
	     fprintf(newstdout,"Sharemode            = %d\n",my_mix_table[j].sharemode);
	     fprintf(newstdout,"Uniform size dist    = %d\n",my_mix_table[j].uniform_file_size_dist);
	     fprintf(newstdout,"Rand dist mode       = %d\n",my_mix_table[j].rand_dist_behavior);
	     fprintf(newstdout,"FS type              = %s\n",my_mix_table[j].fs_type);
	     fprintf(newstdout,"-------------------\n");
	     fflush(newstdout);
	   }
	}
	return(0);
}

/* 
 * This dumps the internal workloads to an ascii file, for editing
 * and later import.
 */
#ifdef HAVE_ANSIC_C
void
dump_table(void)
#else
void
dump_table()
#endif
{
	int i,j;
	for(j=0;j<NUM_WORK_OBJ;j++)
	{
	   printf("Workload name %s\n",my_mix_table[j].workload_name);
	   printf("Percent %-35s %d\n",my_mix_table[j].read_string, my_mix_table[j].percent_read);
	   printf("Percent %-35s %d\n",my_mix_table[j].read_file_string, my_mix_table[j].percent_read_file);
	   printf("Percent %-35s %d\n",my_mix_table[j].mmap_read_string,my_mix_table[j].percent_mmap_read);
	   printf("Percent %-35s %d\n",my_mix_table[j].read_rand_string, my_mix_table[j].percent_read_rand);
	   printf("Percent %-35s %d\n",my_mix_table[j].write_string, my_mix_table[j].percent_write);
	   printf("Percent %-35s %d\n",my_mix_table[j].write_file_string, my_mix_table[j].percent_write_file);
	   printf("Percent %-35s %d\n",my_mix_table[j].mmap_write_string, my_mix_table[j].percent_mmap_write);
	   printf("Percent %-35s %d\n",my_mix_table[j].write_rand_string, my_mix_table[j].percent_write_rand);
	   printf("Percent %-35s %d\n",my_mix_table[j].rmw_string, my_mix_table[j].percent_rmw);
	   printf("Percent %-35s %d\n",my_mix_table[j].mkdir_string, my_mix_table[j].percent_mkdir);
	   printf("Percent %-35s %d\n",my_mix_table[j].unlink_string, my_mix_table[j].percent_unlink);
	   printf("Percent %-35s %d\n",my_mix_table[j].create_string, my_mix_table[j].percent_create);
	   printf("Percent %-35s %d\n",my_mix_table[j].append_string, my_mix_table[j].percent_append);
	   printf("Percent %-35s %d\n",my_mix_table[j].locking_string, my_mix_table[j].percent_locking);
	   printf("Percent %-35s %d\n",my_mix_table[j].access_string, my_mix_table[j].percent_access);
	   printf("Percent %-35s %d\n",my_mix_table[j].stat_string, my_mix_table[j].percent_stat);
	   printf("Percent %-35s %d\n",my_mix_table[j].chmod_string, my_mix_table[j].percent_chmod);
	   printf("Percent %-35s %d\n",my_mix_table[j].readdir_string, my_mix_table[j].percent_readdir);
	   printf("Percent %-35s %d\n",my_mix_table[j].copyfile_string, my_mix_table[j].percent_copyfile);
	   printf("Percent %-35s %d\n",my_mix_table[j].rename_string, my_mix_table[j].percent_rename);
	   printf("Percent %-35s %d\n",my_mix_table[j].statfs_string, my_mix_table[j].percent_statfs);
	   printf("Percent %-35s %d\n",my_mix_table[j].pathconf_string, my_mix_table[j].percent_pathconf);
	   printf("Percent %-35s %d\n",my_mix_table[j].custom1_string, my_mix_table[j].percent_custom1);
	   printf("Percent %-35s %d\n",my_mix_table[j].custom2_string, my_mix_table[j].percent_custom2);
   	   fflush(stdout);
	   for(i=0;i<MAX_DIST_ELEM;i++)
	   {
	     printf("Read elem %d xfer min size %d\n",i,my_mix_table[j].read_dist[i].size_min);
	     printf("Read elem %d xfer max size %d\n",i,my_mix_table[j].read_dist[i].size_max);
	     printf("Read elem %d xfer percent %d\n",i,my_mix_table[j].read_dist[i].percent);
	   }
	   for(i=0;i<MAX_DIST_ELEM;i++)
	   {
	     printf("Write elem %d xfer min size %d\n",i,my_mix_table[j].write_dist[i].size_min);
	     printf("Write elem %d xfer max size %d\n",i,my_mix_table[j].write_dist[i].size_max);
	     printf("Write elem %d xfer percent %d\n",i,my_mix_table[j].write_dist[i].percent);
	   }
	   printf("Percent write commit %d\n",my_mix_table[j].percent_commit);
	   printf("Percent direct %d\n",my_mix_table[j].percent_direct);
	   printf("Align          %d\n",my_mix_table[j].align);
	   printf("Percent osync %d\n",my_mix_table[j].percent_osync);
	   printf("Percent geometric %d\n",my_mix_table[j].percent_geometric);
	   printf("Percent compress %d\n",my_mix_table[j].percent_compress);
	   printf("Background %d\n",my_mix_table[j].background);
	   printf("Sharemode %d\n",my_mix_table[j].sharemode);
	   printf("Uniform size dist %d\n",my_mix_table[j].uniform_file_size_dist);
	   printf("Rand dist behavior %d\n",my_mix_table[j].rand_dist_behavior);
	   printf("FS type       %s\n",my_mix_table[j].fs_type);
   	   fflush(stdout);
	}
}

/*
 * The controlling process uses this function to start the tests on the remote
 * clients.
 */
#ifdef HAVE_ANSIC_C
void
launch_clients(void)
#else
void
launch_clients()
#endif
{
	int i,sret;
	char command[1024];
	double my_oprate = 0.0;
	unsigned int addr_buf[16];
	char name_buf[MAXHNAME];
	
	bzero(name_buf,MAXHNAME);
	my_oprate = op_rate/(float)num_clients; 

	if(mflag) 
	{
		printf("\tLaunching %d processes.\n",num_clients);
   	        fflush(stdout);
		if(log_flag)
		{
		    fprintf(log_out,"\tLaunching %d processes.\n",num_clients);
		    fflush(log_out);
		}
	
	}
	if(sharing_flag)
	  strcpy(option_buf,"-X");
	else
	  bzero(option_buf,4);
	for(i=0;i<num_clients; i++)
	{
		if(client_obj[i].op_rate !=0)
			my_oprate = client_obj[i].op_rate; /* Over-ride from config file. Not uniform distribution */
		else
			my_oprate = op_rate/(float)num_clients;

		if(!localflag)
		{
#if defined(WIN32)
		sprintf(command,"wmic.exe /node:\"%s\" /user:\"%s\" /password:\"%s\" process call create \"%s -A %s -Z %s -Q %d -T %d -m -c %d -d %d -O %.2f -s %lld -n %d -t %d -p %s -e %s -M %s -R %s -U %s -b %lld -B %lld %s -P %d -w %d -W %s -y %d -r %d %s\"",
			client_obj[i].client_name, client_obj[i].client_dom_user, client_obj[i].client_password, client_obj[i].client_execdir, 
			client_obj[i].client_password, client_obj[i].client_dom_user, client_files, client_dirs, num_clients, ppc, my_oprate,  
			fsize, client_obj[i].client_id, run_time, client_obj[i].client_workdir, client_obj[i].client_execdir,master_name, 
			master_results_dir,client_log_dir,client_memsize_mb, agg_set_memsize_mb, var_opt,master_listen_port,warm_time,
			client_obj[i].client_workload_name,ipv6_enable,setfsize,option_buf);
#else
		  sprintf(command,"%s %s -l %s %s -A %s -Z %s -Q %d -T %d -m -c %d -d %d -O %.2f -s %lld -n %d -t %d -p %s -e %s -M %s -R %s -U %s -b %lld -B %lld %s -P %d -w %d -W %s -y %d -r %d %s",
			remsh,client_obj[i].client_name, client_obj[i].client_dom_user, client_obj[i].client_execdir,
			client_obj[i].client_password, client_obj[i].client_dom_user, client_files, client_dirs, num_clients, ppc, my_oprate,  
			fsize, client_obj[i].client_id, run_time, client_obj[i].client_workdir,client_obj[i].client_execdir,master_name, 
			master_results_dir,client_log_dir,client_memsize_mb, agg_set_memsize_mb, var_opt,master_listen_port,warm_time,
			client_obj[i].client_workload_name,ipv6_enable,setfsize,option_buf);
#endif
		}
		else
		{
#if defined(WIN32)
		sprintf(command,"start %s -A %s -Z %s -Q %d -T %d -m -c %d -d %d -O %.2f -s %lld -n %d -t %d -p %s -e %s -M %s -R %s -U %s -b %lld -B %lld %s -P %d -w %d -W %s -z -y %d -r %d %s",
			client_obj[i].client_execdir, client_obj[i].client_password, client_obj[i].client_dom_user, client_files, 
			client_dirs, num_clients, ppc, my_oprate, fsize, client_obj[i].client_id, run_time, client_obj[i].client_workdir,
			client_obj[i].client_execdir,master_name, master_results_dir,client_log_dir,client_memsize_mb,
			agg_set_memsize_mb, var_opt,master_listen_port,warm_time,client_obj[i].client_workload_name,ipv6_enable,setfsize,option_buf);
#else
		  sprintf(command,"%s -A %s -Z %s -Q %d -T %d -m -c %d -d %d -O %.2f -s %lld -n %d -t %d -p %s -e %s -M %s -R %s -U %s -b %lld -B %lld %s -P %d -w %d -W %s -z -y %d -r %d %s",
			client_obj[i].client_execdir, client_obj[i].client_password, client_obj[i].client_dom_user, client_files, client_dirs, num_clients, ppc, my_oprate,  
			fsize, client_obj[i].client_id, run_time, client_obj[i].client_workdir,
			client_obj[i].client_execdir,master_name, master_results_dir,client_log_dir,client_memsize_mb,
			agg_set_memsize_mb, var_opt,master_listen_port,warm_time,client_obj[i].client_workload_name,ipv6_enable,setfsize, option_buf);
#endif
		}
		if(cdebug) 
			printf("Command: %s\n",command);
		/* Get name of child.. using getaddrinfo, and then grab the canonname back */
		get_host_address(client_obj[i].client_name, addr_buf, name_buf);
		if(strlen(name_buf)==0) /* Then use the name from the config file */
			strcpy(name_buf,client_obj[i].client_name);
		printf("\tStarting test client: %d Host: %10s    Workload: %8s  Workdir: %s\n",i, name_buf, client_obj[i].client_workload_name,client_obj[i].client_workdir);
   	        fflush(stdout);
		if(log_flag)
		{
	   		fprintf(log_out,"\tStarting test client: %d Host: %10s    Workload: %8s  Workdir: %s\n",i,name_buf, client_obj[i].client_workload_name,client_obj[i].client_workdir);
			fflush(log_out);
		}
		sret=system(command);
		if(sret != 0)
		{
		   printf("Remote command startup failed.. cleanup and exit\n");
		   kill_clients();
		   exit(99);
		}
	}
}

/*
 * Used to send the kill tests to the remote clients.
 */
#ifdef HAVE_ANSIC_C
void
kill_clients(void)
#else
void
kill_clients()
#endif
{
	int i;
	char command[1024];

	if(mflag) 
	{
		printf("Killing %d clients.\n",num_clients);
   	        fflush(stdout);
	}
	else
		return;

	for(i=0;i<num_clients; i++)
	{
		if(!localflag)
		{

#if defined(WIN32)
		 sprintf(command,"wmic.exe /node:\"%s\" /user:\"%s\" /password:\"%s\" process call create \"%s -A %s -Z %s -k -m -c %d -s %lld -n %d -t %d -p %s -e %s -M %s -R %s -U %s -W %s -y %d -V %d -r %d %s\"",
			client_obj[i].client_name, client_obj[i].client_dom_user, client_obj[i].client_password, client_obj[i].client_execdir, 
			client_obj[i].client_password, client_obj[i].client_dom_user,num_clients, fsize, client_obj[i].client_id, run_time, 
			client_obj[i].client_workdir, client_obj[i].client_execdir, master_name, master_results_dir, client_log_dir,
			client_obj[i].client_workload_name,ipv6_enable,child_pid[i],setfsize,option_buf);
#else
		  sprintf(command,"%s %s -l %s %s -A %s -Z %s -k -m -c %d -s %lld -n %d -t %d -p %s -e %s -M %s -R %s -U %s -W %s -y %d -V %d -r %d %s",
			remsh, client_obj[i].client_name, client_obj[i].client_dom_user, client_obj[i].client_execdir, 
			client_obj[i].client_password, client_obj[i].client_dom_user, num_clients, fsize, client_obj[i].client_id, 
			run_time, client_obj[i].client_workdir, client_obj[i].client_execdir, master_name, master_results_dir, 
			client_log_dir,client_obj[i].client_workload_name,ipv6_enable,child_pid[i],setfsize,option_buf);
#endif
		}
		else
		{
		  sprintf(command,"%s -A %s -Z %s -k -m -c %d -s %lld -n %d -t %d -p %s -e %s -M %s -R %s -U %s -W %s -z -y %d -V %d -r %d %s",
			client_obj[i].client_execdir, client_obj[i].client_password, client_obj[i].client_dom_user, num_clients, fsize, 
			client_obj[i].client_id, run_time, client_obj[i].client_workdir, client_obj[i].client_execdir,master_name, 
			master_results_dir, client_log_dir, client_obj[i].client_workload_name, ipv6_enable,child_pid[i],setfsize, option_buf);
		}
		printf("Killing test on -> %s <-    Client ID %5d     PID %5d\n",client_obj[i].client_name,client_obj[i].client_id,child_pid[i]);
   	        fflush(stdout);
		system(command);
	}
}

static int reverse_bits(int num)
{
	unsigned int size = sizeof(num)*8, i=0;
	int result = 0;
	for (i=0; i<size; i++)
		{
		if (num & (1<<i))
			result |= (1 << (size-i-1));
		}
	if (cdebug)
		fprintf( newstdout, "Number %x is reversed to %x\n", num, result);
	return result;
}

/*
 * This is where the clients start the test.
 */

#ifdef HAVE_ANSIC_C
void
client_startup(void)
#else
void
client_startup()
#endif
{
	int ret,j,loops,pct_complete;
	int prev_pct_complete;
	int save_loops = 1;   /* how many gen_ops done in this time interval */
	char localname[MAXNAME];
	char localtracename[MAXNAME];
	long patt;
	int chosen_fs = 0;
	double res,tmp_time = 0;
	int i_scale = ISCALE; /* auto scaling for sleep injection */
	int inject = 0;   /* loop count for how many possible injections */
	long u,delay_count;
	int got_name=0;
	int y;
	
	int x,b,maxiosize;
	int really = 0;
        double time_per_tick;
	double time_of_last_tick = 0.0;
	double my_last_tick,elap_tick;
	double ach_ops_per_second = 0;
	char command[MAXECOMMAND];
	int retry_count = 0;
	double op_time_ms = 0.0;
	unsigned int delay=0;
	void* context=NULL;

#if defined(WIN32)
	char empty[1] = "";
	char temp[MAXNAME];
	char *xx;
	/*This is the directory where we look for "netmist_pretest.cmd" file*/
	char* pretest_dir=GetParentPath(execdir);
#endif

again:
	x=fork(); /* Become a stand alone child. */
	if(x < 0)
	{
	     if(errno == EAGAIN)
	     {
	       really++;
	       if(really > 10) /* Try ten times with sleep between */
	       {
		 printf("Parent unable to fork !!!. Giving up.\n");
   	         fflush(stdout);
	         exit(EAGAIN);
	       }
	       sleep(2); /* Sleep a bit, and try again */
	       goto again;
	     }
	     printf("Parent unable to fork. Errno %d.\n",errno);
   	     fflush(stdout);
	     exit(-1);
	}
	if(x!=0) /* I am the parent process */
	{
		exit(0); /* Parent exits the system. Let rsh move on */
	}

	prev_heartbeat = 0;
	memset(command,0,MAXECOMMAND);
#if defined(PDSM_R_REGISTRY)
	sprintf(my_pdsm_file,"%s_%d",PDSM_R_REGISTRY,client_id);
#if defined(WIN32)

	/* Usr_dom contains the user's name and domain.*/
 	/* The Password variable contains the user's password.*/

	strcpy(temp,usr_dom); 	/* Make a temporary copy of domain and user */
	xx=strchr(temp,0x5C); 	/* find the backslash */
	if(xx != 0)
	{
		*xx=0;		/* Make that a null seperator */
		xx++; 		/* Point this at user name */
	}
	else
		xx=empty;	/* Point this at an empty string */
	if(Password_flag)
	{

   	   // logon
	   {
              DWORD res;
              TCHAR szUserName[32];
              TCHAR szPassword[32];
              TCHAR szDomain[32];
              HANDLE hToken;

	      strcpy(szDomain,temp); 	/* First half of domain\user 	*/
	      strcpy(szPassword,password); /* Handed in via command line   */
	      strcpy(szUserName,xx); 	/* Second half of domain\user   */

	      fflush(stdout);
	      strcpy((char *)&wmiUser[0],(char *)&szUserName[0]);
	      strcpy((char *)&wmiDomain[0],(char *)&szDomain[0]);
	      strcpy((char *)&wmiPassword[0],(char *)&szPassword[0]);

              // Logon to the local computer
	      if(!localflag)
	      {
                 res = LogonUser(szUserName, szDomain, szPassword, LOGON32_LOGON_NEW_CREDENTIALS, LOGON32_PROVIDER_DEFAULT, &hToken);
                 LogonUser_res = res;
                 if (res == 0) {
		   LogonUser_res = GetLastError();
                   printf("Error %d from LogonUser\n", res);
                 }
                 // Become that user
                 res = ImpersonateLoggedOnUser(hToken);
                 ImpersonateLoggedOnUser_res = res;
                 if (res == 0) {
                   ImpersonateLoggedOnUser_res = GetLastError();
                   printf("Error %d from ImpersonateLoggedOnUser\n", res);
                 }
	      }
       	   }
	}
	if(GetFileAttributes(my_pdsm_file) != INVALID_FILE_ATTRIBUTES)
#else
	if(access(my_pdsm_file,F_OK) == 0)
#endif
		pdsm_stats=(struct pdsm_remote_stats *)
			get_remote_zone(client_id,
			sizeof(struct pdsm_remote_stats));
#if defined(WIN32)
	if(GetFileAttributes(my_pdsm_file) != INVALID_FILE_ATTRIBUTES)
#else
	if(access(my_pdsm_file,F_OK) == 0)
#endif
		pdsm_control=(struct pdsm_remote_control *)
			get_my_master_zone(client_id,
			sizeof(struct pdsm_remote_control));
#endif
	/* Check for existance and executable */
#if defined(WIN32)
	
	if (pretest_dir)
		/*prefix the script file with directory name*/
		sprintf(command,"%s\\%s", pretest_dir, pretest_script);
	else
		sprintf(command,"%s",pretest_script);
	if(GetFileAttributes(command) != INVALID_FILE_ATTRIBUTES)
#else
	if( (access(pretest_script,F_OK) == 0) &&
		(access(pretest_script,X_OK)==0) )
#endif
	{
#if !defined(WIN32)
		/* Build command */
		sprintf(command,"%s",pretest_script);
#endif
	    /* Run command */
	    ret = system(command);
	    if(ret < 0)
	    {
 		printf("%s: ",get_timestamp(&time_string[0]));
		printf("Error returned when starting %s Ret = %d Error = %s\n",
			command, ret,strerror(errno));
   	        fflush(stdout);
	    }
	}

	/*
	 * Check to see if log directory exists, and is writable.
	 */
	if(check_dir_w(client_log_dir) != 0)
	{
		/* If it doesn't exist, try creating it */
		ret = mkdir(client_log_dir,0777);
		if(ret < 0)
		{
 		   printf("%s: ",get_timestamp(&time_string[0]));
		   printf("Unable to create log directory > %s <  Error: %s\n",
			client_log_dir,strerror(errno));
   	           fflush(stdout);
		   exit(1);
		}
	}
	/* Create logs, where children can log output */
        fclose(stdin);
        fclose(stderr);
	/* fclose(stdout); ( freopen() will do this */
	if(Uflag)
	{
#if defined(WIN32)
	   	sprintf(localname,"%s\\netmist_C%d.%s",client_log_dir,
			client_id,"log");
		sprintf(localtracename,"%s\\netmist_trace_C%d.%s",
			client_log_dir,client_id,"csv");
#else
	   	sprintf(localname,"%s/netmist_C%d.%s",client_log_dir,
			client_id,"log");
		sprintf(localtracename,"%s/netmist_trace_C%d.%s",
			client_log_dir,client_id,"csv");
#endif
	}
	else
	{
#if defined(WIN32)
		sprintf(localname, "\\tmp\\netmist_C%d.%s",client_id,"log");
		sprintf(localtracename, "\\tmp\\netmist_trace_C%d.%s",
			client_id,"csv");
#else
		sprintf(localname, "/tmp/netmist_C%d.%s",client_id,"log");
		sprintf(localtracename, "/tmp/netmist_trace_C%d.%s",
			client_id,"csv");
#endif
	}

    	newstdout=freopen(localname,"a+",stdout);
	if(newstdout <= (FILE *)0)
	{
		perror("Unable to open netmist_C..log for stdout\n");
		exit(1);
	}
        newstderr=newstdout;
	if(ppc+2 > max_proc)
	{
		fprintf(newstdout,"Process count exceeds client's max child processes limit\n");
		fflush(newstdout);
		exit(1);
	}	
	
	/*
	 * Start the clients communication channel.
	 */
	if(sockdebug)
	{
	   fprintf(newstdout,"Startup of comm with master.\n");
	   fflush(newstdout);
	}
	child_start_comm(master_name);
	if(sockdebug)
	{
	   fprintf(newstdout,"Startup of comm with master succeeded.\n");
	   fflush(newstdout);
	}
	/* 
	 * Wait for master to say go. All children have 
	 * their comm up and are ready 
	 */
	fprintf(newstdout,"%s Child waiting for master to send Info.\n",
		get_timestamp(&time_string[0]));
	fflush(newstdout);

	child_wait(); 

	if(tracedebug) {
		tracefile=fopen(localtracename,"a+");
		if(tracefile == 0) {
			fprintf(newstdout,"Error opening client trace log: %s\n",localtracename);
			fflush(newstdout);
			exit(1);
		}
		fprintf(tracefile,"Thread,Sequence,State,Op,Action,File,Mode,Offset,IO Size,Access count,Other\n");
		fprintf(tracefile,"%d,%lu,%s,,Trace Log Open,%s,,,,,%s\n",
			client_id,trace_seq++,trace_states[tracestate],localtracename,
			get_timestamp(&time_string[0]));
	}

	/* Setup the workload name to my_workload relationship */
	for(y=0;y<NUM_WORK_OBJ;y++)
	{	
	   if(my_strcasecmp(my_tmp_workload_name,my_mix_table[y].workload_name)==0)
	   {
		my_workload=y;
		got_name=1;
		break;
	   }
	}
	if(got_name == 0)
	{
 		fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Child has bad workload name.\n");
		fflush(newstdout);
                c_eval = BAD_WORKLOAD;
		R_exit(192);
	}
	/* Update the workload object name */

	strcpy(work_obj[my_workload].work_obj_name,my_tmp_workload_name);

	/* This is where we load up the indirect function table. These functions 
  	 * implement all of the POSIX-ish, or other, types of calls that
	 * may be filesystem type dependent.
	 *
 	 * If the fs_type is POSIX, then use the POSIX methods.
	 * If it is some other type, please add it here.
	 * If the type is garbage, then this will default to POSIX.
 	 */

	if( (my_strcasecmp(my_mix_table[my_workload].fs_type, "POSIX")) == 0) 
	{
	     /* The POSIX versions of these virtual functions */
	     Netmist_Ops[my_workload].read=generic_read;
	     Netmist_Ops[my_workload].read_file=generic_read_file;
	     Netmist_Ops[my_workload].mmap_read=generic_mmap_read;
	     Netmist_Ops[my_workload].read_rand=generic_read_rand;
	     Netmist_Ops[my_workload].write=generic_write;
	     Netmist_Ops[my_workload].write_file=generic_write_file;
	     Netmist_Ops[my_workload].mmap_write=generic_mmap_write;
	     Netmist_Ops[my_workload].write_rand=generic_write_rand;
	     Netmist_Ops[my_workload].rmw=generic_rmw;
	     Netmist_Ops[my_workload].mkdir=generic_mkdir;
	     Netmist_Ops[my_workload].unlink=generic_unlink;
	     Netmist_Ops[my_workload].create=generic_create;
	     Netmist_Ops[my_workload].append=generic_append;
	     Netmist_Ops[my_workload].locking=generic_locking;
	     Netmist_Ops[my_workload].access=generic_access;
	     Netmist_Ops[my_workload].stat=generic_stat;
	     Netmist_Ops[my_workload].chmod=generic_chmod;
	     Netmist_Ops[my_workload].readdir=generic_readdir;
	     Netmist_Ops[my_workload].copyfile=generic_copyfile;
	     Netmist_Ops[my_workload].rename=generic_rename;
	     Netmist_Ops[my_workload].statfs=generic_statfs;
	     Netmist_Ops[my_workload].pathconf=generic_pathconf;
	     Netmist_Ops[my_workload].custom1=generic_custom1;
	     Netmist_Ops[my_workload].custom2=generic_custom2;
	     Netmist_Ops[my_workload].makepath=generic_make_path;
	     Netmist_Ops[my_workload].prefix_op=generic_prefix_op;
	     Netmist_Ops[my_workload].postfix_op=generic_postfix_op;
	     Netmist_Ops[my_workload].fill_buf=generic_fill_buf;
	     Netmist_Ops[my_workload].init_dir=generic_init_dir;
	     Netmist_Ops[my_workload].init_file=generic_init_file;
	     Netmist_Ops[my_workload].init_empty_file=generic_init_empty_file;
	     Netmist_Ops[my_workload].stat_workdir=generic_stat_workdir;
	     Netmist_Ops[my_workload].chdir=generic_chdir;
	     Netmist_Ops[my_workload].remove_file=generic_remove_file;
	     Netmist_Ops[my_workload].remove_dir=generic_remove_dir;
	     Netmist_Ops[my_workload].setup = generic_setup;
	     Netmist_Ops[my_workload].teardown = generic_teardown;
	     Netmist_Ops[my_workload].impersonate = generic_impersonate;
	     Netmist_Ops[my_workload].make_impersonate_context = generic_make_impersonate_context;
	     chosen_fs = 1;
	}
	else if( (my_strcasecmp(my_mix_table[my_workload].fs_type, "WINDOWS")) == 0)
	{
		chosen_fs = 1;
	}
	/* Then default to POSIX */
	if(chosen_fs == 0)
	{
	     /* The POSIX versions of these virtual functions */
	     Netmist_Ops[my_workload].read=generic_read;
	     Netmist_Ops[my_workload].read_file=generic_read_file;
	     Netmist_Ops[my_workload].mmap_read=generic_mmap_read;
	     Netmist_Ops[my_workload].read_rand=generic_read_rand;
	     Netmist_Ops[my_workload].write=generic_write;
	     Netmist_Ops[my_workload].write_file=generic_write_file;
	     Netmist_Ops[my_workload].mmap_write=generic_mmap_write;
	     Netmist_Ops[my_workload].write_rand=generic_write_rand;
	     Netmist_Ops[my_workload].rmw=generic_rmw;
	     Netmist_Ops[my_workload].mkdir=generic_mkdir;
	     Netmist_Ops[my_workload].unlink=generic_unlink;
	     Netmist_Ops[my_workload].create=generic_create;
	     Netmist_Ops[my_workload].append=generic_append;
	     Netmist_Ops[my_workload].locking=generic_locking;
	     Netmist_Ops[my_workload].access=generic_access;
	     Netmist_Ops[my_workload].stat=generic_stat;
	     Netmist_Ops[my_workload].chmod=generic_chmod;
	     Netmist_Ops[my_workload].readdir=generic_readdir;
	     Netmist_Ops[my_workload].copyfile=generic_copyfile;
	     Netmist_Ops[my_workload].rename=generic_rename;
	     Netmist_Ops[my_workload].statfs=generic_statfs;
	     Netmist_Ops[my_workload].pathconf=generic_pathconf;
	     Netmist_Ops[my_workload].custom1=generic_custom1;
	     Netmist_Ops[my_workload].custom2=generic_custom2;
	     Netmist_Ops[my_workload].makepath=generic_make_path;
	     Netmist_Ops[my_workload].prefix_op=generic_prefix_op;
	     Netmist_Ops[my_workload].postfix_op=generic_postfix_op;
	     Netmist_Ops[my_workload].fill_buf=generic_fill_buf;
	     Netmist_Ops[my_workload].init_dir=generic_init_dir;
	     Netmist_Ops[my_workload].init_file=generic_init_file;
	     Netmist_Ops[my_workload].init_empty_file=generic_init_empty_file;
	     Netmist_Ops[my_workload].stat_workdir=generic_stat_workdir;
	     Netmist_Ops[my_workload].chdir=generic_chdir;
	     Netmist_Ops[my_workload].remove_file=generic_remove_file;
	     Netmist_Ops[my_workload].remove_dir=generic_remove_dir;
	     Netmist_Ops[my_workload].setup = generic_setup;
	     Netmist_Ops[my_workload].teardown = generic_teardown;
	     Netmist_Ops[my_workload].impersonate = generic_impersonate;
	     Netmist_Ops[my_workload].make_impersonate_context = generic_make_impersonate_context;
	     chosen_fs = 1;
	     fprintf(newstdout,"Filesystem type %s not supported. Defaulting to POSIX\n", 
		my_mix_table[my_workload].fs_type);
	     fflush(newstdout);
	}

	/* we impersonate user correctly before accessing files*/
	if(cdebug)
	{
 		fprintf(newstdout,"%s Child impersonate.!!\n",
			get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	context = (*Netmist_Ops[my_workload].make_impersonate_context)();
	ret = (*Netmist_Ops[my_workload].impersonate)(context);
	if(ret != 0)
	{
	   c_eval = LOGON_FAILURE;
	   c_eflag++;
	   R_exit(ret);
	}

	if(cdebug)
	{
		fprintf(newstdout,"%s Child statdir.!!\n",
			get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	/*
	 * Stat the workdir. If it doesn't exist... exit.
	 * Use virtualized API for portability across file system types.
	 * Try a few times, as some systems stumble at first. 
	 */
tryit:
	ret = (*Netmist_Ops[my_workload].stat_workdir)(workdir);
	if(ret != 0)
	{
	   if(retry_count++ < 10)
	   {
	        sleep(1);
		/* Try to create it */
		(*Netmist_Ops[my_workload].init_dir)(workdir);
		goto tryit;
	   }
	   c_eval = NO_WORKDIR;
	   c_eflag++;
	   R_exit(ret);
	}
		
	/*
	 * Save my pid in a file so that later it can be used to shutdown
	 * the testing. (if needed) This is only used for emergency shutdown.
	 * This is a fall back, if the default shutdown fails.
	 */
	create_kill_file();

	/*
	 * CD to workdir.  Use virtualized API for portability.
	 */
	ret = (*Netmist_Ops[my_workload].chdir)(workdir);
	if(ret < 0)
	{
 		fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Can not chdir to working directory. Error %s\n",
			strerror(errno));
		fflush(newstdout);
		c_eval = NO_WORKDIR;
		c_eflag++;
		R_exit(ret);
	}

	if(fd_caching_limit != 0)
	{
		if(fd_caching_limit < max_open)
		   max_open = fd_caching_limit;
	}

	if(op_rate != 0.0)
	{
		/*calculate average time for each op*/
		op_time_ms = 1000/op_rate;
		if (op_time_ms >=2.0)
			{
			/*seed by moving client id to most signaficate bits*/
			netmist_srand(reverse_bits(client_id)^(int)gettime());
			delay = netmist_rand()%(int)op_time_ms;
			if (cdebug)
				fprintf( newstdout, "Before test starts, the client will sleep for %d ms (max %.2f ms)\n", delay, op_time_ms);
			}

	   fprintf(newstdout,"Client %d: My_workload id %d Name: %s Op_rate = %.2f\n",client_id,my_workload,
		my_mix_table[my_workload].workload_name,op_rate);
	   fflush(newstdout);
	}
	else
	{
	   fprintf(newstdout,"Client %d: My_workload id %d Name: %s\n",client_id,my_workload,
		my_mix_table[my_workload].workload_name);
	   fflush(newstdout);
	}

	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,,assigned workload = %s,,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],
			my_mix_table[my_workload].workload_name);

	/*
	 * Impersonate is used to have all files be shared by 
 	 * all procs, on all clients. By using impersonate all 
 	 * of the files appear to be created, and accessed, by
	 * a single client ID. (client id = 0)
	 */
	if( !(setfsflag) && (!sharing_flag) && my_mix_table[my_workload].sharemode)
		fsize = fsize * (num_clients); /* Re-scale */

	if(my_mix_table[my_workload].sharemode)
	{
		fprintf(newstdout,"Using share mode \n");
		fflush(newstdout);
		sharing_flag =1;
	}

	if(sharing_flag)
	   impersonate = 0;
	else
	   impersonate = client_id;
	/*
	 * Get buffers ready to go;
 	 */
	netmist_srand(INIT_SEED+impersonate);
#ifdef RANDOMIZE_ACROSS_LOAD_POINTS
	init_genrand64((unsigned long long) time(0) + 
		(unsigned long long)impersonate);
#else
	init_genrand64((unsigned long long) (INIT_SEED+impersonate)); 
#endif

	patt=netmist_rand() & 0x7fffffff;

	maxiosize = rsize; /* Default 4 K buffer */

	/*
	 * find the largest xfer size needed 
	 */
	for( b = 0; b < MAX_DIST_ELEM; b++)
	{
	  if( (my_mix_table[my_workload].read_dist[b].percent > 0) || 
		(my_mix_table[my_workload].write_dist[b].percent > 0))
	  {
	    if(my_mix_table[my_workload].read_dist[b].size_max > (unsigned int)maxiosize)
	      maxiosize=my_mix_table[my_workload].read_dist[b].size_max;

	    if(my_mix_table[my_workload].write_dist[b].size_max > (unsigned int)maxiosize)
	      maxiosize=my_mix_table[my_workload].write_dist[b].size_max;
	  }
	}
	max_xfer_size = maxiosize; /* Push to global here */
	/* Allocate buffer, and align for O_DIRECT */
	main_buf= (char *) MALLOC(maxiosize + 4096);
	bzero(main_buf,maxiosize + 4096);
	main_buf = (char *)( ((long)main_buf + 4096) & ~4095); 
	fprintf(newstdout,"Max io size is %d\n",maxiosize);
	fflush(newstdout);

	(*Netmist_Ops[my_workload].fill_buf)(impersonate, patt, maxiosize, main_buf);

	/*
	 * Determine the timer resolution.
	 */
	my_time_res = 999999.99;
	for(j=0;j<2;j++)
	{
	   my_start_time=gettime(); /* Prime path */
	   sleep(1);
	}
	delay_count=1;

	for(j=0;j<10;j++)
	{
over:
	   my_start_time=gettime(); /* Prime path */
	   for(u=0;u<delay_count;u++)
		;
	   my_start_time2=gettime(); /* Prime path */
	   if(my_start_time2-my_start_time <= 0.0)
	   {
		delay_count++;
		goto over;
	   }
	   res=my_start_time2-my_start_time;
	   if(res < 0.0)
		goto over;  /* Yeah, in VMs */
	   if( res < my_time_res)
		my_time_res = res;
	}
        fprintf(newstdout,"Max open files per process %d\n",max_open);
        fprintf(newstdout,"Max PROCs %d\n",max_proc);
        fprintf(newstdout,"Timer resolution %f seconds.\n",my_time_res);
	fflush(newstdout);
	nap(1); /* Prime instruciton and data caches, and clear page faults */
	my_start_time = my_start_time2;
	nap(1); /* Now take the meausrement */
	my_start_time2=gettime();
	fprintf(newstdout,"Nap resolution %f seconds.\n",
		my_start_time2-my_start_time);
        fflush(newstdout);
	
	/*
	 * Randomize the seed so that each client will perform slightly 
	 * different selections of file ops.
	 */
	netmist_srand(impersonate+1);
	/*
	 * Initialize internal data structures..
	 */
	init_tests(test_ops,num_clients);

	/*
	 * Initialize internal data structures..
	 */
	init_dist_tables();

	if(cdebug) 
	{
		fprintf(newstdout,"Client %d performing %d ops. File size %lld kbytes, Record size %ld kbytes.\n",
			client_id, number_ops,fsize,avg_read_size);
		fflush(newstdout);
	}

	/*
	 * Validate that all of the Op types, will actually work on the client.
	 */
	if(tracedebug) {
		tracestate = TRACE_STATE_VALIDATION;
		fflush(tracefile);
	}
	fprintf(newstdout, "my_workload is %d\n", my_workload);
	client_validate_ops();

	/*
	 * Initialize all of the files that will be used by this client.
	 */
	if(tracedebug) {
		tracestate = TRACE_STATE_CREATE;
		fflush(tracefile);
	}
	if( sharing_flag ) {
		client_init_files(0);
	} else {
		client_init_files(client_id);
	}
	fprintf(newstdout,"%s Finished client init files\n",
                                get_timestamp(&time_string[0]));
	fflush(newstdout);

	/*
	 * Wait for the other clients to finish initialization of their data 
	 * set.
 	 */
	tell_master_ready(client_id);
	if(tracedebug) {
		tracestate = TRACE_STATE_RUN;
		fflush(tracefile);
	}

	/*
	 * Wait until the master tells me that everyone is done with init
	 * and ready to begin the test.
 	 */
	child_wait();

	if(sharing_flag)
	   netmist_srand((client_id+1)*3); /* Now re-seed, for the shared file case */
	/*
	 * Now everyone starts at the same time.
	 */
	tick = 0;
        my_last_tick = 0.0;
	prev_pct_complete = 0;
	if(op_rate > 0.0)
	{
		i_scale=op_rate*0.05; /* ops in burst = rate * time -- sets basic interval to 0.05 seconds */
		if(i_scale < 4) 
			i_scale=4; /* Can't let it be 0, and there is no point in hitting every operation at low rates */
	}
/*
	printf("%s Starting warm-up phase.\n",get_timestamp(&time_string[0]));
	fflush(stdout);	
*/
	naps_taken = 0;
	total_seconds_napped = 0.0;
	total_file_ops = 0.0;
	prev_file_ops = 0.0;
	my_start_time=gettime();
	time_of_last_tick = my_start_time;
	loops = 1;
	sleep(5); /* needed so that the master can get into it's listen loop */

	if(client_id == (num_clients-1))
	   tell_master_phase(client_id,WARM_PHASE);

	while(1)
	{
		my_gen_op_time = gettime();
		gen_op(test_ops,impersonate);

		/*-----------------*/
		/* heartbest      */
		/*-----------------*/
		if( heartbeat_flag && 
		   (((int)(my_gen_op_time - my_start_time) % HEARTBEAT_TICK)==0) && 
		   ((int)(my_gen_op_time - my_start_time)!= prev_heartbeat) 
		   && (client_id == (num_clients-1)))
		{
			 tell_master_heartbeat(client_id,WARM_BEAT,ach_ops_per_second);
		         prev_heartbeat = (int)(my_gen_op_time - my_start_time);
		}
		/*------------------------------------------*/
		/* Now check percent of warm time permitted */
		/*------------------------------------------*/
		cur_time = gettime();
		my_run_time = cur_time-my_start_time;
		pct_complete = (int) (((my_run_time + ELAP_T)* 100.0)/warm_time);
		if(pct_complete > 100)
			pct_complete = 100;

                if(pct_complete && (pct_complete != prev_pct_complete)
		     	&& ((pct_complete % 10) == 0) 
			&& (client_id == (num_clients-1)) )
                {
			tell_master_pct_complete(pct_complete,client_id,
				R_PERCENT_W_COMPLETE);
			prev_pct_complete = pct_complete;
                }
		/*------------------------------------------*/
		if(my_last_tick == 0.0 )
		   my_last_tick = gettime();
		elap_tick = my_gen_op_time - my_last_tick;
		if(elap_tick > (double) ELAP_T)
		{
		     my_last_tick=my_gen_op_time;
		     tick = 0;
                     save_loops = loops;
		     loops = 1;
		}
		else
		{
		     tick = 1;
		}
		if(save_loops == 0)
			save_loops = 1;
		my_warmup_ops++;
		if(tick == 0)
		{
		      ops_this_tick = total_file_ops-prev_file_ops;
		      prev_file_ops = total_file_ops;
		      post_time = gettime();


		      if(post_time - my_start_time >= (double)warm_time)
			break;

		      time_per_tick = post_time-time_of_last_tick;
		      time_of_last_tick = post_time;
		      if(time_per_tick == (double)0.0)
			    time_per_tick = (double)0.001;

		      ach_ops_per_second = (double)ops_this_tick/time_per_tick;
		      if(ach_ops_per_second == (double)0.0)
			    ach_ops_per_second = (double)1.000;
#if defined(PDSM_R_REGISTRY)
		      /*
		       * If the PDSM mechansim is active....
		       * then update the stats.
		       */
		      if(pdsm_stats)
		      {
                        pdsm_stats->read_throughput =  (total_read_bytes + 
				(total_meta_ops * META_B))/(float)(cur_time - my_start_time);
                        pdsm_stats->write_throughput =  (total_write_bytes + 
				(total_meta_ops * META_B))/(float)(cur_time - my_start_time);
			pdsm_stats->requested_op_rate = (double)op_rate;
			pdsm_stats->achieved_op_rate = ach_ops_per_second;
			pdsm_stats->run_time = post_time - my_start_time;
			pdsm_stats->cur_state = WARM_BEAT;
		      }
		      if(pdsm_control)
		      {
			set_pdsm_attributes();
		      }
	
#endif

		}
		if(op_rate > 0.0)
		{
			/* Inject a nap every i_scale ops, scaling nap to fit op_rate */
			if(inject%i_scale == 0)
			{
			    if( (double)(total_file_ops)/op_rate > my_run_time )
			    {
				seconds_to_nap = (double)total_file_ops / op_rate - my_run_time;
        			if(cdebug)
				{
				    fprintf( newstdout, "Nap for %f ms %f %d %f %f %d Warmup \n", 
					seconds_to_nap*1000, 
					total_file_ops, i_scale, op_rate, my_run_time, my_warmup_ops );
				    fflush(newstdout);
				}
				/* need to count naps and accumulated nap time to report at end */
				nap( seconds_to_nap*1000 );
				total_seconds_napped += seconds_to_nap;
				naps_taken++;
			    }
			    inject=1;
			}
 			inject++;
		}
		loops++;
	}
        if((pct_complete!=100)  && (prev_pct_complete !=100)
		&& (client_id == (num_clients-1)) )
        {
		tell_master_pct_complete(100,client_id,
				R_PERCENT_W_COMPLETE);
        }
	fprintf( newstdout, "Napped %d times for a total of %f seconds in warmup\n", naps_taken, total_seconds_napped );
	fflush(newstdout);
	naps_taken = 0;
	total_seconds_napped = 0.0;
	total_file_ops = 0.0;
	prev_file_ops=0.0;
	clear_stats();
 	/* Now for the measurement phase */
	loops = 1;
	my_start_time=gettime();

	if(client_id == (num_clients-1))
	   tell_master_phase(client_id,RUN_PHASE);

	while(1)
	{
		my_gen_op_time = gettime();
		gen_op(test_ops,impersonate);
		cur_time = gettime();
		if(tmp_time < (cur_time-my_gen_op_time))
		  tmp_time = ((cur_time-my_gen_op_time)-my_time_res);
		else
		  tmp_time = (cur_time-my_gen_op_time);
		my_run_time = cur_time-my_start_time;

		/*-----------------*/
		/* heartbest       */
		/*-----------------*/
		if( heartbeat_flag && 
		  (((int)(my_gen_op_time - my_start_time) % HEARTBEAT_TICK)==0)
		  && ((int)(my_gen_op_time - my_start_time) != prev_heartbeat)
		  && (client_id == (num_clients-1)))
		{
			tell_master_heartbeat(client_id,RUN_BEAT, ach_ops_per_second);
		        prev_heartbeat = (int)(my_gen_op_time - my_start_time);
		}
		/*----------------------------*/
		/* For the combined workloads */
		/*----------------------------*/
		total_file_op_time += tmp_time;

		/*-----------------------------*/
		/* For the particular workload */
		/*-----------------------------*/
		work_obj[my_workload].total_file_op_time += tmp_time;

                if(my_last_tick == 0.0)
                        my_last_tick = gettime();
                elap_tick = my_gen_op_time - my_last_tick;
                if(elap_tick > (double)ELAP_T)
                {
                        my_last_tick = my_gen_op_time;
                        tick=0;
			save_loops = loops;
			loops = 1;
                }
                else
                {
                     tick=1;
                }

		if(tick == 0)
		{

		     pct_complete = (int) (((my_run_time + ELAP_T) * 100.0)/run_time);
		     if(pct_complete > 100)
			   pct_complete = 100;

                     if(pct_complete && (pct_complete != prev_pct_complete)
		     	&& ((pct_complete % 10) == 0)
		  	&& (client_id == (num_clients-1)))
                     {
			tell_master_pct_complete(pct_complete,client_id,
				R_PERCENT_COMPLETE);
			prev_pct_complete = pct_complete;
                     }
		}

		if (cur_time - my_start_time >= (double)run_time) 
		{
			run_time = cur_time - my_start_time;
			break;
		}
		my_ops++;

		if(tick == 0)
		{
		      ops_this_tick = total_file_ops-prev_file_ops;
		      prev_file_ops = total_file_ops;
		      post_time = cur_time;

		      time_per_tick = post_time-time_of_last_tick;
		      if(time_per_tick == (double)0.0)
			    time_per_tick = (double)0.001;
		      time_of_last_tick = post_time;

		      ach_ops_per_second = (double)ops_this_tick/time_per_tick;
		      if(ach_ops_per_second == (double)0.0)
			    ach_ops_per_second = (double)1.000;
#if defined(PDSM_R_REGISTRY)
		      /*
		       * If the PDSM mechansim is active....
		       * then update the stats.
		       */
		      if(pdsm_stats)
		      {
                        pdsm_stats->read_throughput =  (total_read_bytes + 
				(total_meta_ops * META_B))/(float)(cur_time - my_start_time);
                        pdsm_stats->write_throughput =  (total_write_bytes + 
				(total_meta_ops * META_B))/(float)(cur_time - my_start_time);
			pdsm_stats->requested_op_rate = (double)op_rate;
			pdsm_stats->achieved_op_rate = ach_ops_per_second;
			pdsm_stats->run_time = cur_time - my_start_time;
			pdsm_stats->cur_state = RUN_BEAT;
		      }
		      if(pdsm_control)
		      {
			set_pdsm_attributes();
		      }
#endif
		}
		if(op_rate > 0.0)
		{
			if(inject%i_scale == 0)
			{
                            if(  (double)total_file_ops / op_rate > my_run_time )
                            {
				seconds_to_nap = (double)total_file_ops / op_rate - my_run_time;
				if(cdebug)
                                 {
				    fprintf( newstdout, "Nap for %f ms %f %d %.2f %f %d Run\n",
                                        seconds_to_nap * 1000,
                                        total_file_ops, i_scale, op_rate, my_run_time, my_ops );
				    fflush(newstdout);
				 }
                                 nap( seconds_to_nap * 1000 );
				/* need to count naps and accumulated nap time to report at end */
				total_seconds_napped += seconds_to_nap;
				naps_taken++;
                            }
			    inject=1;
			}
 			inject++;
		}
		loops++;
	}
        if((pct_complete!=100)  && (prev_pct_complete !=100) 
	   && (client_id == (num_clients-1)))
        {
		tell_master_pct_complete(100,client_id,
				R_PERCENT_COMPLETE);
        }
	fprintf( newstdout, "Napped %d times for a total of %f seconds in measurement interval \n", naps_taken, total_seconds_napped );
	fflush(newstdout);

	/*
	 * Testing is done... time to cleanup the files.
	 */
	if(tracedebug) {
		tracestate = TRACE_STATE_CLEANUP;
		fflush(tracefile);
	}

	(*Netmist_Ops[my_workload].teardown)();

/*  	This is now done via the call above. "teardown" 
	if(sharing_flag)
	{
	  if(cleanup_flag )
	     client_remove_files(0);
	}
	else
	{
	   if(cleanup_flag)
	      client_remove_files(client_id);
	}
*/

	/*
	 * Report the statistics of the run to the controlling process.
	 */
	report_stats();

	/*
	 * Remove my kill file as I no longer need to cleanup any files.
	 */
	remove_kill_file();
	/*
	 * Say good night.
	 */
	 /*
	  * send results to master
	  */
	fprintf(newstdout,"%s Client reporting results.\n",get_timestamp(&time_string[0]));
	fflush(newstdout);

	tell_master_results(client_id);

	/*
	 * Close the trace log (if tracing is on)
	 */
	if(tracedebug)
	{
		fprintf(tracefile,"%d,%lu,%s,,Trace Log Close,,,,,,%s\n",client_id,trace_seq++,
			trace_states[tracestate],get_timestamp(&time_string[0]));
		fflush(tracefile);
		fclose(tracefile);
	}

	/* 
	 * Wait for the master to send the terminate message 
	 */
	child_wait(); 

	/*
	 * Log the fact that Elvis has left the building.
	 */
	fprintf(newstdout,"%s Child exiting the system.\n",
		get_timestamp(&time_string[0]));
	fflush(newstdout);
	/*
	 * To great beyond.
	 */
	exit(0);
	/*
	 * I sure hope we never get here :-)
	 */
}

/*
 * Report the client's statistics
 * Send results back to the controlling process's current working directory.
 * Step 1: Put results in the exec directory of the client.
 * Step 2: rcp the results back to the controlling process's working dir.
 */
#ifdef HAVE_ANSIC_C
void
report_stats(void)
#else
void
report_stats()
#endif
{
	char myname[400];
	char mydir[400];
#if defined(USECP)
	char command[500];
#endif
	char deliminator='/';
	FILE *out;
	int x;
	char* ptr;

#if defined(WIN32)
	deliminator = '\\';
#endif

#ifdef EXEC_DIR_RES
	/*
	 * Construct path for results of this client.
	 */
	strcpy(mydir,execdir);
	x=strlen(mydir);
	ptr=&mydir[x-1];

	while(*ptr != deliminator)
		ptr--;
	*ptr=0;
#else
	getcwd(mydir,400);
#endif
	/*remove tail slash*/
	x=strlen(mydir);
	ptr=&mydir[x-1];
	while(*ptr == deliminator)
		ptr--;
	*(ptr+1)=0;

	fprintf(newstdout,"%s Client reporting results.\n",get_timestamp(&time_string[0]));
	fflush(newstdout);

#if defined(WIN32)
	sprintf(myname,"%s\\%s%d_results",mydir,"Client_",client_id);
#else
	sprintf(myname,"%s/%s%d_results",mydir,"Client_",client_id);
#endif

	/*
	 * Remove any old results.
	 */
	/* Can use POSIX, as this is a log and not workload specific */
	unlink(myname);
	out=fopen(myname,"w+");
	if(out==0)
	{
		fprintf(newstdout,"%s Unable to open client output results file %s\n",get_timestamp(&time_string[0]),myname);
		fflush(newstdout);
		c_eval = NO_CREATE;
		c_eflag++;
		R_exit(105);
	}
	fprintf(out,"\nClient %s ID:%d results:\n",client_localname,client_id);
        fprintf(out,"Workload name      : %s\n",work_obj[my_workload].work_obj_name);
	fprintf(out,"Open file cache count %d\n",cached_open_count);
	fprintf(out,"\n");
	fprintf(out,"Client process has %d initialized directories\n",gdir_count);
	fprintf(out,"with a total of %d initialized files. \n", gfile_count);
	fprintf(out,"Client process total active data space = %lld MiBytes\n\n", (fsize*(client_dirs_pop))/(long long)1024);
	fprintf(out,"Client memory size      = %lld MiBytes\n",client_memsize_mb);
	fprintf(out,"Adjustable aggregate data set value = %lld MiBytes\n",agg_set_memsize_mb);
	fprintf(out,"Warmup time %d seconds.\n",warm_time);
	fprintf(out,"%d total logical ops. File size %lld kbytes, Record size %ld kbytes. \n",my_ops,fsize,avg_read_size);
	fprintf(out,"Runtime of %d seconds\n",run_time);
	fprintf(out,"Min Direct I/O size %d bytes\n\n",(int)min_direct_size);

	if( (work_obj[my_workload].read_op_count != 0) || (work_obj[my_workload].read_rand_op_count != 0)  ||
	    (work_obj[my_workload].mmap_read_op_count != 0) || (work_obj[my_workload].read_file_op_count != 0))
	{
	      fprintf(out,"Read throughput        = %10.2f kbytes/sec.  \n",
		/* 
		 * Credit all meta-data ops with META_B bytes of data, and 
                 * calculate the actual reads & writes with the average sizes.
		 */
		((total_read_bytes + (total_meta_ops * META_B))/run_time)/1024);
                read_throughput =  (total_read_bytes + (total_meta_ops * META_B))/run_time;

	      fprintf(out,"Read %10.2f kbytes total. \n",total_read_bytes/1024);


	}
	else
	{
	      fprintf(out,"Read throughput   = No reads were performed. Increase requested ops\n");
	}

	if( (work_obj[my_workload].write_op_count != 0) || (work_obj[my_workload].write_rand_op_count != 0) ||
	     (work_obj[my_workload].mmap_write_op_count != 0)|| (work_obj[my_workload].write_file_op_count != 0) )
	{
	      fprintf(out,"Write throughput       = %10.2f kbytes/sec.\n",
		/* 
		 * Credit all meta-data ops with META_B kbytes of data, and 
                 * calculate the actual reads & writes with the average sizes.
		 */
		((total_write_bytes + (total_meta_ops * META_B))/run_time)/1024);
                write_throughput =  (total_write_bytes + (total_meta_ops * META_B))/run_time;

	      fprintf(out,"Write %10.2f kbytes total.\n",total_write_bytes/1024);
	}
	else
	{
	      fprintf(out,"Write throughput  = No writes were performed. Increase requested ops\n");
	}

	fprintf(out,"%s \n",get_timestamp(&time_string[0]));
	fprintf(out,"Ops distribution:\n");
	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].read_string, 
			work_obj[my_workload].read_op_count);
	if((work_obj[my_workload].read_op_count != 0) &&
		(work_obj[my_workload].read_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n",work_obj[my_workload].read_op_time/(double)work_obj[my_workload].read_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].read_file_string, 
			work_obj[my_workload].read_file_op_count);
	if((work_obj[my_workload].read_file_op_count != 0) &&
		(work_obj[my_workload].read_file_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n",work_obj[my_workload].read_file_op_time/(double)work_obj[my_workload].read_file_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].read_rand_string, 
		work_obj[my_workload].read_rand_op_count);
	if((work_obj[my_workload].read_rand_op_count!=0) &&
		(work_obj[my_workload].read_rand_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n",
		work_obj[my_workload].read_rand_op_time / (double)work_obj[my_workload].read_rand_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].rmw_string, 
			work_obj[my_workload].rmw_op_count);
	if((work_obj[my_workload].rmw_op_count!=0) &&
		(work_obj[my_workload].rmw_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n",
		work_obj[my_workload].rmw_op_time / (double)work_obj[my_workload].rmw_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld","Open file",
		work_obj[my_workload].open_op_count);
	fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld","Close file",
		work_obj[my_workload].close_op_count);
	fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].write_string, 
		work_obj[my_workload].write_op_count);
	if((work_obj[my_workload].write_op_count != 0) &&
		(work_obj[my_workload].write_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].write_op_time / (double)work_obj[my_workload].write_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].write_file_string, 
			work_obj[my_workload].write_file_op_count);
	if((work_obj[my_workload].write_file_op_count != 0) &&
		(work_obj[my_workload].write_file_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].write_file_op_time / (double)work_obj[my_workload].write_file_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].mmap_write_string, 
			work_obj[my_workload].mmap_write_op_count);
	if((work_obj[my_workload].mmap_write_op_count != 0) &&
		(work_obj[my_workload].mmap_write_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].mmap_write_op_time / (double)work_obj[my_workload].mmap_write_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].mmap_read_string, 
			work_obj[my_workload].mmap_read_op_count);
	if((work_obj[my_workload].mmap_read_op_count != 0) &&
		(work_obj[my_workload].mmap_read_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].mmap_read_op_time / (double)work_obj[my_workload].mmap_read_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].write_rand_string, 
			work_obj[my_workload].write_rand_op_count);
	if((work_obj[my_workload].write_rand_op_count!=0) &&
		(work_obj[my_workload].write_rand_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].write_rand_op_time/(double)work_obj[my_workload].write_rand_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].mkdir_string, 
		work_obj[my_workload].mkdir_op_count);
	if((work_obj[my_workload].mkdir_op_count!=0) &&
		(work_obj[my_workload].mkdir_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n",work_obj[my_workload].mkdir_op_time / (double)work_obj[my_workload].mkdir_op_count);
	else
	   fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].unlink_string, 
			work_obj[my_workload].unlink_op_count);
	if((work_obj[my_workload].unlink_op_count != 0) &&
		(work_obj[my_workload].unlink_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f \n", work_obj[my_workload].unlink_op_time / (double)work_obj[my_workload].unlink_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].create_string, 
			work_obj[my_workload].create_op_count);
	if((work_obj[my_workload].create_op_count != 0) &&
		(work_obj[my_workload].create_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f \n", work_obj[my_workload].create_op_time / (double)work_obj[my_workload].create_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].append_string, 
			work_obj[my_workload].append_op_count);
	if((work_obj[my_workload].append_op_count!=0) &&
		(work_obj[my_workload].append_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f \n", work_obj[my_workload].append_op_time / (double)work_obj[my_workload].append_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].locking_string, 
			work_obj[my_workload].lock_op_count);
	if((work_obj[my_workload].lock_op_count!=0) &&
		(work_obj[my_workload].lock_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].lock_op_time / (double)work_obj[my_workload].lock_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].access_string, 
			work_obj[my_workload].access_op_count);
	if((work_obj[my_workload].access_op_count!=0) &&
		(work_obj[my_workload].access_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].access_op_time / (double)work_obj[my_workload].access_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].stat_string, 
			work_obj[my_workload].stat_op_count);
	if((work_obj[my_workload].stat_op_count!=0) &&
		(work_obj[my_workload].stat_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n",work_obj[my_workload].stat_op_time / (double)work_obj[my_workload].stat_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].chmod_string, 
			work_obj[my_workload].chmod_op_count);
	if((work_obj[my_workload].chmod_op_count!=0) &&
		(work_obj[my_workload].chmod_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].chmod_op_time / (double)work_obj[my_workload].chmod_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].readdir_string, 
			work_obj[my_workload].readdir_op_count);
	if((work_obj[my_workload].readdir_op_count!=0) &&
		(work_obj[my_workload].readdir_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].readdir_op_time / (double)work_obj[my_workload].readdir_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].copyfile_string, 
			work_obj[my_workload].copyfile_op_count);
	if((work_obj[my_workload].copyfile_op_count!=0) &&
		(work_obj[my_workload].copyfile_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].copyfile_op_time / (double)work_obj[my_workload].copyfile_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].rename_string, 
			work_obj[my_workload].rename_op_count);
	if((work_obj[my_workload].rename_op_count!=0) &&
		(work_obj[my_workload].rename_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].rename_op_time / (double)work_obj[my_workload].rename_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].statfs_string, 
			work_obj[my_workload].statfs_op_count);
	if((work_obj[my_workload].statfs_op_count!=0) &&
		(work_obj[my_workload].statfs_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].statfs_op_time / (double)work_obj[my_workload].statfs_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].pathconf_string, 
			work_obj[my_workload].pathconf_op_count);
	if((work_obj[my_workload].pathconf_op_count!=0) &&
		(work_obj[my_workload].pathconf_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].pathconf_op_time / (double)work_obj[my_workload].pathconf_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].custom1_string, 
			work_obj[my_workload].custom1_op_count);
	if((work_obj[my_workload].custom1_op_count!=0) &&
		(work_obj[my_workload].custom1_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n", work_obj[my_workload].custom1_op_time / (double)work_obj[my_workload].custom1_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n");

	fprintf(out,"\t%-35s ops = %10ld", my_mix_table[my_workload].custom2_string, 
			work_obj[my_workload].custom2_op_count);
	if((work_obj[my_workload].custom2_op_count!=0) &&
		(work_obj[my_workload].custom2_op_time!=0.0))
	      fprintf(out," \tAvg Latency: %10.6f\n\n", work_obj[my_workload].custom2_op_time / (double)work_obj[my_workload].custom2_op_count);
	else
	      fprintf(out," \tAvg Latency:   No_count.\n\n");

        fprintf(out,"\tOps/sec              = %10.2f\n",
                work_obj[my_workload].total_file_ops/run_time);
	fprintf(out,"\tTotal file ops       = %10.0f\n",total_file_ops);
	fprintf(out,"\tAverage latency      = %7.2f milli seconds\n",
			(total_file_op_time/total_file_ops)*(double)1000);
	fprintf(out,"\n");
	fflush(out);	
	/*
	 * Testing is done... time to dump the hot-files list.
	 */
	if(dump_files_flag)
	   dump_file_objects(out);

	if(Op_lat_flag)
	   child_dump_hist(out);
	/*
	 * Construct command to copy results back to master.
	 */
#if defined(USECP)
	if(!localflag)
	{
#if defined(WIN32)
	  sprintf(command,"%s %s %s", "copy",myname,map_to_network_path(master_name, master_results_dir));
#else
	  sprintf(command,"%s %s %s:%s",remcp,myname,master_name,master_results_dir);
#endif
	}
	else
	{
#if defined(WIN32)
	  sprintf(command,"%s %s %s","copy",myname,master_results_dir);
#else
	  sprintf(command,"%s %s %s","cp",myname,master_results_dir);
#endif
	}
	if(ldebug) 
	{
 	   fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
	   fprintf(newstdout,"Copy results to master command: %s\n",command);
	   fflush(newstdout);
	}
	/*
	 * Copy results back to master.
	 */
	system(command);
#endif
	fflush(out);	
	fclose(out);
}

/*
 * Allocate the space for the internal test array, then fill it in according
 * to the distribution mix.
 */
#ifdef HAVE_ANSIC_C
void
init_tests(int num_test_ops,int num_clients)
#else
void
init_tests(num_test_ops, num_clients)
int num_test_ops, num_clients;
#endif
{
	struct test *testd;
	/* Dummy guard band of unused memory */
	testd=(struct test *)MALLOC(sizeof(struct test) * num_test_ops);
	if(testd == 0)
	{
		fprintf(newstdout,"Malloc failed.\n");
	  	fflush(newstdout);
		exit(1);
	}
	tests=(struct test *)MALLOC(sizeof(struct test) * num_test_ops);
	if(tests == 0)
	{
		fprintf(newstdout,"Malloc failed.\n");
	  	fflush(newstdout);
		exit(1);
	}
	bzero(tests, (sizeof(struct test) * num_test_ops));
	/* Dummy guard band of unused memory */
	testd=(struct test *)MALLOC(sizeof(struct test) * num_test_ops);
	if(testd == 0)
	{
		fprintf(newstdout,"Malloc failed.\n");
	  	fflush(newstdout);
		exit(1);
	}
	create_tests(num_test_ops,num_clients);
}

/*
 * Initialize the test array. There are "num_test_ops" entries in the array
 * and this code distributes the percentages of each type of
 * test across the array. Note: my_workload is already set by now !!!.
 */
#ifdef HAVE_ANSIC_C
void
create_tests(int num_test_ops,int num_clients)
#else
void
create_tests( num_test_ops, num_clients)
int num_test_ops,num_clients;
#endif
{
        int x,j;
	int k;
	struct test *tp;
	if(ldebug) 
	{
 	        fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Init %d test ops\n",num_test_ops);
		fflush(newstdout);
	}
	k=0;
	/*
	 * Initialize "writes" in the test array.
         * x=(num_test_ops*PERCENT_WRITE)/100;	
	 */
        x=(num_test_ops * my_mix_table[my_workload].percent_write)/100;	
	if(ldebug) 
	{
		fprintf(newstdout,"Init %d write test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_WRITE;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "write file ops" in the test array.
	 */
        x=(num_test_ops * my_mix_table[my_workload].percent_write_file)/100;	
	if(ldebug) 
	{
		fprintf(newstdout,"Init %d write file test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_WRITE_FILE;
		tp->status_flags=NONE;
	}

	/*
	 * Initialize "mmap writes" in the test array.
         * x=(num_test_ops*PERCENT_MM_WRITE)/100;	
	 */
        x=(num_test_ops * my_mix_table[my_workload].percent_mmap_write)/100;	
	if(ldebug) 
	{
		fprintf(newstdout,"Init %d mmap write test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_MM_WRITE;
		tp->status_flags=NONE;
	}

	/*
	 * Initialize "write rand" in the test array.
         * x=(num_test_ops*PERCENT_WRITE_RAND)/100;	
	 */
        x=(num_test_ops * my_mix_table[my_workload].percent_write_rand)/100;	
	if(ldebug) 
	{
		fprintf(newstdout,"Init %d write rand test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_WRITE_RAND;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "reads" in the test array.
	 */
        x=(num_test_ops*PERCENT_READ)/100;	
	if(ldebug) 
	{
		fprintf(newstdout,"Init %d read test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_READ;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "read file ops" in the test array.
	 */
        x=(num_test_ops*PERCENT_READ_FILE)/100;	
	if(ldebug) 
	{
		fprintf(newstdout,"Init %d read file test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_READ_FILE;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "mmap reads" in the test array.
         * x=(num_test_ops*PERCENT_MM_READ)/100;	
	 */
        x=(num_test_ops * my_mix_table[my_workload].percent_mmap_read)/100;	
	if(ldebug) 
	{
		fprintf(newstdout,"Init %d mmap read test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_MM_READ;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "read rand" in the test array.
	 */
        x=(num_test_ops*PERCENT_READ_RAND)/100;	
	if(ldebug) 
	{
		fprintf(newstdout,"Init %d read rand test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_READ_RAND;
		tp->status_flags=NONE;
	}

	/*
	 * Initialize "rmw" in the test array.
	 */
        x=(num_test_ops*PERCENT_RMW)/100;	
	if(ldebug) 
	{
		fprintf(newstdout,"Init %d rmw test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_RMW;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "mkdir" in the test array.
	 */
        x=(num_test_ops*PERCENT_MKDIR)/100;	
	if(ldebug) 
	{
		fprintf(newstdout,"Init %d mkdir test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_MKDIR;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "unlink" in the test array.
	 */
        x=(num_test_ops*PERCENT_UNLINK)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d unlink test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_UNLINK;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "Create" in the test array.
	 */
        x=(num_test_ops*PERCENT_CREATE)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d create test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_CREATE;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "stat" in the test array.
	 */
        x=(num_test_ops*PERCENT_STAT)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d stat test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_STAT;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "append" in the test array.
	 */
        x=(num_test_ops*PERCENT_APPEND)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d append test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_APPEND;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "locking" in the test array.
	 */
        x=(num_test_ops*PERCENT_LOCKING)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d locking test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_LOCKING;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "access" in the test array.
	 */
        x=(num_test_ops*PERCENT_ACCESS)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d access test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_ACCESS;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "chmod" in the test array.
	 */
        x=(num_test_ops*PERCENT_CHMOD)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d chmod test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_CHMOD;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "readdir" in the test array.
	 */
        x=(num_test_ops*PERCENT_READDIR)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d readdir test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_READDIR;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "copyfile" in the test array.
	 */
        x=(num_test_ops*PERCENT_COPYFILE)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d copyfile test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_COPYFILE;
		tp->status_flags=NONE;
	}

	/*
	 * Initialize "Rename" in the test array.
	 */
        x=(num_test_ops*PERCENT_RENAME)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d rename test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_RENAME;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "statfs" in the test array.
	 */
        x=(num_test_ops*PERCENT_STATFS)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d statfs test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_STATFS;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "pathconf" in the test array.
	 */
        x=(num_test_ops*PERCENT_PATHCONF)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d pathconf test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_PATHCONF;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "custom1" in the test array.
	 */
        x=(num_test_ops*PERCENT_CUSTOM1)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d custom1 test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_CUSTOM1;
		tp->status_flags=NONE;
	}
	/*
	 * Initialize "custom2" in the test array.
	 */
        x=(num_test_ops*PERCENT_CUSTOM2)/100;	
	if(ldebug)
	{
		fprintf(newstdout,"Init %d custom2 test ops\n",x);
		fflush(newstdout);
	}
	for(j=0;(j < x && k < num_test_ops);j++)
	{
		tp=&tests[k++];
		tp->op_type=OP_CUSTOM2;
		tp->status_flags=NONE;
	}
	if(ldebug)
	{
		fprintf(newstdout,"Initialized %d tests\n",k);
		fflush(newstdout);
	}
}

/*
 * This is the call that the client makes to pick its next 
 * operation to perform. It will pick a random entry from
 * the test array and perform that operation.
 *
 * Some operations have work that needs to be done before
 * the operation can be performed, others need some work
 * to be done after an operation has been performed. 
 * The generic_prefix_op() and generic_postfix_op() handle this work.
 */
#ifdef HAVE_ANSIC_C
int
gen_op(int num_test_ops, int client_id)
#else
int
gen_op(num_test_ops, client_id)
int num_test_ops, client_id;
#endif
{
	int x;
	struct test *tp;
	struct file_object *fp;

	/* 
	 * Pick a random operation to perform from the test array.
	 */
	x=((unsigned int)netmist_rand()) % num_test_ops;
	tp=&tests[x];
	if(cdebug)
	{
		fprintf(newstdout,"Client picks test number %d OP = %d\n",x,tp->op_type);
		fflush(newstdout);
	}
	switch(tp->op_type){
	case OP_WRITE:
		/* Remove the file first ? */
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].write)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_WRITE_FILE:
		/* Remove the file first ? */
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].write_file)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_MM_WRITE:
		/* Remove the file first ? */
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].mmap_write)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_MM_READ:
		/* Remove the file first ? */
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].mmap_read)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_WRITE_RAND:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].write_rand)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_READ:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].read)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_READ_FILE:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].read_file)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_READ_RAND:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].read_rand)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_RMW:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].rmw)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_MKDIR:
		/* Remove the directory first ? */
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].mkdir)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_UNLINK:
		/* Put back after unlink ??? */
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].unlink)(fp,0);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_CREATE:
		/* Put back after unlink ??? */
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].create)(fp,0);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_STAT:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].stat)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_CHMOD:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].chmod)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_READDIR:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].readdir)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_ACCESS:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].access)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_LOCKING:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].locking)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_APPEND:
		/* Remove the file first ? */
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].append)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_COPYFILE:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].copyfile)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_RENAME:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].rename)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_STATFS:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].statfs)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_PATHCONF:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].pathconf)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_CUSTOM1:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].custom1)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	case OP_CUSTOM2:
		fp=(*Netmist_Ops[my_workload].makepath)(client_id,x);
		(*Netmist_Ops[my_workload].prefix_op)(x,fp);
		(*Netmist_Ops[my_workload].custom2)(fp);
		(*Netmist_Ops[my_workload].postfix_op)(x,fp);
		break;

	default:
		fprintf(newstdout,"Unknown Op type \n");
		fflush(newstdout);
		break;
	}
	return(x); /* Return index in tests array */
}

/*
 * Create the directory path and file path for this test entry
 * for this client, or child of client. Store the path in the
 * test array entry associated with this operation.
 */
#ifdef HAVE_ANSIC_C
struct file_object *
generic_make_path(int client_id, int test_num)
#else
struct file_object *
generic_make_path(client_id, test_num)
int client_id, test_num;
#endif
{
	int dir,filenum,op_type=0;
	int sel;
	struct test *tp;
	char buf[1024];
	char leaf_name[MAXNAME];
	struct file_object *file_ptr = NULL;
	struct file_object *file_array_ptr = NULL;
	unsigned long jj = 0;

	/*
	 * Pick a type of OP
	 */
	tp=&tests[test_num];
	/*
	 * Pick a random directory
	 */
	dir=netmist_rand()%client_dirs;
	/*
	 * Pick a random file within this directory.
	 */
	/* filenum=netmist_rand()%client_files;*/

	jj = (netmist_rand() % 100) + 1;
	if((my_mix_table[my_workload].percent_geometric) && 
		(jj <= my_mix_table[my_workload].percent_geometric))
	   filenum=geometric_rand(client_files);  /* jj % of the time */
	else
	   filenum=netmist_rand()%client_files;  /* remaining % of time */

/*	fprintf(newstdout,"value %d\n",filenum);
	fprintf(newstdout,"dir %d File %d\n",dir,filenum);
	fflush(newstdout);
*/
	/*
	 * Pick a file for op type
	 */
	switch(tp->op_type) {
	case OP_WRITE:
		/* Pick seq read or write */
                sel = netmist_rand() % 2;
		switch(sel) {
		case 0:
		  file_array_ptr=op_X_array[OP_WRITE];
		  file_ptr = &file_array_ptr[(dir*client_files)+filenum]; 
                  op_type = OP_WRITE;
		  break;
		case 1:
		  file_array_ptr=op_X_array[OP_READ];
		  file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                  op_type = OP_READ;
		  break;
		}
		break;
	case OP_WRITE_FILE:
		/* Pick seq read or write */
                sel = netmist_rand() % 2;
		switch(sel) {
		case 0:
		  file_array_ptr=op_X_array[OP_WRITE_FILE];
		  file_ptr = &file_array_ptr[(dir*client_files)+filenum]; 
                  op_type = OP_WRITE_FILE;
		  break;
		case 1:
		  file_array_ptr=op_X_array[OP_READ_FILE];
		  file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                  op_type = OP_READ_FILE;
		  break;
		}
		break;
	case OP_MM_WRITE:
		/* Pick seq mmap read or mmap write */
                sel = netmist_rand() % 2;
		switch(sel) {
		case 0:
		  file_array_ptr=op_X_array[OP_MM_WRITE];
		  file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                  op_type = OP_MM_WRITE;
		  break;
		case 1:
		  file_array_ptr=op_X_array[OP_MM_READ];
		  file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                  op_type = OP_MM_READ;
		  break;
		}
		break;
	case OP_MM_READ:
		/* Pick seq mmap read or seq mmap write */
                sel = netmist_rand() % 2;
		switch(sel) {
		case 0:
		  file_array_ptr=op_X_array[OP_MM_READ];
		  file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                  op_type = OP_MM_READ;
		  break;
		case 1:
		  file_array_ptr=op_X_array[OP_MM_WRITE];
		  file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                  op_type = OP_MM_WRITE;
		  break;
		}
		break;
	case OP_READ:
		/* Pick seq read or seq write */
                sel = netmist_rand() % 2;
		switch(sel) {
		case 0:
		   file_array_ptr=op_X_array[OP_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                   op_type = OP_READ;
		   break;
		case 1: 
		   file_array_ptr=op_X_array[OP_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                   op_type = OP_WRITE;
		   break;
		}
		break;
	case OP_READ_FILE:
		/* Pick seq read or seq write */
                sel = netmist_rand() % 2;
		switch(sel) {
		case 0:
		   file_array_ptr=op_X_array[OP_READ_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                   op_type = OP_READ_FILE;
		   break;
		case 1: 
		   file_array_ptr=op_X_array[OP_WRITE_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                   op_type = OP_WRITE_FILE;
		   break;
		}
		break;
	case OP_WRITE_RAND:
		/* Pick rand read or rand write */
                sel = netmist_rand() % 2;
		switch(sel) {
		case 0:
		   file_array_ptr=op_X_array[OP_WRITE_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                   op_type = OP_WRITE_RAND;
		   break;
		case 1:
		   file_array_ptr=op_X_array[OP_READ_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                   op_type = OP_READ_RAND;
		   break;
		}
		break;
	case OP_READ_RAND:
		/* Pick rand read or rand write */
                sel = netmist_rand() % 2;
		switch(sel) {
		case 0:
		   file_array_ptr=op_X_array[OP_READ_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                   op_type = OP_READ_RAND;
		   break;
		case 1:
		   file_array_ptr=op_X_array[OP_WRITE_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
                   op_type = OP_WRITE_RAND;
		   break;
		}
		break;
	case OP_RMW:
		/* Pick rand read or rand write, or rmw  */
		sel = netmist_rand() % 3;
                if(sel==0)
		{
		   file_array_ptr=op_X_array[OP_READ_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_RAND;
		}
                if(sel==1)
		{
		   file_array_ptr=op_X_array[OP_WRITE_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_RAND;
		}
                if(sel==2)
		{
		   file_array_ptr=op_X_array[OP_RMW];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_RMW;
		}
		break;
	case OP_MKDIR:
		file_array_ptr=op_X_array[OP_MKDIR];
		file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		op_type = tp->op_type;
		break;
	case OP_UNLINK:
		file_array_ptr=op_X_array[OP_UNLINK];
		file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		op_type = tp->op_type;
		break;
	case OP_CREATE:
		file_array_ptr=op_X_array[OP_CREATE];
		file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		op_type = tp->op_type;
		break;
	case OP_STAT:
		/* Pick from: write, read, append, rand write, 
		   rand read, access, append, or lock 
		   Don't share with unlink or rename, as things might
	           disappear, in the shared files mode. */
		sel = netmist_rand() % 18;
		switch(sel) {
		case 0: 
		   file_array_ptr=op_X_array[OP_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE;
		   break;
		case 1:
		   file_array_ptr=op_X_array[OP_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ;
		   break;
                case 2:
		   file_array_ptr=op_X_array[OP_APPEND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_APPEND;
		   break;
		case 3:
		   file_array_ptr=op_X_array[OP_WRITE_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_RAND;
		   break;
		case 4:
		   file_array_ptr=op_X_array[OP_READ_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_RAND;
		   break;
                case 5:
		   file_array_ptr=op_X_array[OP_ACCESS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_ACCESS;
		   break;
                case 6:
		   file_array_ptr=op_X_array[OP_RMW];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_RMW;
		   break;
                case 7:
		   file_array_ptr=op_X_array[OP_COPYFILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_COPYFILE;
	 	   break;
                case 8:
		   file_array_ptr=op_X_array[OP_STAT];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STAT;
		   break;
                case 9:
		   file_array_ptr=op_X_array[OP_CHMOD];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_CHMOD;
		   break;
		case 10:
		   file_array_ptr=op_X_array[OP_MM_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_READ;
		   break;
                case 11:
		   file_array_ptr=op_X_array[OP_MM_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_WRITE;
		   break;
		case 12:
		   file_array_ptr=op_X_array[OP_LOCKING];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_LOCKING;
		   break;
                case 13:
		   file_array_ptr=op_X_array[OP_PATHCONF];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_PATHCONF;
		   break;
                case 14:
		   file_array_ptr=op_X_array[OP_STATFS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STATFS;
		   break;
                case 15:
		   file_array_ptr=op_X_array[OP_READDIR];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READDIR;
		   break;
                case 16:
		   file_array_ptr=op_X_array[OP_WRITE_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_FILE;
		   break;
                case 17:
		   file_array_ptr=op_X_array[OP_READ_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_FILE;
		   break;
		}
		break;
	case OP_APPEND:
		file_array_ptr=op_X_array[OP_APPEND];
		file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		op_type = tp->op_type;
		break;
	case OP_LOCKING:
		file_array_ptr=op_X_array[OP_LOCKING];
		file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		op_type = tp->op_type;
		break;
	case OP_ACCESS:
		/* Pick from: write, read, append, rand write, 
		   rand read, access, append, copyfile */
		sel = netmist_rand() % 18;
		switch(sel) {
                case 0:
		   file_array_ptr=op_X_array[OP_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE;
		   break;
                case 1:
		   file_array_ptr=op_X_array[OP_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ;
		   break;
                case 2:
		   file_array_ptr=op_X_array[OP_APPEND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_APPEND;
		   break;
                case 3:
		   file_array_ptr=op_X_array[OP_WRITE_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_RAND;
		   break;
		case 4:
		   file_array_ptr=op_X_array[OP_READ_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_RAND;
		   break;
                case 5:
		   file_array_ptr=op_X_array[OP_ACCESS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_ACCESS;
		   break;
                case 6:
		   file_array_ptr=op_X_array[OP_RMW];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_RMW;
		   break;
                case 7:
		   file_array_ptr=op_X_array[OP_COPYFILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_COPYFILE;
		   break;
                case 8:
		   file_array_ptr=op_X_array[OP_STAT];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STAT;
		   break;
                case 9:
		   file_array_ptr=op_X_array[OP_CHMOD];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_CHMOD;
		   break;
                case 10:
		   file_array_ptr=op_X_array[OP_MM_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_READ;
		   break;
                case 11:
		   file_array_ptr=op_X_array[OP_MM_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_WRITE;
		   break;
                case 12:
		   file_array_ptr=op_X_array[OP_LOCKING];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_LOCKING;
		   break;
                case 13:
		   file_array_ptr=op_X_array[OP_PATHCONF];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_PATHCONF;
		   break;
                case 14:
		   file_array_ptr=op_X_array[OP_STATFS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STATFS;
		   break;
                case 15:
		   file_array_ptr=op_X_array[OP_READDIR];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READDIR;
		   break;
                case 16:
		   file_array_ptr=op_X_array[OP_WRITE_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_FILE;
		   break;
                case 17:
		   file_array_ptr=op_X_array[OP_READ_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_FILE;
		   break;
		}
		break;
	case OP_CHMOD:
		/* Pick from: write, read, append, rand write, 
		   rand read, access, append, or lock */
		sel = netmist_rand() % 18;
		switch(sel) {
                case 0:
		   file_array_ptr=op_X_array[OP_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE;
	   	   break;
                case 1:
		   file_array_ptr=op_X_array[OP_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ;
		   break;
                case 2:
		   file_array_ptr=op_X_array[OP_APPEND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_APPEND;
		   break;
                case 3:
		   file_array_ptr=op_X_array[OP_WRITE_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_RAND;
		   break;
                case 4:
		   file_array_ptr=op_X_array[OP_READ_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_RAND;
		   break;
                case 5:
		   file_array_ptr=op_X_array[OP_ACCESS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_ACCESS;
		   break;
                case 6:
		   file_array_ptr=op_X_array[OP_RMW];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_RMW;
		   break;
		case 7:
		   file_array_ptr=op_X_array[OP_COPYFILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_COPYFILE;
		   break;
                case 8:
		   file_array_ptr=op_X_array[OP_STAT];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STAT;
		   break;
		case 9:
		   file_array_ptr=op_X_array[OP_CHMOD];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_CHMOD;
		   break;
                case 10:
		   file_array_ptr=op_X_array[OP_MM_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_READ;
		   break;
                case 11:
		   file_array_ptr=op_X_array[OP_MM_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_WRITE;
		   break;
                case 12:
		   file_array_ptr=op_X_array[OP_LOCKING];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_LOCKING;
		   break;
                case 13:
		   file_array_ptr=op_X_array[OP_PATHCONF];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_PATHCONF;
		   break;
                case 14:
		   file_array_ptr=op_X_array[OP_STATFS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STATFS;
		   break;
                case 15:
		   file_array_ptr=op_X_array[OP_READDIR];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READDIR;
		   break;
                case 16:
		   file_array_ptr=op_X_array[OP_WRITE_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_FILE;
		   break;
                case 17:
		   file_array_ptr=op_X_array[OP_READ_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_FILE;
		   break;
		}
		break;
	case OP_READDIR:
		/* Pick from: write, read, append, rand write, 
		   rand read, access, append, or lock */
		sel = netmist_rand() % 18;
		switch(sel) {
                case 0:
		   file_array_ptr=op_X_array[OP_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE;
		   break;
                case 1:
		   file_array_ptr=op_X_array[OP_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ;
		   break;
                case 2:
		   file_array_ptr=op_X_array[OP_APPEND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_APPEND;
		   break;
                case 3:
		   file_array_ptr=op_X_array[OP_WRITE_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_RAND;
		   break;
                case 4:
		   file_array_ptr=op_X_array[OP_READ_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_RAND;
		   break;
                case 5:
		   file_array_ptr=op_X_array[OP_ACCESS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_ACCESS;
		   break;
                case 6:
		   file_array_ptr=op_X_array[OP_RMW];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_RMW;
		   break;
                case 7:
		   file_array_ptr=op_X_array[OP_COPYFILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_COPYFILE;
		   break;
                case 8:
		   file_array_ptr=op_X_array[OP_STAT];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STAT;
		   break;
                case 9:
		   file_array_ptr=op_X_array[OP_CHMOD];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_CHMOD;
		   break;
                case 10:
		   file_array_ptr=op_X_array[OP_MM_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_READ;
		   break;
                case 11:
		   file_array_ptr=op_X_array[OP_MM_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_WRITE;
		   break;
                case 12:
		   file_array_ptr=op_X_array[OP_LOCKING];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_LOCKING;
		   break;
                case 13:
		   file_array_ptr=op_X_array[OP_PATHCONF];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_PATHCONF;
		   break;
                case 14:
		   file_array_ptr=op_X_array[OP_STATFS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STATFS;
		   break;
                case 15:
		   file_array_ptr=op_X_array[OP_READDIR];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READDIR;
		   break;
                case 16:
		   file_array_ptr=op_X_array[OP_WRITE_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_FILE;
		   break;
                case 17:
		   file_array_ptr=op_X_array[OP_READ_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_FILE;
		   break;
		}
		break;
	case OP_COPYFILE:
		file_array_ptr=op_X_array[OP_COPYFILE];
		file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		op_type = tp->op_type;
		break;
	case OP_RENAME:
		file_array_ptr=op_X_array[OP_RENAME];
		file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		op_type = tp->op_type;
		break;
	case OP_STATFS:
		/* Pick from: write, read, append, rand write, 
		   rand read, access, append, or lock */
		sel = netmist_rand() % 18;
		switch(sel) {
                case 0:
		   file_array_ptr=op_X_array[OP_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE;
	   	   break;
                case 1:
		   file_array_ptr=op_X_array[OP_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ;
		   break;
                case 2:
		   file_array_ptr=op_X_array[OP_APPEND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_APPEND;
		   break;
                case 3:
		   file_array_ptr=op_X_array[OP_WRITE_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_RAND;
		   break;
                case 4:
		   file_array_ptr=op_X_array[OP_READ_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_RAND;
		   break;
                case 5:
		   file_array_ptr=op_X_array[OP_ACCESS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_ACCESS;
		   break;
                case 6:
		   file_array_ptr=op_X_array[OP_RMW];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_RMW;
		   break;
		case 7:
		   file_array_ptr=op_X_array[OP_COPYFILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_COPYFILE;
		   break;
                case 8:
		   file_array_ptr=op_X_array[OP_STAT];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STAT;
		   break;
		case 9:
		   file_array_ptr=op_X_array[OP_CHMOD];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_CHMOD;
		   break;
                case 10:
		   file_array_ptr=op_X_array[OP_MM_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_READ;
		   break;
                case 11:
		   file_array_ptr=op_X_array[OP_MM_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_WRITE;
		   break;
                case 12:
		   file_array_ptr=op_X_array[OP_LOCKING];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_LOCKING;
		   break;
                case 13:
		   file_array_ptr=op_X_array[OP_PATHCONF];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_PATHCONF;
		   break;
                case 14:
		   file_array_ptr=op_X_array[OP_STATFS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STATFS;
		   break;
                case 15:
		   file_array_ptr=op_X_array[OP_READDIR];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READDIR;
		   break;
                case 16:
		   file_array_ptr=op_X_array[OP_WRITE_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_FILE;
		   break;
                case 17:
		   file_array_ptr=op_X_array[OP_READ_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_FILE;
		   break;
		}
		break;
	case OP_PATHCONF:
		/* Pick from: write, read, append, rand write, 
		   rand read, access, append, or lock */
		sel = netmist_rand() % 18;
		switch(sel) {
		case 0: 
		   file_array_ptr=op_X_array[OP_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE;
		   break;
		case 1:
		   file_array_ptr=op_X_array[OP_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ;
		   break;
                case 2:
		   file_array_ptr=op_X_array[OP_APPEND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_APPEND;
		   break;
		case 3:
		   file_array_ptr=op_X_array[OP_WRITE_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_RAND;
		   break;
		case 4:
		   file_array_ptr=op_X_array[OP_READ_RAND];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_RAND;
		   break;
                case 5:
		   file_array_ptr=op_X_array[OP_ACCESS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_ACCESS;
		   break;
                case 6:
		   file_array_ptr=op_X_array[OP_RMW];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_RMW;
		   break;
                case 7:
		   file_array_ptr=op_X_array[OP_COPYFILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_COPYFILE;
	 	   break;
                case 8:
		   file_array_ptr=op_X_array[OP_STAT];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STAT;
		   break;
                case 9:
		   file_array_ptr=op_X_array[OP_CHMOD];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_CHMOD;
		   break;
		case 10:
		   file_array_ptr=op_X_array[OP_MM_READ];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_READ;
		   break;
                case 11:
		   file_array_ptr=op_X_array[OP_MM_WRITE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_MM_WRITE;
		   break;
		case 12:
		   file_array_ptr=op_X_array[OP_LOCKING];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_LOCKING;
		   break;
                case 13:
		   file_array_ptr=op_X_array[OP_PATHCONF];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_PATHCONF;
		   break;
                case 14:
		   file_array_ptr=op_X_array[OP_STATFS];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_STATFS;
		   break;
                case 15:
		   file_array_ptr=op_X_array[OP_READDIR];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READDIR;
		   break;
                case 16:
		   file_array_ptr=op_X_array[OP_WRITE_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_WRITE_FILE;
		   break;
                case 17:
		   file_array_ptr=op_X_array[OP_READ_FILE];
		   file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		   op_type = OP_READ_FILE;
		   break;
		}
		break;
	case OP_CUSTOM1:
		file_array_ptr=op_X_array[OP_CUSTOM1];
		file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		op_type = tp->op_type;
		break;
	case OP_CUSTOM2:
		file_array_ptr=op_X_array[OP_CUSTOM2];
		file_ptr=&file_array_ptr[(dir*client_files)+filenum]; 
		op_type = tp->op_type;
		break;
	}

	make_name(leaf_name,
	   (client_id*client_dirs*NUM_OP_TYPES * client_files) +
	   (dir*NUM_OP_TYPES*client_files)+(op_type*client_files)+filenum,
	   TYPE_FILE,filenum);

	if(sharing_flag)
	{
#if defined(WIN32)
	   sprintf(buf,"%s\\SM_%s\\Dir%d\\bucket%d\\%s",workdir,
		work_obj[my_workload].work_obj_name,dir,op_type,leaf_name);
#else
	   sprintf(buf,"%s/SM_%s/Dir%d/bucket%d/%s",workdir,
		work_obj[my_workload].work_obj_name,dir,op_type,leaf_name);
#endif
	}
	else
	{
#if defined(WIN32)
	   sprintf(buf,"%s\\CL%d_%s\\Dir%d\\bucket%d\\%s",workdir,client_id,
		work_obj[my_workload].work_obj_name,dir,op_type,leaf_name);
#else
	   sprintf(buf,"%s/CL%d_%s/Dir%d/bucket%d/%s",workdir,client_id,
		work_obj[my_workload].work_obj_name,dir,op_type,leaf_name);
#endif
	}

	if(cdebug) 
	{
		fprintf(newstdout,"Client got path %s\n",buf);
		fflush(newstdout);
	}
	strcpy(file_ptr->filename,buf);

	file_ptr->access_count++; /* Bump access count for this file */

	if(sharing_flag)
	{
#if defined(WIN32)
	   sprintf(buf,"%s\\SM_%s\\Dir%d\\bucket%d",workdir,
		work_obj[my_workload].work_obj_name,dir,op_type);
#else
	   sprintf(buf,"%s/SM_%s/Dir%d/bucket%d",workdir,
		work_obj[my_workload].work_obj_name,dir,op_type);
#endif
	}
	else
	{
#if defined(WIN32)
	   sprintf(buf,"%s\\CL%d_%s\\Dir%d\\bucket%d",workdir,client_id,
		work_obj[my_workload].work_obj_name,dir,op_type);
#else
	   sprintf(buf,"%s/CL%d_%s/Dir%d/bucket%d",workdir,client_id,
		work_obj[my_workload].work_obj_name,dir,op_type);
#endif
	}
	if(cdebug) 
	{
		fprintf(newstdout,"Client got dirname %s\n",buf);
		fflush(newstdout);
	}
	strcpy(file_ptr->dirname,buf);
	return((void*)file_ptr);
}

/*
 * It is the case that some pre-op work needs to be 
 * done for some operations. This function would get called 
 * just before the work was to start. No time would be 
 * accumlated for this action.
 */
#ifdef HAVE_ANSIC_C
void
generic_prefix_op(int test_num, struct file_object *fp)
#else
void
generic_prefix_op(test_num, fp)
int test_num;
struct file_object *fp;
#endif
{
	struct test *tp;
	int ret;
#if defined(WIN32)
	DWORD wret = 0;
#endif

#if !defined(WIN32)
  #ifdef _LARGEFILE64_SOURCE
  /* 
   * Heavy magic happens with the size of the stat and statfs structures
   * about the time that 10.6 came out.
   */
    #if defined(_macos_)
      #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  > 1050
  	struct stat stbuf;
      #endif
      #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  < 1050
	struct stat stbuf;
      #endif
      #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  == 1050
  	struct stat64 stbuf;
      #endif
    #else
	struct stat64 stbuf;
    #endif
  #else
	struct stat stbuf;
  #endif
#endif
	tp=&tests[test_num];
	switch(tp->op_type){
	case OP_WRITE:	 
		break;
	case OP_WRITE_FILE: 
		break;
	case OP_READ:
		break;
	case OP_READ_FILE:
		break;
	case OP_RMW:
		break;
	case OP_LOCKING:
		break;
	case OP_ACCESS:
		break;
	case OP_STAT:
		break;
	case OP_CHMOD:
		break;
	case OP_WRITE_RAND:	 /* Need to unlink the file before writing it. ??? */
		break;
	case OP_READ_RAND:
		break;
	case OP_APPEND:
		/* Truncate back to original size */
		I_TRUNC(fp->filename,(fp->Original_fsize*(long long)1024));
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,append,(prefix) Truncate,%s,,%lld,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->Original_fsize*(long long)1024,
				fp->access_count);
		total_file_ops+=1;
		work_obj[my_workload].total_file_ops+=1;
		work_obj[my_workload].open_op_count++; /* Count somewhere */
		work_obj[my_workload].close_op_count++; /* Count somewhere */
		break;
	case OP_MKDIR:
		/* Need to remove the directory before creating a new one */
		if(cdebug)
		{
			fprintf(newstdout,"Prefix rmdir before mkdir \n");
			fflush(newstdout);
		}
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,mkdir,(prefix) Stat,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->dirname,fp->access_count);
#if defined(WIN32)
                wret = GetFileAttributes(fp->dirname);
                if(wret != INVALID_FILE_ATTRIBUTES)
                {
                        if(tracedebug)
                                fprintf(tracefile,"%d,%lu,%s,mkdir,(prefix) Rmdir,%s,,,,%ld,\n",
                                        client_id,trace_seq++,trace_states[tracestate],
                                        fp->access_count,fp->dirname);
                        ret=rmdir(fp->dirname);
                        if( ret != 0 )
                        {
                           fprintf(newstdout,"generic_prefix_op rmdir failed\n");
                           fflush(newstdout);
                        }
                }
#else
           	ret = I_STAT(fp->dirname, &stbuf);
		if( ret == 0 )
		{
			if(tracedebug)
				fprintf(tracefile,"%d,%lu,%s,mkdir,(prefix) Rmdir,%s,,,,%ld,\n",
					client_id,trace_seq++,trace_states[tracestate],
					fp->dirname,fp->access_count);
			ret=rmdir(fp->dirname);
		        if( ret < 0 )
			{
 			   fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
			   fprintf(newstdout,"generic_prefix_op rmdir failed: %s\n",strerror(errno));
			   fflush(newstdout);
			}
		}
#endif
		/*
	 	 * It is normal for this to fail. The directory should not
 		 * exist, but could have, if things were left behind from
		 * a previous run. 
		 */

		total_file_ops+=1;
		work_obj[my_workload].total_file_ops+=1;
		work_obj[my_workload].unlink_op_count++; /* Count somewhere ? */
		break;
	case OP_UNLINK:
		break;
	case OP_CREATE:
		/* Prefix is to remove any from previous create */
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,create,(prefix) Unlink,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
		unlink(fp->filename);
		break;
	case OP_READDIR:
		break;
	case OP_COPYFILE:
		break;
	case OP_RENAME:
		break;
	case OP_STATFS:
		break;
	case OP_PATHCONF:
		break;
	case OP_CUSTOM1:
		break;
	case OP_CUSTOM2:
		break;
	}
}

/*
 * It may be the case that some post-op work needs to 
 * be done after some operations. This function 
 * is for that purpose. No time will be accumlated for this
 * action.
 */
#ifdef HAVE_ANSIC_C
void
generic_postfix_op(int test_num,struct file_object *fp)
#else
void
generic_postfix_op(test_num, fp)
int test_num;
struct file_object *fp;
#endif
{
	struct test *tp;
	char altname[MAXNAME]; /* for Rename */

	tp=&tests[test_num];
	switch(tp->op_type){
	case OP_WRITE:
		break;
	case OP_WRITE_FILE:
		break;
	case OP_READ:
		break;
	case OP_READ_FILE:
		break;
	case OP_RMW:
		break;
	case OP_READDIR:
		break;
	case OP_LOCKING:
		/* Lock is already released */
		break;
	case OP_STAT:
		break;
	case OP_CHMOD:
		break;
	case OP_ACCESS:
		break;
	case OP_APPEND:
		break;
	case OP_MKDIR:
		break;
	case OP_WRITE_RAND:
		break;
	case OP_READ_RAND:
		break;
	case OP_UNLINK:
		/* Need to put the file back after the unlink */
		if(cdebug)
		{
			fprintf(newstdout,"Postfix re-create after unlink \n");
			fflush(newstdout);
		}
		/* Create an empty file */
		if(tracedebug) /* this is a dup if init logging is on */
			fprintf(tracefile,"%d,%lu,%s,unlink,(postfix) Create,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
		(*Netmist_Ops[my_workload].init_empty_file)(fp->filename);
		break;
	case OP_CREATE:
		/* Post-fix is to remove any from previous create */
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,create,(postfix) Unlink,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
		unlink(fp->filename);
		break;
	case OP_COPYFILE:
		/* The old way, we deleted after every op. The new
                   way, we delete at the end of the run.
	        strcpy(altname,fp->filename);
		strcat(altname,"_2");
		unlink(altname);
		*/
		break;
	case OP_RENAME:
	        strcpy(altname,fp->filename);
		strcat(altname,"_2");
		/* Put it back */
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,rename,(postfix) Rename,%s,,,,%ld,newname=%s\n",
				client_id,trace_seq++,trace_states[tracestate],
				altname,fp->access_count,fp->filename);
		rename(altname,fp->filename);
		break;
	case OP_STATFS:
		break;
	case OP_PATHCONF:
		break;
	case OP_CUSTOM1:
		break;
	case OP_CUSTOM2:
		break;
	}
}


/*
 * This is the code the client will call to initialize
 * the directory structure and files.
 */
#ifdef HAVE_ANSIC_C
void
client_init_files(int client_id)
#else
void
client_init_files(client_id)
int client_id;
#endif
{
#if defined(WIN32)
	DWORD ret=0;
#else
	int ret;
#endif
	int i,j,k,ret1;
	int dir_count=0;
	int file_count=0;
	int empty_count=0;
	int tot,pct_complete=0,prev_pct_complete=0;
	int locked_dir = 0;
	char buf[1024];
	char dir_buf[MAXNAME];
	char leaf_name[MAXNAME];
	struct file_object *file_ptr;
	struct file_object *file_array_ptr;
	struct file_object tmp_file;
#if !defined(WIN32)
 #ifdef _LARGEFILE64_SOURCE
  /* 
   * Heavy magic happens with the size of the stat and statfs structures
   * about the time that 10.6 came out.
   */
   #if defined(_macos_)
     #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  > 1050
  	struct stat stbuf;
     #endif
     #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  < 1050
	struct stat stbuf;
     #endif
     #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  == 1050
  	struct stat64 stbuf;
     #endif
   #else
	struct stat64 stbuf;
   #endif
 #else
	struct stat stbuf;
 #endif
#endif

	prev_heartbeat = 0;
 	fprintf(newstdout,"%s Client starting init\n",get_timestamp(&time_string[0]));
	fflush(newstdout);
#if defined(PDSM_R_REGISTRY)
	if(pdsm_stats)
        {
	   pdsm_stats->cur_state = (int)INIT_BEAT;
	   pdsm_stats->requested_op_rate = (double)op_rate;
	   pdsm_stats->achieved_op_rate = (double)0.0;
	   pdsm_stats->run_time =(double) 0.0;
        }
#endif
	init_phase = 1;
	if(client_id == (num_clients-1))
	   tell_master_phase(real_client_id,INIT_PHASE);
	my_start_time = gettime();
	for(k=0;k<NUM_OP_TYPES;k++)
	{
	   if(ldebug)
	   {
	       fprintf(newstdout,"Initializing files for op type %d \n",k);
	       fflush(newstdout);
           }
	   if(k != OP_MKDIR)
	   {
	      /* Create all directories */
	       for(i=0;i<client_dirs;i++)
	       {
		  if(sharing_flag)
		  {
#if defined(WIN32)
		    sprintf(buf,"%s\\SM_%s",workdir,
			work_obj[my_workload].work_obj_name);
#else
		    sprintf(buf,"%s/SM_%s",workdir,
			work_obj[my_workload].work_obj_name);
#endif
		  }
		  else
		  {
#if defined(WIN32)
		    sprintf(buf,"%s\\CL%d_%s",workdir,client_id,
			work_obj[my_workload].work_obj_name);
#else
		    sprintf(buf,"%s/CL%d_%s",workdir,client_id,
			work_obj[my_workload].work_obj_name);
#endif
		  }
	   	  if(ldebug)
	   	  {
	       	    fprintf(newstdout,"Mkdir %s \n",buf);
	       	    fflush(newstdout);
           	  }

		  /* Use virtualized interface to create the directory. */
		  ret=(*Netmist_Ops[my_workload].init_dir)(buf);

		  if((ret < 0) && (errno != EEXIST))
		  {
		    ret1=errno;
 	   	    fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		    fprintf(newstdout,"Mkdir failed. Error %s\n",strerror(errno));
		    fflush(newstdout);
                    c_eval = LOAD_GEN_ERR;
		    R_exit(ret1);
		  }
		  if(sharing_flag)
		  {
#if defined(WIN32)
		    sprintf(buf,"%s\\SM_%s\\Dir%d",workdir,
			work_obj[my_workload].work_obj_name,i);
#else
		    sprintf(buf,"%s/SM_%s/Dir%d",workdir,
			work_obj[my_workload].work_obj_name,i);
#endif
		  }
		  else
		  {
#if defined(WIN32)
		    sprintf(buf,"%s\\CL%d_%s\\Dir%d",workdir,
			client_id,work_obj[my_workload].work_obj_name,i);
#else
		    sprintf(buf,"%s/CL%d_%s/Dir%d",workdir,
			client_id,work_obj[my_workload].work_obj_name,i);
#endif
		  }
	   	  if(ldebug)
	   	  {
	       	    fprintf(newstdout,"Mkdir %s \n",buf);
	       	    fflush(newstdout);
           	  }
		  /* Use virtualized, in case this function is workload specific */
		  ret=(*Netmist_Ops[my_workload].init_dir)(buf);

		  if((ret < 0) && (errno != EEXIST))
		  {
		    ret1=errno;
 	   	    fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		    fprintf(newstdout,"Mkdir failed. Error %s\n",strerror(errno));
		    fflush(newstdout);
                    c_eval = LOAD_GEN_ERR;
		    R_exit(ret1);
		  }
		  if(sharing_flag)
		  {
#if defined(WIN32)
		    sprintf(buf,"%s\\SM_%s\\Dir%d\\bucket%d",workdir,
			work_obj[my_workload].work_obj_name,i,k);
#else
		    sprintf(buf,"%s/SM_%s/Dir%d/bucket%d",workdir,
			work_obj[my_workload].work_obj_name,i,k);
#endif
	  	  }
		  else
		  {
#if defined(WIN32)
		    sprintf(buf,"%s\\CL%d_%s\\Dir%d\\bucket%d",workdir,
			client_id,work_obj[my_workload].work_obj_name,i,k);
#else
		    sprintf(buf,"%s/CL%d_%s/Dir%d/bucket%d",workdir,
			client_id,work_obj[my_workload].work_obj_name,i,k);
#endif
		  }
	   	  if(ldebug)
	   	  {
	       	    fprintf(newstdout,"Mkdir %s \n",buf);
	       	    fflush(newstdout);
           	  }
		  /* Use virtualized interface to create the directory. */
		  ret=(*Netmist_Ops[my_workload].init_dir)(buf);
		  if((ret < 0) && (errno != EEXIST))
		  {
		    ret1=errno;
 	   	    fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		    fprintf(newstdout,"Mkdir failed. Error %s\n",strerror(errno));
		    fflush(newstdout);
                    c_eval = LOAD_GEN_ERR;
		    R_exit(ret1);
		  }
		  dir_count++;
		  if(cdebug) 
		  {
		     fprintf(newstdout,"mkdir %s\n",buf);
		     fflush(newstdout);
		  }

		  /*
		   * Logic to reduce STAT calls while performing a parallel 
		   * initialization of files within buckets. There arre 
		   * potential race conditions, but its ok if one hits
		   * the race. It would only cause a few extra calls to 
		   * stat().
		   */
		  if(sharing_flag) 
		  {
		     /* Create a lock file for the directory level */
#if defined(WIN32)
		     sprintf(dir_buf,"%s\\SM_%s\\Dir%d\\bucket%d\\%s",workdir,
			work_obj[my_workload].work_obj_name,i,k,"dir_lock");
#else
		     sprintf(dir_buf,"%s/SM_%s/Dir%d/bucket%d/%s",workdir,
			work_obj[my_workload].work_obj_name,i,k,"dir_lock");
#endif
#if defined(WIN32)
                     ret = GetFileAttributes(dir_buf);
                     if(ret != INVALID_FILE_ATTRIBUTES)
                     {
#else

           	     ret = I_STAT(dir_buf, &stbuf);
		     if( ret == 0 )
		     {
#endif
		         locked_dir = 0;
		         /* Percent done */
		         tot=NUM_OP_TYPES*client_dirs*client_files;
		         pct_complete=(file_count*100)/tot;
			 if(pct_complete > 100)
				pct_complete = 100;
                 	 if(pct_complete && (pct_complete != prev_pct_complete)
		     		&& ((pct_complete % 10) == 0)
				&& (client_id == (num_clients-1)))
                     	{
				tell_master_pct_complete(pct_complete,real_client_id,
			   		R_PERCENT_I_COMPLETE);
				prev_pct_complete = pct_complete;
                         }
		     }
		     else
		     {
		        /* Create a file, to indicate that this dir is being worked */
			push_name(dir_buf); /* save name for delayed unlink */
		        strcpy(tmp_file.filename,dir_buf);
		        (*Netmist_Ops[my_workload].init_empty_file)(dir_buf);
			if(c_eval ==0)
		         locked_dir = 1; /* Locked by me */
			else
		         locked_dir = 0; /* Locked by some other */
		     }
		  }

		  /* Create all files */
		  for(j=0;j<client_files;j++)
		  {
		     /*-----------------*/
		     /* heartbest       */
		     /*-----------------*/
		     cur_time = gettime();
		     if( heartbeat_flag && 
			((int)(cur_time - my_start_time) != prev_heartbeat) &&
			(((int)(cur_time - my_start_time) % HEARTBEAT_TICK)==0)
			&& (client_id == (num_clients -1 )))
		     {
			 tell_master_heartbeat(real_client_id,INIT_BEAT,(double)0.0);
		         prev_heartbeat = (int)(cur_time - my_start_time);
#if defined(PDSM_R_REGISTRY)
		         if(pdsm_stats)
        	         {
	   		   pdsm_stats->cur_state = (int)INIT_BEAT;
	   		   pdsm_stats->requested_op_rate = (double)op_rate;
	   		   pdsm_stats->achieved_op_rate = (double)0.0;
	   		   pdsm_stats->run_time =(double) prev_heartbeat;
        	         }
#endif
		     }
		     /*-----------------*/

		     file_count++;

		     /* Percent done */
		     tot=NUM_OP_TYPES*client_dirs*client_files;
		     pct_complete=(file_count*100)/tot;
		     if(pct_complete > 100)
		         pct_complete = 100;
             	     if(pct_complete && (pct_complete != prev_pct_complete)
		     	&& ((pct_complete % 10) == 0)
			&& (client_id == (num_clients -1 )))
                     {
			tell_master_pct_complete(pct_complete,real_client_id,
			   R_PERCENT_I_COMPLETE);
			prev_pct_complete = pct_complete;
		        retry1_counter=0;
		        retry2_counter=0;
                     }

		     make_name(leaf_name,
			 (client_id*client_dirs*NUM_OP_TYPES*client_files)
			+(i*NUM_OP_TYPES*client_files)+(k*client_files)
			+j,TYPE_FILE,j);

		     /* 
		      * Save the Original file size information 
		      * k = to the type of op.
		      * i = which directory, client_files is the number of files
 		      * per directory, and j is which file in the lowest level dir.
		      */
		     file_array_ptr=op_X_array[k]; /* op type */
		     file_ptr = &file_array_ptr[(i*client_files)+j]; 
		     /*
		      * The fsize slots will create a range of file sizes that
		      * will average out to be that of the calculated fsize.
		      */

    		     if((my_mix_table[my_workload].uniform_file_size_dist == NO_UNIFORM_DIST) &&
			  (client_files > FSIZE_MIN_SLOTS))
		     {
		        file_ptr->Original_fsize = get_next_fsize(fsize*1024)/1024;
		        file_ptr->Original_fsize = (file_ptr->Original_fsize + ((long long)rsize-1)) & ~((long long)rsize-1);
		        file_ptr->Current_fsize  = file_ptr->Original_fsize;
		        /*
			file_ptr->Current_fsize  = get_next_fsize(fsize);
		        file_ptr->Original_fsize = (file_ptr->Original_fsize + ((long long)rsize-1)) & ~((long long)rsize-1);
		        file_ptr->Current_fsize  = (file_ptr->Original_fsize + ((long long)rsize-1)) & ~((long long)rsize-1); 
			fprintf(newstdout,"Original Fsize %lld\n",file_ptr->Original_fsize);
			fprintf(newstdout,"Current Fsize %lld\n",file_ptr->Current_fsize);
			*/
		     }
		     else
		     {
		        file_ptr->Original_fsize = ((fsize +((long long)rsize-1)) & ~((long long)rsize-1));
		        file_ptr->Current_fsize  = ((fsize +((long long)rsize-1)) & ~((long long)rsize-1));
/*
			fprintf(newstdout,"Original Fsize %lld\n",file_ptr->Original_fsize);
			fprintf(newstdout,"Current Fsize %lld\n",file_ptr->Current_fsize);
			fflush(newstdout);
*/
		     }

		     if(sharing_flag)
		     {
#if defined(WIN32)
		       sprintf(buf,"%s\\SM_%s\\Dir%d\\bucket%d\\%s",workdir,
			work_obj[my_workload].work_obj_name,i,k,leaf_name);
#else
		       sprintf(buf,"%s/SM_%s/Dir%d/bucket%d/%s",workdir,
			work_obj[my_workload].work_obj_name,i,k,leaf_name);
#endif
		     }
		     else
		     {
#if defined(WIN32)
		       sprintf(buf,"%s\\CL%d_%s\\Dir%d\\bucket%d\\%s",workdir,
			  client_id, work_obj[my_workload].work_obj_name,i,k,
			  leaf_name);
#else
		       sprintf(buf,"%s/CL%d_%s/Dir%d/bucket%d/%s",workdir,client_id,
			work_obj[my_workload].work_obj_name,i,k,leaf_name);
#endif
		     }
		     if(ldebug) 
		     {
			fprintf(newstdout,"Creating file %s\n",buf);
			fflush(newstdout);
		     }
		     if(k==OP_UNLINK || k==OP_STAT || k==OP_LOCKING || 
			k==OP_ACCESS || k==OP_CHMOD || k==OP_READDIR || 
		        k==OP_STATFS || k==OP_CUSTOM1 || k==OP_CUSTOM2 ||
			k==OP_PATHCONF || k==OP_RENAME || k==OP_CREATE)
		     {
		        /* 
		         * Jump out to next directory, some other thread is currently doing 
		         * this directory.
		         */
		        if(sharing_flag && (locked_dir == 0))
			{
				empty_count++;
				continue;	
			}
/*
			if( k == OP_CREATE)
				continue;
*/
			/* Create an empty file. Use virtualized interface  */
		        (*Netmist_Ops[my_workload].init_empty_file)(buf);
			empty_count++;
		     }
		     else	
		     {
		        /* 
		         * Jump out to next directory, some other thread is currently doing 
		         * this directory.
		         */
		        if(sharing_flag && (locked_dir == 0))
				continue;
			/* Create a file and fill it up with data. Use virtualized interface */
		       (*Netmist_Ops[my_workload].init_file)(buf,k,i,j);
		     }
		   }
	       }
	   }
	}
	if((pct_complete!=100) && (prev_pct_complete !=100)
			&& (client_id == (num_clients -1 )))
       	{
		tell_master_pct_complete(100,real_client_id,
	   		R_PERCENT_I_COMPLETE);
        }
	if(sharing_flag)
	{
		drain_list();
	}
/*
	fprintf(newstdout,"Flushed Cache size %lld\n",cache_file_space);
	fflush(newstdout);
*/
	gdir_count=dir_count; /* total directory count */
	gfile_count=file_count; /* total file count of files that have data */
	gempty_count=empty_count; /* total file count of empty files */
	init_phase = 0;
}

/*
 * At the end of the test it is a good idea to cleanup the 
 * huge collection of files that this client created.
 * Also a good idea to call this from a signal handler for sigint.
 */
#ifdef HAVE_ANSIC_C
void
client_remove_files(int client_id)
#else
void
client_remove_files(client_id)
int client_id;
#endif
{
	int i,j,k,ret;
	char buf[1024];
	char leaf_name[MAXNAME];
	char alt_name[MAXNAME];

	prev_heartbeat = 0;
	close_cached_file_desc();

	/*for(k=0;k<NUM_OP_TYPES;k++) for all op types */
	for(i=0;i<client_dirs;i++)
	{
		/* for all directories of this child */
	 	/*for(i=0;i<client_dirs;i++) */
	 	for(k=0;k<NUM_OP_TYPES;k++) 
		{
		
		   if(k == OP_MKDIR )
			goto rem_dir;  /* There are no files */
		   /* For all files in each directory */
		   for(j=0;j<client_files;j++) 
		   {
		     cur_time = gettime();
		     if( heartbeat_flag && !kflag &&
			((int)(cur_time - my_start_time) != prev_heartbeat) &&
			(((int)(cur_time - my_start_time) % HEARTBEAT_TICK)==0)
			&& (client_id == (num_clients -1 )))
		     {
			 tell_master_heartbeat(real_client_id,CLEAN_BEAT,(double)0.0);
		         prev_heartbeat = (int)(cur_time - my_start_time);
#if defined(PDSM_R_REGISTRY)
		         if(pdsm_stats)
        	         {
	   		   pdsm_stats->cur_state = (int)CLEAN_BEAT;
	   		   pdsm_stats->requested_op_rate = (double)0.0;
	   		   pdsm_stats->achieved_op_rate = (double)0.0;
	   		   pdsm_stats->run_time =(double) prev_heartbeat;
        	         }
#endif
		     }
		     make_name(leaf_name,
			(client_id*client_dirs*NUM_OP_TYPES*client_files)
			+(i*NUM_OP_TYPES*client_files)
			+(k*client_files)
			+j,TYPE_FILE,j);
		     if(sharing_flag)
		     {
#if defined(WIN32)
		        sprintf(buf,"%s\\SM_%s\\Dir%d\\bucket%d\\%s",workdir,
				work_obj[my_workload].work_obj_name,i,k,
				leaf_name);
#else
		        sprintf(buf,"%s/SM_%s/Dir%d/bucket%d/%s",workdir,
				work_obj[my_workload].work_obj_name,i,k,
				leaf_name);
#endif
		     }
		     else
	 	     {
#if defined(WIN32)
		        sprintf(buf,"%s\\CL%d_%s\\Dir%d\\bucket%d\\%s",workdir,client_id,
				work_obj[my_workload].work_obj_name,i,k,
				leaf_name);
#else
		        sprintf(buf,"%s/CL%d_%s/Dir%d/bucket%d/%s",workdir,client_id,
				work_obj[my_workload].work_obj_name,i,k,
				leaf_name);
#endif
		     }

		     /* Special cleanup for copyfile,or Rename destination files */
		     if((k == OP_COPYFILE) || (k == OP_RENAME))
		     {
			strcpy(alt_name,buf);
			strcat(alt_name,"_2");
			(*Netmist_Ops[my_workload].remove_file)(alt_name);
		     }
		     /* 
		      * First unlink the files and then remove the directories 
		      */
		     (*Netmist_Ops[my_workload].remove_file)(buf);
		     if(cdebug) 
		     {
			fprintf(newstdout,"Remove file %s\n",buf);
			fflush(newstdout);
		     }
		   }
rem_dir:
			/* Remove all directories */
		   if(sharing_flag)
		   {
#if defined(WIN32)
		      sprintf(buf,"%s\\SM_%s\\Dir%d\\bucket%d",workdir,
			work_obj[my_workload].work_obj_name,i,k); 
#else
		      sprintf(buf,"%s/SM_%s/Dir%d/bucket%d",workdir,
			work_obj[my_workload].work_obj_name,i,k); 
#endif
		   }
		   else
		   {
#if defined(WIN32)
		      sprintf(buf,"%s\\CL%d_%s\\Dir%d\\bucket%d",workdir,client_id,
			work_obj[my_workload].work_obj_name,i,k); 
#else
		      sprintf(buf,"%s/CL%d_%s/Dir%d/bucket%d",workdir,client_id,
			work_obj[my_workload].work_obj_name,i,k); 
#endif
		   }
		   ret = (*Netmist_Ops[my_workload].remove_dir)(buf);
		   if(ret < 0)
		   {
 			fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
			fprintf(newstdout,"remove_client_files: remove_dir (bucket) failed: %s Name %s\n",
				strerror(errno),buf);
			fflush(newstdout);
		   }
		}
		if(sharing_flag)
		{
#if defined(WIN32)
		     sprintf(buf,"%s\\SM_%s\\Dir%d",workdir,
			work_obj[my_workload].work_obj_name,i); 
#else
		     sprintf(buf,"%s/SM_%s/Dir%d",workdir,
			work_obj[my_workload].work_obj_name,i); 
#endif
		}
		else
		{
#if defined(WIN32)
		     sprintf(buf,"%s\\CL%d_%s\\Dir%d",workdir,client_id,
			work_obj[my_workload].work_obj_name,i); 
#else
		     sprintf(buf,"%s/CL%d_%s/Dir%d",workdir,client_id,
			work_obj[my_workload].work_obj_name,i); 
#endif
		}
		ret = (*Netmist_Ops[my_workload].remove_dir)(buf);
		if(ret < 0)
		{
 		   fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		   fprintf(newstdout,"remove_client_files: remove_dir (Dir) failed: %s Name %s\n",
			strerror(errno),buf);
		   fflush(newstdout);
		}
			/* Remove all directories */
	}
	if(sharing_flag)
	{
#if defined(WIN32)
	      sprintf(buf,"%s\\SM_%s",workdir,
			work_obj[my_workload].work_obj_name); 
#else
	      sprintf(buf,"%s/SM_%s",workdir,
			work_obj[my_workload].work_obj_name); 
#endif
	}
	else
	{
#if defined(WIN32)
	      sprintf(buf,"%s\\CL%d_%s",workdir,client_id,
			work_obj[my_workload].work_obj_name); 
#else
	      sprintf(buf,"%s/CL%d_%s",workdir,client_id,
			work_obj[my_workload].work_obj_name); 
#endif
	}
	if(cdebug) 
	{
		fprintf(newstdout,"rmdir %s\n",buf);
	        fflush(newstdout);
	}
	ret = (*Netmist_Ops[my_workload].remove_dir)(buf);
	if(ret < 0)
	{
 		fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"client_remove_files: remove_dir (CL) failed: %s Name %s\n",
			strerror(errno),buf);
	        fflush(newstdout);
	}
}

/*
 * This function reads a file so that cache space is consumed. Once 
 * a sufficient amount of files has been fed through this, the caches
 * no longer contain cached file data that will be used. The amount
 * of data pumped through here is automatically scaled so that it
 * will break the caches.
 */
void
purge_file(struct file_object *file_ptr, char *filename)
{
	char *buf;
	int ret;
	int fd;
	long long i;
	buf=main_buf;
#if defined(INIT_TRACE_DEBUG)
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,purge_file,Open,%s,0666,,,O_RDONLY=1\n",
			client_id,trace_seq++,trace_states[tracestate],filename);
#endif
        fd=I_OPEN(filename,O_RDONLY,0666);
        if(fd < 0)
        {
	      c_eval=errno;
	      fprintf(newstdout,"%s Open for read failed. Filename = %s %s\n",
                  get_timestamp(&time_string[0]),filename,strerror(errno));
	      fflush(newstdout);
	      c_eflag++;
	      R_exit(106);
              return;
        }
        for(i=0 ; i<(file_ptr->Original_fsize*1024) ; i+=(max_xfer_size))
        {
#if defined(INIT_TRACE_DEBUG)
	      if(tracedebug)
	      	fprintf(tracefile,"%d,%lu,%s,purge_file,Read,%s,,%lld,%d,,\n",
			client_id,trace_seq++,trace_states[tracestate],filename,i,
			max_xfer_size);
#endif
	      ret=read(fd,buf,max_xfer_size);
	      if(ret < 0)
	      {
		c_eval=errno;
	   	fprintf(newstdout,"%s Read file failed to file %s Error %s\n",
                  get_timestamp(&time_string[0]),filename,strerror(errno));
		fflush(newstdout);
		c_eflag++;
		close(fd);
		R_exit(107);
	      }
        }
	cache_file_space+=(file_ptr->Original_fsize);
#if defined(INIT_TRACE_DEBUG)
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,purge_file,Close,%s,,,,,\n",client_id,
			trace_seq++,trace_states[tracestate],filename);
#endif
        close(fd);
        return;
}
/*
 * This is the code that implements the "write" op. It will
 * write a file with the name that is in the test entry and 
 * collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_write(struct file_object *fp)
#else
void
generic_write(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	HANDLE fd;
	LARGE_INTEGER largeoffset;
	BOOL err=0;
	DWORD ret=0;
	DWORD retry_xfer_count=0;
#else
	int fd;
	int ret;
	int retry_xfer_count=0;
#endif
	char *buf;
	char *err_str=NULL;
	double lat = 0.0;
	unsigned int trans;
	int flags = 0;
	int flags2 = 0;
	int use_cached_fd = 0;

	if(cdebug) 
	{
		fprintf(newstdout,"Doing write op\n");
		fflush(newstdout);
	}
	buf = main_buf;
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
	if( ((work_obj[my_workload].write_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
#if defined(WIN32)
		flags|=FILE_FLAG_NO_BUFFERING;
#else
		flags|=O_DIRECT;
#endif
		flags2 = 1;
	}
	if( ((work_obj[my_workload].write_op_count % 100) < my_mix_table[my_workload].percent_osync) )
	{
#if defined(WIN32)
		flags|=FILE_FLAG_WRITE_THROUGH;
#else
		flags|=O_SYNC;
#endif
		flags2 |= 2;
	}	
	if( fp->file_desc )
	{
		use_cached_fd=1;
		fd = fp->file_desc;
		if(tracedebug)
		{
			fprintf(tracefile,"%d,%lu,%s,write,Using cached fd,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
				client_id,trace_seq++,trace_states[tracestate],fp->filename,
				fp->access_count,
				((flags2 & 1)?1:0),
				((flags2 & 2)?1:0));
		}
	}
	else
	{
#if defined(WIN32)
	  fd=CreateFile(fp->filename, 
		GENERIC_READ | GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, 
		flags, 
		NULL);
	  if(fd == INVALID_HANDLE_VALUE)
	  {
		  c_eval = GetLastError();
		  err_str = win32_strerror(c_eval);
#else
	 fd=I_OPEN(fp->filename,O_CREAT|O_RDWR|flags,0666);
	 if(fd <0)

	 {
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for generic write op failed. Filename = %s Error: %s Error Value: %d\n",
			fp->filename,err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(109);
	 }
	 if(tracedebug)
	 {
	 	fprintf(tracefile,"%d,%lu,%s,write,Opened file,%s,0666,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
				fp->access_count,
				((flags2 & 1)?1:0),
				((flags2 & 2)?1:0));
	 }
	  if( (cached_open_count+10 < max_open) && !(fp->file_desc))
	  {
	     cached_open_count++;
	     fp->file_desc = fd;
	     use_cached_fd = 1;
	  }
	} /* thus ends the opening of the file */
#if defined(_solaris_)
	if( ((work_obj[my_workload].write_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    directio(fd,DIRECTIO_ON);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    directio(fd,DIRECTIO_OFF);
		}
	}
#endif
#if defined(_macos_)
	if( ((work_obj[my_workload].write_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    fcntl(fd,F_NOCACHE,1);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    fcntl(fd,F_NOCACHE,0);
		}
	}
#endif
	if(!use_cached_fd)
	{
	   total_file_ops +=1;		/* open */
	   work_obj[my_workload].total_file_ops+=1;  /* The open */
	   work_obj[my_workload].open_op_count++;
	}
#if defined(WIN32)
	largeoffset.QuadPart = fp->seqwrite_offset;
    	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	I_LSEEK(fd,fp->seqwrite_offset,SEEK_SET);
#endif
#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   fp->seqwrite_offset = fp->seqwrite_offset & ~4095 ;
#if defined(WIN32)
	   largeoffset.QuadPart = fp->seqwrite_offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->seqwrite_offset, SEEK_SET);
#endif
	}
	if(my_mix_table[my_workload].align)
	{
	   fp->seqwrite_offset = fp->seqwrite_offset & (off64_t) (~(my_mix_table[my_workload].align-1));
#if defined(WIN32)
	   largeoffset.QuadPart = fp->seqwrite_offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->seqwrite_offset, SEEK_SET);
#endif
	}
	if(gg_flag)
	{
		if((fp->seqwrite_offset+gg_offset)/(long long)1024 > (fp->Original_fsize))
		    fp->seqwrite_offset = gg_offset;
		else
		    fp->seqwrite_offset += gg_offset;
#if defined(WIN32)
	   	largeoffset.QuadPart = fp->seqwrite_offset;
	   	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   	I_LSEEK(fd, fp->seqwrite_offset, SEEK_SET);
#endif	
	}

	/* Select write size to use */
	trans = write_dist_array[(netmist_rand() % RW_ARRAY)];


#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   trans = trans & ~(min_direct_size-1);
	   if(trans == 0)
		trans=min_direct_size;
	}

/*
	make_non_dedup((long *)buf,trans);
*/
#if defined(WIN32)
try_large_xfer:
	err=WriteFile(fd, buf, trans, &ret, NULL);
	if((FALSE == err) || (ret <= 0))
	{
		c_eval=GetLastError();
		/*
		 * Special case for O_DIRECT with client rejecting some xfer sizes. 
	  	 */
		if((c_eval == ERROR_INVALID_PARAMETER) && (flags & FILE_FLAG_NO_BUFFERING) 
			&& (retry_xfer_count==0) && (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = win32_strerror(c_eval);
#else
try_large_xfer:
	ret=write(fd,buf,trans);
	if(ret<0)
	{
		if((errno == EINVAL) && (flags & O_DIRECT) && (retry_xfer_count==0)
			&& (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
		
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Write failed to file %s Error: %s Error Eval %d Ret %d\n",
			fp->filename,err_str,c_eval,ret);
	        fflush(newstdout);
		c_eflag++;
	     	if(use_cached_fd)
		{
	           cached_open_count--;
		   fp->file_desc=0;
		}
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(110);
	}
	else
	{
		if(tracedebug)
		{
			fprintf(tracefile,"%d,%lu,%s,write,Write,%s,,%lld,%d,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,(long long)fp->seqwrite_offset,trans,
				fp->access_count);
		}
	}
#if defined(WIN32)
	/*query current file seqwrite_offset*/
	largeoffset.QuadPart = 0;
	SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_CURRENT);
	fp->seqwrite_offset = largeoffset.QuadPart;
#else
	fp->seqwrite_offset = I_LSEEK(fd,0,SEEK_CUR);
#endif
	if((fp->seqwrite_offset/(long long)1024) >= (fp->Original_fsize))
	   fp->seqwrite_offset=0;
	fp->op_count++;

	/* global (across all files in this work object) write count */
	total_file_ops += 1;
	work_obj[my_workload].total_file_ops+=1; /* Write */

	work_obj[my_workload].write_op_count++;



/*
	if ( flush_flag && ((fp->op_count % 3) == 0)) 
*/

	if( flush_flag && 
             ((work_obj[my_workload].write_op_count % 100) < my_mix_table[my_workload].percent_commit) )
	{
		I_FSYNC(fd);
		if(tracedebug)
		{
			fprintf(tracefile,"%d,%lu,%s,write,Sync file,%s,,,,%ld,\n",client_id,trace_seq++,
				trace_states[tracestate],fp->filename,fp->access_count);
		}
	}

	total_write_bytes += (double)(ret);
	/* 
	 * After writes are done then flush to make sure the 
	 * client cache is not creating an effect.
	 */
	if(lflag && flush_flag)
		I_FSYNC(fd);  /* Also helps make local filesystem testing act like NFS */
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].write_op_time += lat;
	}
	if(!use_cached_fd)
	{
#if defined(WIN32)
	  CloseHandle(fd);
#else
	  close(fd);
#endif
	  if(tracedebug)
	  {
	  	fprintf(tracefile,"%d,%lu,%s,write,Closed file,%s,,,,%ld,\n",client_id,trace_seq++,
	  		trace_states[tracestate],fp->filename,fp->access_count);
	  }
	}
	if(!use_cached_fd)
	{
	   work_obj[my_workload].close_op_count++;
	   total_file_ops += 1.0;	 /* Close */
	   work_obj[my_workload].total_file_ops+=1.0; /* Close */
	}

	/*free(buf);*/
}	

/*
 * This is the code that implements the "write file" op. It will
 * write a whole file with the name that is in the test entry and 
 * collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_write_file(struct file_object *fp)
#else
void
generic_write_file(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	HANDLE fd;
	LARGE_INTEGER largeoffset;
	DWORD ret=0;
	DWORD retry_xfer_count=0;
	BOOL err=0;
#else
	int fd;
	int ret;
	int retry_xfer_count=0;
#endif
	char *buf;
	char *err_str=NULL;
	double lat = 0.0;
	unsigned int trans;
	unsigned long long transferred = 0;
	int flags = 0;
	int flags2 = 0;
	int use_cached_fd = 0;

	if(cdebug) 
	{
		fprintf(newstdout,"Doing write file op\n");
		fflush(newstdout);
	}
	if(big_buf == 0)
	{
		big_buf = (char *)MALLOC((1024*1024) + 8192);
		big_buf = (char *)( ((long)big_buf + 4096) & ~4095); 
	}
	buf = big_buf;
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
	if( ((work_obj[my_workload].write_file_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
#if defined(WIN32)
		flags|=FILE_FLAG_NO_BUFFERING;
#else
		flags|=O_DIRECT;
#endif
		flags2=1;
	}
	if( ((work_obj[my_workload].write_file_op_count % 100) < my_mix_table[my_workload].percent_osync) )
	{
#if defined(WIN32)
		flags|=FILE_FLAG_WRITE_THROUGH;
#else
		flags|=O_SYNC;
#endif
		flags2 |= 2;
	}
	if( fp->file_desc )
	{
	  use_cached_fd=1;
	  fd = fp->file_desc;
	  if(tracedebug)
	  	fprintf(tracefile,"%d,%lu,%s,write_file,Using cached fd,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	}
	else
	{
#if defined(WIN32)
	  fd=CreateFile(fp->filename, 
		GENERIC_READ | GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, 
		flags, 
		NULL);
	  if(fd == INVALID_HANDLE_VALUE)
	  {
		  c_eval = GetLastError();
		  err_str = win32_strerror(c_eval);
#else
	 fd=I_OPEN(fp->filename,O_CREAT|O_RDWR|flags,0666);
	  if(fd <0)

	  {
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for generic write file op failed. Filename = %s Error: %s Error Value: %d\n",
			fp->filename,err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(109);
	  }
	  if(tracedebug)
	  {
		fprintf(tracefile,"%d,%lu,%s,write_file,Opened file,%s,0666,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	  }
	  if( (cached_open_count+10 < max_open) && !(fp->file_desc))
	  {
	     cached_open_count++;
	     fp->file_desc = fd;
	     use_cached_fd = 1;
	  }
	}
#if defined(_solaris_)
	if( ((work_obj[my_workload].write_file_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    directio(fd,DIRECTIO_ON);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    directio(fd,DIRECTIO_OFF);
		}
	}
#endif
#if defined(_macos_)
	if( ((work_obj[my_workload].write_file_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    fcntl(fd,F_NOCACHE,1);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    fcntl(fd,F_NOCACHE,0);
		}
	}
#endif
	if(!use_cached_fd)
	{
	   total_file_ops +=1;		/* open */
	   work_obj[my_workload].total_file_ops+=1;  /* The open */
	   work_obj[my_workload].open_op_count++;
	}

	/* Select write size to use */
#ifdef OLDWAY
	trans = 1024 * 1024; /* Use large write size */
#else
        trans = max_xfer_size;  /* Trans and max_xfer_size are in bytes */
        if(max_xfer_size/1024 > fp->Original_fsize)
                trans = fp->Original_fsize*1024;
#endif
	if((fp->wholefilewrite_offset + trans) >= (fp->Original_fsize * 1024))
	      fp->wholefilewrite_offset=0;
#if defined(WIN32)
	largeoffset.QuadPart = fp->wholefilewrite_offset;
    	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	I_LSEEK(fd,fp->wholefilewrite_offset,SEEK_SET);
#endif
#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   fp->wholefilewrite_offset = fp->wholefilewrite_offset & ~4095 ;
#if defined(WIN32)
	   largeoffset.QuadPart = fp->wholefilewrite_offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->wholefilewrite_offset, SEEK_SET);
#endif
	}
	if(my_mix_table[my_workload].align)
	{
	   fp->wholefilewrite_offset = fp->wholefilewrite_offset & (off64_t) (~(my_mix_table[my_workload].align-1));
#if defined(WIN32)
	   largeoffset.QuadPart = fp->wholefilewrite_offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->wholefilewrite_offset, SEEK_SET);
#endif
	}
	if(gg_flag)
	{
		if((fp->wholefilewrite_offset+gg_offset)/(long long)1024 > (fp->Original_fsize))
		    fp->wholefilewrite_offset = gg_offset;
		else
		    fp->wholefilewrite_offset += gg_offset;
#if defined(WIN32)
	   	largeoffset.QuadPart = fp->wholefilewrite_offset;
	   	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   	I_LSEEK(fd, fp->wholefilewrite_offset, SEEK_SET);
#endif	
	}


#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif	
	{
	   trans = trans & ~(min_direct_size-1);
	   if(trans == 0)
		trans=min_direct_size;
	}

	while( transferred < trans)
	{
	   make_non_dedup((long *)buf,trans);
#if defined(WIN32)
try_large_xfer:
	   err=WriteFile(fd, buf, trans, &ret, NULL);
	   if((FALSE == err) || (ret <= 0))
	   {
		c_eval=GetLastError();
		if((c_eval == ERROR_INVALID_PARAMETER) && (flags & FILE_FLAG_NO_BUFFERING) 
			&& (retry_xfer_count==0) && (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = win32_strerror(c_eval);
#else
try_large_xfer:
	   ret=write(fd,buf,trans);
	   if(ret<0)
	   {
		if((errno == EINVAL) && (flags & O_DIRECT) && (retry_xfer_count==0)
			&& (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
		
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Write file: Write failed to file %s Error: %s Eval %d Ret %d\n",
			fp->filename,err_str,c_eval,ret);
	        fflush(newstdout);
		c_eflag++;
	     	if(use_cached_fd)
		{
	           cached_open_count--;
		   fp->file_desc=0;
		}
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(110);
	   }
	   if(tracedebug)
	   	fprintf(tracefile,"%d,%lu,%s,write_file,Write file,%s,,%lld,%d,%ld,transferred=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			(long long)fp->wholefilewrite_offset,trans,fp->access_count,ret);
#if defined(WIN32)
	   /*query current file wholefilewrite_offset*/
	   largeoffset.QuadPart = 0;
	   SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_CURRENT);
	   fp->wholefilewrite_offset = largeoffset.QuadPart;
#else
	   fp->wholefilewrite_offset = I_LSEEK(fd,0,SEEK_CUR);
#endif
	   if((fp->wholefilewrite_offset/(long long)1024) >= (fp->Original_fsize))
	   {
	      fp->wholefilewrite_offset=0;
#if defined(WIN32)
	      largeoffset.QuadPart = fp->wholefilewrite_offset;
    	      SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	      I_LSEEK(fd,fp->wholefilewrite_offset,SEEK_SET);
#endif
	   }

	   transferred+=ret;
	}
	fp->op_count++;  /* for this file's personal op count */
	/* global (across all files in this work object) write count */
	total_file_ops += 1;
	work_obj[my_workload].total_file_ops+=1;     /* Total op count */
	work_obj[my_workload].write_file_op_count++; /* Write file op count */

	if( flush_flag && 
             ((work_obj[my_workload].write_file_op_count % 100) < my_mix_table[my_workload].percent_commit) )
	{
		I_FSYNC(fd);
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,write_file,Sync file,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
	}

	total_write_bytes += (double)(transferred);
	/* 
	 * After writes are done then flush to make sure the 
	 * client cache is not creating an effect.
	 */
	if(lflag && flush_flag)
		I_FSYNC(fd);  /* Also helps make local filesystem testing act like NFS */
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].write_file_op_time += lat;
	}
	if(!use_cached_fd)
	{
#if defined(WIN32)
	  CloseHandle(fd);
#else
	  close(fd);
#endif
	  if(tracedebug)
	  {
		fprintf(tracefile,"%d,%lu,%s,write_file,Closed file,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
	  }
	}
	if(!use_cached_fd)
	{
	   work_obj[my_workload].close_op_count++;
	   total_file_ops += 1.0;	 /* Close */
	   work_obj[my_workload].total_file_ops+=1.0; /* Close */
	}

	/*free(buf);*/
}	



/*
 * This is the code that implements the "write rand file" op. It will
 * write a file with the name that is in the test entry and 
 * collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_write_rand(struct file_object *fp)
#else
void
generic_write_rand(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	HANDLE fd;
	LARGE_INTEGER largeoffset;
        DWORD ret = 0;
        DWORD retry_xfer_count = 0;
        BOOL err = 0;
#else
	int fd;
	int ret;
        int retry_xfer_count = 0;
#endif
	char *buf;
	char *err_str = NULL;
	double lat = 0.0;
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t where;
#else
	off_t where;
#endif
	int flags = 0;
	int flags2 = 0;
	unsigned int trans;
	int use_cached_fd = 0;

	if(cdebug) 
	{
		fprintf(newstdout,"Doing rand write op\n");
		fflush(newstdout);
	}
	buf = main_buf;
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
	if( ((work_obj[my_workload].write_rand_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
#if	defined(WIN32)
		flags|=FILE_FLAG_NO_BUFFERING;
#else
		flags|=O_DIRECT;
#endif
		flags2=1;
	}
	if( ((work_obj[my_workload].write_rand_op_count % 100) < my_mix_table[my_workload].percent_osync))
	{
#if	defined(WIN32)
		flags|=FILE_FLAG_WRITE_THROUGH;
#else
		flags|=O_SYNC;
#endif
		flags2 |= 2;
	}
	if( fp->file_desc)
	{
	   use_cached_fd = 1;
	   fd = fp->file_desc;
	   if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,write_rand,Using cached fd,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	}
	else
	{
#if defined(WIN32)
	fd=CreateFile(fp->filename, 
		GENERIC_READ | GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, 
		flags, 
		NULL);
	if(fd == INVALID_HANDLE_VALUE)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else
	  fd=I_OPEN(fp->filename,O_CREAT|O_RDWR|flags,0666);
	  if(fd <0)
	  {
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for rand write op failed. Error value = %d, Filename = %s Error: %s\n",c_eval, 
		   fp->filename, err_str);
		fflush(newstdout);
		c_eflag++;
		R_exit(111);
	  }
	  if(tracedebug)
	 	fprintf(tracefile,"%d,%lu,%s,write_rand,Opened file,%s,0666,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	  if( (cached_open_count+10 < max_open) && !(fp->file_desc))
	  {
	     cached_open_count++;
	     fp->file_desc = fd;
	     use_cached_fd = 1;
	  }
	}
#if defined(_solaris_)
	if( ((work_obj[my_workload].write_rand_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    directio(fd,DIRECTIO_ON);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    directio(fd,DIRECTIO_OFF);
		}
	}
#endif
#if defined(_macos_)
	if( ((work_obj[my_workload].write_rand_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    fcntl(fd,F_NOCACHE,1);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    fcntl(fd,F_NOCACHE,0);
		}
	}
#endif
	if(!use_cached_fd)
	{
	   total_file_ops += 1.0;
	   work_obj[my_workload].total_file_ops+=1.0; /* Open */
	   work_obj[my_workload].open_op_count++;
	}
	/* Select write size to use */
	trans = write_dist_array[(netmist_rand() % RW_ARRAY)];

#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   trans = trans & ~(min_direct_size-1);
	   if(trans == 0)
		trans=min_direct_size; 
	}

    	/* Uniform Random access */
    	if(my_mix_table[my_workload].rand_dist_behavior == UNIFORM_RAND)
    	{
        	/* Pick random offset within the file on 8k boundary */
                if(fp->Original_fsize == 1)
                        where = 0;
                else
        		where = (netmist_llrand() % (fp->Original_fsize-1));
    	}
    	else
    		/* geometric Random access */
    	{
                if(fp->Original_fsize == 1)
                        where = 0;
                else
        		where = (netmist_llrand() % (fp->Original_fsize-1));
        	/* Pick random offset within the file on 8k boundary */

        	spot_size = (fp->Original_fsize * (long long)SPOT_PCT)/(long long)100;
        	if(spot_size <= 0)
            		spot_size = (long long)1;

		hit_in_spot = (spot_size / 8) * 5; /* 5 times as many 8k records 
						    * as fits in a spot */
		if(hit_in_spot ==0)
		   hit_in_spot=1;

		/* 
	 	* Special logic for DB_TABLE, dividing into spots for geometric, and 
	 	* jumping out of geometric PCT percent of the time. DB_LOG is the
		* log writer, and mainly sequential.
	 	*/
		if( (my_strcasecmp(my_mix_table[my_workload].workload_name,"DB_TABLE")==0)
			|| (my_strcasecmp(my_mix_table[my_workload].workload_name,"DB_LOG")==0 ))
		{
           		if(((fp->db_access_count) % hit_in_spot) == 0)
              			fp->which_spot = (long long)(geometric_rand(NSPOTS)%NSPOTS);
	   		fp->which_spot = picker(fp->which_spot,NSPOTS); /* a bit of jitter */
           		fp->db_access_count++;
			}
		else
		{
            		fp->which_spot = (long long)(geometric_rand(NSPOTS)%NSPOTS);
		}
                if(spot_size ==0)
                   offset_in_spot = 0;
                else
        	   offset_in_spot = netmist_llrand() % spot_size;
        	where = (fp->which_spot * spot_size) + offset_in_spot;
		where= where & ~7 ; /* Where is in Kbytes */
        	if(cdebug)
        	{
            		fprintf(newstdout,"%lld Offset %lld Spsize %lld Whichsp %lld fCount \n",
                		(long long)where,spot_size,(long long)(fp->which_spot),
                		(long long)(fp->db_access_count));
            		fflush(newstdout);
        	}
    	}
        /* If it is beyond eof, start over */
	if(where + (trans/(long long)1024) > (fp->Original_fsize))
	   where = 0;
#if defined(WIN32)
	largeoffset.QuadPart = (where*1024);
    	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	I_LSEEK(fd, (where*(long long)1024), SEEK_SET);
#endif
	fp->offset = (where * 1024);

#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   fp->offset = fp->offset & ~4095 ;
#if defined(WIN32)
	   largeoffset.QuadPart = fp->offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->offset, SEEK_SET);
#endif
	   
	}
	if(my_mix_table[my_workload].align)
	{
	   fp->offset = fp->offset & (off64_t) (~(my_mix_table[my_workload].align-1));
#if defined(WIN32)
	   largeoffset.QuadPart = fp->offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->offset, SEEK_SET);
#endif
	}
	if(gg_flag)
	{
		if((fp->offset+gg_offset)/(long long)1024 > (fp->Original_fsize))
		    fp->offset = gg_offset;
		else
		    fp->offset += gg_offset;
#if defined(WIN32)
	   	largeoffset.QuadPart = fp->offset;
	   	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   	I_LSEEK(fd, fp->offset, SEEK_SET);
#endif
	}

#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align min size for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align min size for O_DIRECT */
#endif
	{
	   trans = trans & ~(min_direct_size-1);
	   if(trans == 0)
		trans=min_direct_size;
	}

	make_non_dedup((long *)buf,trans);

#if defined(WIN32)
try_large_xfer:
	err=WriteFile(fd, buf, trans, &ret, NULL);
	if((FALSE == err) || (ret <= 0))
	{
		c_eval=GetLastError();
		if((c_eval == ERROR_INVALID_PARAMETER) && (flags & FILE_FLAG_NO_BUFFERING) 
			&& (retry_xfer_count==0) && (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = win32_strerror(c_eval);
#else
try_large_xfer:
	ret=write(fd,buf,trans);
	if(ret<0)
	{
		c_eval=errno;
		if((errno == EINVAL) && (flags & O_DIRECT) && (retry_xfer_count==0)
			&& (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = strerror(c_eval);
#endif
 	    	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Write rand failed to file %s Error: %s Error value %d Ret %d\n",
			fp->filename, err_str, c_eval, ret);
	        fflush(newstdout);
		c_eflag++;
	     	if(use_cached_fd)
		{
		  cached_open_count--;
		  fp->file_desc=0;
		}
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(112);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,write_rand,Random write,%s,,%lld,%d,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,(long long)fp->offset,trans,fp->access_count);
#if defined(WIN32)
	largeoffset.QuadPart = 0;
	SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_CURRENT);
    	fp->offset = (unsigned int)largeoffset.QuadPart;
#else
	fp->offset = I_LSEEK(fd,0,SEEK_CUR);
#endif
	fp->op_count++; /* write in file counter */

	/* global (across all files in this work object) write count */
	work_obj[my_workload].write_rand_op_count++;

	if( flush_flag && 
             ((work_obj[my_workload].write_rand_op_count % 100) < my_mix_table[my_workload].percent_commit) )
	{
		I_FSYNC(fd);
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,write_rand,Sync file,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
	}
	total_write_rand_bytes+=(double)(ret);

	total_write_bytes += (double)(ret);

	total_file_ops += 1.0; /* Write */
	work_obj[my_workload].total_file_ops+=1.0; /* Write */
	/* 
	 * After writes are done then flush to make sure the 
	 * client cache is not creating an effect.
	 */
	if(lflag && flush_flag)
		I_FSYNC(fd);  /* Also helps make local filesystem testing act like NFS */
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].write_rand_op_time += lat;
	}

	if(!use_cached_fd)
	{
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		if(tracedebug)
		{
			fprintf(tracefile,"%d,%lu,%s,write_rand,Closed file,%s,,,,%ld,\n",client_id,trace_seq++,
				trace_states[tracestate],fp->filename,fp->access_count);
		}
	}
	if(!use_cached_fd)
	{
	   total_file_ops += 1.0; /* Close */
	   work_obj[my_workload].total_file_ops+=1.0; /* Close */
	   work_obj[my_workload].close_op_count++;
	}

	/*free(buf);*/
}	

/*
 * This is the code that implements the "read" op. It will
 * read a file with the name that is in the test entry and 
 * collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_read(struct file_object *fp)
#else
void
generic_read(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	HANDLE fd;
	LARGE_INTEGER largeoffset;
	DWORD ret;
	DWORD retry_xfer_count=0;
	BOOL err=0;
#else
	int fd,ret;
	int retry_xfer_count=0;
#endif
	char *buf;
	char *err_str = NULL;
	unsigned int trans;
	double lat = 0.0;
	int flags = 0;
	int flags2 = 0;
	int use_cached_fd=0;

	if(cdebug) 
	{
		fprintf(newstdout,"Doing read op\n");
		fflush(newstdout);
	}
	buf = main_buf;
	purge_cache();
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
	if( ((work_obj[my_workload].read_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
#if defined(WIN32)
		flags|=FILE_FLAG_NO_BUFFERING;
#else
		flags|=O_DIRECT;
#endif
		flags2=1;
	}
	if( ((work_obj[my_workload].read_op_count % 100) < my_mix_table[my_workload].percent_osync) ) 
	{
#if defined(WIN32)
		flags|=FILE_FLAG_WRITE_THROUGH;
#else
		flags|=O_SYNC;
#endif
		flags2 |= 2;
	}
	if( fp->file_desc )
	{
	  use_cached_fd = 1;
	  fd = fp->file_desc;
	  if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,read,Using cached fd,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	}
	else
	{
#if defined(WIN32)
	fd=CreateFile(fp->filename, 
		GENERIC_READ | GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, 
		flags, 
		NULL);
	if(fd == INVALID_HANDLE_VALUE)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else 
	  fd=I_OPEN(fp->filename,O_CREAT|O_RDWR|flags,0666);
	  if(fd <0)
	  {
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for read op failed. Filename = %s Error: %s Error value: %d\n",
			fp->filename, err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(113);
	  }
	  if( (cached_open_count+10 < max_open) && !(fp->file_desc))
	  {
	     cached_open_count++;
	     fp->file_desc = fd;
	     use_cached_fd = 1;
	  }
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,read,Opened file,%s,0666,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
#if defined(_solaris_)
	if( ((work_obj[my_workload].read_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    directio(fd,DIRECTIO_ON);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    directio(fd,DIRECTIO_OFF);
		}
	}
#endif
#if defined(_macos_)
	if( ((work_obj[my_workload].read_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    fcntl(fd,F_NOCACHE,1);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    fcntl(fd,F_NOCACHE,0);
		}
	}
#endif
	if(!use_cached_fd)
	{
	   total_file_ops += 1.0; /* Open */
	   work_obj[my_workload].total_file_ops+=1.0; /* Open */
	   work_obj[my_workload].open_op_count++;
	}

	/* Select write size to use */
	trans = read_dist_array[(netmist_rand() % RW_ARRAY)];

#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   fp->seqread_offset = fp->seqread_offset & ~4095 ;
#if defined(WIN32)
	   largeoffset.QuadPart = fp->seqread_offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->seqread_offset, SEEK_SET);
#endif
	}

	if((fp->seqread_offset/1024) + (trans/(long long)1024) > (fp->Original_fsize))
	   fp->seqread_offset = 0;

#if defined(WIN32)
	largeoffset.QuadPart = fp->seqread_offset;
	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	I_LSEEK(fd, fp->seqread_offset, SEEK_SET);
#endif
	if(my_mix_table[my_workload].align)
	{
	   fp->seqread_offset = fp->seqread_offset & (off64_t) (~(my_mix_table[my_workload].align-1));
#if defined(WIN32)
	   largeoffset.QuadPart = fp->seqread_offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->seqread_offset, SEEK_SET);
#endif
	}

	if(gg_flag)
	{
		if((fp->seqread_offset+gg_offset)/(long long)1024 > (fp->Original_fsize))
		    fp->seqread_offset = gg_offset;
		else
		    fp->seqread_offset += gg_offset;
#if defined(WIN32)
	   	largeoffset.QuadPart = fp->seqread_offset;
	   	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   	I_LSEEK(fd, fp->seqread_offset, SEEK_SET);
#endif
	}

#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   trans = trans & ~(min_direct_size-1);
	   if(trans == 0)
		trans=min_direct_size;
	}
#if defined(WIN32)
try_large_xfer:
	err=ReadFile(fd, buf, trans, &ret, NULL);
	if( (FALSE == err) || (ret < 0))
	{
		c_eval=GetLastError();
		if((c_eval == ERROR_INVALID_PARAMETER) && (flags & FILE_FLAG_NO_BUFFERING) 
			&& (retry_xfer_count==0) && (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = win32_strerror(c_eval);
		if((err == 0) && (ret <= 0))
		{
 	          fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		  fprintf(newstdout,"Generic read op failed file %s: Error %s, Error value %d Ret %d\n",
			fp->filename,err_str,c_eval,ret);
		  fflush(newstdout);
		}
#else
try_large_xfer:
	ret=read(fd,buf,trans);
	if(ret <= 0)
	{
		c_eval=errno;
		if((errno == EINVAL) && (flags & O_DIRECT) && (retry_xfer_count==0) 
			&& (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Generic read op failed file %s Error: %s, Eval %d, Ret %d\n",
			fp->filename,err_str,c_eval, ret);
		fflush(newstdout);
		c_eflag++;
		free(buf);
	     	if(use_cached_fd)
		{
		  fp->file_desc = 0;
		  cached_open_count--;
		}
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(114);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,read,Read,%s,,%lld,%d,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			(long long)fp->seqread_offset,trans,fp->access_count);
#if defined(WIN32)
	largeoffset.QuadPart = 0;
	SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_CURRENT);
    	fp->seqread_offset = largeoffset.QuadPart;
#else
	fp->seqread_offset = I_LSEEK(fd,0,SEEK_CUR);
#endif
	if((fp->seqread_offset/(long long)1024) >= (fp->Original_fsize))
	   fp->seqread_offset = 0;
	total_read_bytes += (double)(ret);
	total_file_ops += 1.0; /* Read */
	work_obj[my_workload].total_file_ops+=1.0; /* Read */
	fp->op_count++;
	/* global (across all files in this work object) read count */
	work_obj[my_workload].read_op_count++;
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].read_op_time += lat;
	}
	if(!use_cached_fd)
	{
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		if(tracedebug)
		{
			fprintf(tracefile,"%d,%lu,%s,read,Closed file,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
		}
	}			
	if(!use_cached_fd)
	{
	   total_file_ops += 1.0;
	   work_obj[my_workload].total_file_ops+=1.0; /* Close */
	   work_obj[my_workload].close_op_count++;
	}
	/*free(buf);*/
}	

/*
 * This is the code that implements the "read whole file" op. It will
 * read a file with the name that is in the test entry and 
 * collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_read_file(struct file_object *fp)
#else
void
generic_read_file(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	HANDLE fd;
	LARGE_INTEGER largeoffset;
	DWORD ret;
	DWORD retry_xfer_count=0;
	BOOL err=0;
#else
	int fd,ret;
	int retry_xfer_count=0;
#endif
	char *buf;
	char *err_str = NULL;
	unsigned int trans;
	unsigned long long transferred = 0;
	double lat = 0.0;
	int flags = 0;
	int flags2 = 0;
	int use_cached_fd=0;

	if(cdebug) 
	{
		fprintf(newstdout,"Doing read file op\n");
		fflush(newstdout);
	}
	if(big_buf == 0)
	{
		big_buf = (char *)MALLOC((1024*1024) + 4096);
		big_buf = (char *)( ((long)big_buf + 4096) & ~4095); 
	}
	buf = big_buf;
	purge_cache();
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
	if( ((work_obj[my_workload].read_file_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
#if defined(WIN32)
		flags|=FILE_FLAG_NO_BUFFERING;
#else
		flags|=O_DIRECT;
#endif
		flags2 = 1;
	}
	if( ((work_obj[my_workload].read_file_op_count % 100) < my_mix_table[my_workload].percent_osync) ) 
	{
#if defined(WIN32)
		flags|=FILE_FLAG_WRITE_THROUGH;
#else
		flags|=O_SYNC;
#endif
		flags2 |= 2;
	}

	trans = 1024*1024; /* One Meg transfers */

	if((fp->wholefileread_offset/1024) >= fp->Original_fsize)
	   fp->wholefileread_offset = 0;
	/* 
	 * If trans is larger than remaining data in the file.
	 **/
	if(trans >= ((fp->Original_fsize*1024)  - fp->wholefileread_offset))
		trans = (fp->Original_fsize - (fp->wholefileread_offset/1024))* 1024;

	if( fp->file_desc )
	{
	  use_cached_fd = 1;
	  fd = fp->file_desc;
	  if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,read_file,Using cached fd,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	}
	else
	{

#if defined(WIN32)
	fd=CreateFile(fp->filename, 
		GENERIC_READ | GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, 
		flags, 
		NULL);
	if(fd == INVALID_HANDLE_VALUE)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else 
	  fd=I_OPEN(fp->filename,O_CREAT|O_RDWR|flags,0666);
	  if(fd <0)
	  {
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for read file op failed. Filename = %s Error: %s Error value: %d \n",
			fp->filename, err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(113);
	  }
	  if( (cached_open_count+10 < max_open) && !(fp->file_desc))
	  {
	     cached_open_count++;
	     fp->file_desc = fd;
	     use_cached_fd = 1;
	  }
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,read_file,Opened file,%s,0666,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
#if defined(_solaris_)
	if( ((work_obj[my_workload].read_file_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    directio(fd,DIRECTIO_ON);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    directio(fd,DIRECTIO_OFF);
		}
	}
#endif
#if defined(_macos_)
	if( ((work_obj[my_workload].read_file_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    fcntl(fd,F_NOCACHE,1);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    fcntl(fd,F_NOCACHE,0);
		}
	}
#endif
	if(!use_cached_fd)
	{
	   total_file_ops += 1.0; /* Open */
	   work_obj[my_workload].total_file_ops+=1.0; /* Open */
	   work_obj[my_workload].open_op_count++;
	}

	/* Seek to last known wholefileread_offset */
#if defined(WIN32)
	largeoffset.QuadPart = fp->wholefileread_offset;
	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	I_LSEEK(fd, fp->wholefileread_offset, SEEK_SET);
#endif

#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   fp->wholefileread_offset = fp->wholefileread_offset & ~4095 ;
#if defined(WIN32)
	   largeoffset.QuadPart = fp->wholefileread_offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->wholefileread_offset, SEEK_SET);
#endif
	}

	if((fp->wholefileread_offset/1024) + (trans/(long long)1024) > (fp->Original_fsize))
	   fp->wholefileread_offset = 0;

#if defined(WIN32)
	largeoffset.QuadPart = fp->wholefileread_offset;
	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	I_LSEEK(fd, fp->wholefileread_offset, SEEK_SET);
#endif
	if(my_mix_table[my_workload].align)
	{
	   fp->wholefileread_offset = fp->wholefileread_offset & (off64_t) (~(my_mix_table[my_workload].align-1));
#if defined(WIN32)
	   largeoffset.QuadPart = fp->wholefileread_offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->wholefileread_offset, SEEK_SET);
#endif
	}

	if(gg_flag)
	{
		if((fp->wholefileread_offset+gg_offset)/(long long)1024 > (fp->Original_fsize))
		    fp->wholefileread_offset = gg_offset;
		else
		    fp->wholefileread_offset += gg_offset;
#if defined(WIN32)
	   	largeoffset.QuadPart = fp->wholefileread_offset;
	   	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   	I_LSEEK(fd, fp->wholefileread_offset, SEEK_SET);
#endif
	}

#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   trans = trans & ~(min_direct_size-1);
	   if(trans == 0)
		trans=min_direct_size;
	}
	while(transferred < trans)
	{
#if defined(WIN32)
try_large_xfer:
	   err=ReadFile(fd, buf, trans, &ret, NULL);
	   if((FALSE == err) || (ret < 0))
	   {
		c_eval=GetLastError();
		if((c_eval == ERROR_INVALID_PARAMETER) && (flags & FILE_FLAG_NO_BUFFERING) 
			&& (retry_xfer_count==0) && (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = win32_strerror(c_eval);
#else
try_large_xfer:
	   ret=read(fd,buf,trans);
	   if(ret < 0)
	   {
		c_eval=errno;
		if((errno == EINVAL) && (flags & O_DIRECT) && (retry_xfer_count==0)
			&& (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Read file op failed file %s Error: %s Error value %d Ret %d\n",fp->filename,
			err_str,c_eval,ret);
		fflush(newstdout);
		c_eflag++;
		free(buf);
	     	if(use_cached_fd)
		{
		  fp->file_desc = 0;
		  cached_open_count--;
		}
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(114);
	   }
	   if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,read_file,Read file,%s,,%lld,%d,%ld,transferred=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			(long long)fp->wholefileread_offset,trans,fp->access_count,ret);
	   if(ret < trans)
	   {
	        fp->wholefileread_offset = 0;
#if defined(WIN32)
	        largeoffset.QuadPart = fp->wholefileread_offset;
	        SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	        I_LSEEK(fd, fp->wholefileread_offset, SEEK_SET);
#endif
	   	total_read_bytes += (double)(ret);
	   	transferred += ret;
		break;
	   } 
	   /*
 	    *  Save the concurency.
	    */
#if defined(WIN32)
	   largeoffset.QuadPart = 0;
	   SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_CURRENT);
           fp->wholefileread_offset = largeoffset.QuadPart;
#else
	   fp->wholefileread_offset = I_LSEEK(fd,0,SEEK_CUR);
#endif
	   total_read_bytes += (double)(ret);
	   transferred += ret;
	}
	total_file_ops += 1.0; /* Read file op */
	work_obj[my_workload].total_file_ops+=1.0; /* total ops */
	fp->op_count++; /* Read file op count for this file */
	work_obj[my_workload].read_file_op_count++; /* Global read_file op count */
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].read_file_op_time += lat;
	}
	if(!use_cached_fd)
	{
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		if(tracedebug)
		{
			fprintf(tracefile,"%d,%lu,%s,read_file,Closed file,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,fp->access_count);
		}
	}
	if(!use_cached_fd)
	{
	   total_file_ops += 1.0;
	   work_obj[my_workload].total_file_ops+=1.0; /* Close */
	   work_obj[my_workload].close_op_count++;
	}
	/*free(buf);*/
}	

/*
 * This is the code that implements the "read rand file" op. It will
 * read a file with the name that is in the test entry and 
 * collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_read_rand(struct file_object *fp)
#else
void
generic_read_rand(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	HANDLE fd;
	LARGE_INTEGER largeoffset;
	DWORD ret=0;
	DWORD retry_xfer_count=0;
	BOOL err=0;
#else
	int fd,ret;
	int retry_xfer_count=0;
#endif
	char *buf;
	char *err_str = NULL;
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t where;
#else
	off_t where;
#endif
	unsigned int trans;
	double lat = 0.0;
	int flags = 0;
	int flags2 = 0;
	int use_cached_fd = 0;

	if(cdebug) 
	{
		fprintf(newstdout,"Doing read rand op\n");
		fflush(newstdout);
	}
	buf = main_buf;
	purge_cache();
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
	if( ((work_obj[my_workload].read_rand_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
#if defined(WIN32)
		flags|=FILE_FLAG_NO_BUFFERING;
#else
		flags|=O_DIRECT;
#endif
		flags2 = 1;
	}
	if( ((work_obj[my_workload].read_rand_op_count % 100) < my_mix_table[my_workload].percent_osync) ) 
	{
#if defined(WIN32)
		flags|=FILE_FLAG_WRITE_THROUGH;
#else
		flags|=O_SYNC;
#endif
		flags2 |= 2;
	}
	if(fp->file_desc)
	{
	  use_cached_fd = 1;
	  fd = fp->file_desc;
	  if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,read_rand,Using cached fd,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	}
	else
	{
#if defined(WIN32)
	  fd=CreateFile(fp->filename, 
		GENERIC_READ | GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, 
		flags, 
		NULL);
	  if(fd == INVALID_HANDLE_VALUE)
	  {
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else
	  fd=I_OPEN(fp->filename,O_CREAT|O_RDWR|flags,0666);
	  if(fd <0)
	  {
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for read rand op failed. Error value = %d, Filename = %s Error: %s\n",c_eval, 
			fp->filename, err_str);
		fflush(newstdout);
		c_eflag++;
		R_exit(115);
	  }
	  if( (cached_open_count+10 < max_open) && !(fp->file_desc))
	  {
	     cached_open_count++;
	     fp->file_desc = fd;
	     use_cached_fd = 1;
	  }
	  if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,read_rand,Opened file,%s,0666,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	}
#if defined(_solaris_)
	if( ((work_obj[my_workload].read_rand_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    directio(fd,DIRECTIO_ON);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    directio(fd,DIRECTIO_OFF);
		}
	}
#endif
#if defined(_macos_)
	if( ((work_obj[my_workload].read_rand_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    fcntl(fd,F_NOCACHE,1);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    fcntl(fd,F_NOCACHE,0);
		}
	}
#endif
	if(!use_cached_fd)
	{
	   total_file_ops += 1.0; /* Open */
	   work_obj[my_workload].total_file_ops+=1.0; /* Open */
	   work_obj[my_workload].open_op_count++;
	}

	/* Select read size to use */
	trans = read_dist_array[(netmist_rand() % RW_ARRAY)];

	/* Uniform Random access */
	if(my_mix_table[my_workload].rand_dist_behavior == UNIFORM_RAND) 
	{
           /* Pick random offset within the file on 8k boundary */
           if(fp->Original_fsize == 1)
                where =0;
           else
           	where = (netmist_llrand() % (fp->Original_fsize-1));
	}
	else
	/* geometric Random access */
	{
	   /* Pick random offset within the file on 8k boundary */
           spot_size = (fp->Original_fsize * (long long)SPOT_PCT)/(long long)100;
	   if(spot_size <= 0)
		  spot_size = (long long)1;

	   hit_in_spot = (spot_size / 8) * 5; /* 5 times as many 8k records 
					       * as fits in a spot */
	   if(hit_in_spot == 0)
		hit_in_spot = 1;
	   /* 
	    * Special logic for DATABASE, dividing into spots for geometric, and 
	    * jumping out of geometric PCT percent of the time 
	    */
	   if( my_strcasecmp(my_mix_table[my_workload].workload_name,"DB_TABLE")==0)
	   {
              if(((fp->db_access_count) % hit_in_spot) == 0)
                  fp->which_spot = (long long)(geometric_rand(NSPOTS)%NSPOTS);
			/* Sometimes jump to other spot */
	      fp->which_spot = picker(fp->which_spot,NSPOTS);  
              fp->db_access_count++;
	   }
	   else
	   {
               fp->which_spot = (long long)(geometric_rand(NSPOTS)%NSPOTS);
	   }
           if(spot_size ==0)
              offset_in_spot = 0;
           else
              offset_in_spot = netmist_llrand() % spot_size;
           where = (fp->which_spot * spot_size) + offset_in_spot;
	   where= where & ~7 ; /* where is in Kbytes here */
	   if(cdebug)
	   {
		fprintf(newstdout,"%s Offset %lld Original size %lld Spotsize %lld Which spot %lld Fcount %lld \n",
		   fp->filename, (long long)where,fp->Original_fsize,spot_size,(long long)(fp->which_spot), 
		   (long long)(fp->db_access_count));
		fflush(newstdout);
	   }
	}
        /* If it is beyond eof, start over */
    	if(where + (trans/(long long)1024) > (fp->Original_fsize))
        	where = 0;
#if defined(WIN32)
	largeoffset.QuadPart = where*1024;
    	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
    	I_LSEEK(fd,(where*(long long)1024),SEEK_SET);
#endif
	fp->offset = (where * 1024);

#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING ) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   fp->offset = fp->offset & ~4095 ;
#if defined(WIN32)
	   largeoffset.QuadPart = fp->offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->offset, SEEK_SET);
#endif
	}
	if(my_mix_table[my_workload].align)
	{
	   fp->offset = fp->offset & (off64_t) (~(my_mix_table[my_workload].align-1));
#if defined(WIN32)
	   largeoffset.QuadPart = fp->offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->offset, SEEK_SET);
#endif
	}
	if(gg_flag)
	{
		if((fp->offset+gg_offset)/(long long)1024 > (fp->Original_fsize))
		    fp->offset = gg_offset;
		else
		    fp->offset += gg_offset;
#if defined(WIN32)
	   	largeoffset.QuadPart = fp->offset;
	   	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   	I_LSEEK(fd, fp->offset, SEEK_SET);
#endif
	}
#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING ) /* align min size for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align min size for O_DIRECT */
#endif
	{
	   trans = trans & ~(min_direct_size-1);
	   if(trans == 0)
		trans=min_direct_size;
	}
#if defined(WIN32)
try_large_xfer:
	err=ReadFile(fd,buf,trans,&ret,NULL);
	if((FALSE == err) || (ret < 0))
	{
		c_eval=GetLastError();
		if((c_eval == ERROR_INVALID_PARAMETER) && (flags & FILE_FLAG_NO_BUFFERING) 
			&& (retry_xfer_count==0) && (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = win32_strerror(c_eval);
#else
try_large_xfer:
	ret=read(fd,buf,trans);
	if(ret < 0)
	{
		c_eval=errno;
		if((errno == EINVAL) && (flags & O_DIRECT) && (retry_xfer_count==0)
			&& (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Read file rand op failed file %s Error: %s Error value %d Ret %d\n",
			fp->filename,err_str,c_eval, ret);
		fflush(newstdout);
		/*free(buf);*/
		c_eflag++;
	     	if(use_cached_fd)
		{
		  fp->file_desc=0;
		  cached_open_count--;
		}
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(116);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,read_rand,Random read,%s,,%lld,%d,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,(long long)fp->offset,trans,fp->access_count);
	total_read_bytes += (double)(ret);
	total_read_rand_bytes+=(double)(ret);
	total_file_ops += 1.0; /*Read*/
	work_obj[my_workload].total_file_ops+=1.0; /*Read*/
#if defined(WIN32)
	largeoffset.QuadPart = 0;
	SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_CURRENT);
    	fp->offset = (unsigned int)largeoffset.QuadPart;
#else
	fp->offset = I_LSEEK(fd,0,SEEK_CUR);
#endif
	
	fp->op_count++;
	work_obj[my_workload].read_rand_op_count++;
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].read_rand_op_time += lat;
	}
	if(!use_cached_fd)
	{
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		if(tracedebug)
		{
			fprintf(tracefile,"%d,%lu,%s,read_rand,Closed file,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
		}
	}
	if(!use_cached_fd)
	{
	   total_file_ops += 1.0; /* Close */
	   work_obj[my_workload].total_file_ops+=1.0; /* Close */
	   work_obj[my_workload].close_op_count++;
	}

	/*free(buf);*/
}	

/*
 * This is the code that implements the "read modify write file" op. It will
 * read a file with the name that is in the test entry and 
 * collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_rmw(struct file_object *fp)
#else
void
generic_rmw(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	HANDLE fd;
	LARGE_INTEGER largeoffset;
	BOOL err=0;
	DWORD read_ret=0;
	DWORD write_ret=0;
	DWORD retry_xfer_count=0;
#else
	int fd,read_ret,write_ret;
	int retry_xfer_count=0;
#endif
	char *buf;
	char *err_str = NULL;
#if defined(_LARGEFILE64_SOURCE) || defined(WIN32)
	off64_t where;
#else
	off_t where;
#endif
	unsigned int trans;
	double lat = 0.0;
	int flags = 0;
	int flags2 = 0;
	int use_cached_fd = 0;

	if(cdebug) 
	{
		fprintf(newstdout,"Doing rmw op\n");
		fflush(newstdout);
	}
	buf = main_buf;
	purge_cache();
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
	if( ((work_obj[my_workload].rmw_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
#if defined(WIN32)
		flags|=FILE_FLAG_NO_BUFFERING;
#else
		flags|=O_DIRECT;
#endif
		flags2 = 1;
	}
	if( ((work_obj[my_workload].rmw_op_count % 100) < my_mix_table[my_workload].percent_osync) )
	{
#if defined(WIN32)
		flags|=FILE_FLAG_WRITE_THROUGH;
#else
		flags|=O_SYNC;
#endif
		flags2 |= 2;
	}

	if(fp->file_desc)
	{
	  	use_cached_fd = 1;
	  	fd = fp->file_desc;
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,rmw,Using cached fd,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
				client_id,trace_seq++,trace_states[tracestate],fp->filename,
				fp->access_count,
				((flags2 & 1)?1:0),
				((flags2 & 2)?1:0));

	}
	else
	{
#if defined(WIN32)
	fd=CreateFile(fp->filename, 
		GENERIC_READ | GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, 
		flags, 
		NULL);
	if(fd == INVALID_HANDLE_VALUE)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else
	  	fd=I_OPEN(fp->filename,O_CREAT|O_RDWR|flags,0666);
	  	if(fd <0)
	  	{
		   c_eval=errno;
		   err_str = strerror(c_eval);
#endif
 	           fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		   fprintf(newstdout,"Open for rmw op failed. Filename = %s Error: %s Error Value: %d\n",
			fp->filename, err_str,c_eval);
		   fflush(newstdout);
		   c_eflag++;
		   R_exit(115);
	  	}
	  	if( (cached_open_count+10 < max_open) && !(fp->file_desc))
	  	{
	     	   cached_open_count++;
	     	   fp->file_desc = fd;
	     	   use_cached_fd = 1;
	  	}
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,rmw,Opened file,%s,0666,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
#if defined(_solaris_)
	if( ((work_obj[my_workload].rmw_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    directio(fd,DIRECTIO_ON);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    directio(fd,DIRECTIO_OFF);
		}
	}
#endif
#if defined(_macos_)
	if( ((work_obj[my_workload].rmw_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    fcntl(fd,F_NOCACHE,1);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    fcntl(fd,F_NOCACHE,0);
		}
	}
#endif
	if(!use_cached_fd)
	{
	   total_file_ops += 1.0; /* Open */
	   work_obj[my_workload].total_file_ops+=1.0; /* Open */
	   work_obj[my_workload].open_op_count++;
	}

	/* Select read size to use */
	trans = read_dist_array[(netmist_rand() % RW_ARRAY)];

	/* Pick random offset within the file on 8k boundary */
        if(fp->Original_fsize == 1)
           where = 0;
        else
	   where = (netmist_llrand() % (fp->Original_fsize-1));

	/* If it is beyond eof, start over */
	if((where*1024) + trans > (fp->Original_fsize * (long long)1024))
	   where = 0;
#if defined(WIN32)
	largeoffset.QuadPart = where*1024;
    	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
    	I_LSEEK(fd,(where*(long long)1024),SEEK_SET);
/*fprintf(newstdout,"%lld\n",where);*/
#endif
	fp->rmw_offset = (where * (long long)1024);

#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   fp->rmw_offset = fp->rmw_offset & ~4095 ;
	   where  = (fp->rmw_offset/(long long)1024); /* Save where for later use by write. */
#if defined(WIN32)
	   largeoffset.QuadPart = fp->rmw_offset;
       	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
           I_LSEEK(fd, fp->rmw_offset, SEEK_SET);
#endif
	}
	if(my_mix_table[my_workload].align)
	{
	   fp->rmw_offset = fp->rmw_offset & (off64_t) (~(my_mix_table[my_workload].align-1));
#if defined(WIN32)
	   largeoffset.QuadPart = fp->rmw_offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->rmw_offset, SEEK_SET);
#endif
	}

	if(gg_flag)
	{
		if((fp->rmw_offset+gg_offset)/(long long)1024 > (fp->Original_fsize))
		{
		    fp->rmw_offset = gg_offset;
	   	    where  = (fp->rmw_offset/(long long)1024); /* Save where for later use by write. */
		}
		else
		{
		    fp->rmw_offset += gg_offset;
	   	    where  = (fp->rmw_offset/(long long)1024); /* Save where for later use by write. */
		}
#if defined(WIN32)
	   	largeoffset.QuadPart = fp->rmw_offset;
       		SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
       		I_LSEEK(fd, fp->rmw_offset, SEEK_SET);
#endif
	}

#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   trans = trans & ~(min_direct_size-1);
	   if(trans == 0)
		trans=min_direct_size;
	}
#if defined(WIN32)
try_large_xfer:
	err=ReadFile(fd,buf,trans,&read_ret,NULL);
	if((FALSE == err) || (read_ret < 0))
	{
		c_eval=GetLastError();
		if((c_eval == ERROR_INVALID_PARAMETER) && (flags & FILE_FLAG_NO_BUFFERING) 
			&& (retry_xfer_count==0) && (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = win32_strerror(c_eval);
#else
try_large_xfer:
	read_ret=read(fd,buf,trans);
	if(read_ret < 0)
	{
		if((errno == EINVAL) && (flags & O_DIRECT) && (retry_xfer_count==0)
			&& (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Rmw file op failed file %s Error: %s Error value %d Ret %d\n",
			fp->filename,err_str,c_eval,read_ret);
		fflush(newstdout);
		/*free(buf);*/
		c_eflag++;
	     	if(use_cached_fd)
		{
	           cached_open_count--;
		   fp->file_desc=0;
		}
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(116);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,rmw,RMW read,%s,,%lld,%d,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,(long long)fp->rmw_offset,trans,fp->access_count);
	total_rmw_bytes+=(double)(read_ret);
	total_file_ops += 1.0; /* Read */
	work_obj[my_workload].total_file_ops+=1.0; /* Read */
	/*work_obj[my_workload].rmw_op_count++;*/
#if defined(WIN32)
	largeoffset.QuadPart = 0;
	SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_CURRENT);
	fp->rmw_offset = largeoffset.QuadPart;
#else
	fp->rmw_offset = I_LSEEK(fd,0,SEEK_CUR);
#endif

	/* Move back to previous location */
	fp->rmw_offset -= trans;
	/* Go back for the write cycle */
#if defined(WIN32)
	largeoffset.QuadPart = where*1024;
    	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
    	I_LSEEK(fd,(where*(long long)1024),SEEK_SET);
#endif
	/*is this statement needed? ==> fp->rmw_offset = where*1024;*/
	if(gg_flag)
	{
		if((fp->rmw_offset+gg_offset)/(long long)1024 > (fp->Original_fsize))
		{
		    fp->rmw_offset = gg_offset;
	   	    where  = (fp->rmw_offset/(long long)1024); /* Save where for later use by write. */
		}
		else
		{
		    fp->rmw_offset += gg_offset;
	   	    where  = (fp->rmw_offset/(long long)1024); /* Save where for later use by write. */
		}
#if defined(WIN32)
	   	largeoffset.QuadPart = fp->rmw_offset;
       		SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
       		I_LSEEK(fd, fp->rmw_offset, SEEK_SET);
#endif
	}

	make_non_dedup((long *)buf,trans);
#if defined(WIN32)
	err = WriteFile(fd,buf,trans,&write_ret,NULL);
	if((FALSE == err) || (write_ret <= 0))
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else
	write_ret=write(fd,buf,trans);
	if(write_ret <= 0)
	{
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Rmw file op failed file %s Error: %s Error value %d Ret %d\n",
			fp->filename,err_str,c_eval,write_ret);
		fflush(newstdout);
		/*free(buf);*/
		c_eflag++;
	     	if(use_cached_fd)
		{
	           cached_open_count--;
		   fp->file_desc=0;
		}
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(116);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,rmw,RMW write,%s,,%lld,%d,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,(long long)fp->rmw_offset,trans,fp->access_count);
	fp->op_count++; /* Read */
	total_file_ops += 1.0; /* Write */
	work_obj[my_workload].total_file_ops+=1.0; /* Write */
	/*work_obj[my_workload].write_op_count++;*/
	fp->op_count++; /* write */
	if(!use_cached_fd)
	{
	   total_file_ops += 1.0; /* Close */
	   work_obj[my_workload].total_file_ops+=1.0; /* Close */
	   work_obj[my_workload].close_op_count++;
	}

	total_rmw_bytes+=(double)(write_ret);
#if defined(WIN32)
	largeoffset.QuadPart = 0;
	SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_CURRENT);
	fp->rmw_offset = largeoffset.QuadPart;
#else
	fp->rmw_offset = I_LSEEK(fd,0,SEEK_CUR);
#endif

	work_obj[my_workload].rmw_op_count++;
	/*********************************************************/
	/* Increment so that the totals of Op counts will add up */
	total_file_ops += 1.0; /* RMW */
	work_obj[my_workload].total_file_ops+=1.0; /* RMW */
	/*********************************************************/
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].rmw_op_time += lat;
	}
	total_rmw_ops += 1.0;
	if( flush_flag && 
             ((work_obj[my_workload].rmw_op_count % 100) < my_mix_table[my_workload].percent_commit) )
	{
		I_FSYNC(fd);
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,rmw,Sync file,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
	}
	
	if(!use_cached_fd)
	{
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		if(tracedebug)
		{
			fprintf(tracefile,"%d,%lu,%s,rmw,Closed file,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
		}
	}
}	

/*
 * This is the code that implements the "stat" op. It will
 * perform a "stat" call on a file with the name that is in the 
 * test entry and collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_stat(struct file_object *fp)
#else
void
generic_stat(fp)
struct file_object *fp;
#endif
{
#ifdef _LARGEFILE64_SOURCE
  /* 
   * Heavy magic happens with the size of the stat and statfs structures
   * about the time that 10.6 came out.
   */
  #if defined(_macos_)
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  > 1050
  	struct stat stbuf;
    #endif
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  < 1050
	struct stat stbuf;
    #endif
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  == 1050
  	struct stat64 stbuf;
    #endif
  #else
	struct stat64 stbuf;
  #endif
#else
  #if !defined(WIN32)
	struct stat stbuf;
  #endif
#endif
#if defined(WIN32)
	DWORD ret=0;
	WIN32_FILE_ATTRIBUTE_DATA    fileInfo;
	GET_FILEEX_INFO_LEVELS fInfoLevelId = GetFileExInfoStandard;
#else
	int ret;
#endif
	char *err_str=NULL;
	double lat = 0.0;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing stat op\n");
		fflush(newstdout);
	}
#if defined(WIN32)
        /*
           Using 'Ex' so that it more closely resembles the functionality
           of Unix systems.  Unix stat() gets more of the attributes than the
           standard GetFileAttributes() call does.
         */
        ret = GetFileAttributesEx(fp->filename,fInfoLevelId,&fileInfo);
        if(ret == INVALID_FILE_ATTRIBUTES)
        {
                c_eval = GetLastError();
                err_str = win32_strerror(c_eval);
                fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
                fprintf(newstdout,"Stat failed on file %s Error: %s Error value: %d\n",fp->filename,err_str,c_eval);
                fflush(newstdout);
                c_eflag++;
                R_exit(117);
        }
#else
	ret=I_STAT(fp->filename, &stbuf);
	if(ret != 0)
	{
		c_eval=errno;
		err_str = strerror(errno);
 	        fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Stat failed on file %s Error: %s\n",fp->filename,err_str);
		fflush(newstdout);
		c_eflag++;
		R_exit(117);
	}
#endif
	if(tracedebug)
	{
		fprintf(tracefile,"%d,%lu,%s,stat,Stat,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,fp->access_count);
	}
	work_obj[my_workload].stat_op_count++;
	total_file_ops += 1.0;
	work_obj[my_workload].total_file_ops+=1.0;
	total_meta_ops += 1.0;
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].stat_op_time += lat;
	}
}	

/*
 * This is the code that implements the "chmod" op. It will
 * perform a "chmod" call on a file with the name that is in the 
 * test entry and collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_chmod(struct file_object *fp)
#else
void
generic_chmod(fp)
struct file_object *fp;
#endif
{
	int ret;
	double lat = 0.0;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing chmod op\n");
		fflush(newstdout);
	}
	ret=chmod(fp->filename, 0666);
	if(ret < 0)
	{
		c_eval=errno;
 	        fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"chmod failed on file %s Error: %s\n",
			fp->filename,strerror(errno));
		fflush(newstdout);
		c_eflag++;
		R_exit(118);
	}
	if(tracedebug)
	{
		fprintf(tracefile,"%d,%lu,%s,chmod,Chmod,%s,0666,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
	}
	work_obj[my_workload].chmod_op_count++;
	total_file_ops += 1.0;
	work_obj[my_workload].total_file_ops+=1.0;
	total_meta_ops += 1.0;
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].chmod_op_time += lat;
	}
}	

/*
 * This is the code that implements the "readdir" op. It will
 * perform a "readdir" call on a file with the name that is in the 
 * test entry and collect statistics for the operation.
 */
#if !defined(WIN32)
#ifdef HAVE_ANSIC_C
void
generic_readdir(struct file_object *fp)
#else
void
generic_readdir(fp)
struct file_object *fp;
#endif
{
	struct dirent *ret;
	DIR *dp;
	double lat = 0.0;
	long seek_loc = 0;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing readdir op\n");
		fflush(newstdout);
	}
	dp = opendir(fp->dirname);
	if(dp == 0)
	{
		c_eval=errno;
		fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Opendir failed on file %s\n",fp->filename);
		fflush(newstdout);
		c_eflag++;
		R_exit(119);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,readdir,Opendir,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->dirname,
			fp->access_count);

	 /* Seek to random location within the directory */
	seek_loc = netmist_rand()%client_files-1;
	seekdir(dp,seek_loc);
	/*seekdir(dp,(netmist_rand()%client_files-1));*/
	ret = readdir(dp);
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,readdir,Readdir,%s,,%ld,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->dirname,
			seek_loc,fp->access_count);
	if(ret < (struct dirent *)0)
	{
		c_eval=errno;
 	        fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Opendir failed on file %s\n",fp->filename);
		fflush(newstdout);
		c_eflag++;
		R_exit(119);
	}
	closedir(dp);
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,readdir,Closedir,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->dirname,
			fp->access_count);
	work_obj[my_workload].readdir_op_count++;
	total_file_ops += 1.0;
	work_obj[my_workload].total_file_ops+=1.0;
	total_meta_ops += 1.0;
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].readdir_op_time += lat;
	}
}	
#else

#ifdef HAVE_ANSIC_C
void
generic_readdir(struct file_object *fp)
#else
void
generic_readdir(fp)
struct file_object *fp;
#endif
{
	HANDLE hFind = INVALID_HANDLE_VALUE;
	double lat = 0.0;
	char searchexpr[MAX_PATH];
    WIN32_FIND_DATAA ffd;
	int	randomloc = 0;
	if(cdebug)
	{
		fprintf(newstdout,"Doing readdir op\n");
		fflush(newstdout);
	}
	/* Windows use expression to search files instead of readdir, so postfix \* to dir name */
	sprintf(searchexpr, "%s\\*", fp->dirname);

    hFind = FindFirstFile(searchexpr, &ffd);

    if (hFind == INVALID_HANDLE_VALUE)
    {
		c_eval=GetLastError();
 	        fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Opendir (FindFirstFile) failed on directory %s, Error value=%d\n",fp->dirname, c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(119);
    }
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,readdir,Opendir,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],searchexpr,
			fp->access_count);
	/* Seek to random location within the directory */
	randomloc = netmist_rand()%client_files-1;
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,readdir,Readdir,%s,,%d,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],searchexpr,
			randomloc,fp->access_count);
	/*simulate seekdir */
	while (randomloc-- > 0)
	{
		if (FindNextFile(hFind, &ffd) == 0)
		{
			c_eval=GetLastError();
			if (c_eval == ERROR_NO_MORE_FILES)
				break; /*WE can't fail the test, otherwise client_validate_ops() will fail*/
 			fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
			fprintf(newstdout,"Opendir (FindNextFile) failed on directory %s, Error value=%d",fp->dirname,c_eval);
			fflush(newstdout);
			c_eflag++;
			R_exit(119);
		}
	}
	/*
	if(FindNextFile(hFind, &ffd) == 0)
	{
		c_eval=GetLastError();
 	    fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Opendir failed on file %s\n",fp->dirname);
		fflush(newstdout);
		c_eflag++;
        FindClose(hFind);
		R_exit(119);
	}*/

	FindClose(hFind);
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,readdir,Closedir,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],searchexpr,
			fp->access_count);

	work_obj[my_workload].readdir_op_count++;
	total_file_ops += 1.0;
	total_meta_ops += 1.0;
	work_obj[my_workload].total_file_ops+=1.0;

	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].readdir_op_time += lat;
	}
}
#endif /*!defined(WIN32)*/
/*
 * This is the code that implements the "access" op. It will
 * perform a "access" call on a file with the name that is in the 
 * test entry and collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_access(struct file_object *fp)
#else
void
generic_access(fp)
struct file_object *fp;
#endif
{
	int ret;
	double lat = 0.0;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing access op\n");
		fflush(newstdout);
	}
	ret=access(fp->filename,F_OK);
	if(ret < 0)
	{
		c_eval=errno;
 	   	fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Access failed on file %s Error %s\n",
			fp->filename,strerror(errno));
		fflush(newstdout);
		c_eflag++;
		R_exit(120);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,access,Access,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
	total_file_ops += 1.0;
	work_obj[my_workload].total_file_ops+=1.0;
	work_obj[my_workload].access_op_count++;
	total_meta_ops += 1.0;
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].access_op_time += lat;
	}
}	

/*
 * This is the code that implements the "mkdir" op. It will
 * perform a "mkdir" call on a directory  with the name that is in the 
 * test entry and collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_mkdir(struct file_object *fp)
#else
void
generic_mkdir(fp)
struct file_object *fp;
#endif
{
	int ret;
	double lat = 0.0;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing mkdir op\n");
		fflush(newstdout);
	}
	ret=mkdir(fp->dirname,0777);
	if((ret<0) && (errno != EEXIST))
	{
		c_eval=errno;
 	   	fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Mkdir failed to file %s Error %s\n",
			fp->dirname,strerror(errno));
		fflush(newstdout);
		c_eflag++;
		R_exit(121);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,mkdir,Mkdir,%s,0777,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->dirname,
			fp->access_count);
	total_file_ops += 1.0;
	work_obj[my_workload].total_file_ops+=1.0;
	total_meta_ops += 1.0;
	work_obj[my_workload].mkdir_op_count++;
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].mkdir_op_time += lat;
	}
}	

/*
 * THis is used to create the directories during the client_init_files
 * phase.
 */
#ifdef HAVE_ANSIC_C
int
generic_init_dir(char *dirname)
#else
int
generic_init_dir(dirname)
char *dirname;
#endif
{
	int ret;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing init_dir op\n");
		fflush(newstdout);
	}
	ret=mkdir(dirname,0777);
#if defined(INIT_TRACE_DEBUG)
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,init_dir,Mkdir,%s,0777,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],dirname);
#endif
	return(ret);
}	

/*
 * This is the code that implements the "append" op. It will
 * perform an "append" on a file  with the name that is in the 
 * test entry and collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_append(struct file_object *fp)
#else
void
generic_append(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	HANDLE fd;
	LARGE_INTEGER largeoffset;
        DWORD ret = 0;
        DWORD retry_xfer_count = 0;
        BOOL err = 0;
#else
	int fd;
        int retry_xfer_count = 0;
	int ret;
#endif
	char *buf;
	char *err_str=NULL;
	double lat = 0.0;
	unsigned int trans;
	int flags = 0;
	int flags2 = 0;

	if(cdebug) 
	{
		fprintf(newstdout,"Doing append op\n");
		fflush(newstdout);
	}
	buf = main_buf;
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
/*
 * Linux systems may not permit O_DIRECT and O_APPEND over NFS, so 
 * let's not go there for now.
 */
#ifdef COMBINE_DIRECT_APPEND
	if( ((work_obj[my_workload].append_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
 #if defined(WIN32)
		flags|=FILE_FLAG_NO_BUFFERING;
 #else
		flags|=O_DIRECT;
 #endif
		flags = 1;
	}
#endif
	if( ((work_obj[my_workload].append_op_count % 100) < my_mix_table[my_workload].percent_osync) )
	{
#if defined(WIN32)
		flags|=FILE_FLAG_WRITE_THROUGH;
#else
		flags|=O_SYNC;
#endif
		flags2 |= 2;
	}
#if defined(WIN32)
	fd=CreateFile(fp->filename, 
		GENERIC_READ | GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, 
		flags, 
		NULL);
	if(fd == INVALID_HANDLE_VALUE)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else
	fd=I_OPEN(fp->filename,O_APPEND|O_RDWR|flags,0666);
	if(fd <0)
	{
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for append op failed. Filename = %s Error: %s Error value: %d\n",
			fp->filename, err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(122);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,append,Opened file,%s,0666,,,%ld,O_DIRECT=%d O_SYNC=%d O_APPEND=1\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
#if defined(_solaris_)
	if( ((work_obj[my_workload].append_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    directio(fd,DIRECTIO_ON);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    directio(fd,DIRECTIO_OFF);
		}
	}
#endif
#if defined(_macos_)
	if( ((work_obj[my_workload].append_op_count % 100) < 
		my_mix_table[my_workload].percent_direct) )
	{
		if(!(fp->directio_on))
		{
		    fp->directio_on=1;
		    fcntl(fd,F_NOCACHE,1);
		}
	}
	else
	{
		if(fp->directio_on)
		{
		    fp->directio_on=0;
		    fcntl(fd,F_NOCACHE,0);
		}
	}
#endif
	total_file_ops += 1.0; /* Open */
	work_obj[my_workload].total_file_ops+=1.0; /* Open */
	work_obj[my_workload].open_op_count++;
	/*I_LSEEK(fd,0,SEEK_END);*/ /*this is redundant*/
#if defined(WIN32)
	largeoffset.QuadPart = 0;
    	SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_END);
	fp->offset = largeoffset.QuadPart;
#else
	fp->offset = I_LSEEK(fd,0,SEEK_END);
#endif
#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   fp->offset = fp->offset & ~4095;
#if defined(WIN32)
	   largeoffset.QuadPart = fp->offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd,fp->offset,SEEK_SET);
#endif
	}
	if(my_mix_table[my_workload].align)
	{
	   fp->offset = fp->offset & (off64_t) (~(my_mix_table[my_workload].align-1));
#if defined(WIN32)
	   largeoffset.QuadPart = fp->offset;
	   SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fd, fp->offset, SEEK_SET);
#endif
	}
	/* Select write size to use */
	trans = write_dist_array[(netmist_rand() % RW_ARRAY)];
#if defined(WIN32)
	if(flags & FILE_FLAG_NO_BUFFERING) /* align things for O_DIRECT */
#else
	if(flags & O_DIRECT) /* align things for O_DIRECT */
#endif
	{
	   trans = trans & ~(min_direct_size-1);
	   if(trans == 0)
		trans=min_direct_size;
	}
	make_non_dedup((long *)buf,trans);
#if defined(WIN32)
try_large_xfer:
	err=WriteFile(fd, buf, trans, &ret, NULL);
	if((FALSE == err) || (ret <= 0))
	{
		c_eval=GetLastError();
		if((c_eval == ERROR_INVALID_PARAMETER) && (flags & FILE_FLAG_NO_BUFFERING) 
			&& (retry_xfer_count==0) && (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = win32_strerror(c_eval);
#else
try_large_xfer:
	ret=write(fd,buf,trans);
	if(ret<0)
	{
		c_eval=errno;
		if((errno == EINVAL) && (flags & O_DIRECT) && (retry_xfer_count==0)
			&& (min_direct_size == MIN_DIRECT_SIZE))
		{
			fprintf(newstdout,"Detected 4096 byte sector device\n");
			fflush(newstdout);
			min_direct_size = MIN2_DIRECT_SIZE;
			retry_xfer_count++;
	   		trans = trans & ~(min_direct_size-1);
	   		if(trans == 0)
				trans=min_direct_size; 
			goto try_large_xfer;
		}
		err_str = strerror(c_eval);
#endif
 	   	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Append failed to file %s Error: %s Error value %d Ret %d\n",
			fp->filename,err_str,c_eval, ret);
		fflush(newstdout);
		c_eflag++;
#if defined(WIN32)
	  CloseHandle(fd);
#else
	  close(fd);
#endif
		R_exit(123);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,append,Write,%s,,%lld,%d,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,(long long)fp->offset,trans,fp->access_count);
	total_file_ops += 1.0; /* Write/append */
	work_obj[my_workload].total_file_ops+=1.0; /* Write/append */
	fp->op_count++;
#if defined(WIN32)
	largeoffset.QuadPart = 0;
    	SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_END);
	fp->offset = largeoffset.QuadPart;
#else
	fp->offset = I_LSEEK(fd,0,SEEK_END);
	fp->Current_fsize = (I_LSEEK(fd,0,SEEK_END))/1024;
#endif

	/* global (across all files in this work object) write count */
	work_obj[my_workload].append_op_count++;

	if ( flush_flag &&
             ((work_obj[my_workload].append_op_count % 100) < my_mix_table[my_workload].percent_commit) )
	{
		I_FSYNC(fd);
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,append,Sync file,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
	}
	if(lflag && flush_flag)
		I_FSYNC(fd);
#if defined(WIN32)
	  CloseHandle(fd);
#else
	  close(fd);
#endif
	if(tracedebug)
	{
		fprintf(tracefile,"%d,%lu,%s,append,Closed file,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
	}
	work_obj[my_workload].close_op_count++;
	total_write_bytes+=(double)(ret);
	total_file_ops += 1.0; /* Close */
	work_obj[my_workload].total_file_ops+=1.0; /* Close */
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].append_op_time += lat;
	}
	/*free(buf);*/
}	

/*
 * This is the code that implements the "unlink" op. It will
 * perform an "unlink" on a file  with the name that is in the 
 * test entry and collect statistics for the operation.
 * postfix() will put the file back.
 */
#ifdef HAVE_ANSIC_C
void
generic_unlink(struct file_object *fp, int force)
#else
void
generic_unlink(fp,force)
struct file_object *fp;
int force;
#endif
{
	int ret;
	double lat = 0.0;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing unlink op\n");
		fflush(newstdout);
	}
	/* Let's not do this while sharing all files */
	if((!force) && (sharing_flag)) 
	{
	   if(Op_lat_flag)
	   {
	      lat = gettime() - my_gen_op_time;
	      hist_insert(lat);
	      work_obj[my_workload].unlink_op_time += lat;
	   }
	   return;
	}
	ret=unlink(fp->filename);
	if((ret < 0) && !sharing_flag)
	{
		c_eval=errno;
 	   	fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Unlink failed to file %s Error: %s\n",
			fp->filename,strerror(errno));
		fflush(newstdout);
		c_eflag++;
		R_exit(124);
	}
	if(ret == 0)
	{
	   if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,unlink,Unlink,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
	   work_obj[my_workload].unlink_op_count++;
	   total_file_ops += 1.0;
	   work_obj[my_workload].total_file_ops+=1.0;
	   total_meta_ops += 1.0;
	   if(Op_lat_flag)
	   {
	      lat = gettime() - my_gen_op_time;
	      hist_insert(lat);
	      work_obj[my_workload].unlink_op_time += lat;
	   }
	}
}	
/*
 * This is the code that implements the "create" op. It will
 * perform an "create" of a file  with the name that is in the 
 * test entry and collect statistics for the operation.
 * postfix() will put remove the file.
 */
#ifdef HAVE_ANSIC_C
void
generic_create(struct file_object *fp, int force)
#else
void
generic_create(fp,force)
struct file_object *fp;
int force;
#endif
{
#if defined(WIN32)
	HANDLE fd;
	char *err_str=NULL;
#else
	int fd;
#endif

	int flags = 0;
	double lat = 0.0;

#if !defined(WIN32)
		/* for Windows, these flags need to be specified as different */
		/* arguments */
        flags = O_CREAT|O_RDWR;
#endif

#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
	if(cdebug) 
	{
		fprintf(newstdout,"Doing create op\n");
		fflush(newstdout);
	}
	/* Let's not do this while sharing all files */
	if((!force) && (sharing_flag)) 
	{
	   if(Op_lat_flag)
	   {
	      lat = gettime() - my_gen_op_time;
	      hist_insert(lat);
	      work_obj[my_workload].create_op_time += lat;
	   }
	   return;
	}
#if defined(WIN32)
	fd=CreateFile(fp->filename, 
		GENERIC_READ|GENERIC_WRITE, /*equivalent to O_RDWR*/
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, /*equivalent to O_CREAT*/
		flags, NULL);
	if(fd == INVALID_HANDLE_VALUE)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
 	   	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for create op failed. Filename = %s Error: %s Error value: %d\n", 
			fp->filename,err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(126);
	}
#else 
	fd=I_OPEN(fp->filename,flags,0666);
	if((fd < 0) && !sharing_flag)
	{
		c_eval=errno;
 	   	fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Create failed to file %s Error: %s\n",
			fp->filename,strerror(errno));
		fflush(newstdout);
		c_eflag++;
		R_exit(124);
	}
#endif
	if(fd > 0)
	{
	   if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,create,Create,%s,0666,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
	   work_obj[my_workload].create_op_count++;
	   total_file_ops += 1.0;
	   work_obj[my_workload].total_file_ops+=1.0;
	   total_meta_ops += 1.0;
	   if(Op_lat_flag)
	   {
	      lat = gettime() - my_gen_op_time;
	      hist_insert(lat);
	      work_obj[my_workload].create_op_time += lat;
	   }
	}
#if defined(WIN32)
	CloseHandle(fd);
#else
	close(fd);
#endif
	if(tracedebug)
	{
		fprintf(tracefile,"%d,%lu,%s,create,Closed file,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
	}
}	

/*
 * This is the code that implements the "lock" op. It will
 * perform a "lock" call on a file with the name that is in the 
 * test entry and collect statistics for the operation.
 */
#ifdef HAVE_ANSIC_C
void
generic_locking(struct file_object *fp)
#else
void
generic_locking(fp)
struct file_object *fp;
#endif
{
	int ret;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing lock op\n");
		fflush(newstdout);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,locking,Calling lock_file(),%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
	ret=lock_file(fp->filename,fp);
	if(ret < 0)
	{
		c_eval=errno;
 	   	fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Lock failed on file %s Error %s\n",
			fp->filename,strerror(errno));
		fflush(newstdout);
		c_eflag++;
		R_exit(125);
	}
}	

/*
 * This is the code that implements the "copyfile" op. It will
 * perform a "copyfile" call on a file with the name that is in the 
 * test entry and collect statistics for the operation.
 */
#define COPY_CHUNK 2
#define CHUNK_SIZE (32 * 1024)

#ifdef HAVE_ANSIC_C
void
generic_copyfile(struct file_object *fp)
#else
void
generic_copyfile(fp)
struct file_object *fp;
#endif
{

#ifdef _LARGEFILE64_SOURCE
  /* 
   * Heavy magic happens with the size of the stat and statfs structures
   * about the time that 10.6 came out.
   */
  #if defined(_macos_)
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  > 1050
  	struct stat stbuf;
    #endif
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  < 1050
	struct stat stbuf;
    #endif
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  == 1050
  	struct stat64 stbuf;
    #endif
  #else
	struct stat64 stbuf;
  #endif
#else
  #if defined(WIN32)
	struct _stati64 stbuf;
  #else
	struct stat stbuf;
  #endif
#endif
#if defined(WIN32)
	HANDLE fdi,fdo;
	LARGE_INTEGER largeoffset;
	DWORD read_ret;
	DWORD write_ret=0;
	BOOL err=0;
#else
	int fdo,fdi;
	int read_ret,write_ret;
#endif
	int ret;
	int read_transferred=0; 
	int write_transferred=0;
	size_t my_size;
	int i;
	int my_mode;
	int iflags=0,oflags=0; /*initialized flags to 0*/
	int iflags2=0,oflags2=0;
	double lat = 0.0;
	char alt_name[MAXNAME];
	char *err_str=NULL;

	/* Allocate and align buffer for possible use in O_DIRECT operatino */
	if(copy_file_buf == NULL) /* If you have one, just use it, else get */
	{
	   copy_file_buf= (char *) MALLOC(CHUNK_SIZE + 4096);
	   copy_file_buf = (char *)( ((long)copy_file_buf + 4096) & ~4095); 
	}
#if !defined(WIN32)
		/* for Windows, these flags need to be specified as different */
		/* arguments */
        iflags = O_RDONLY;
        oflags = O_CREAT|O_WRONLY;
#endif
#if defined(WIN32)
	iflags|=FILE_FLAG_POSIX_SEMANTICS;
	oflags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
        if( ((work_obj[my_workload].copyfile_op_count % 100)
                < my_mix_table[my_workload].percent_osync) )
	{
#if defined(WIN32)
		iflags|=FILE_FLAG_WRITE_THROUGH;
		oflags|=FILE_FLAG_WRITE_THROUGH;
#else
		iflags|=O_SYNC;
		oflags|=O_SYNC;
#endif
		iflags2|=2;
		oflags2|=2;
	}
        if( ((work_obj[my_workload].copyfile_op_count % 100) <
                my_mix_table[my_workload].percent_direct) )
	{
#if defined(WIN32)
		iflags|=FILE_FLAG_NO_BUFFERING;
		oflags|=FILE_FLAG_NO_BUFFERING;
#else
		iflags|=O_DIRECT;
		oflags|=O_DIRECT;
#endif
		iflags2|=1;
		oflags2|=1;
	}

	if(cdebug) 
	{
		fprintf(newstdout,"Doing copyfile op\n");
		fflush(newstdout);
	}
	strcpy(alt_name,fp->filename);
	strcat(alt_name,"_2");

	/* Get the size and mode */
	ret=I_STAT(fp->filename, &stbuf);
        my_size = stbuf.st_size;
        my_mode = stbuf.st_mode;
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,copyfile,Stat,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);

	/* If dest does not exist then start from beginning */
	ret=I_STAT(alt_name, &stbuf); 
	if(ret != 0)
		fp->copyfile_offset = 0;	
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,copyfile,Stat,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],alt_name,
			fp->access_count);

	/* Open the input file */
#if defined(WIN32)
	fdi=CreateFile(fp->filename, 
		GENERIC_READ, /*equivalent to O_RDONLY*/
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_EXISTING, /*only open existing file*/
		iflags, NULL);
	if(fdi == INVALID_HANDLE_VALUE)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
 	   	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for copyfile op failed. Filename = %s Error: %s Error value: %d\n", 
			fp->filename,err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(126);
	}
#else 
	fdi=I_OPEN(fp->filename,iflags,0);
	if(fdi <0)
	{
		c_eval=errno;
		err_str = strerror(c_eval);
 	   	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for copyfile op failed. Filename = %s Error: %s\n", 
			fp->filename,strerror(c_eval));
		fflush(newstdout);
		c_eflag++;
		R_exit(126);
	}
#endif
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,copyfile,Opened file,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d O_RDONLY=1\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((iflags2 & 1)?1:0),
			((iflags2 & 2)?1:0));
	/* Seek to last known copyfile_offset */
#if defined(WIN32)
	largeoffset.QuadPart = fp->copyfile_offset;
	SetFilePointerEx(fdi, largeoffset, NULL, FILE_BEGIN);
#else
	I_LSEEK(fdi, fp->copyfile_offset, SEEK_SET);
#endif
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,copyfile,Seek,%s,,%lld,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			(long long)fp->copyfile_offset,fp->access_count);
	/* Now let's go the destination file */
#if defined(WIN32)
	fdo=CreateFile(alt_name, 
		GENERIC_WRITE, /*equivalent to O_WRONLY*/
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, /*equivalent to O_CREAT*/
		oflags, 
		NULL);
	if(fdo == INVALID_HANDLE_VALUE)
	{
		c_eval = GetLastError();
		err_str = win32_strerror(c_eval);
 	   	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for copyfile op failed. Filename = %s Error: %s Error value: %d\n",
			alt_name,err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(126);
	}
#else
	fdo=I_OPEN(alt_name,oflags,my_mode);
	if(fdo <0)
	{
		c_eval=errno;
		err_str = strerror(c_eval);
 	   	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for copyfile op failed. Filename = %s %s\n",
			alt_name,strerror(c_eval));
		fflush(newstdout);
		c_eflag++;
		R_exit(126);
	}
#endif
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,copyfile,Opened file,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d O_WRONLY=1\n",
			client_id,trace_seq++,trace_states[tracestate],alt_name,
			fp->access_count,
			((oflags2 & 1)?1:0),
			((oflags2 & 2)?1:0));
#if defined(_solaris_)
        if( ((work_obj[my_workload].copyfile_op_count % 100) <
                my_mix_table[my_workload].percent_direct) )
        {
                if(!(fp->directio_on))
                {
                    fp->directio_on=1;
                    directio(fdo,DIRECTIO_ON);
                }
        }
        else
        {
                if(fp->directio_on)
                {
                    fp->directio_on=0;
                    directio(fdo,DIRECTIO_OFF);
                }
        }
#endif
#if defined(_macos_)
        if( ((work_obj[my_workload].copyfile_op_count % 100) <
                my_mix_table[my_workload].percent_direct) )
        {
                if(!(fp->directio_on))
                {
                    fp->directio_on=1;
                    fcntl(fdo,F_NOCACHE,1);
                }
        }
        else
        {
                if(fp->directio_on)
                {
                    fp->directio_on=0;
                    fcntl(fdo,F_NOCACHE,0);
                }
        }
#endif
	/* Seek to same copyfile_offset in the output file */
#if defined(WIN32)
	   largeoffset.QuadPart = fp->copyfile_offset;
	   SetFilePointerEx(fdo, largeoffset, NULL, FILE_BEGIN);
#else
	   I_LSEEK(fdo, fp->copyfile_offset, SEEK_SET);
#endif
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,copyfile,Seek,%s,,%lld,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],alt_name,
			(long long)fp->copyfile_offset,fp->access_count);

	/* Now copy a chunk */	
	for(i=0;i<COPY_CHUNK;i++)
	{
	   /*the copy should proceed if the copyfile_offset+CHUNK_SIZE == my_size*/
	   if( fp->copyfile_offset + CHUNK_SIZE <= my_size )
	   {
#if defined(WIN32)
	        err=ReadFile(fdi, copy_file_buf, CHUNK_SIZE, &read_ret, NULL);
		if((FALSE == err) || (read_ret < 0))
	        {
		   c_eval=GetLastError();
		   err_str = win32_strerror(c_eval);
                   c_eval=errno;
                   fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
                   fprintf(newstdout,"Copyfile: Read failed from file %s Error: %s Error value %d Ret %d\n",
			fp->filename,err_str,c_eval,read_ret);
                   fflush(newstdout);
                   c_eflag++;
		   CloseHandle(fdi);
		   CloseHandle(fdo);
		   fp->copyfile_offset = 0;
		   unlink(alt_name);
                   R_exit(110);
	        }
#else
	        read_ret=read(fdi,copy_file_buf,CHUNK_SIZE);
	        if(read_ret <= 0)
	        {
		   c_eval=errno;
		   err_str = strerror(c_eval);
                   c_eval=errno;
                   fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
                   fprintf(newstdout,"Copyfile: Read failed from file %s Error: %s\n",fp->filename,strerror(errno));
                   fflush(newstdout);
                   c_eflag++;
		   close(fdi);
		   close(fdo);
		   fp->copyfile_offset = 0;
		   unlink(alt_name);
                   R_exit(110);
	      }
#endif
	      else
	      {
		   read_transferred+=read_ret;
		   if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,copyfile,Read chunk,%s,,%lld,%d,%ld,transferred=%d\n",
				client_id,trace_seq++,trace_states[tracestate],fp->filename,
				(long long)fp->copyfile_offset,CHUNK_SIZE,fp->access_count,read_ret);
#if defined(WIN32)
	           err=WriteFile(fdo, copy_file_buf, read_ret, &write_ret, NULL);
		   if((FALSE == err) || (write_ret <= 0))
	           {
		           c_eval=GetLastError();
		           err_str = win32_strerror(c_eval);
                           c_eval=errno;
                           fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
                           fprintf(newstdout,"Copyfile: Write failed to file %s Error: %s Error value %d Ret %d\n",
				alt_name,err_str,c_eval,write_ret);
                           fflush(newstdout);
                           c_eflag++;
		           CloseHandle(fdi);
		           CloseHandle(fdo);
		           unlink(alt_name);
                           R_exit(110);
		   }
#else
	           write_ret=write(fdo,copy_file_buf,read_ret);
	           if(write_ret<0)
	           {
		           c_eval=errno;
		           err_str = strerror(c_eval);
                           c_eval=errno;
                           fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
                           fprintf(newstdout,"Copyfile: Write failed to file %s Error: %s\n",alt_name,err_str);
                           fflush(newstdout);
                           c_eflag++;
		           close(fdi);
		           close(fdo);
		           unlink(alt_name);
                           R_exit(110);
		   }
#endif
		   if(tracedebug)
		   	fprintf(tracefile,"%d,%lu,%s,copyfile,Write chunk,%s,,%lld,%d,%ld,transferred=%d\n",
				client_id,trace_seq++,trace_states[tracestate],alt_name,
				(long long)fp->copyfile_offset,CHUNK_SIZE,
				fp->access_count,write_ret);

         	   fp->copyfile_offset += write_ret;
		   write_transferred+=write_ret;
	      }
	   } 
	   else
	   {
		/* We're done, so unlink the destination and head for home */
		fp->copyfile_offset = 0;
		unlink(alt_name);
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,copyfile,Unlink,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],alt_name,
				fp->access_count);
		/*need a "break" statement here?? otherwise if the condition is satisfied in the first iteration, it will do the copy again*/
		break;
	   }
	}

	/* global (across all files in this work object) write count */
	work_obj[my_workload].copyfile_op_count++;

        if( flush_flag &&
             ((work_obj[my_workload].copyfile_op_count % 100) < my_mix_table[my_workload].percent_commit) )
	{
                I_FSYNC(fdo);
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,copyfile,Sync file,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				alt_name,fp->access_count);
	}

        total_copyfile_bytes+=(double)((read_transferred+write_transferred));
        total_read_bytes+=(double)(read_transferred);
        total_write_bytes+=(double)(write_transferred);

	total_file_ops += 2.0; /* Open of both files */
	work_obj[my_workload].total_file_ops+=2.0; /* Open */

	work_obj[my_workload].open_op_count+=2;
	work_obj[my_workload].close_op_count+=2;

	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].copyfile_op_time += lat;
	}
#if defined(WIN32)
	CloseHandle(fdi);
	CloseHandle(fdo);
#else
	close(fdi);
	close(fdo);
#endif
	if(tracedebug)
	{
		fprintf(tracefile,"%d,%lu,%s,copyfile,Closed file,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
		fprintf(tracefile,"%d,%lu,%s,copyfile,Closed file,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],alt_name,
			fp->access_count);
	}
	total_file_ops += 2.0; /* Close */
	work_obj[my_workload].total_file_ops+=2.0; /* Close */
}	

#ifdef HAVE_ANSIC_C
void
generic_rename(struct file_object *fp)
#else
void
generic_rename(fp)
struct file_object *fp;
#endif
{
	int ret;
	double lat = 0.0;
	char alt_name[MAXNAME];
	
	if(cdebug) 
	{
		fprintf(newstdout,"Doing rename op\n");
		fflush(newstdout);
	}
	if(sharing_flag)
		return; /* Let's not do this while sharing all files */

	strcpy(alt_name,fp->filename);
	strcat(alt_name,"_2");

	ret = rename(fp->filename,alt_name);
	if((ret < 0) && !sharing_flag)
	{
		c_eval=errno;
 	   	fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Rename failed to file %s. Error: %s\n",
			fp->filename, strerror(c_eval));
		fflush(newstdout);
		c_eflag++;
		R_exit(125);
	}
	if(ret == 0)
	{
	   if(tracedebug)
	  	 fprintf(tracefile,"%d,%lu,%s,rename,Rename,%s,,,,%ld,newname=%s\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,alt_name);
	   /* global (across all files in this work object) write count */
	   work_obj[my_workload].rename_op_count++;
	   total_file_ops += 1.0; 

	   if(Op_lat_flag)
	   {
	      lat = gettime() - my_gen_op_time;
	      hist_insert(lat);
	      work_obj[my_workload].rename_op_time += lat;
	   }
	}
}
/*
 * This is the code that implements the "statfs" op. It will
 * perform a "statfs" call on a file with the name that is in the 
 * test entry and collect statistics for the operation.
 */
#if defined(_bsd_) || defined(_macos_) || defined(_solaris_) || defined(_cygwin_)
#include <sys/statvfs.h>
#elif !defined(WIN32)
#include <sys/vfs.h>
#endif

#ifdef HAVE_ANSIC_C
void
generic_statfs(struct file_object *fp)
#else
void
generic_statfs(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	char root_path[MAX_PATH]={0}; 
	DWORD SectorsPerCluster;
	DWORD BytesPerSector;
	DWORD NumberOfFreeClusters;
	DWORD TotalNumberOfClusters;
#else
  #if defined(_bsd_) || defined(_solaris_) || defined(_cygwin_)
    #ifdef _LARGEFILE64_SOURCE
	struct statvfs64 stvfsbuf;
    #else
	struct statvfs stvfsbuf;
    #endif
  #else
    #ifdef _LARGEFILE64_SOURCE
      /* 
       * Heavy magic happens with the size of the stat and statfs structures
       * about the time that 10.6 came out.
       */
       #if defined(_macos_)
         #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  > 1050
	     struct statfs stfsbuf;
         #endif
         #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  < 1050
	     struct statfs stfsbuf;
         #endif
         #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  == 1050
	     struct statfs64 stfsbuf;
         #endif
       #else
     	     struct statfs64 stfsbuf;
       #endif
     #else
     	     struct statfs stfsbuf;
     #endif
   #endif
	int ret;
#endif /*WIN32*/
	double lat = 0.0;

	if(cdebug) 
	{
		fprintf(newstdout,"Doing statfs op\n");
		fflush(newstdout);
	}
#if defined(WIN32)
	/*GetDiskFreeSpace() may be the closest one*/
	if (!GetRootPathName(fp->filename, root_path) ||
		!GetDiskFreeSpace(root_path,
							&SectorsPerCluster,
							&BytesPerSector,
							&NumberOfFreeClusters,
							&TotalNumberOfClusters))
	{
 	   	fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"generic_statfs() failed: file name = %s, root path=%s, error value:%d\n",fp->filename, root_path, GetLastError());
		fflush(newstdout);
                c_eval = LOAD_GEN_ERR;
		c_eflag++;
		R_exit(126);
	}
#else
#if defined(_bsd_) || defined(_solaris_) || defined(_cygwin_)
	ret = I_STATVFS(fp->filename,&stvfsbuf);
	if(ret != 0)
	{
 	   	fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"generic_statfs() failed: file name = %s, error string: %s\n",fp->filename, strerror(errno));
		fflush(newstdout);
		c_eval = errno;
		c_eflag++;
		R_exit(126);
	}
#else
	ret = I_STATFS(fp->filename,&stfsbuf);
	if(ret != 0)
	{
 	   	fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"generic_statfs() failed: file name = %s, error code: %s\n",fp->filename, strerror(errno));
		fflush(newstdout);
		c_eval = errno;
		c_eflag++;
		R_exit(126);
	}
#endif
#endif
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,statfs,Statfs,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
	/* global (across all files in this work object) write count */
	work_obj[my_workload].statfs_op_count++;

	total_file_ops += 1.0; /* for the statfs call */
	work_obj[my_workload].total_file_ops+=1.0; /* Open */

	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].statfs_op_time += lat;
	}
}	

#ifdef HAVE_ANSIC_C
void
generic_pathconf(struct file_object *fp)
#else
void
generic_pathconf(fp)
struct file_object *fp;
#endif
{
	
	double lat = 0.0;
#if defined(WIN32)
	char root_path[MAX_PATH]; 
	DWORD MaxComponentLength;
	char VolumeName[MAX_PATH+1];
#else
	int ret;
#endif
	if(cdebug) 
	{
		fprintf(newstdout,"Doing pathconf op\n");
		fflush(newstdout);
	}
#if defined(WIN32)
	if (!GetRootPathName(fp->filename, root_path))
	{
 	   	fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"GetRootPathName(): invalid file name: %s\n",fp->filename);
		fflush(newstdout);
		c_eval = LOAD_GEN_ERR;
		c_eflag++;
		R_exit(126);
	}
	GetVolumeInformation(root_path, VolumeName,MAX_PATH+1,NULL,&MaxComponentLength,NULL,NULL,0);
#else
	errno=0; /* Yeah, this is the goofy way that pathconf works */
#if defined(_bsd_) || defined(_cygwin_)
	ret = pathconf(fp->filename,_PC_ASYNC_IO);
#else
	ret = pathconf(fp->filename,_PC_NAME_MAX);
#endif
	if((ret < 0) && (errno != 0))
	{
 	   	fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Pathconf() failed. Error %s\n",strerror(errno));
		fflush(newstdout);
		c_eval = errno;
		c_eflag++;
		R_exit(126);
	}
#endif
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,pathconf,Pathconf,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count);
	/* global (across all files in this work object) write count */
	work_obj[my_workload].pathconf_op_count++;

	total_file_ops += 1.0; /* for the pathconf call */
	work_obj[my_workload].total_file_ops+=1.0; /* Open */

	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].pathconf_op_time += lat;
	}
}	

#ifdef HAVE_ANSIC_C
void
generic_custom1(struct file_object *fp)
#else
void
generic_custom1(fp)
struct file_object *fp;
#endif
{ 
	/* Implement Custom1 op type here */
}

#ifdef HAVE_ANSIC_C
void
generic_custom2(struct file_object *fp)
#else
void
generic_custom2(fp)
struct file_object *fp;
#endif
{ 
	/* Implement Custom2 op type here */
}


#ifdef HAVE_ANSIC_C
int
lock_file(char *filename,struct file_object *fp)
#else
int
lock_file(filename,fp)
char *filename;
struct file_object *fp;
#endif
{
	int flags=0; /*initialized flags to 0*/
#if defined(WIN32)
	HANDLE fd;
	OVERLAPPED overlapped;
#else
    struct flock myflock;
	int fd;
#endif
    int lret;
	double lat = 0.0;
	char *err_str = NULL;
	
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
	fd = CreateFile(filename,
                       GENERIC_READ | GENERIC_WRITE,		//O_RDWR
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,    //Allow other to read/write
                       NULL,								// no security attributes
                       OPEN_ALWAYS ,					    // attempt to open existing file, create a new one if the file does not exist
                       flags,									// not overlapped index/O
                       NULL);
	if(INVALID_HANDLE_VALUE == fd)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else
	fd=I_OPEN(filename,O_CREAT|O_RDWR|flags,0666);
	if(fd <0)
	{
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	   	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for lock op failed. Filename = %s Error: %s Error value: %d\n",
			filename,err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(126);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,lock_file,Opened file,%s,0666,,,%ld,\n",client_id,
			trace_seq++,trace_states[tracestate],filename,fp->access_count);
                 /* Generic lock the whole file */
	total_file_ops += 1.0;
	work_obj[my_workload].total_file_ops+=1.0;
	work_obj[my_workload].open_op_count++;
	total_meta_ops += 1.0;
#if defined(WIN32)
	/* initialize overlapped structure */
	ZeroMemory(&overlapped, sizeof(overlapped));
	/*lock start from beginning*/
	overlapped.Offset = (DWORD)0;
	overlapped.OffsetHigh = (DWORD)0;
	if (FALSE == LockFileEx(fd,			// file handle
							  LOCKFILE_EXCLUSIVE_LOCK|LOCKFILE_FAIL_IMMEDIATELY, //write lock
							  0,			//reserved, must be 0
							  (DWORD)-1,	//The low-order 32 bits of the length of the byte range to lock.
							  (DWORD)-1,	//The high-order 32 bits of the length of the byte range to lock
							  &overlapped))
	{
		lret = -1;
#else
    	myflock.l_type=F_WRLCK; /* Apply write lock */
    	myflock.l_whence=SEEK_SET;
    	myflock.l_start=0;
    	myflock.l_len=0; /* The whole file */
    	myflock.l_pid=getpid();
    	lret=fcntl(fd,F_SETLKW, &myflock);
    	if(lret < 0)
    	{
#endif
 	   	fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"lock op failed. Error %s\n",strerror(errno));
		fflush(newstdout);
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		return(lret);
	}
	fp->op_count++;
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,lock_file,Lock,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],filename,fp->access_count);
	total_file_ops += 1.0; /* fcntl */
	work_obj[my_workload].total_file_ops+=1.0; /* fcntl */
	work_obj[my_workload].lock_op_count++;
#if defined(WIN32)
	/* initialize overlapped structure */
	ZeroMemory(&overlapped, sizeof(overlapped));
	/*Unlock from beginning*/
	overlapped.Offset = (DWORD)0;
	overlapped.OffsetHigh = (DWORD)0;
	UnlockFileEx(fd,			// file handle
				0,			//reserved, must be 0
				(DWORD)-1,	//The low-order 32 bits of the length of the byte range to lock.
				(DWORD)-1,	//The high-order 32 bits of the length of the byte range to lock
				&overlapped);
	CloseHandle(fd);
#else
    	myflock.l_type=F_UNLCK; /* Unlock */
    	lret=fcntl(fd,F_SETLKW, &myflock);
	if(lret < 0)
	{
 	   	fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Unlock failed. Error %s\n",strerror(errno));
		fflush(newstdout);
	}
	close(fd);/* This will also release the lock */
#endif
	if(tracedebug)
	{
		fprintf(tracefile,"%d,%lu,%s,lock_file,Closed file,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],filename,fp->access_count);
	}
	total_file_ops += 1.0; /* Close */
	work_obj[my_workload].total_file_ops+=1.0; /* Close */
	work_obj[my_workload].close_op_count++;
	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].lock_op_time += lat;
	}
    return(0);
}

/* This is only used in ShareMode to prevent multiple
 * writers in the same file, at the same time, during INIT.
 * As the file size would end up being wrong.
 */
#if defined(WIN32)
int
try_lock_file_hold(HANDLE fileptr)
#else
#ifdef HAVE_ANSIC_C
int
try_lock_file_hold(int fileptr)
#else
int
try_lock_file_hold(fileptr)
int fileptr;
#endif
#endif /*WIN32*/
{
#if !defined(WIN32)
        struct flock myflock;
        int lret;
	
        myflock.l_type=F_WRLCK; /* Apply write lock */
        myflock.l_whence=SEEK_SET;
        myflock.l_start=0;
        myflock.l_len=0; /* The whole file */
        myflock.l_pid=getpid();
        lret=fcntl(fileptr,F_SETLK, &myflock);
	if(lret < 0)
	{
		return(-1);
	}
#else
	if (FALSE == LockFile(fileptr, 0, 0, (DWORD)-1, (DWORD)-1))
		return (-1);
#endif
        return(0);
}

#ifdef HAVE_ANSIC_C
void
clear_stats(void)
#else
void
clear_stats()
#endif
{
	int i;

	total_read_bytes=0;
	total_write_bytes=0;
	total_copyfile_bytes = total_file_ops=total_read_rand_bytes=0;
	total_file_op_time=total_meta_ops=0;
	total_rmw_ops=0;
	for(i=0;i<NUM_WORK_OBJ;i++)
	{
		/* Raw counters */
	    work_obj[i].write_op_count=0;
	    work_obj[i].write_file_op_count=0;
	    work_obj[i].read_op_count=0;
	    work_obj[i].read_file_op_count=0;
	    work_obj[i].mkdir_op_count=0;
	    work_obj[i].unlink_op_count=0;
	    work_obj[i].create_op_count=0;
	    work_obj[i].stat_op_count=0;
	    work_obj[i].copyfile_op_count=0;
	    work_obj[i].rename_op_count=0;
	    work_obj[i].statfs_op_count=0;
	    work_obj[i].pathconf_op_count=0;
	    work_obj[i].custom1_op_count=0;
	    work_obj[i].custom2_op_count=0;
	    work_obj[i].append_op_count=0;
	    work_obj[i].lock_op_count=0;
	    work_obj[i].access_op_count=0;
	    work_obj[i].chmod_op_count=0;
	    work_obj[i].readdir_op_count=0;
	    work_obj[i].write_rand_op_count=0;
	    work_obj[i].read_rand_op_count=0;
	    work_obj[i].mmap_write_op_count=0;
	    work_obj[i].mmap_read_op_count=0;
	    work_obj[i].rmw_op_count=0; 
	    work_obj[i].open_op_count=0;
	    work_obj[i].close_op_count=0;

		/* Raw time accumulators */
	    work_obj[i].write_op_time=0.0;
	    work_obj[i].write_file_op_time=0.0;
	    work_obj[i].read_op_time=0.0;
	    work_obj[i].read_file_op_time=0.0;
	    work_obj[i].mkdir_op_time=0.0;
	    work_obj[i].unlink_op_time=0.0;
	    work_obj[i].create_op_time=0.0;
	    work_obj[i].stat_op_time=0.0;
	    work_obj[i].copyfile_op_time=0.0;
	    work_obj[i].rename_op_time=0.0;
	    work_obj[i].statfs_op_time=0.0;
	    work_obj[i].pathconf_op_time=0.0;
	    work_obj[i].custom1_op_time=0.0;
	    work_obj[i].custom2_op_time=0.0;
	    work_obj[i].append_op_time=0.0;
	    work_obj[i].lock_op_time=0.0;
	    work_obj[i].access_op_time=0.0;
	    work_obj[i].chmod_op_time=0.0;
	    work_obj[i].readdir_op_time=0.0;
	    work_obj[i].write_rand_op_time=0.0;
	    work_obj[i].read_rand_op_time=0.0;
	    work_obj[i].rmw_op_time=0.0;
	    work_obj[i].mmap_write_op_time=0.0;
	    work_obj[i].mmap_read_op_time=0.0;
	    work_obj[i].total_file_op_time = 0.0;
	    work_obj[i].total_file_ops = 0.0;
	}
	for(i=0;i<BUCKETS;i++)
	{
	  global_buckets[i]=0LL;
	  buckets[i]=0LL;
	}
}

/*
 * Return time in seconds. Floating poiint.
 */
#ifdef HAVE_ANSIC_C
double
gettime(void)
#else
double
gettime()
#endif
{
  struct timeval tp;

  if(pit_hostname[0]!=0)
  {
#if defined(_solaris_)
     if(pit_gettimeofday((struct timeval *)&tp, NULL, pit_hostname, 
	  pit_service) == -1){
#else
     if(pit_gettimeofday((struct timeval *)&tp, (struct timezone *) NULL, pit_hostname, 
	  pit_service) == -1){
#endif
	     fprintf(newstderr,"%s Pit_gettimeofday error\n",get_timestamp(&time_string[0]));
	     fflush(newstderr);
	     c_eflag++;
	     c_eval = BAD_PIT_CONN;
	     R_exit(BAD_PIT_CONN);
     }
     return ((double) (tp.tv_sec)) + (((double) tp.tv_usec) * 0.000001 );
  }
  else
  {
#if defined(_solaris_)
    return(m_gettimeofday((struct timeval *)&tp, (long)NULL));
#else
    return(m_gettimeofday((struct timeval *)&tp, (struct timezone *) NULL));
#endif
  }
}


/*
 * Called by signal handler and by exit paths where
 * things are not going well and the test needs to
 * cleanup and exit.
 */
#ifdef HAVE_ANSIC_C
void
pipe_elog(int val)
#else
void
pipe_elog(val)
int val;
#endif
{
	int i;
	struct stat buf;
	if(mflag)
	{
		printf("Prime got a SIGPIPE. Value: %d\n",val);
		printf("This indicates that the operating system is unexpectedly closing sockets.\n");
		printf("This may indicate an overload of the clients in the solution under test.\n");
		fflush(stdout);
		kill_clients();
	}
	else
	{
		fprintf(newstdout,"Child got a SIGPIPE. Value: %d\n",val);
		fprintf(newstdout,"This indicates that the operating system is unexpectedly closing sockets.\n");
		fprintf(newstdout,"This may indicate an overload of the clients in the solution under test.\n");
		fflush(newstdout);
		for(i=0;i<10;i++)
		{
			if((i==child_listen_socket) ||
				(i==child_socket_val))
				continue;
			/*		
			 * If a valid file descriptor.
			 */
			if(fstat(i,&buf)==0)
				close(i);
		}
		(*Netmist_Ops[my_workload].teardown)();
		remove_kill_file();
		c_eflag++;
		c_eval = val;
		if(c_eflag)
			tell_master_error(client_id,c_eval);
	}
	exit(val);
}


#ifdef HAVE_ANSIC_C
void
clean_exit(int val)
#else
void
clean_exit(val)
int val;
#endif
{
	int i;
	struct stat buf;

	if(mflag)
	{
		printf("Exit with cleanup\n");
		fflush(stdout);
		kill_clients();
	}
	else
	{
		for(i=0;i<10;i++)
		{
			if((i==child_listen_socket) ||
				(i==child_socket_val))
				continue;
			/*		
			 * If a valid file descriptor.
			 */
			if(fstat(i,&buf)==0)
				close(i);
		}
		(*Netmist_Ops[my_workload].teardown)();

/* This is now done via the call above. teardown() 
		if(sharing_flag)
		{
		   if(cleanup_flag)
		       client_remove_files(0);
		}
		else
		{
		   if(cleanup_flag)
		      client_remove_files(client_id);
		}
*/
		remove_kill_file();
		if(c_eflag)
			tell_master_error(client_id,c_eval);
	}
	exit(val);
}

/*
 * Called by exit paths on remote children. Indicates
 * things are not going well and the test needs to
 * cleanup and exit. 
 * 
 * If c_eflag is set, then c_eval is the cause and
 * will be handed to the master for reporting.
 *
 * The passed in parameter will be this processes exit status.
 */
#ifdef HAVE_ANSIC_C
void
R_exit(int val)
#else
void
R_exit(val)
int val;
#endif
{
#if !defined(WIN32)
	int i;
	struct stat buf;
#endif

	if(mflag) /* If I am master */
	{
		printf("Exit with cleanup. Terminating remote processes.\n");
		fflush(stdout);
		kill_clients();
	}
	else
	{
		/* Tell the master the cause, if you have one */
		if(c_eflag)
			tell_master_error(client_id,c_eval);
#if !defined(WIN32) 
		/*This does not work for Windows, will cause app to crash*/
		for(i=0;i<10;i++)
		{
			if((i==child_listen_socket) ||
				(i==child_socket_val))
				continue;
			/*		
			 * If a valid file descriptor.
			 */
			if(fstat(i,&buf)==0)
				close(i);
		}
#endif
		(*Netmist_Ops[my_workload].teardown)();

/* This is now done via teardown() above 
		if( sharing_flag )
		{
		   client_remove_files(0);
		}
		else
		{
		      client_remove_files(client_id);
		}
*/
		remove_kill_file();
	}

#if defined(WIN32)
	/* close network connection */
	win32_close();
#endif
	exit(val);
}


/*
 * Your generic help screen.
 */
#ifdef HAVE_ANSIC_C
void
usage(void)
#else
void
usage()
#endif
{
	int segsize = 0;
	unsigned short sum_val = 0;
	printf("\n\t%s: \n\n",short_testname);
	printf("\tUser options:\n");
	printf("\tRequired:\n");
	printf("\t[-b #[mMgGtT]] ......... Client memory size in [MGT]iGytes.\n");
	printf("\t[-B #[mMgGtT]] ......... Adjustable aggregate data set value in [MGT]iBytes.\n");
	printf("\t[-d #]  ................ Number of processes/client.\n");
	printf("\t[-g client_token_file] . Path to file that contains token based info.\n");
	printf("\tor\n");
#if defined(DEPRECATED)
	printf("\t[-f client_filename] ... Path to file that contains client names. (deprecated)\n");
#endif
	printf("\t\n");
	printf("\tOptional:\n");
	printf("\t[-t #] ................. Seconds to run.\n");
	printf("\t[-w #] ................. Seconds to warmup.\n");
	printf("\t[-r #[kKmMgG] .......... Force average file size to this size.\n");
	printf("\t[-G]   ................. Enable heartbeat notifications.\n");
	printf("\t[-N]   ................. Enable Op latency collection.\n");
	printf("\t[-O #] ................. Requested Op/sec rate.\n");
	printf("\t[-q heartbeat_log] ..... Save all heartbeats in this log.\n");
	printf("\t[-Q #] ................. Number of files per dir.\n");
	printf("\t[-R masters dir ]  ..... Master's result directory name.\n");
	printf("\t[-T #] ................. Number of dirs per proc.\n");
	printf("\t[-U log_path ] ...       Directory path for clients logs.\n");
	printf("\t[-l log_filename] ...... Master's log filename.\n");
	printf("\t[-i] ................... Information without run.\n");
	printf("\t[-k ]  ................. Kill the testing on all clients.\n");
	printf("\t[-v ]  ................. Version information.\n");
	printf("\t[-D ]  ................. Install benchmark on clients.\n");
	printf("\t[-E ]  ................. Export mix file to stdout.\n");
	printf("\t[-I filename]  ......... Import mix from file.\n");
	printf("\t[-H name ]  ............ Pit_server's hostname.\n");
	printf("\t[-S # ]  ............... Pit_server's port.\n");
	printf("\t[-F ]  ................. Disable all fsyncs.\n");
	printf("\t[-K ]  ................. Disable file unlink.\n");
	printf("\t[-u ]  ................. Use rsh/rcp instead of ssh/scp.\n");
	printf("\t[-y 6] ................. Use IPv6.\n");
	printf("\t[-z ]  ................. Run locally only.\n");
	printf("\t[-a offset]  ........... Addional byte offset for xfers.\n");
	printf("\t[-Y ]  ................. Dump file access list to the client log.\n");
	printf("\t[-J #]  ................ Limit open file desc caching to this value.\n");
	printf("\t[-X ]  ................. Sharing mode.\n");
	printf("\t[-f name]  ............. External monitor script name.\n");
	printf("\t[-j args]  ............. External monitor script arguments.\n");
	printf("\t[-x ]  ................. Enable socket debug.\n");
	printf("\t[-2 ]  ................. Enable trace debug.\n");
	printf("\t[-h ]  ................. Help menu.\n");
	printf("\n");
	printf("\tInternal use only. Do not use.\n");
	printf("\t[-m ]  ................. Master flag.\n");
	printf("\t[-e execdir ]........... Location of executable on client.\n");
	printf("\t[-c #] ................. Total Number of clients.\n");
	printf("\t[-n #] ................. Client identification number.\n");
	printf("\t[-p directory path] .... Path to working directory.\n");
	printf("\t[-C client_name ]  ..... Client name.\n");
	printf("\t[-M master's name ] .... Master's name.\n");
	printf("\t[-s #]  ................ File size in kbytes.\n");
	printf("\t[-o #]  ................ Ops/sec.\n");
	printf("\t[-P #]  ................ Master's listen port.\n");
	printf("\t[-W]    ................ Workload name.\n");
	printf("\t[-L]    ................ Debug on local filesystem.\n\n");
	if(sizeof(struct pdsm_remote_stats) > sizeof(struct pdsm_remote_control))		
	   segsize = sizeof(struct pdsm_remote_stats);
	else
	   segsize = sizeof(struct pdsm_remote_control);
	segsize = (segsize+getpagesize()-1) & ~(getpagesize()-1);
	printf("\tMin PDSM size %d bytes..The minimum PDSM seg size.\n",segsize);
	sum_val = check_sum();
	if(modified_run)
	    sum_val = 0;
	if(licensed)
	   printf("\tRegistered Finger Print     %5u \n",(unsigned int)sum_val);
	else
	   printf("\tPublic Finger Print         %5u \n",(unsigned int)sum_val);
	printf("\n\n");
	fflush(stdout);
}

/*
 * This is the code that reads the client file that contains
 * the descriptions of the clients that will be running the test.
 * Format of the file is:
 * 
 * client_name  work_dir  exec_dir workload_type
 *
 * Space delimited.
 */

int is_empty_line(char* buf)
{
	while (buf != NULL && *buf != '\0')
	{
		if (*buf == ' ' ||
			*buf == '\t' ||
			*buf == '\n' ||
			*buf == '\r')
		{
			buf++;
		}
		else return 0;
	}
	return 1;
}

int is_comment(char* buf)
{
	if (buf != NULL && buf[0] == '#')
		return 1;
	else
		return 0;
}

#ifdef HAVE_ANSIC_C
void
read_clientfile(void)
#else
void
read_clientfile()
#endif
{
	char buf[600];
	int ret;
	int client_count;
	char *cret;
	char workdir[MAXNAME], lexecdir[MAXNAME], name[50], workload_name[20];
	char dom_user[MAXNAME],password[MAXNAME];
	FILE *fd;

	client_count=0;

	fd=fopen(client_filename,"r");
	if(fd==0)
	{
		printf("Bad client filename > %s <\n",client_filename);
		fflush(stdout);
		exit(0);
	}
	while(1)
	{
		bzero(buf,MAXNAME);
		cret=fgets(buf,600,fd);
		if(cret == (char *)0)
		{
			break;
		}
		/*we also ignore empty line*/
		if(buf[0]=='#' || is_empty_line(buf)) 
			continue;
		ret=sscanf(buf,"%s %s %s %s %s %s\n",name, dom_user, password, workdir, lexecdir, workload_name);
		if(ret==-1)
		{
			break;
		}
		if(ret != NUM_TOKENS)
		{
		   printf("Error reading configuration file.\n");
		   printf("Found %d tokens but expecting %d tokens.\n",ret,NUM_TOKENS);
		   printf("Format: Host   User   Workdir   Exec_dir   Workload\n");
		   fflush(stdout);
		   exit(1);
		}
		strcpy(client_obj[client_count].client_name,name);
		strcpy(client_obj[client_count].client_dom_user,dom_user);
		strcpy(client_obj[client_count].client_password,password);
		strcpy(client_obj[client_count].client_workdir,workdir);
		strcpy(client_obj[client_count].client_execdir,lexecdir);
		strcpy(client_obj[client_count].client_workload_name,workload_name);
		client_obj[client_count].client_id=client_count;
		if(xdebug) printf("Client name %s Workdir %s execdir %s\n",name,workdir,lexecdir);
		client_count++;
		if(strlen(workload_name)==0)
		{
			printf("Master: Missing workload name in configuration file.\n");
		   	fflush(stdout);
			exit(1);
		}
		if(client_count > MAXCLIENTS)
		{
			printf("Master: Too many clients. Increase MAXCLIENTS and recompile.\n");
		   	fflush(stdout);
			exit(1);
		}
		if(license_check() == 0)
		{
		   if(client_count > max_licensed_clients)
		   {
			printf("Too many client procs for un-licensed version.\n ");
		   	fflush(stdout);
			exit(1);
		   }
		   if(im_flag)
		   {
			printf("Import of Workload objects not permitted in the un-licensed version.\n ");
		   	fflush(stdout);
			exit(1);
		   }
		}
	}
	num_clients=client_count;
	if (num_clients <= 0)
	{
		/*Microsofot: terminate the test if no client definition is found*/
		printf("No client found in client file: %s, test aborted\n", client_filename);
		fflush(stdout);
		exit(1);
	}
}
#ifdef HAVE_ANSIC_C
void
read_client_tok_file(void)
#else
void
read_client_tok_file()
#endif
{
	int client_count;
	char buf[600];
	FILE *fd;
	int ret,j;
	char *cret;

	client_count=0;

	fd=fopen(client_tok_filename,"r");
	if(fd==0)
	{
		printf("Bad client filename > %s <\n",client_tok_filename);
		fflush(stdout);
		exit(0);
	}
	while(1)
	{
		bzero(buf,600);
		cret=fgets(buf,600,fd);
		if(cret == (char *)0)
		{
			break;
		}
		/*We also ignore empty line*/
		if(buf[0]=='#' || is_empty_line(buf)) 
			continue;
		if(buf[strlen(buf)-1]=='\n')
			buf[strlen(buf)-1]=0;;
        	ret = parse_all_tokens(buf);
		if(ret)
		{
	  	     printf("Error parsing config\n");
		     fflush(stdout);
		     exit(1);
		}
		ret = validate_all_tokens(client_count);
		if(ret)
		{
	  	     printf("Error validating config\n");
		     fflush(stdout);
		     exit(1);
		}
		client_obj[client_count].client_id=client_count;

		if( client_obj[client_count].instances != 0)
		{
		   j = client_obj[client_count].instances;
		   while( j > 1)
		   {
			/* 
			 * Now clone all of the fields from this entry to however 
		         * many was specified 
			 */
		       client_obj[client_count].client_id=client_count;
		       client_obj[client_count+1].client_id=client_count+1;
	               strcpy(client_obj[client_count+1].client_name,client_obj[client_count].client_name);
		       strcpy(client_obj[client_count+1].client_dom_user, client_obj[client_count].client_dom_user);
		       strcpy(client_obj[client_count+1].client_password, client_obj[client_count].client_password);
		       strcpy(client_obj[client_count+1].client_workdir,client_obj[client_count].client_workdir);
		       strcpy(client_obj[client_count+1].client_execdir, client_obj[client_count].client_execdir);
		       strcpy(client_obj[client_count+1].client_workload_name,client_obj[client_count].client_workload_name);
		       client_obj[client_count+1].op_rate=client_obj[client_count].op_rate;
		       client_obj[client_count+1].instances = 0;
		       client_count++;
		       j--;
		       if(client_count > MAXCLIENTS)
		       {
			 printf("Master: Error: Increase MAXCLIENTS.\n");
		         fflush(stdout);
			 exit(1);
		       }
		       if(license_check() == 0)
		       {
		          if(client_count > max_licensed_clients)
		          {
			      printf("Too many client procs for un-licensed version.\n ");
		              fflush(stdout);
			      exit(1);
		          }
		       }
		   }
		}
		client_count++;

		if(client_count > MAXCLIENTS)
		{
			printf("Master: Too many clients. Increase MAXCLIENTS and recompile.\n");
		        fflush(stdout);
			exit(1);
		}
		if(license_check() == 0)
		{
		    if(client_count > max_licensed_clients)
		    {
		        printf("Too many client procs for un-licensed version.\n ");
		        fflush(stdout);
			exit(1);
		    }
		}
	}
	num_clients=client_count;
	if (num_clients <= 0)
	{
		/*Terminate the test if no client definition is found*/
		printf("No client found in client token file: %s, test aborted\n", client_filename);
		fflush(stdout);
		exit(1);
	}
}

/*
 * If the testing is completed and the cleanup is done
 * then remove the kill file.
 */
#ifdef HAVE_ANSIC_C
void
remove_kill_file(void)
#else
void
remove_kill_file()
#endif
{
	char buf[500];
#if defined(WIN32)
	sprintf(buf,"%s\\id%d_killfile",client_log_dir,client_id);
#else
	sprintf(buf,"%s/id%d_killfile",client_log_dir,client_id);
#endif
	/* Can use POSIX, as this is a log and not workload specific */
	unlink(buf);
}

/*
 * Create a kill file that contains this client's pid. If 
 * a termination is requested then this file is used to 
 * send a SIGINT to the copy of the benchmark that is running
 * and tell it to perform a clean_exit.
 */
#ifdef HAVE_ANSIC_C
void
create_kill_file(void)
#else
void
create_kill_file()
#endif
{
	char buf[500];
	FILE *fd;
#if defined(WIN32)
	sprintf(buf,"%s\\id%d_killfile",client_log_dir,client_id);
#else
	sprintf(buf,"%s/id%d_killfile",client_log_dir,client_id);
#endif

	unlink(buf);

	fd=fopen(buf,"w+");
	if(fd == 0)
	{
 	   	fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
	   	fprintf(newstdout,"Unable to open kill file for client %d\n",client_id);
		fflush(newstdout);
		c_eflag = 1;
		c_eval = NO_CREATE;
		R_exit(NO_CREATE);
	}
	fprintf(fd,"%d\n",(int)getpid());
	fflush(fd);
	fclose(fd);
}

/*
 * If a kill of the benchmark has been requested then find the
 * kill file for this client and send him/her a SIGINT so he/she
 * will perform a clean_exit().
 */
#ifdef HAVE_ANSIC_C
void
use_kill_file(void)
#else
void
use_kill_file()
#endif
{
	char buf[500];
	FILE *fd;
	int mypid;

	/* Speical short cut.. use PID handed to master at JOIN time */
	if(Vflag && ( my_child_pid != 0 ) ) /* If you can, try using the pid that was */
				            /* handed back when the child started     */
	{
	    kill((pid_t)my_child_pid,SIGINT);	
	    return;
	}
#if defined(WIN32)
	sprintf(buf,"%s\\id%d_killfile",client_log_dir,client_id);
#else
	sprintf(buf,"%s/id%d_killfile",client_log_dir,client_id);
#endif
	fd=fopen(buf,"r");
	if(fd == 0)
	{
		exit(5); /* No pid to kill */
	}
	fscanf(fd,"%d\n",&mypid);
	fclose(fd);
	kill(mypid,SIGINT);	
}


#ifdef HAVE_ANSIC_C
void
distribute(void)
#else
void
distribute()
#endif
{
	int i;
	char command[1024];
	for(i=0;i<num_clients;i++)
	{
		printf("Installing %s on client %s in %s\n",
			short_testname,client_obj[i].client_name,
			client_obj[i].client_execdir);
		fflush(stdout);
#if !defined(WIN32)
		sprintf(command,"%s %s %s:%s",
			remcp,bin_testname, client_obj[i].client_name,
			client_obj[i].client_execdir);
#else
		sprintf(command,"mkdir %s & copy /y %s.exe %s", GetParentPath(map_to_network_path(client_obj[i].client_name,
			client_obj[i].client_execdir)), bin_testname, map_to_network_path(client_obj[i].client_name,
			client_obj[i].client_execdir));
#endif
		printf("%s\n", command);
		fflush(stdout);
		system(command);
	}
}

#ifdef HAVE_ANSIC_C
void
purge_cache(void)
#else
void
purge_cache()
#endif
{
	int j;
	long long memsize;
	
	/* Disabled for now. It drives Linux into a VM nightmare */
	if(1)
	  return;

	memsize = fsize;
	if(cdebug) 
	{
	   fprintf(newstdout,"Client proc memsize %lld Kbytes\n",memsize);
	   fflush(newstdout);
	}

	if(!mem_buf)
	{
	    mem_buf=(long *)MALLOC(memsize * 1024);
	    bzero(mem_buf,memsize * 1024);
	}

	for(j=0; j < memsize * 1024; j+=1024)
	{
	   mem_buf[j]=0;
	}
}	

/****************************************************************************/
/* COMM SECTION */
/****************************************************************************/
#ifdef HAVE_ANSIC_C
int
master_wait_barrier(int num)
#else
int
master_wait_barrier(num)
int num;
#endif
{
	int i;
	struct master_command *mc;
	struct result_object;

	mc=&master_rcv_buf;
	
	i = 0;
	while(i<num)
	{
		bzero(mc,sizeof(struct master_command));
		master_listen(master_listen_socket);
		if( mc->m_command == R_FLAG_DATA)
		{
			i++;
			if(sockdebug)
				printf(" %s Master got ready message from %d. Remaining %d\n",
				  get_timestamp(&time_string[0]),mc->m_client_number,num-i);
			continue;
		}
		if( mc->m_command == INIT_PHASE)
		{
		   if(mc->m_client_number == (num_clients-1))
		   {
		     	printf("\t%s Starting INIT phase\n", 
			   get_timestamp(&time_string[0]));
			fflush(stdout);
			if(log_flag)
			{
				fprintf(log_out,"\t%s Starting INIT phase\n", 
			     get_timestamp(&time_string[0]));
				fflush(log_out);
			}
		   }
		   continue;
		}
		if( mc->m_command == RUN_PHASE)
		{
		   if(mc->m_client_number == (num_clients-1))
		   {
		        printf("\t%s Starting RUN phase\n", 
			   get_timestamp(&time_string[0]));
			fflush(stdout);
			if(log_flag)
			{
			   fprintf(log_out,"\t%s Starting RUN phase\n", 
			   get_timestamp(&time_string[0]));
			   fflush(log_out);
			}
			/* Let the consumer defined monitor script know we're 
   			   entering RUN */
			if(extern_mon(SCHECK,"RUNNING")==0)
	   		   extern_mon(SRUN,"RUNNING");

			/* Let the user defined monitor script know we're 
			   entering RUN */
			if(u_extern_mon(SCHECK,"RUNNING")==0)
	   		   u_extern_mon(SRUN,"RUNNING");

			/* Consumer defined stat script. This is so vendor 
			   can zero counters */
			if(extern_stat(SCHECK,"ZERO")==0)
	   		   extern_stat(SRUN,"ZERO");
		   }
		   continue;
		}
		if( mc->m_command == WARM_PHASE)
		{
		   if(mc->m_client_number == (num_clients-1))
		   {
		        printf("\t%s Starting WARM phase\n", 
			   get_timestamp(&time_string[0]));
			fflush(stdout);
			if(log_flag)
			{
			   fprintf(log_out,"\t%s Starting WARM phase\n", 
			   get_timestamp(&time_string[0]));
			   fflush(log_out);
			}
		   }
		   continue;
		}
		if( mc->m_command == R_PERCENT_I_COMPLETE)
		{
		   if(mc->m_client_number == (num_clients-1))
		   {
		     printf("\t%s Init %d percent complete\n",
			get_timestamp(&time_string[0]),mc->m_percent_complete);
			fflush(stdout);
			if(log_flag)
			{
			  fprintf(log_out,"\t%s Init %d percent complete\n",
			  get_timestamp(&time_string[0]),mc->m_percent_complete);
			  fflush(log_out);
			}
		   }
		   continue;
		}
		if( mc->m_command == R_PERCENT_COMPLETE)
		{
		   if(mc->m_client_number == (num_clients-1)) /* grab highest client id */
		   {
		     printf("\t%s Run %d percent complete\n",
			get_timestamp(&time_string[0]),mc->m_percent_complete);
			fflush(stdout);
			if(log_flag)
			{
			  fprintf(log_out,"\t%s Run %d percent complete\n",
			  get_timestamp(&time_string[0]),mc->m_percent_complete);
			  fflush(log_out);
			}
		   }
		   continue;
		}
		if( mc->m_command == R_PERCENT_W_COMPLETE)
		{
		   if(mc->m_client_number == (num_clients-1))
		   {
		     printf("\t%s Warm-up %d percent complete\n",
			get_timestamp(&time_string[0]),mc->m_percent_complete);
			fflush(stdout);
			if(log_flag)
			{
			  fprintf(log_out,"\t%s Warm-up %d percent complete\n",
			  get_timestamp(&time_string[0]),mc->m_percent_complete);
			  fflush(log_out);
			}
		   }
		   continue;
		}

		current_time = time(0);
		if( (mc->m_command == R_HEARTBEAT) && heartbeat_flag )
		{
		   if(log_beats && (current_time - delivered_log_heartbeat >= HEARTBEAT_TICK))
		   {
		     switch(mc->m_heartbeat_type) {
		     case WARM_BEAT:
		        fprintf(log_beat_fd,"\t%s Warm Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
		        fflush(log_beat_fd);
			delivered_log_heartbeat = time(0);
			break;
		     case PRE_WARM_BEAT:
		        fprintf(log_beat_fd,"\t%s Pre-Warm Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
		        fflush(log_beat_fd);
			delivered_log_heartbeat = time(0);
			break;
		     case INIT_BEAT:
		        if(mc->m_client_number == (num_clients-1))
		        {
		             fprintf(log_beat_fd,"\t%s Init Heartbeat __/\\_/\\__ Client %d\n", 
			         get_timestamp(&time_string[0]),
			         mc->m_client_number);
		             fflush(log_beat_fd);
		        }
			delivered_log_heartbeat = time(0);
			break;
		     case CLEAN_BEAT:
		        if(mc->m_client_number == (num_clients-1))
		        {
		           fprintf(log_beat_fd,"\t%s Cleanup Heartbeat __/\\_/\\__ Client %d\n", 
			       get_timestamp(&time_string[0]),
			       mc->m_client_number);
		           fflush(log_beat_fd);
		        }
			delivered_log_heartbeat = time(0);
			break;
		     case RUN_BEAT:
		        fprintf(log_beat_fd,"\t%s Run  Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
		        fflush(log_beat_fd);
			delivered_log_heartbeat = time(0);
			break;
		     }
		   }
		   current_time = time(0);

		   if(current_time - delivered_heartbeat >= HEARTBEAT_TICK)
		   {
		     switch(mc->m_heartbeat_type) {
		     case WARM_BEAT:
		        printf("\t%s Warm Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
			delivered_heartbeat = time(0);
		        fflush(stdout);
			break;
		     case PRE_WARM_BEAT:
		        printf("\t%s Pre-Warm Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
			delivered_heartbeat = time(0);
		        fflush(stdout);
			break;
		     case INIT_BEAT:
		        if(mc->m_client_number == (num_clients-1))
		        {
		           printf("\t%s Init Heartbeat __/\\_/\\__\n", 
			       get_timestamp(&time_string[0]));
			   delivered_heartbeat = time(0);
		           fflush(stdout);
		        }
			break;
		     case CLEAN_BEAT:
		        if(mc->m_client_number == (num_clients-1))
		        {
		           printf("\t%s Cleanup Heartbeat __/\\_/\\__\n", 
			       get_timestamp(&time_string[0]));
			   delivered_heartbeat = time(0);
		           fflush(stdout);
		        }
			break;
		     case RUN_BEAT:
		        printf("\t%s Run  Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
			delivered_heartbeat = time(0);
		        fflush(stdout);
			break;
		     }
		   }
		   continue;
		}
		if( mc->m_command == R_FAILED)
		{
		  printf("Master: Failure message client %d Error = %d\n",
				mc->m_client_number,mc->m_client_eval);
		  fflush(stdout);
		  if(mc->m_client_eval==TOO_SHORT)
		  {
		     printf("Master: Runtime too short on client %d\n",
			mc->m_client_number);
		     fflush(stdout);
		     return(1);
		  }
		  if(mc->m_client_eval==WRONG_VERS)
		  {
		     printf("Master: Client %d is running wrong version of software.\n",
			mc->m_client_number);
		     fflush(stdout);
		     return(1);
		  }
		  if(mc->m_client_eval==BAD_MIX)
		  {
		     printf("Master: Client %d reporting bad mix.\n",
			mc->m_client_number);
		     fflush(stdout);
		     return(1);
		  }
		  if(mc->m_client_eval==LOGON_FAILURE)
		  {
		     printf("Master: Client %d reporting LOGON Failure.\n",
			mc->m_client_number);
		     fflush(stdout);
		     return(1);
		  }
		  if(mc->m_client_eval==LOAD_GEN_ERR)
		  {
		     printf("Master: Client %d reporting LOAD_GEN_ERR error.\n",
			mc->m_client_number);
		     fflush(stdout);
		     return(1);
		  }
		  if(mc->m_client_eval==BAD_DNS)
		  {
		     printf("Master: Client %d reporting DNS is broken.\n",
			mc->m_client_number);
		     fflush(stdout);
		     return(1);
		  }
		  if(mc->m_client_eval==NO_CREATE)
		  {
		     printf("Master: Client %d Unable to create file.\n",
			mc->m_client_number);
		     fflush(stdout);
		     if(log_flag)
		     {
		       fprintf(log_out,"Master: Client %d Unable to create file.\n",
			mc->m_client_number);
		       fflush(log_out);
		     }
			
		     return(1);
		  }
		  if(mc->m_client_eval==NO_WORKDIR)
		  {
		     printf("Master: Client %d NO_WORKDIR.\n",
			mc->m_client_number);
		     fflush(stdout);
		     return(1);
		  }
		  if(mc->m_client_eval==BAD_PIT_CONN)
		  {
		     printf("Master: Client %d BAD_PIT_CONN.\n",
			mc->m_client_number);
		     fflush(stdout);
		     return(1);
		  }
		  if(mc->m_client_eval != 0)
		  {
		     printf("Master: Client %d returned error %d Possibly: %s\n",
			mc->m_client_number,mc->m_client_eval,strerror(mc->m_client_eval));
		     fflush(stdout);
		     return(1);
		  }

		  printf("Master got unknown error type %d from client %d\n",mc->m_client_eval, mc->m_client_number);
		  fflush(stdout);
		  return(1);
		}
		if(mc->m_command == R_SENDING_RESULTS)
		{
		     printf("Master: Client %d Should NOT be sending results while master is waiting at a sync barrier.\n",
			 mc->m_client_number);
		     fflush(stdout);
		     return(1);
		}
		else
		{
			printf("Master got unknown command type %d from client %d\n",mc->m_command, mc->m_client_number);
		  	fflush(stdout);
			return(1);
		}

	}
	return(0);
}

#ifdef HAVE_ANSIC_C
int
master_wait_results(int num)
#else
int
master_wait_results(num)
int num;
#endif
{
	struct master_command *mc;
	struct result_object *res_obj;
	int completed = 0;
	int u,tmp_client_id;

	mc=&master_rcv_buf;
	
	while(completed < num)
	{
		bzero(mc,sizeof(struct master_command));
		master_listen(master_listen_socket);
		/* 
		 * Deliver percent complete updates to the screen from 
                 * the first child 
	 	 */
		if( mc->m_command == R_PERCENT_COMPLETE)
		{
		   if(mc->m_client_number == (num_clients-1)) /* Use last client id */
		   {
		     printf("\t%s Run %d percent complete\n",
			get_timestamp(&time_string[0]),mc->m_percent_complete);
		     fflush(stdout);
		     if(log_flag)
		     {
			  fprintf(log_out,"\t%s Run %d percent complete\n",
			  get_timestamp(&time_string[0]),mc->m_percent_complete);
			  fflush(log_out);
		     }
		     if((mc->m_percent_complete > POWER_CHECK_LEVEL) && (power_once++==0))
		     {
			if(power_mon(SCHECK,"RUN")==0)
			   power_mon(SRUN,"RUN");
		     }
		   }
		   continue;
		}
		if( mc->m_command == INIT_PHASE)
		{
		   if(mc->m_client_number == (num_clients-1))
		   {
		     printf("\t%s Starting INIT phase\n", 
			get_timestamp(&time_string[0]));
		     fflush(stdout);
		   }
		   continue;
		}
		if( mc->m_command == RUN_PHASE)
		{
		   if(mc->m_client_number == (num_clients-1))
		   {
		     printf("\t%s Starting RUN phase\n", 
			get_timestamp(&time_string[0]));
		     fflush(stdout);
		     if(log_flag)
		     {
			   fprintf(log_out,"\t%s Starting RUN phase\n", 
			   get_timestamp(&time_string[0]));
			   fflush(log_out);
		     }
		     /* Let the consumer defined monitor script know we're 
   			   entering RUN */
		     if(extern_mon(SCHECK,"RUNNING")==0)
	   		   extern_mon(SRUN,"RUNNING");

		     /* Let the user defined monitor script know we're 
			   entering RUN */
		     if(u_extern_mon(SCHECK,"RUNNING")==0)
	   		   u_extern_mon(SRUN,"RUNNING");

		     /* Consumer defined stat script. This is so vendor 
			   can zero counters */
		     if(extern_stat(SCHECK,"ZERO")==0)
	   		   extern_stat(SRUN,"ZERO");
		   }
		   continue;
		}
		if( mc->m_command == WARM_PHASE)
		{
		   if(mc->m_client_number == (num_clients-1))
		   {
		     printf("\t%s Starting WARM phase\n", 
			get_timestamp(&time_string[0]));
		     fflush(stdout);
	  	     if(log_flag)
		     {
			   fprintf(log_out,"\t%s Starting WARM phase\n", 
			   get_timestamp(&time_string[0]));
			   fflush(log_out);
		     }
		   }
		   continue;
		}
		if( mc->m_command == R_PERCENT_I_COMPLETE)
		{
		   if(mc->m_client_number == (num_clients-1))
		   {
		     printf("\t%s Init %d percent complete\n",
			get_timestamp(&time_string[0]),mc->m_percent_complete);
		     fflush(stdout);
		     if(log_flag)
		     {
			  fprintf(log_out,"\t%s Init %d percent complete\n",
			  get_timestamp(&time_string[0]),mc->m_percent_complete);
			  fflush(log_out);
		     }
		   }
		   continue;
		}
		if( mc->m_command == R_PERCENT_W_COMPLETE)
		{
		   if(mc->m_client_number == (num_clients-1))
		   {
		     printf("\t%s Warm-up %d percent complete\n",
			get_timestamp(&time_string[0]),mc->m_percent_complete);
		     fflush(stdout);
		     if(log_flag)
		     {
			  fprintf(log_out,"\t%s Warm-up %d percent complete\n",
			  get_timestamp(&time_string[0]),mc->m_percent_complete);
			  fflush(log_out);
		     }
		   }
		   continue;
		}
		current_time = time(0);
		if( heartbeat_flag && (mc->m_command == R_HEARTBEAT))
		{
		   if(log_beats && 
			(current_time - delivered_log_heartbeat >= HEARTBEAT_TICK))
		   {
		     switch(mc->m_heartbeat_type) {
		     case WARM_BEAT:
		        fprintf(log_beat_fd,"\t%s Warm Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
		        fflush(log_beat_fd);
			delivered_log_heartbeat = time(0);
			break;
		     case PRE_WARM_BEAT:
		        fprintf(log_beat_fd,"\t%s Pre-Warm Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
		        fflush(log_beat_fd);
			delivered_log_heartbeat = time(0);
			break;
		     case INIT_BEAT:
		        if(mc->m_client_number == (num_clients-1))
		        {
		           fprintf(log_beat_fd,"\t%s Init Heartbeat __/\\_/\\__ Client %d\n", 
			       get_timestamp(&time_string[0]),
			       mc->m_client_number);
		           fflush(log_beat_fd);
		        }
			delivered_log_heartbeat = time(0);
			break;
		     case CLEAN_BEAT:
		        if(mc->m_client_number == (num_clients-1))
		        {
		           fprintf(log_beat_fd,"\t%s Cleanup Heartbeat __/\\_/\\__ Client %d\n", 
			       get_timestamp(&time_string[0]),
			       mc->m_client_number);
		           fflush(log_beat_fd);
		        }
			delivered_log_heartbeat = time(0);
			break;
		     case RUN_BEAT:
		        fprintf(log_beat_fd,"\t%s Run  Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
		        fflush(log_beat_fd);
			delivered_log_heartbeat = time(0);
			break;
		     }
		   }
		   current_time = time(0);
		   if(current_time - delivered_heartbeat >= HEARTBEAT_TICK)
		   {
		     switch(mc->m_heartbeat_type) {
		     case WARM_BEAT:
		        printf("\t%s Warm Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
			delivered_heartbeat = time(0);
		        fflush(stdout);
			break;
		     case PRE_WARM_BEAT:
		        printf("\t%s Pre-Warm Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
			delivered_heartbeat = time(0);
		        fflush(stdout);
			break;
		     case INIT_BEAT:
		        if(mc->m_client_number == (num_clients-1))
		        {
		           printf("\t%s Init Heartbeat __/\\_/\\__\n", 
			       get_timestamp(&time_string[0]));
			   delivered_heartbeat = time(0);
		           fflush(stdout);
		        }
			break;
		     case CLEAN_BEAT:
		        if(mc->m_client_number == (num_clients-1))
		        {
		           printf("\t%s Cleanup Heartbeat __/\\_/\\__\n", 
			       get_timestamp(&time_string[0]));
			   delivered_heartbeat = time(0);
		           fflush(stdout);
		        }
			break;
		     case RUN_BEAT:
		        printf("\t%s Run  Heartbeat Client %d: %6.2f Ops/sec\n", 
			    get_timestamp(&time_string[0]),
			    mc->m_client_number,
			    mc->m_current_op_rate);
			delivered_heartbeat = time(0);
		        fflush(stdout);
			break;
		     }
		   }
		   continue;
		}
		/* 
		 *Build a linked list of results 
		 */
		if( mc->m_command == R_SENDING_RESULTS)
		{
		     
		   /*
		    * Old zombie re-awakening.. :-(
		    */
		   if(mc->m_generation != generation)
		   {
			printf("Unknown children are attempting to report results !!!\n");
		  	fflush(stdout);
			continue;
		   }

		   tmp_client_id= mc->m_client_number;
		   res_obj = &client_obj[tmp_client_id].results;

		   if(sockdebug)
		   {
		      printf("\t%s Master receiving results from Process %d\n",get_timestamp(&time_string[0]),
			tmp_client_id);
		  	fflush(stdout);
		   }

		   strcpy(res_obj->client_name, mc->m_client_name);
		   strcpy(res_obj->work_obj_name, mc->m_workload_name);
		   res_obj->work_obj_index = mc->m_work_obj_index;

		   res_obj->run_time = mc->m_run_time;
		   res_obj->min_direct_size = mc->m_min_direct_size;
		   res_obj->total_file_op_time = mc->m_total_file_op_time;
		   res_obj->total_file_ops = mc->m_total_file_ops;
		   res_obj->ops_per_second = mc->m_ops_per_second;
		   res_obj->average_latency = mc->m_average_latency;
		   res_obj->read_throughput = mc->m_read_throughput;
		   res_obj->write_throughput = mc->m_write_throughput;
		   res_obj->client_id = mc->m_client_number;
		   res_obj->modified_run = mc->m_modified_run;
		   res_obj->background = mc->m_background;
		   res_obj->sharemode = mc->m_sharemode;
		   res_obj->uniform_file_size_dist = mc->m_uniform_file_size_dist;
		   res_obj->rand_dist_behavior = mc->m_rand_dist_behavior;
		   /* Move banding data */
		   for(u=0;u<BUCKETS;u++)
		     res_obj->bands[u] = mc->m_bands[u];


		   res_obj->read_count = mc->m_read_count;
		   strcpy(res_obj->read_string, mc->m_read_string);

		   res_obj->read_file_count = mc->m_read_file_count;
		   strcpy(res_obj->read_file_string, mc->m_read_file_string);

		   res_obj->read_rand_count = mc->m_read_rand_count;
		   strcpy(res_obj->read_rand_string, mc->m_read_rand_string);

		   res_obj->write_count = mc->m_write_count;
		   strcpy(res_obj->write_string, mc->m_write_string);

		   res_obj->write_file_count = mc->m_write_file_count;
	   	   strcpy(res_obj->write_file_string, mc->m_write_file_string);

		   res_obj->open_count = mc->m_open_count;
		   res_obj->close_count = mc->m_close_count;

		   res_obj->mmap_write_count = mc->m_mmap_write_count;
	   	   strcpy(res_obj->mmap_write_string, mc->m_mmap_write_string);

		   res_obj->mmap_read_count = mc->m_mmap_read_count;
	   	   strcpy(res_obj->mmap_read_string, mc->m_mmap_read_string);

		   res_obj->write_rand_count = mc->m_write_rand_count;
	   	   strcpy(res_obj->write_rand_string, mc->m_write_rand_string);

		   res_obj->rmw_count = mc->m_rmw_count;
	   	   strcpy(res_obj->rmw_string, mc->m_rmw_string);

		   res_obj->mkdir_count = mc->m_mkdir_count;
	   	   strcpy(res_obj->mkdir_string, mc->m_mkdir_string);

		   res_obj->unlink_count = mc->m_unlink_count;
	   	   strcpy(res_obj->unlink_string, mc->m_unlink_string);

		   res_obj->create_count = mc->m_create_count;
	   	   strcpy(res_obj->create_string, mc->m_create_string);

		   res_obj->append_count = mc->m_append_count;
		   strcpy(res_obj->append_string, mc->m_append_string);

		   res_obj->lock_count = mc->m_locking_count;
	   	   strcpy(res_obj->lock_string, mc->m_locking_string);

		   res_obj->access_count = mc->m_access_count;
	   	   strcpy(res_obj->access_string, mc->m_access_string);

		   res_obj->chmod_count = mc->m_chmod_count;
	   	   strcpy(res_obj->chmod_string, mc->m_chmod_string);

		   res_obj->readdir_count = mc->m_readdir_count;
	   	   strcpy(res_obj->readdir_string, mc->m_readdir_string);

		   res_obj->stat_count = mc->m_stat_count;
	   	   strcpy(res_obj->stat_string, mc->m_stat_string);

		   res_obj->copyfile_count = mc->m_copyfile_count;
	   	   strcpy(res_obj->copyfile_string, mc->m_copyfile_string);

		   res_obj->rename_count = mc->m_rename_count;
	   	   strcpy(res_obj->rename_string, mc->m_rename_string);

		   res_obj->statfs_count = mc->m_statfs_count;
	   	   strcpy(res_obj->statfs_string, mc->m_statfs_string);

		   res_obj->pathconf_count = mc->m_pathconf_count;
	   	   strcpy(res_obj->pathconf_string, mc->m_pathconf_string);

		   res_obj->custom1_count = mc->m_custom1_count;
	   	   strcpy(res_obj->custom1_string, mc->m_custom1_string);

		   res_obj->custom2_count = mc->m_custom2_count;
	   	   strcpy(res_obj->custom2_string, mc->m_custom2_string);

		   if(Op_lat_flag)
		   {
			      res_obj->read_time = mc->m_read_time;
			      res_obj->read_file_time = mc->m_read_file_time;
			      res_obj->read_rand_time = mc->m_read_rand_time;
			      res_obj->write_time = mc->m_write_time;
			      res_obj->write_file_time = mc->m_write_file_time;
			      res_obj->mmap_write_time = mc->m_mmap_write_time;
			      res_obj->mmap_read_time = mc->m_mmap_read_time;
			      res_obj->write_rand_time = mc->m_write_rand_time;
			      res_obj->rmw_time = mc->m_rmw_time;
			      res_obj->mkdir_time = mc->m_mkdir_time;
			      res_obj->unlink_time = mc->m_unlink_time;
			      res_obj->create_time = mc->m_create_time;
			      res_obj->append_time = mc->m_append_time;
			      res_obj->lock_time = mc->m_locking_time;
			      res_obj->access_time = mc->m_access_time;
			      res_obj->chmod_time = mc->m_chmod_time;
			      res_obj->readdir_time = mc->m_readdir_time;
			      res_obj->stat_time = mc->m_stat_time;
			      res_obj->copyfile_time = mc->m_copyfile_time;
			      res_obj->rename_time = mc->m_rename_time;
			      res_obj->statfs_time = mc->m_statfs_time;
			      res_obj->pathconf_time = mc->m_pathconf_time;
			      res_obj->custom1_time = mc->m_custom1_time;
			      res_obj->custom2_time = mc->m_custom2_time;
		   }

		   if(mdebug)
		   {
				  printf("\tChild %d:\n",mc->m_client_number);
   				  printf("\tWorkload Name    %s\n",mc->m_workload_name);
   				  printf("\tRun time         %ld\n",mc->m_run_time);
   				  printf("\tMin Direct Size  %ld\n",mc->m_min_direct_size);
				  printf("\tTotal file ops   %10.0f\n",mc->m_total_file_ops);
				  printf("\tOps/sec          %10.2f\n",mc->m_ops_per_second);
				  printf("\tAvg Latency      %7.3f\n",mc->m_average_latency);
				  printf("\t%-35s  %d\n",res_obj->read_string, mc->m_read_count);
				  printf("\t%-35s  %d\n",res_obj->read_file_string, mc->m_read_file_count);
				  printf("\t%-35s  %d\n",res_obj->read_rand_string,mc->m_read_rand_count);
				  printf("\t%-35s  %d\n",res_obj->write_string,mc->m_write_count);
				  printf("\t%-35s  %d\n",res_obj->write_file_string,mc->m_write_file_count);
				  printf("\t%-35s  %d\n",res_obj->mmap_write_string,mc->m_mmap_write_count);
				  printf("\t%-35s  %d\n",res_obj->mmap_read_string,mc->m_mmap_read_count);
				  printf("\t%-35s  %d\n",res_obj->write_rand_string,mc->m_write_rand_count);
				  printf("\t%-35s  %d\n",res_obj->rmw_string,mc->m_rmw_count);
				  printf("\t%-35s  %d\n",res_obj->mkdir_string,mc->m_mkdir_count);
				  printf("\t%-35s  %d\n",res_obj->unlink_string,mc->m_unlink_count);
				  printf("\t%-35s  %d\n",res_obj->create_string,mc->m_create_count);
				  printf("\t%-35s  %d\n",res_obj->append_string,mc->m_append_count);
				  printf("\t%-35s  %d\n",res_obj->lock_string,mc->m_locking_count);
				  printf("\t%-35s  %d\n",res_obj->access_string,mc->m_access_count);
				  printf("\t%-35s  %d\n",res_obj->chmod_string,mc->m_chmod_count);
				  printf("\t%-35s  %d\n",res_obj->readdir_string,mc->m_readdir_count);
				  printf("\t%-35s  %d\n",res_obj->stat_string,mc->m_stat_count);
				  printf("\tOpen_count       %d\n",mc->m_open_count);
				  printf("\tClose_count      %d\n",mc->m_close_count);
				  fflush(stdout);
		   }
		   completed++;
		   continue;
		}
		/* 
		 * Child reporting that it is finished.
		 */
		if( mc->m_command == R_IM_DONE)
		{
			completed++;
			continue;
		}
		/* 
		 * Child reporting an error.
		 */
		if( mc->m_command == R_FAILED)
		{
		  printf("Master: Failure message client %d Error = %d\n",
				mc->m_client_number,mc->m_client_eval);
		  fflush(stdout);
		  if(mc->m_client_eval==TOO_SHORT)
		  {
		     printf("Master: Runtime too short on client %d\n",
			mc->m_client_number);
		     fflush(stdout);
		  }
		  if(mc->m_client_eval==WRONG_VERS)
		  {
		     printf("Master: Client %d is running wrong version of software.\n",
			mc->m_client_number);
		     fflush(stdout);
		  }
		  if(mc->m_client_eval==BAD_MIX)
		  {
		     printf("Master: Client %d reporting bad mix.\n",
			mc->m_client_number);
		     fflush(stdout);
		  }
		  if(mc->m_client_eval==LOGON_FAILURE)
		  {
		     printf("Master: Client %d reporting LOGON Failure.\n",
			mc->m_client_number);
		     fflush(stdout);
		  }
		  if(mc->m_client_eval==LOAD_GEN_ERR)
		  {
		     printf("Master: Client %d reporting LOAD_GEN_ERR error.\n",
			mc->m_client_number);
		     fflush(stdout);
		  }
		  if(mc->m_client_eval==BAD_DNS)
		  {
		     printf("Master: Client %d reporting DNS is broken.\n",
			mc->m_client_number);
		     fflush(stdout);
		  }
		  if(mc->m_client_eval==NO_CREATE)
		  {
		     printf("Master: Client %d Unable to create file.\n",
			mc->m_client_number);
		     fflush(stdout);
		     if(log_flag)
		     {
		       fprintf(log_out,"Master: Client %d Unable to create file.\n",
			mc->m_client_number);
		       fflush(log_out);
		     }

		  }
		  if(mc->m_client_eval==NO_WORKDIR)
		  {
		     printf("Master: Client %d NO_WORKDIR.\n",
			mc->m_client_number);
		     fflush(stdout);
		     return(1);
		  }
		  if(mc->m_client_eval==BAD_PIT_CONN)
		  {
		     printf("Master: Client %d BAD_PIT_CONN.\n",
			mc->m_client_number);
		     fflush(stdout);
		  }
		  /* 
		  * Child reporting some other error.
		  */
		  if(mc->m_client_eval != 0)
		  {
		     printf("Master: Client %d returned error %d Possibly: %s\n",
			mc->m_client_number,
			mc->m_client_eval,
			strerror(mc->m_client_eval));
		     fflush(stdout);
		  }
		  return(1);
		}
		else
		{
			printf("Master got unknown command type %d from client %d\n",mc->m_command, mc->m_client_number);
		        fflush(stdout);
			return(1);
		}
	}
	return(0);
}


/*
 * Setup the master's listening port.
 */
#ifdef HAVE_ANSIC_C
void
start_master_init(void)
#else
void
start_master_init()
#endif
{
	int ret;
	while(1)
	{
		ret = gethostname(c_master_name,MAXHNAME);
		if(ret == 0)
			break;
		perror("gethostname failure 2");
	}
	/*
	 * Start master listen 
	 */
	if(sockdebug) printf("%s Start master listen\n",get_timestamp(&time_string[0]));
	master_listen_socket=start_master_listen();
}

/*
 * Setup all of the comm to talk with all of the 
 * clients. As the children start they will send
 * a JOIN message. This message also contains their
 * port that they wish to use for inbound traffic.
 */
#ifdef HAVE_ANSIC_C
void
master_start_comm(int number_of_children)
#else
void
master_start_comm(number_of_children)
int number_of_children;
#endif
{
	struct master_command *mc;
	int chid;
	int child_port;
	char child_name[MAXHNAME];
	int i;
	mc = (struct master_command *)&master_rcv_buf;
	/*
	 * wait for child to send join message
	 */
	for(i=0;i<number_of_children;i++)
	{
		bzero(child_name,MAXHNAME);
		child_port=0;
		master_listen(master_listen_socket);
		if(mc->m_command == R_CHILD_JOIN)
		{
			strcpy(child_name,mc->m_client_name);
			chid = mc->m_client_number;
			c_chid = chid;
			child_port = mc->m_child_port;
			child_pid[c_chid] = mc->m_child_pid;
			if(sockdebug) 
			   printf("%s Master: Child %s joined. Port = %d ID = %d PID: %d\n",
				get_timestamp(&time_string[0]),child_name, child_port, chid,child_pid[c_chid]);
		}
		else
		{
			printf("Unexpected message %d from child %d\n",
			   mc->m_command,mc->m_client_number);
		        fflush(stdout);
		}
		if(sockdebug) 
		   printf("%s Start master send channel %d for %s\n",get_timestamp(&time_string[0]),
		     i,child_name);
		strcpy(&don_master_send_socket_name[map_index * MAXHNAME],child_name);
		don_master_send_socket_port[map_index]= child_port;
		map_index++;
	}
	sleep(2);
	master_send_go(number_of_children);
}

/*
 * The master is sending a GO message to all of the 
 * clients. They will all be waiting in child_wait()
 */
#ifdef HAVE_ANSIC_C
void
master_send_go(int number_of_children)
#else
void
master_send_go(number_of_children)
int number_of_children;
#endif
{
	struct client_command cc;
	int i;
	/*
	* Zero out the structure as not all fields in the structure will be initialized, 
	* client(s) may receive garbage and cause memory corruption
	*/
	bzero(&cc, sizeof(struct client_command)); 
	strcpy(cc.c_client_version,version);
	cc.c_generation=generation; /* Master tell child the generation # */
	cc.c_flush_flag = flush_flag;
	cc.c_sockdebug = sockdebug;
	cc.c_cleanup = cleanup_flag;
	cc.c_dump_files_flag = dump_files_flag;
	cc.c_tracedebug = tracedebug;
	cc.c_ipv6_enable = ipv6_enable;
	cc.c_licensed = licensed;
	cc.c_heartbeat = heartbeat_flag;
	cc.c_fd_caching_limit = fd_caching_limit;
	cc.c_op_lat_flag = Op_lat_flag;
	cc.c_gg_offset = gg_offset;
	cc.c_gg_flag = gg_flag;
	cc.c_command=R_GO;
	export_pit(&cc);
	export_dist_table(&cc);
	for(i=0;i<number_of_children;i++)
	{
		if(sockdebug) printf("%s master send go %d\n",get_timestamp(&time_string[0]),i);
		if(don_master_send_socket_port[i])
		{
		     Master_send(&don_master_send_socket_name[i * MAXHNAME],
			don_master_send_socket_port[i],
			(char *)&cc);
		}
	}
}

/*
 * This terminates the master's communication with all of the 
 * clients, and sends a message to all of the children
 * telling them that they should terminate.
 */
#ifdef HAVE_ANSIC_C
void
master_stop_comm(int number_of_children)
#else
void
master_stop_comm(number_of_children)
int number_of_children;
#endif
{
	struct client_command cc;
	int i;
	cc.c_command=R_TERMINATE;
	export_dist_table(&cc);
	for(i=0;i<number_of_children;i++)
	{
		if(sockdebug) printf("%s Master send terminate %d\n",get_timestamp(&time_string[0]),i);
		if(don_master_send_socket_port[i])
		{
		     Master_send(&don_master_send_socket_name[i * MAXHNAME],
		           don_master_send_socket_port[i],(char *)&cc);
		}
	}
	stop_master_listen(master_listen_socket);
}


/*
 * The client uses this to send a message to the master. This
 * message contains the JOIN and the port that the client
 * will use for inbound traffic.
 */
#ifdef HAVE_ANSIC_C
void
child_start_comm(char *c_master_name)
#else
void
child_start_comm(c_master_name)
char *c_master_name;
#endif
{
	struct master_command mc;
	char client_name[MAXHNAME];
	int ret;
	bzero(&mc,sizeof(struct master_command));
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child child_start_comm calling my_gethostname()\n",
			get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	while(1)
	{
		ret = my_gethostname(client_name,MAXHNAME);
		if(ret==0)
			break;
		/* Save globally */
		fprintf(newstdout,"my_gethostname failed: %s\n",strerror(errno));
		fflush(newstdout);
	}
	strncpy(client_localname,client_name,MAXHNAME);
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child child_start_comm returning from my_gethostname()\n",
			get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child calling start_child_listen()\n",
			get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
        child_listen_socket=start_child_listen();
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child returned from start_child_listen()\n",
			get_timestamp(&time_string[0]));
		fflush(newstdout);
	}

	/* Let master get into its accept call */
	sleep( (netmist_rand() % 5)+2);

	/* send join message */
	mc.m_command=R_CHILD_JOIN;
	mc.m_child_port = child_port;
	mc.m_child_pid = getpid();
	mc.m_client_number = client_id;
	strcpy(mc.m_client_name,client_name);
	if(sockdebug) 
	{
		fprintf(newstdout,"%s Child: sends join.\n",get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	Child_send(master_name, (char *)&mc);

}

/*
 * Clients use this to wait for a message from the master.
 * The message can be either GO or TERMINATE.
 */
#ifdef HAVE_ANSIC_C
void
child_wait(void)
#else
void
child_wait()
#endif
{
	struct client_command *cc;
	/*sleep(1);*/
	if(sockdebug ) 
	{
		fprintf(newstdout,"%s Child Wait.\n",get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	child_listen(child_listen_socket);
	if(sockdebug) 
	{
		fprintf(newstdout,"%s Child Wait back from listen.\n",get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	cc = (struct client_command *)&child_rcv_buf;
	if(cc->c_command == R_TERMINATE)
	{
		if(cdebug || 1) printf("%s Child got terminate message\n",
                                get_timestamp(&time_string[0]));
		stop_child_listen(child_listen_socket);
		return;
	}
	if(strcmp(cc->c_client_version,version))
	{
		c_eflag=1;
		c_eval=WRONG_VERS;
		tell_master_error(client_id,c_eval);
		R_exit(WRONG_VERS);
	}
	if(import_dist_table(cc) != 0)
	{
		c_eflag=1;
		c_eval=BAD_MIX;
		tell_master_error(client_id,c_eval);
		R_exit(BAD_MIX);
	}
	if(import_pit(cc) != 0)
	{
                c_eflag=1;
		c_eval=BAD_PIT;
		tell_master_error(client_id,c_eval);
		R_exit(BAD_PIT);
        }
	if(bad_name_option)
	{
                c_eflag=1;
		c_eval=BAD_WORKLOAD;
		tell_master_error(client_id,c_eval);
		R_exit(BAD_WORKLOAD);
        }
	if(cc->c_command == R_GO)
	{
		if(cdebug || 1) 
		{
		     fprintf(newstdout,"%s Child received GO.\n",
                                get_timestamp(&time_string[0]) );
		     fflush(newstdout);
		}
		sockdebug = cc->c_sockdebug;
		flush_flag = cc->c_flush_flag;
		cleanup_flag = cc->c_cleanup;
		dump_files_flag = cc->c_dump_files_flag;
		tracedebug = cc->c_tracedebug;
		ipv6_enable = cc->c_ipv6_enable;
		licensed = cc->c_licensed;
		heartbeat_flag = cc->c_heartbeat;
		fd_caching_limit = cc->c_fd_caching_limit;
		Op_lat_flag = cc->c_op_lat_flag;
		gg_offset = cc->c_gg_offset;
		gg_flag = cc->c_gg_flag;
		/* Child remembers the generation for later replies */
		generation = cc->c_generation;
		return;
	}
	if(cdebug || 1) 
	{
 	   	fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Child received %d command\n",cc->c_command);
		fflush(newstdout);
	}

	/*
	 * All other messages just cause the child to return
	 */
}

/*
 * Allocate the master listening port that 
 * all children will use to send messages to the master.
 */
#ifdef HAVE_ANSIC_C
int
start_master_listen(void)
#else
int
start_master_listen()
#endif
{
	int s,rc,tmp_port;
	int sockerr;
	int recv_buf_size=65536;
	int optval=1;
#if defined(_KEEPALIVE_)
	int sol_optval=1;
#endif
	struct sockaddr_in addr;
	struct sockaddr_in6 addr6;
	int mretry = 0;

hereL:
	if(ipv6_enable)
	{
          s = socket(AF_INET6, SOCK_STREAM, 0);
	}
	else
	{
          s = socket(AF_INET, SOCK_STREAM, 0);
	}
        if (s < 0)
        {
                perror("socket failed:");
		if(mretry++ < 300)
		{
			printf("Retry master socket call.\n");
		        fflush(stdout);
			sleep(1);
			goto hereL;
		}
		mretry=0;
		goto hereL;
        }
	sockerr = setsockopt (s, SOL_SOCKET, SO_RCVBUF, (char *)
		&recv_buf_size, sizeof(int));
	if ( sockerr == -1 ) {
		perror("Error in setsockopt\n");
	}
#if !defined(_vmware_)
	sockerr = setsockopt (s, SOL_SOCKET, SO_REUSEADDR, 
		(char *)&optval, sizeof(int));
	if ( sockerr == -1 ) {
		perror("Error in setsockopt\n");
	}
#endif
#if defined(_KEEPALIVE_)
	/*
	 * In case one is behind a NAT :-)
	 */
   	sol_optval = 1;
   	if(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,&sol_optval,sizeof(int)) < 0) 
	{
      		perror("setsockopt(SO_KEEPALIVE)");
   	}
#endif
here4:

#if defined(Windows) || defined(_bsd_) || defined(_solaris_)  || defined(_linux_) ||defined(WIN32) || defined(_cygwin_)
	/* Let's avoid REUSEADDR problems in the universe */
	netmist_srand((int)time(0)+rot++);
	tmp_port=HOST_LIST_PORT + (netmist_rand() % PORT_RANGE); /* Jump about */
#else
	tmp_port=HOST_LIST_PORT;
#endif
	if(ipv6_enable)
	{
           bzero(&addr6, sizeof(struct sockaddr_in6));
           addr6.sin6_port = htons(tmp_port);
           addr6.sin6_family = AF_INET6;
           memset(&addr6.sin6_addr.s6_addr[0],0,16);
	}
	else
	{
           bzero(&addr, sizeof(struct sockaddr_in));
           addr.sin_port = htons(tmp_port);
           addr.sin_family = AF_INET;
           memset(&addr.sin_addr.s_addr,0,4);
	}
        rc = -1;
        while (rc < 0)
        {
		if(ipv6_enable)
                   rc = bind(s, (struct sockaddr *)&addr6,
                                sizeof(struct sockaddr_in6));
		else
                   rc = bind(s, (struct sockaddr *)&addr,
                                sizeof(struct sockaddr));
		if(rc < 0)
		{
			tmp_port++;
			if(tmp_port > (HOST_LIST_PORT + PORT_RANGE))
			{
			   if(sockdebug)
			      printf("%s Re-start master init\n",get_timestamp(&time_string[0]));
			   sleep(1);
			   goto here4;
			}
			if(ipv6_enable)
                	   addr6.sin6_port=htons(tmp_port);
			else
                	   addr.sin_port=htons(tmp_port);
			nap(100);
			continue;
		}
		if(ipv6_enable)
		  master_listen_port = ntohs(addr6.sin6_port);
		else
		  master_listen_port = ntohs(addr.sin_port);
        }
	if(rc < 0)
	{
		perror("bind failed\n");
		exit(20);
	}
	if(cdebug )
	{
	   printf("Start master listen returns socket\n");
	}
	return(s);
}

#define SEND_HANDSHAKE
#define SEND_MSG

/*
 * Master listens for messages and blocks until
 * something arrives.
 */
struct sockaddr_in listener_sync_sock;
struct sockaddr_in6 listener_sync_sock6;
#ifdef HAVE_ANSIC_C
void
master_listen(int sock)
#else
void
master_listen(sock)
int sock;
#endif
{
	int tsize,read_size;
	int s;
	int rc;
	int rcvd;
	struct sockaddr_in *addr = NULL;
	struct sockaddr_in6 *addr6 = NULL;
	unsigned int me,me6;
	int ns,ret;
	char *where;
#if defined(SEND_HANDSHAKE)
	char ok[4] = "ok";
#endif

	if(ipv6_enable)
	   me6 = sizeof(struct sockaddr_in6);
	else
	   me = sizeof(struct sockaddr_in);
	tsize = sizeof(struct ext_master_command);
again:
	bzero(&ext_master_rcv_buf,tsize);
	s = sock;

	if(ipv6_enable)
	   addr6=&listener_sync_sock6;
	else
	   addr=&listener_sync_sock;
	if(sockdebug)
	  printf("%s Master in listening mode\n",get_timestamp(&time_string[0]));
        ret=listen(s,MAXCLIENTS*2);
	if(ret != 0)
	{
		perror("Master: listen returned error\n");
		printf("trying to recover\n");
		fflush(stdout);
		goto again;
	}
	if(sockdebug)
	  printf("%s Master in accepting connection\n",get_timestamp(&time_string[0]));
	if(ipv6_enable)
           ns=accept(s,(void *)addr6,(socklen_t *)&me6);
	else
           ns=accept(s,(void *)addr,(socklen_t *)&me);
	if(ns < 0)
	{
		perror("Master: accept returned error\n");
		printf("trying to recover\n");
		fflush(stdout);
		goto again;
	}
	if(sockdebug)
	  printf("%s Master in reading from connection\n",get_timestamp(&time_string[0]));

	where = (char *) &ext_master_rcv_buf;
	rcvd=0;
	read_size = tsize;
	while(rcvd < tsize)
	{

#ifdef SEND_MSG
	   rc = recv(ns,where, read_size,0);
	   if(rc < 0)
	   {
		perror("Master recv error \n");
		printf("trying to recover\n");
		fflush(stdout);
		goto again;
	   }
#else
	   rc = read(ns,where, read_size);
	   if(rc < 0)
	   {
		perror("Master listen error \n");
		printf("trying to recover\n");
		fflush(stdout);
		goto again;
	   }
#endif
	   rcvd += rc;
	   where+=rcvd;
	   read_size-=rc;
	}
#ifdef SEND_MSG
#if defined(SEND_HANDSHAKE)
	rc = send(ns, ok, 2,0);
	if (rc < 0)
	{
		perror("Master ok error \n");
		printf("trying to recover\n");
		fflush(stdout);
		goto again;
	}
#endif
#else
#if defined(SEND_HANDSHAKE)
	rc = write(ns, ok, 2);
	if (rc < 0)
	{
		perror("Master ok error \n");
		printf("trying to recover\n");
		fflush(stdout);
		goto again;
	}
#endif
#endif
	master_ext_to_int(&master_rcv_buf,&ext_master_rcv_buf);
#if defined(_SHUTDOWN_)
#if !defined(WIN32)
	fsync(ns);
	shutdown(ns,SHUT_RDWR);
#else
	shutdown(ns,SD_BOTH);
#endif

#endif
	close(ns);
}

/*
 * Child sends message to master.
 */

/*
 * Client setting up the channel for sending messages to the master.
 */
#ifdef HAVE_ANSIC_C
int
Child_send(char *controlling_host_name, char *send_buffer)
#else
int
Child_send(controlling_host_name, send_buffer)
char *controlling_host_name;
char *send_buffer;
#endif
{
	int rc,xx;
	/* struct hostent *he; */
	int tmp_port;
	/*struct in_addr *ip;*/
	struct sockaddr_in cs_addr,cs_raddr;
	struct sockaddr_in6 cs_addr6,cs_raddr6;
	struct ext_master_command emc;
#if defined(_KEEPALIVE_)
	int sol_optval = 1;
#endif 

#ifdef REUSE_ADDR
	int optval = 1;
	int sockerr;
#endif
	char name_buf[MAXHNAME];
	unsigned int addr_buf[16];
#if defined(SEND_HANDSHAKE)
	char ok[4];
#endif

	bzero(name_buf,MAXHNAME);
	if(sockdebug)
	{
           fprintf(newstdout,"%s Start_child_send: %s \n",get_timestamp(&time_string[0]),
		controlling_host_name);
   	   fflush(newstdout);
	}
/*
        he = gethostbyname(controlling_host_name);
        if (he == NULL)
        {
                exit(22);
        }
        ip = (struct in_addr *)he->h_addr_list[0];
*/
	
        xx=get_host_address(controlling_host_name, addr_buf, name_buf);
	if(xx != 0)
	{
		c_eflag = 1;
		c_eval = BAD_DNS;
		R_exit(BAD_DNS);
	}

overL:
	if(ipv6_enable)
	{
           cs_raddr6.sin6_family = AF_INET6;
           cs_raddr6.sin6_port = htons(master_p);
	   /* The next line needs work */
           memcpy(&cs_raddr6.sin6_addr.s6_addr[0],addr_buf,16);
           child_socket_val = socket(AF_INET6, SOCK_STREAM, 0);
	}
	else
	{
           cs_raddr.sin_family = AF_INET;
           cs_raddr.sin_port = htons(master_p);
           memcpy(&cs_raddr.sin_addr.s_addr,addr_buf,4);
           /*cs_raddr.sin_addr.s_addr = ip->s_addr;*/
           child_socket_val = socket(AF_INET, SOCK_STREAM, 0);
	}
        if (child_socket_val < 0)
        {
		fprintf(newstdout,"Child: Socket failed: %s\n",strerror(errno));
		fflush(newstdout);
                exit(23);
        }
#ifdef REUSE_ADDR
#if !defined(_vmware_)
	sockerr = setsockopt (child_socket_val, SOL_SOCKET, SO_REUSEADDR, 
		(char *)&optval, sizeof(int));
	if ( sockerr == -1 ) {
		fprintf(newstdout,"Child: Error in setsockopt: %s\n",
			strerror(errno));
		fflush(newstdout);
	}
#endif
#endif
#if defined(_KEEPALIVE_)
	/*
	 * In case one is behind a NAT :-)
	 */
   	sol_optval = 1;
   	if(setsockopt(child_socket_val,SOL_SOCKET,SO_KEEPALIVE,
		&sol_optval,sizeof(int)) < 0) 
	{
		fprintf(newstdout,"Child: Error in setsockopt(KEEPALIVE): %s\n",
			strerror(errno));
		fflush(newstdout);
   	}
#endif

	if(ipv6_enable)
           bzero(&cs_addr6, sizeof(struct sockaddr_in6));
	else
           bzero(&cs_addr, sizeof(struct sockaddr_in));

here5:

#if defined(Windows) || defined(_bsd_) || defined(_solaris_) || defined(_linux_)||defined(WIN32) || defined(_cygwin_)
	/*
	 * Holy cow Batman, WTH is this rand() for?  Robin, this is the 
         * workaround for those with very poor implementations of SO_REUSEADDR. 
	 * It turns out that SO_REUSEADDR doesn't work correctly on some 
	 * systems. Instead, it returns errors Address in Use
	 * when in reality, one has specified SO_REUSEADDR. 
         * Someone thought that one should ignore the user's setsockopt()
         * if the silly port in the bind() is the same as one that 
         * was in use, or was in use and is in timewait. The bind() succeeded,
	 * but then later calls to connect() fail. :-( This is just
         * GOOFY!. 
	 * So.. we are forced to once again play hide and seek with
         * some poor & broken implementations! 
	 * Fortunately, obscurity seems to defeat the goofyness, thus 
         * the rand().
	 */
	netmist_srand((int)time(0)+rot++); /* yeah, stir the pot, based on time */
	tmp_port= CHILD_ESEND_PORT + (netmist_rand() % PORT_RANGE);
#else
	tmp_port= CHILD_ESEND_PORT;
#endif
	if(sockdebug)
	{
	   fprintf(newstdout,"%s Child port %d\n",get_timestamp(&time_string[0]),
		tmp_port);
	   fflush(newstdout);
	}
	if(ipv6_enable)
	{
           cs_addr6.sin6_port = htons(tmp_port);
           cs_addr6.sin6_family = AF_INET6;
	   /* The next line needs work */
           memset(&cs_addr6.sin6_addr.s6_addr[0],0,16);
	}
	else
	{
           cs_addr.sin_port = htons(tmp_port);
           cs_addr.sin_family = AF_INET;
           memset(&cs_addr.sin_addr.s_addr,0,4);
	}
        rc = -1;
        while (rc < 0)
        {
		if(ipv6_enable)
                   rc = bind(child_socket_val, (struct sockaddr *)&cs_addr6,
                                                sizeof(struct sockaddr_in6));
		else
                   rc = bind(child_socket_val, (struct sockaddr *)&cs_addr,
                                                sizeof(struct sockaddr_in));
		if(rc < 0)
		{
                	tmp_port++;
			if(tmp_port> (CHILD_ESEND_PORT + PORT_RANGE))
			{
			   if(sockdebug)
			   {
			      fprintf(newstdout,"%s Child socket restart\n",get_timestamp(&time_string[0]));
			      fflush(newstdout);
			   }
			   sleep(1);
			   goto here5;
			}
			if(ipv6_enable)
                	   cs_addr6.sin6_port=htons(tmp_port);
			else
                	   cs_addr.sin_port=htons(tmp_port);
			nap(100);
			continue;
		}
        }
        if (rc < 0)
        {
	   fprintf(newstdout,"%s Child: bind failed: %s\n", get_timestamp(&time_string[0]), strerror(errno));
	   fflush(newstdout);
           exit(24);
        }
	if(sockdebug)
	{
	   fprintf(newstdout,"%s Child sender bound to port %d Master port %d \n",
		get_timestamp(&time_string[0]), tmp_port,master_p);
	   fflush(newstdout);
	}

	nap(10);

/*
again:
*/
	if(ipv6_enable)
           rc = connect(child_socket_val, (struct sockaddr *)&cs_raddr6, 
			sizeof(struct sockaddr_in6));
	else
           rc = connect(child_socket_val, (struct sockaddr *)&cs_raddr, 
			sizeof(struct sockaddr_in));
        if (rc < 0)
        {
                if((ecount++ < 200) && (errno != EISCONN))
                {
			nap(10);
	  		if(sockdebug)
			{
	  		   fprintf(newstdout,"%s Child connect error %s \n",
				get_timestamp(&time_string[0]),strerror(errno));
			   fflush(newstdout);
			}
                        if(errno == EPERM)
			{
	  		   fprintf(newstdout,"Child connect permission error %s \n",
				strerror(errno));
	  		   fprintf(newstdout,"Possibly local firewall issues ? \n");
			   fflush(newstdout);
			}
                        if(errno == ECONNREFUSED)
			{
	  		   fprintf(newstdout,"Child connect refused. Sleep(x).. try again.\n");
			   fflush(newstdout);
			   /* Master not in the listen yet ? */
			   sleep( (netmist_rand() % 5)+1);
			}
			retry1_counter++;
#if defined(_SHUTDOWN_)
#if !defined(WIN32)
			fsync(child_socket_val);
			shutdown(child_socket_val,SHUT_RDWR);
#else
			shutdown(child_socket_val,SD_BOTH);
#endif
#endif
			close(child_socket_val);
			nap(100);
			goto overL;
                }
		if(sockdebug)
		{
                   fprintf(newstdout,"%s Child: Connect failed %s\n",
		     get_timestamp(&time_string[0]),strerror(errno));
		   fflush(newstdout);
		}
#if defined(_SHUTDOWN_)
#if !defined(WIN32)
		fsync(child_socket_val);
		shutdown(child_socket_val,SHUT_RDWR);
#else
		shutdown(child_socket_val,SD_BOTH);
#endif
#endif
		close(child_socket_val);
		ecount=0;
		retry2_counter++;
		nap(100);
		goto overL;
        }
	ecount=0;
	if(sockdebug)
	{
	   fprintf(newstdout,"%s Child connected\n",get_timestamp(&time_string[0]));
	   fflush(newstdout);
	}

	/* Now send */

	if(cdebug)
	{
	   fprintf(newstdout,"Child_send\n");
	   fflush(newstdout);
	}
	bzero(&emc,sizeof(struct ext_master_command));
	master_int_to_ext((struct master_command *)send_buffer, &emc);
#ifdef SEND_MSG
	rc=send(child_socket_val,&emc,sizeof(struct ext_master_command),0);
        if (rc < 0)
        {
		fprintf(newstdout,"Child: Send failed %s\n",strerror(errno));
		fflush(newstdout);
                exit(26);
        }
#if defined(SEND_HANDSHAKE)
	rc = recv(child_socket_val,ok, 2,0);
	/* could check here that ok == "ok" */
	if (rc < 0 || (rc != 2))
	{
		fprintf(newstdout,"Child: ok failed %s\n",strerror(errno));
		fflush(newstdout);
                exit(26);
	}
#endif
#else

	rc=write(child_socket_val,&emc,sizeof(struct ext_master_command));
        if (rc < 0)
        {
		fprintf(newstdout,"Child: Write failed %s\n",strerror(errno));
		fflush(newstdout);
                exit(26);
        }
#if defined(SEND_HANDSHAKE)
	rc = read(child_socket_val,ok, 2);
	/* could check here that ok == "ok" */
	if (rc < 0 || (rc != 2))
	{
		fprintf(newstdout,"Child: ok failed %s\n",strerror(errno));
		fflush(newstdout);
                exit(26);
	}
#endif
#endif
	nap(10);

#if defined(_SHUTDOWN_)
#if !defined(WIN32)
	fsync(child_socket_val);
	shutdown(child_socket_val,SHUT_RDWR);
#else
	shutdown(child_socket_val,SD_BOTH);
#endif
#endif
	close(child_socket_val);
	if(sockdebug)
	{
	   fprintf(newstdout,"%s Child_send completed\n",get_timestamp(&time_string[0]));
	   fflush(newstdout);
	}

	return (0);
}

/*
 * Close the childs listening port for messages from the master.
 */
#ifdef HAVE_ANSIC_C
void
stop_child_listen(int child_socket_val)
#else
void
stop_child_listen(child_socket_val)
int child_socket_val;
#endif
{
#if defined(_SHUTDOWN_)
/* Really need to make sure that the child's initial receive socket
   gets closed, and not just with close() but fully closed so that
   it does not stall in a TIME_WAIT state.  If it hangs out there, 
   then the next invocation of the tool will get VERY confused.
*/
#if !defined(WIN32)
	fsync(child_socket_val);
	shutdown(child_socket_val,SHUT_RDWR);
#else
	shutdown(child_socket_val,SD_BOTH);
#endif
#endif

#if defined(WIN32)
	closesocket(child_socket_val);
#else
	close(child_socket_val);
#endif

	fprintf(newstdout,"%s Child: Closing child's listening socket\n",
		get_timestamp(&time_string[0]));
	fflush(newstdout);
#ifdef FOO
	sleep(60); /* Wait for listening sockets in WAIT to shutdown */
		   /* This is a guessing game, an donly interesting on */
		   /* boxes where the shutdown() above did not work. */
#endif
	fprintf(newstdout,"%s Child: Listen socket shutdown complete\n", 
		get_timestamp(&time_string[0]));
	fflush(newstdout);
}


/*
 * Close the masters listening channel for all clients messages.
 */
#ifdef HAVE_ANSIC_C
void
stop_master_listen(int master_socket_val)
#else
void
stop_master_listen(master_socket_val)
int master_socket_val;
#endif
{
#if defined(_SHUTDOWN_)
/* Really need to make sure that the child's initial receive socket
   gets closed, and not just with close() but fully closed so that
   it does not stall in a TIME_WAIT state.  If it hangs out there, 
   then the next invocation of the tool will get VERY confused.
*/
#if !defined(WIN32)
	fsync(master_socket_val);
	shutdown(master_socket_val,SHUT_RDWR);
#else
	shutdown(master_socket_val,SD_BOTH);
#endif
#endif

#if defined(WIN32)
	closesocket(master_socket_val);
#else
	close(master_socket_val);
#endif

	printf("\tShutting down clients, and communications layer...\n");
	fflush(stdout);
#ifdef FOO
	sleep(120); /* Give socket layer more time to exit TIME_WAIT
                      or CLOSE_WAIT states */
#endif

}


/*
 * Start the childs listening service for messages from the master.
 */

#ifdef HAVE_ANSIC_C
int
start_child_listen(void)
#else
int
start_child_listen()
#endif
{
	int s,rc,tmp_port;
	struct sockaddr_in addr;
	struct sockaddr_in6 addr6;
#if defined(_KEEPALIVE_)
	int sol_optval = 1;
#endif
#ifdef REUSE_ADDR
	int sockerr;
	int optval = 1;
#endif
	int retry = 0;

	if(sockdebug)
	{
		fprintf(newstdout,"%s Child enters start_child_listen()\n",get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
hereL:
	if(ipv6_enable)
           s = socket(AF_INET6, SOCK_STREAM, 0);
	else
           s = socket(AF_INET, SOCK_STREAM, 0);
        if (s < 0)
        {
		if(retry++ >= 300)
		{
                   fprintf(newstdout,"Child socket create failed: Error %s",strerror(errno));
		   fflush(newstdout);
		   retry=0;
		   sleep(5); /* and swing chicken about your head, mumbling voo-doo-ish sounds */
		   goto hereL;
		}
		else
		{
            fprintf(newstdout,"Retry #%d of Child socket create. Error = %s",retry,
			strerror(errno));
		   fflush(newstdout);
		   sleep(5); /* and swing chicken about your head, mumbling voo-doo-ish sounds */
		   goto hereL;
		}
        }
	retry=0;
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child socket allocated\n",get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
#ifdef REUSE_ADDR
#if !defined(_vmware_)
	sockerr = setsockopt (s, SOL_SOCKET, SO_REUSEADDR, 
		(char *)&optval, sizeof(int));
	if ( sockerr == -1 ) {
		fprintf(newstdout,"Error in setsockopt Error %s\n",strerror(errno));
	}
#endif
#endif
#if defined(_KEEPALIVE_)
	/*
	 * In case one is behind a NAT :-)
	 */
   	sol_optval = 1;
   	if(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE, &sol_optval,sizeof(int)) < 0) 
	{
		fprintf(newstdout,"Child: Error in setsockopt(KEEPALIVE): %s\n",
			strerror(errno));
		fflush(newstdout);
   	}
#endif
	tmp_port= CHILD_LIST_PORT+client_id;
	if(ipv6_enable)
	{
           bzero(&addr6, sizeof(struct sockaddr_in6));
           addr6.sin6_port = htons(tmp_port);
           addr6.sin6_family = AF_INET6;
	   /* The next line needs work */
           memset(&addr6.sin6_addr.s6_addr[0],0,16);
	}
	else
	{
           bzero(&addr, sizeof(struct sockaddr_in));
           addr.sin_port = htons(tmp_port);
           addr.sin_family = AF_INET;
           memset(&addr.sin_addr.s_addr,0,4);
	}
        rc = -1;
        while (rc < 0)
        {
		if(ipv6_enable)
                   rc = bind(s, (struct sockaddr *)&addr6,
                                sizeof(struct sockaddr_in6));
		else
                   rc = bind(s, (struct sockaddr *)&addr,
                                sizeof(struct sockaddr));
		if(rc < 0)
		{
                	tmp_port++;
			if(tmp_port > CHILD_LIST_PORT+MAXCLIENTS)
			{
			   if(sockdebug)
			   {
			     fprintf(newstdout,"%s Child: Out of ports, retrying\n",get_timestamp(&time_string[0]));
			     fflush(newstdout);
			   }
			   sleep(1);
			   tmp_port = CHILD_LIST_PORT+client_id;
			}
			if(ipv6_enable)
                	  addr6.sin6_port=htons(tmp_port);
			else
                	  addr.sin_port=htons(tmp_port);
			nap(100);
			if(sockdebug)
			{
				fprintf(newstdout,"%s Child trying next port %d\n",get_timestamp(&time_string[0]),tmp_port);
				fflush(newstdout);
			}
			continue;
		}
        }
	if(ipv6_enable)
	   child_port = ntohs(addr6.sin6_port);
	else
	   child_port = ntohs(addr.sin_port);
	if(rc < 0)
	{
		fprintf(newstdout,"bind failed. Error %s\n",strerror(errno));
		fflush(newstdout);
		exit(20);
	}
	if(sockdebug)
	{
		fprintf(newstdout,"%s Start child listen.. return success\n",
			get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	return(s);
}

/*
 * The clients use this to block waiting for a message from
 * the master.
 */
struct sockaddr_in child_listener_sock;
struct sockaddr_in6 child_listener_sock6;

#ifdef HAVE_ANSIC_C
void
child_listen(int sock)
#else
void
child_listen(sock)
int sock;
#endif
{
	int tsize;
	int s;
	int rc;
	unsigned int me,me6;
	int listen_count = 0;
	int ns,ret;
	char *ptr;
	struct sockaddr_in *addr = NULL;
	struct sockaddr_in6 *addr6 = NULL;
	s = sock;

	if(ipv6_enable)
	{
	  me6 = sizeof(struct sockaddr_in6);
	  addr6=&child_listener_sock6;
	}
	else
	{
	  me = sizeof(struct sockaddr_in);
	  addr=&child_listener_sock;
	}
	listen_count = 0;
lagain:
        ret=listen(s,MAXCLIENTS*2);
	if(ret != 0)
	{
		fprintf(newstdout,"Child listen failed: %s (retry)\n",strerror(errno));
		fflush(newstdout);
		sleep(2); /* Hope that the impossible "Socket already has a listener" */
			  /* goes away. This should "never" happen, but yet it does! */
		if(listen_count++ < 30) /* Give up after 1 minute of retries */
			goto lagain;
		else
		{
		  fprintf(newstdout,"Child listen failed: %s ( You're dead! )\n",strerror(errno));
		  fprintf(newstdout,"Some other child process is holding a needed TCP port busy!!\n");
		  fprintf(newstdout,"You may need to manually terminate all children and start over.\n");
		  fflush(newstdout);
		  exit(22);
		}
		listen_count = 0;
	}
	if(ipv6_enable)
          ns=accept(s,(void *)addr6,(socklen_t *)&me6);
	else
          ns=accept(s,(void *)addr,(socklen_t *)&me);
	if(ns < 0)
		perror("Child: accept failed\n");

	tsize=sizeof(struct ext_client_command);

	/* zero things out */
	bzero(&ext_child_rcv_buf,sizeof(struct ext_client_command));
	bzero(&child_rcv_buf, sizeof(struct client_command));
	/* zero things out */

	ptr = (char *)&ext_child_rcv_buf;
	while(tsize !=0 )
	{
	   rc=read(ns,ptr,tsize);
	   if(rc < 0 )
	   {
		if(cdebug || 1)
		{
		    fprintf(newstdout,"Child Read failed rc %d errno: %s\n",rc,strerror(errno));
		    fflush(newstdout);
		}
		exit(21);
	   }
	   if(cdebug)
	   {
	        fprintf(newstdout,"Child Read %d bytes \n",rc);
		fflush(newstdout);
	   }
	   ptr+=rc;
	   tsize-=rc;
	}
	client_ext_to_int(&child_rcv_buf,&ext_child_rcv_buf);
	if(cdebug )
	{
	   fprintf(newstdout,"Child received message\n");
	   fflush(newstdout);
	}
#if defined(_SHUTDOWN_)
#if !defined(WIN32)
	fsync(ns);
	shutdown(ns,SHUT_RDWR);
#else
	shutdown(ns,SD_BOTH);
#endif
#endif
	close(ns);
}

/*
 * Start the channel for the master to send a message to 
 * a particular child on a particular port that the child
 * has created for the parent to use to communicate.
 */
#ifdef HAVE_ANSIC_C
int
Master_send(char *child_host_name, int child_port, char *send_buffer)
#else
int
Master_send(child_host_name, child_port, send_buffer)
char *send_buffer;
char *child_host_name; 
int child_port;
#endif
{
	int rc,master_socket_val;
	/*struct hostent *he; */
	int tmp_port;
	/*struct in_addr *ip; */
	struct sockaddr_in ms_addr,ms_raddr;
	struct sockaddr_in6 ms_addr6,ms_raddr6;
	struct ext_client_command ecc;
#if defined(_KEEPALIVE_)
	int sol_optval = 1;
#endif

#ifdef REUSE_ADDR
	int optval = 1;
	int sockerr;
#endif
	int xx;
	unsigned int addr_buf[16];
	char name_buf[MAXHNAME];

/*
        he = gethostbyname(child_host_name);
        if (he == NULL)
        {
                if(xdebug) printf("Master: Bad hostname %s\n",child_host_name);
		if(xdebug) fflush(stderr);
                exit(22);
        }
        ip = (struct in_addr *)he->h_addr_list[0];
*/

	bzero(name_buf,MAXHNAME);
        xx=get_host_address(child_host_name, addr_buf, name_buf);
	if(xx != 0)
	{
		printf("DNS problem. Unable to determine the IP address for the child: %s\n",child_host_name);
		fflush(stdout);
		exit(BAD_DNS);
	}
overL:
	tmp_port=child_port;
	if(ipv6_enable)
	{
          ms_raddr6.sin6_family = AF_INET6;
          ms_raddr6.sin6_port = htons(tmp_port);
	  /* The next line needs work */
          memcpy(&ms_raddr6.sin6_addr.s6_addr[0],addr_buf,16);
          /*ms_raddr6.sin6_addr.s6_addr[0] = ip->s_addr;*/
          master_socket_val = socket(AF_INET6, SOCK_STREAM, 0);
	}
	else
	{
          ms_raddr.sin_family = AF_INET;
          ms_raddr.sin_port = htons(tmp_port);
          memcpy(&ms_raddr.sin_addr.s_addr,addr_buf,4);
          /*ms_raddr.sin_addr.s_addr = ip->s_addr; */
          master_socket_val = socket(AF_INET, SOCK_STREAM, 0);
	}
        if (master_socket_val < 0)
        {
                perror("Master: socket failed:");
                exit(23);
        }
#ifdef REUSE_ADDR
#if !defined(_vmware_)
	sockerr = setsockopt (master_socket_val, SOL_SOCKET, SO_REUSEADDR, 
		(char *)&optval, sizeof(int));
	if ( sockerr == -1 ) {
		perror("Error in setsockopt\n");
	}
#endif
#endif
#if defined(_KEEPALIVE_)
	/*
	 * In case one is behind a NAT :-)
	 */
   	sol_optval = 1;
   	if(setsockopt(master_socket_val,SOL_SOCKET,SO_KEEPALIVE,
		&sol_optval,sizeof(int)) < 0) 
	{
		printf("Master: Error in setsockopt(KEEPALIVE): %s\n", 
			strerror(errno));
   	}
#endif
	if(ipv6_enable)
           bzero(&ms_addr6, sizeof(struct sockaddr_in6));
	else
           bzero(&ms_addr, sizeof(struct sockaddr_in));

here6:

#if defined(Windows) || defined(_bsd_) || defined(_solaris_) || defined(_linux_)||defined(WIN32) || defined(_cygwin_)
	netmist_srand((int)time(0)+rot++); /* yeah, that's ugly */
	tmp_port= HOST_ESEND_PORT + (netmist_rand() % PORT_RANGE);
#else
	tmp_port=HOST_ESEND_PORT;
#endif
	if(ipv6_enable)
	{
          ms_addr6.sin6_port = htons(tmp_port);
          ms_addr6.sin6_family = AF_INET6;
	  /* The next line needs work */
          memset(&ms_addr6.sin6_addr.s6_addr[0],0,16);
	}
	else
	{
          ms_addr.sin_port = htons(tmp_port);
          ms_addr.sin_family = AF_INET;
          memset(&ms_addr.sin_addr.s_addr,0,4);
	}
        rc = -1;
        while (rc < 0)
        {
	 	if(ipv6_enable)
                  rc = bind(master_socket_val, (struct sockaddr *)&ms_addr6,
                                                sizeof(struct sockaddr_in6));
		else
                  rc = bind(master_socket_val, (struct sockaddr *)&ms_addr,
                                                sizeof(struct sockaddr_in));
		if(rc < 0)
		{
			tmp_port++;
			if(tmp_port>(HOST_ESEND_PORT + PORT_RANGE))
			{
			   if(sockdebug)
			     printf("%s Master send port restart\n",get_timestamp(&time_string[0]));
			   sleep(1);
			   goto here6;
			}
	 		if(ipv6_enable)
                	  ms_addr6.sin6_port=htons(tmp_port);
			else
                	  ms_addr.sin_port=htons(tmp_port);
			nap(100);
			continue;
		}
        }
        if (rc < 0)
        {
                perror("Master: bind failed\n");
                exit(24);
        }
	if(sockdebug)
		printf("%s Master used port %d to bind to %d\n",
			get_timestamp(&time_string[0]),tmp_port,child_port);
#if defined(Windows) || defined(_bsd_) || defined(_solaris_)||defined(WIN32) || defined(_cygwin_)
#else
again:
#endif
	nap(10);
	if(ipv6_enable)
           rc = connect(master_socket_val, (struct sockaddr *)&ms_raddr6, 
			sizeof(struct sockaddr_in6));
	else
           rc = connect(master_socket_val, (struct sockaddr *)&ms_raddr, 
			sizeof(struct sockaddr_in));
        if (rc < 0)
        {
                if((ecount++ < 200) && (errno != EISCONN))
                {
			nap(10);
                        if(errno == EPERM)
			{
	  		   printf("Master_send connect permission error %s \n",
				strerror(errno));
	  		   printf("Possibly local firewall issues ? \n");
			   fflush(stdout);
			}
                        if(errno == ECONNREFUSED)
			{
	  		   printf("Master connect refused. Sleep(x).. try again.\n");
			   fflush(stdout);
			   sleep( (netmist_rand() % 5)+1);
			}
#if defined(Windows) || defined(_bsd_) || defined(_solaris_)||defined(WIN32)  || defined(_cygwin_)
			close(master_socket_val);
	  		printf("Master_send connect error %s. Retry operation.\n",strerror(errno));
			fflush(stdout);
			nap(100);
			goto overL;
#else
	  		printf("Master_send connect error %s. Retry operation. \n",strerror(errno));
			fflush(stdout);
			nap(100);
                        goto again;
#endif
                }
		if(sockdebug)
		{
                   perror("Master: connect failed...\n");
                   printf("Error: %s\n",strerror(errno));
                   printf("%s trying again\n",get_timestamp(&time_string[0]));
		   fflush(stdout);
		}
		close(master_socket_val);
		nap(100);
		ecount=0;
		goto overL;
        }
	ecount=0;
	if(sockdebug)
	{
	   printf("%s Master connected\n",get_timestamp(&time_string[0]));
	   fflush(stdout);
	}

	/* Now send it */
	bzero(&ecc,sizeof(struct ext_client_command));
	client_int_to_ext((struct client_command *)send_buffer, &ecc);
        rc = write(master_socket_val, &ecc, sizeof(struct ext_client_command));
        if (rc < 0)
        {
                perror("write failed\n");
                exit(26);
        }
	if(sockdebug)
	{
	  printf("%s Master write complete\n",get_timestamp(&time_string[0]));
	}
	nap(10);
#if defined(_SHUTDOWN_)
#if !defined(WIN32)
	fsync(master_socket_val);
	shutdown(master_socket_val,SHUT_RDWR);
#else
	shutdown(master_socket_val,SD_BOTH);
#endif

#endif
	close(master_socket_val);
	return (0);
}



/*
 * Clients tell the master that I am this percent complete 
 */
#ifdef HAVE_ANSIC_C
void
tell_master_pct_complete(int percent, int chid, int type)
#else
void
tell_master_pct_complete(percent,chid,type)
int percent,chid,type;
#endif
{
	struct master_command mc;
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child: Tell master percent complete\n",
		   get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	bzero(&mc,sizeof(struct master_command));
	mc.m_generation = generation;
	mc.m_command = type;
	mc.m_client_number = (int)chid; 
	mc.m_percent_complete = percent;
	Child_send(master_name, (char *)&mc);
}

/*
 * Clients tell the master that I am in this phase
 */
#ifdef HAVE_ANSIC_C
void
tell_master_phase(int chid, int phase)
#else
void
tell_master_phase(chid,phase)
int chid,phase;
#endif
{
	struct master_command mc;
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child: Tell master phase\n",
		   get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	bzero(&mc,sizeof(struct master_command));
	mc.m_generation = generation;
	mc.m_command = phase;
	mc.m_client_number = (int)chid; 
	Child_send(master_name, (char *)&mc);
}
/*
 * Clients tell the master that I am alive..heartbeat
 */
#ifdef HAVE_ANSIC_C
void
tell_master_heartbeat(int chid,int type, double rate)
#else
void
tell_master_heartbeat(chid,type, rate)
int chid;
int type;
double rate;
#endif
{
	struct master_command mc;
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child %d: Tell master heartbeat\n",
		   get_timestamp(&time_string[0]),chid);
		fflush(newstdout);
	}
	bzero(&mc,sizeof(struct master_command));
	mc.m_generation = generation;
	mc.m_command = R_HEARTBEAT;
	mc.m_current_op_rate = rate; 
	mc.m_client_number = (int)chid; 
	mc.m_heartbeat_type = (int)type; 
	Child_send(master_name, (char *)&mc);
}

/*
 * Clients tell the master that I am this percent complete 
 */
#ifdef HAVE_ANSIC_C
void
tell_master_results(int chid)
#else
void
tell_master_results(chid)
int chid;
#endif
{
	struct master_command mc;
	int u;
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child: Tell master results\n",
		   get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	bzero(&mc,sizeof(struct master_command));


	mc.m_generation = generation;
	mc.m_command = R_SENDING_RESULTS;
	mc.m_client_number = (int)chid; 
	mc.m_work_obj_index = (int)my_workload; 
	strcpy(mc.m_client_name, client_localname);
	strcpy(mc.m_workload_name, work_obj[my_workload].work_obj_name);

	mc.m_read_count = work_obj[my_workload].read_op_count;
	strcpy(mc.m_read_string, my_mix_table[my_workload].read_string);

	mc.m_read_file_count = work_obj[my_workload].read_file_op_count;
	strcpy(mc.m_read_file_string, my_mix_table[my_workload].read_file_string);

	mc.m_read_rand_count = work_obj[my_workload].read_rand_op_count;
	strcpy(mc.m_read_rand_string, my_mix_table[my_workload].read_rand_string);

	mc.m_write_count = work_obj[my_workload].write_op_count;
	strcpy(mc.m_write_string, my_mix_table[my_workload].write_string);

	mc.m_write_file_count = work_obj[my_workload].write_file_op_count;
	strcpy(mc.m_write_file_string, my_mix_table[my_workload].write_file_string);

	mc.m_open_count = work_obj[my_workload].open_op_count;
	mc.m_close_count = work_obj[my_workload].close_op_count;

	mc.m_mmap_write_count = work_obj[my_workload].mmap_write_op_count;
	strcpy(mc.m_mmap_write_string, my_mix_table[my_workload].mmap_write_string);

	mc.m_mmap_read_count = work_obj[my_workload].mmap_read_op_count;
	strcpy(mc.m_mmap_read_string, my_mix_table[my_workload].mmap_read_string);

	mc.m_write_rand_count = work_obj[my_workload].write_rand_op_count;
	strcpy(mc.m_write_rand_string, my_mix_table[my_workload].write_rand_string);

	mc.m_rmw_count = work_obj[my_workload].rmw_op_count;
	strcpy(mc.m_rmw_string, my_mix_table[my_workload].rmw_string);

	mc.m_mkdir_count = work_obj[my_workload].mkdir_op_count;
	strcpy(mc.m_mkdir_string, my_mix_table[my_workload].mkdir_string);

	mc.m_unlink_count = work_obj[my_workload].unlink_op_count;
	strcpy(mc.m_unlink_string, my_mix_table[my_workload].unlink_string);

	mc.m_create_count = work_obj[my_workload].create_op_count;
	strcpy(mc.m_create_string, my_mix_table[my_workload].create_string);

	mc.m_append_count = work_obj[my_workload].append_op_count;
	strcpy(mc.m_append_string, my_mix_table[my_workload].append_string);

	mc.m_locking_count = work_obj[my_workload].lock_op_count;
	strcpy(mc.m_locking_string, my_mix_table[my_workload].locking_string);

	mc.m_access_count = work_obj[my_workload].access_op_count;
	strcpy(mc.m_access_string, my_mix_table[my_workload].access_string);

	mc.m_chmod_count = work_obj[my_workload].chmod_op_count;
	strcpy(mc.m_chmod_string, my_mix_table[my_workload].chmod_string);

	mc.m_readdir_count = work_obj[my_workload].readdir_op_count;
	strcpy(mc.m_readdir_string, my_mix_table[my_workload].readdir_string);

	mc.m_stat_count = work_obj[my_workload].stat_op_count;
	strcpy(mc.m_stat_string, my_mix_table[my_workload].stat_string);

	mc.m_copyfile_count = work_obj[my_workload].copyfile_op_count;
	strcpy(mc.m_copyfile_string, my_mix_table[my_workload].copyfile_string);

	mc.m_rename_count = work_obj[my_workload].rename_op_count;
	strcpy(mc.m_rename_string, my_mix_table[my_workload].rename_string);

	mc.m_statfs_count = work_obj[my_workload].statfs_op_count;
	strcpy(mc.m_statfs_string, my_mix_table[my_workload].statfs_string);

	mc.m_pathconf_count = work_obj[my_workload].pathconf_op_count;
	strcpy(mc.m_pathconf_string, my_mix_table[my_workload].pathconf_string);

	mc.m_custom1_count = work_obj[my_workload].custom1_op_count;
	strcpy(mc.m_custom1_string, my_mix_table[my_workload].custom1_string);

	mc.m_custom2_count = work_obj[my_workload].custom2_op_count;
	strcpy(mc.m_custom2_string, my_mix_table[my_workload].custom2_string);
/* HHHH */

	if(Op_lat_flag)
	{
	   mc.m_read_time = work_obj[my_workload].read_op_time;
	   mc.m_read_file_time = work_obj[my_workload].read_file_op_time;
	   mc.m_read_rand_time = work_obj[my_workload].read_rand_op_time;
	   mc.m_rmw_time = work_obj[my_workload].rmw_op_time;
	   mc.m_write_time = work_obj[my_workload].write_op_time;
	   mc.m_write_file_time = work_obj[my_workload].write_file_op_time;
	   mc.m_mmap_write_time = work_obj[my_workload].mmap_write_op_time;
	   mc.m_mmap_read_time = work_obj[my_workload].mmap_read_op_time;
	   mc.m_write_rand_time = work_obj[my_workload].write_rand_op_time;
	   mc.m_rmw_time = work_obj[my_workload].rmw_op_time;
	   mc.m_mkdir_time = work_obj[my_workload].mkdir_op_time;
	   mc.m_unlink_time = work_obj[my_workload].unlink_op_time;
	   mc.m_create_time = work_obj[my_workload].create_op_time;
	   mc.m_append_time = work_obj[my_workload].append_op_time;
 	   mc.m_access_time = work_obj[my_workload].access_op_time;
	   mc.m_chmod_time = work_obj[my_workload].chmod_op_time;
	   mc.m_readdir_time = work_obj[my_workload].readdir_op_time;
	   mc.m_stat_time = work_obj[my_workload].stat_op_time;
	   mc.m_copyfile_time = work_obj[my_workload].copyfile_op_time;
	   mc.m_locking_time = work_obj[my_workload].lock_op_time;
	   mc.m_rename_time = work_obj[my_workload].rename_op_time;
	   mc.m_statfs_time = work_obj[my_workload].statfs_op_time;
	   mc.m_pathconf_time = work_obj[my_workload].pathconf_op_time;
	   mc.m_custom1_time = work_obj[my_workload].custom1_op_time;
	   mc.m_custom2_time = work_obj[my_workload].custom2_op_time;

	   for(u=0;u<BUCKETS;u++)
	   {
	     	mc.m_bands[u] = buckets[u];
		if(cdebug) {
		  fprintf(newstdout,"band[%d] = %lld\n",u,buckets[u]);
	          fflush(newstdout);
		}
	   }
	}
	mc.m_background =  my_mix_table[my_workload].background;	
	mc.m_sharemode =  my_mix_table[my_workload].sharemode;	
	mc.m_uniform_file_size_dist =  my_mix_table[my_workload].uniform_file_size_dist;	
	mc.m_rand_dist_behavior =  my_mix_table[my_workload].rand_dist_behavior;	
	mc.m_total_file_ops = work_obj[my_workload].total_file_ops;
	mc.m_total_file_op_time = work_obj[my_workload].total_file_op_time;
	mc.m_run_time = run_time; /* My run time */
	mc.m_min_direct_size = min_direct_size; 
/*
	mc.m_ops_per_second = 
	    work_obj[my_workload].total_file_ops/work_obj[my_workload].total_file_op_time;
*/
	mc.m_ops_per_second = 
	    (float)work_obj[my_workload].total_file_ops/(float)run_time;

	mc.m_average_latency =
	   (work_obj[my_workload].total_file_op_time/work_obj[my_workload].total_file_ops)*(double)1000;
	mc.m_read_throughput = read_throughput;
	mc.m_write_throughput = write_throughput;
	mc.m_modified_run = modified_run;

	Child_send(master_name, (char *)&mc);
}

/*
 * Clients tell the master that I am at the barrier and ready
 * for the message to start work.
 */
#ifdef HAVE_ANSIC_C
void
tell_master_ready(int chid)
#else
void
tell_master_ready(chid)
int chid;
#endif
{
	struct master_command mc;
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child %d Tell master ready\n",
		   get_timestamp(&time_string[0]),chid);
		fflush(newstdout);
	}
	bzero(&mc,sizeof(struct master_command));
	mc.m_generation = generation;
	mc.m_command = R_FLAG_DATA;
	mc.m_child_flag = CHILD_STATE_READY; 
	mc.m_client_number = (int)chid; 
	Child_send(master_name, (char *)&mc);
}

/*
 * Clients tell the master that I am done
 */
#ifdef HAVE_ANSIC_C
void
tell_master_done(int chid)
#else
void
tell_master_done(chid)
int chid;
#endif
{
	struct master_command mc;
	if(sockdebug)
	{
	   fprintf(newstdout,"%s Child: %d Tell master done\n",get_timestamp(&time_string[0]),chid);
	   fflush(newstdout);
	}
	bzero(&mc,sizeof(struct master_command));
	mc.m_generation = generation;
	mc.m_command = R_IM_DONE;
	mc.m_child_flag = CHILD_STATE_READY; 
	mc.m_client_number = (int)chid; 
	Child_send(master_name, (char *)&mc);
}

/*
 * Clients tell the master that I have an error.
 */
#ifdef HAVE_ANSIC_C
void
tell_master_error(int chid, int eval )
#else
void
tell_master_error(chid,eval)
int chid,eval;
#endif
{
	struct master_command mc;
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child: Tell master error\n",
		   get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	bzero(&mc,sizeof(struct master_command));
	mc.m_generation = generation;
	mc.m_command = R_FAILED;
	mc.m_child_flag = CHILD_STATE_READY; 
	mc.m_client_number = (int)chid; 
	mc.m_client_eval = (int)eval; 
	Child_send(master_name, (char *)&mc);
}

/*
 * Clients wait at a barrier for the master to tell them
 * to begin work. This is the function where they wait.
 */
#ifdef HAVE_ANSIC_C
void
wait_for_master_go(int chid)
#else
wait_for_master_go(chid)
int chid;
#endif
{
	struct client_command *cc;
	if(sockdebug)
	{
		fprintf(newstdout,"%s Child: Wait for master Go\n",
		   get_timestamp(&time_string[0]));
		fflush(newstdout);
	}
	child_listen(child_listen_socket);
	cc = (struct client_command *)&child_rcv_buf;
	if(cc->c_command == R_TERMINATE)
	{
 	   	fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
		fprintf(newstdout,"%s Child: Received terminate\n",
		   get_timestamp(&time_string[0]));
		fflush(newstdout);
		exit(1);
	}
}
void
master_int_to_ext(struct master_command *mc, struct ext_master_command *emc)
{
	int u;
	strcpy(emc->m_host_name,mc->m_host_name);
	strcpy(emc->m_client_name,mc->m_client_name);
	sprintf(emc->m_client_eval,"%x",mc->m_client_eval);
	sprintf(emc->m_client_number,"%x",mc->m_client_number);
	sprintf(emc->m_work_obj_index,"%x",mc->m_work_obj_index);
	sprintf(emc->m_child_port,"%x",mc->m_child_port);
	sprintf(emc->m_child_pid,"%x",mc->m_child_pid);
	sprintf(emc->m_child_flag,"%x",mc->m_child_flag);
	sprintf(emc->m_command,"%x",mc->m_command);
	sprintf(emc->m_generation,"%x",mc->m_generation);
	sprintf(emc->m_percent_complete,"%d",mc->m_percent_complete);
	/* now for the results */
	strcpy(emc->m_workload_name,mc->m_workload_name);
	sprintf(emc->m_run_time,"%ld",mc->m_run_time);
	sprintf(emc->m_min_direct_size,"%ld",mc->m_min_direct_size);
	sprintf(emc->m_total_file_ops,"%10.0f",mc->m_total_file_ops);
	sprintf(emc->m_total_file_op_time,"%10.2f",mc->m_total_file_op_time);
	sprintf(emc->m_ops_per_second,"%10.2f",mc->m_ops_per_second);
	sprintf(emc->m_average_latency,"%7.3f",mc->m_average_latency);
	sprintf(emc->m_read_throughput,"%10.3f",mc->m_read_throughput);
	sprintf(emc->m_write_throughput,"%10.3f",mc->m_write_throughput);
	sprintf(emc->m_current_op_rate,"%10.3f",mc->m_current_op_rate);
/* HERE */
	sprintf(emc->m_read_count,"%x",mc->m_read_count);
	sprintf(emc->m_read_string,"%s",mc->m_read_string);
	sprintf(emc->m_read_file_count,"%x",mc->m_read_file_count);
	sprintf(emc->m_read_file_string,"%s",mc->m_read_file_string);
	sprintf(emc->m_read_rand_count,"%x",mc->m_read_rand_count);
	sprintf(emc->m_read_rand_string,"%s",mc->m_read_rand_string);
	sprintf(emc->m_write_count,"%x",mc->m_write_count);
	sprintf(emc->m_write_string,"%s",mc->m_write_string);
	sprintf(emc->m_write_file_count,"%x",mc->m_write_file_count);
	sprintf(emc->m_write_file_string,"%s",mc->m_write_file_string);
	sprintf(emc->m_open_count,"%x",mc->m_open_count);
	sprintf(emc->m_close_count,"%x",mc->m_close_count);
	sprintf(emc->m_mmap_write_count,"%x",mc->m_mmap_write_count);
	sprintf(emc->m_mmap_write_string,"%s",mc->m_mmap_write_string);
	sprintf(emc->m_mmap_read_count,"%x",mc->m_mmap_read_count);
	sprintf(emc->m_mmap_read_string,"%s",mc->m_mmap_read_string);
	sprintf(emc->m_write_rand_count,"%x",mc->m_write_rand_count);
	sprintf(emc->m_write_rand_string,"%s",mc->m_write_rand_string);
	sprintf(emc->m_rmw_count,"%x",mc->m_rmw_count);
	sprintf(emc->m_rmw_string,"%s",mc->m_rmw_string);
	sprintf(emc->m_mkdir_count,"%x",mc->m_mkdir_count);
	sprintf(emc->m_mkdir_string,"%s",mc->m_mkdir_string);
	sprintf(emc->m_unlink_count,"%x",mc->m_unlink_count);
	sprintf(emc->m_unlink_string,"%s",mc->m_unlink_string);
	sprintf(emc->m_create_count,"%x",mc->m_create_count);
	sprintf(emc->m_create_string,"%s",mc->m_create_string);
	sprintf(emc->m_append_count,"%x",mc->m_append_count);
	sprintf(emc->m_append_string,"%s",mc->m_append_string);
	sprintf(emc->m_locking_count,"%x",mc->m_locking_count);
	sprintf(emc->m_locking_string,"%s",mc->m_locking_string);
	sprintf(emc->m_access_count,"%x",mc->m_access_count);
	sprintf(emc->m_access_string,"%s",mc->m_access_string);
	sprintf(emc->m_chmod_count,"%x",mc->m_chmod_count);
	sprintf(emc->m_chmod_string,"%s",mc->m_chmod_string);
	sprintf(emc->m_readdir_count,"%x",mc->m_readdir_count);
	sprintf(emc->m_readdir_string,"%s",mc->m_readdir_string);
	sprintf(emc->m_stat_count,"%x",mc->m_stat_count);
	sprintf(emc->m_stat_string,"%s",mc->m_stat_string);
	sprintf(emc->m_copyfile_count,"%x",mc->m_copyfile_count);
	sprintf(emc->m_copyfile_string,"%s",mc->m_copyfile_string);
	sprintf(emc->m_rename_count,"%x",mc->m_rename_count);
	sprintf(emc->m_rename_string,"%s",mc->m_rename_string);
	sprintf(emc->m_statfs_count,"%x",mc->m_statfs_count);
	sprintf(emc->m_statfs_string,"%s",mc->m_statfs_string);
	sprintf(emc->m_pathconf_count,"%x",mc->m_pathconf_count);
	sprintf(emc->m_pathconf_string,"%s",mc->m_pathconf_string);
	sprintf(emc->m_custom1_count,"%x",mc->m_custom1_count);
	sprintf(emc->m_custom1_string,"%s",mc->m_custom1_string);
	sprintf(emc->m_custom2_count,"%x",mc->m_custom2_count);
	sprintf(emc->m_custom2_string,"%s",mc->m_custom2_string);
	sprintf(emc->m_heartbeat_type,"%x",mc->m_heartbeat_type);
	sprintf(emc->m_modified_run,"%x",mc->m_modified_run);
	sprintf(emc->m_background,"%d",mc->m_background);
	sprintf(emc->m_sharemode,"%d",mc->m_sharemode);
	sprintf(emc->m_uniform_file_size_dist,"%d",mc->m_uniform_file_size_dist);
	sprintf(emc->m_rand_dist_behavior,"%d",mc->m_rand_dist_behavior);

	if(Op_lat_flag)
	{
	   sprintf(emc->m_read_time,"%10.6f",mc->m_read_time);
	   sprintf(emc->m_read_file_time,"%10.6f",mc->m_read_file_time);
	   sprintf(emc->m_read_rand_time,"%10.6f",mc->m_read_rand_time);
	   sprintf(emc->m_write_time,"%10.6f",mc->m_write_time);
	   sprintf(emc->m_write_file_time,"%10.6f",mc->m_write_file_time);
	   sprintf(emc->m_mmap_write_time,"%10.6f",mc->m_mmap_write_time);
	   sprintf(emc->m_mmap_read_time,"%10.6f",mc->m_mmap_read_time);
	   sprintf(emc->m_write_rand_time,"%10.6f",mc->m_write_rand_time);
	   sprintf(emc->m_rmw_time,"%10.6f",mc->m_rmw_time);
	   sprintf(emc->m_mkdir_time,"%10.6f",mc->m_mkdir_time);
	   sprintf(emc->m_unlink_time,"%10.6f",mc->m_unlink_time);
	   sprintf(emc->m_create_time,"%10.6f",mc->m_create_time);
	   sprintf(emc->m_append_time,"%10.6f",mc->m_append_time);
	   sprintf(emc->m_locking_time,"%10.6f",mc->m_locking_time);
	   sprintf(emc->m_access_time,"%10.6f",mc->m_access_time);
	   sprintf(emc->m_chmod_time,"%10.6f",mc->m_chmod_time);
	   sprintf(emc->m_readdir_time,"%10.6f",mc->m_readdir_time);
	   sprintf(emc->m_stat_time,"%10.6f",mc->m_stat_time);
	   sprintf(emc->m_copyfile_time,"%10.6f",mc->m_copyfile_time);
	   sprintf(emc->m_rename_time,"%10.6f",mc->m_rename_time);
	   sprintf(emc->m_statfs_time,"%10.6f",mc->m_statfs_time);
	   sprintf(emc->m_pathconf_time,"%10.6f",mc->m_pathconf_time);
	   sprintf(emc->m_custom1_time,"%10.6f",mc->m_custom1_time);
	   sprintf(emc->m_custom2_time,"%10.6f",mc->m_custom2_time);
	   for(u=0;u<BUCKETS;u++)
	   {
	     sprintf(&emc->m_bands[u][0],"%lld",mc->m_bands[u]);
	     if(cdebug)
	     {
	        fprintf(newstdout,"c->m Band[%d] = %s\n",u,&emc->m_bands[u][0]);
	        fflush(newstdout);
	     }
	   }
	}
}
void
master_ext_to_int(struct master_command *mc, struct ext_master_command *emc)
{
	int u;
	strcpy(mc->m_host_name,emc->m_host_name);
	strcpy(mc->m_client_name,emc->m_client_name);
	sscanf(emc->m_client_eval,"%x",&mc->m_client_eval);
	sscanf(emc->m_client_number,"%x",&mc->m_client_number);
	sscanf(emc->m_work_obj_index,"%x",&mc->m_work_obj_index);
	sscanf(emc->m_child_port,"%x",&mc->m_child_port);
	sscanf(emc->m_child_pid,"%x",&mc->m_child_pid);
	sscanf(emc->m_child_flag,"%x",&mc->m_child_flag);
	sscanf(emc->m_command,"%x",&mc->m_command);
	sscanf(emc->m_generation,"%x",&mc->m_generation);
	sscanf(emc->m_percent_complete,"%d",&mc->m_percent_complete);
	/* now for the results */
	strcpy(mc->m_client_name,emc->m_client_name);
	strcpy(mc->m_workload_name,emc->m_workload_name);
	sscanf(emc->m_run_time,"%ld",&mc->m_run_time);
	sscanf(emc->m_min_direct_size,"%ld",&mc->m_min_direct_size);
	   /* Yeah, that weird %lf is really needed for doubles!!! */
	sscanf(emc->m_total_file_ops,"%lf",&mc->m_total_file_ops);
	sscanf(emc->m_total_file_op_time,"%lf",&mc->m_total_file_op_time);
	sscanf(emc->m_ops_per_second,"%lf",&mc->m_ops_per_second);
	sscanf(emc->m_average_latency,"%lf",&mc->m_average_latency);
	sscanf(emc->m_read_throughput,"%lf",&mc->m_read_throughput);
	sscanf(emc->m_write_throughput,"%lf",&mc->m_write_throughput);
	sscanf(emc->m_current_op_rate,"%lf",&mc->m_current_op_rate);
	   /* Yeah, that weird %lf is really needed for doubles!!! */
/* HERE */
	sscanf(emc->m_read_count,"%x",&mc->m_read_count);
	sscanf(emc->m_read_string,"%s",mc->m_read_string);
	sscanf(emc->m_read_file_count,"%x",&mc->m_read_file_count);
	sscanf(emc->m_read_file_string,"%s",mc->m_read_file_string);
	sscanf(emc->m_read_rand_count,"%x",&mc->m_read_rand_count);
	sscanf(emc->m_read_rand_string,"%s",mc->m_read_rand_string);
	sscanf(emc->m_write_count,"%x",&mc->m_write_count);
	sscanf(emc->m_write_string,"%s",mc->m_write_string);
	sscanf(emc->m_write_file_count,"%x",&mc->m_write_file_count);
	sscanf(emc->m_write_file_string,"%s",mc->m_write_file_string);
	sscanf(emc->m_open_count,"%x",&mc->m_open_count);
	sscanf(emc->m_close_count,"%x",&mc->m_close_count);
	sscanf(emc->m_mmap_write_count,"%x",&mc->m_mmap_write_count);
	sscanf(emc->m_mmap_write_string,"%s",mc->m_mmap_write_string);
	sscanf(emc->m_mmap_read_count,"%x",&mc->m_mmap_read_count);
	sscanf(emc->m_mmap_read_string,"%s",mc->m_mmap_read_string);
	sscanf(emc->m_write_rand_count,"%x",&mc->m_write_rand_count);
	sscanf(emc->m_write_rand_string,"%s",mc->m_write_rand_string);
	sscanf(emc->m_rmw_count,"%x",&mc->m_rmw_count);
	sscanf(emc->m_rmw_string,"%s",mc->m_rmw_string);
	sscanf(emc->m_mkdir_count,"%x",&mc->m_mkdir_count);
	sscanf(emc->m_mkdir_string,"%s",mc->m_mkdir_string);
	sscanf(emc->m_unlink_count,"%x",&mc->m_unlink_count);
	sscanf(emc->m_unlink_string,"%s",mc->m_unlink_string);
	sscanf(emc->m_create_count,"%x",&mc->m_create_count);
	sscanf(emc->m_create_string,"%s",mc->m_create_string);
	sscanf(emc->m_append_count,"%x",&mc->m_append_count);
	sscanf(emc->m_append_string,"%s",mc->m_append_string);
	sscanf(emc->m_locking_count,"%x",&mc->m_locking_count);
	sscanf(emc->m_locking_string,"%s",mc->m_locking_string);
	sscanf(emc->m_access_count,"%x",&mc->m_access_count);
	sscanf(emc->m_access_string,"%s",mc->m_access_string);
	sscanf(emc->m_chmod_count,"%x",&mc->m_chmod_count);
	sscanf(emc->m_chmod_string,"%s",mc->m_chmod_string);
	sscanf(emc->m_readdir_count,"%x",&mc->m_readdir_count);
	sscanf(emc->m_readdir_string,"%s",mc->m_readdir_string);
	sscanf(emc->m_stat_count,"%x",&mc->m_stat_count);
	sscanf(emc->m_stat_string,"%s",mc->m_stat_string);
	sscanf(emc->m_copyfile_count,"%x",&mc->m_copyfile_count);
	sscanf(emc->m_copyfile_string,"%s",mc->m_copyfile_string);
	sscanf(emc->m_rename_count,"%x",&mc->m_rename_count);
	sscanf(emc->m_rename_string,"%s",mc->m_rename_string);
	sscanf(emc->m_statfs_count,"%x",&mc->m_statfs_count);
	sscanf(emc->m_statfs_string,"%s",mc->m_statfs_string);
	sscanf(emc->m_pathconf_count,"%x",&mc->m_pathconf_count);
	sscanf(emc->m_pathconf_string,"%s",mc->m_pathconf_string);
	sscanf(emc->m_custom1_count,"%x",&mc->m_custom1_count);
	sscanf(emc->m_custom1_string,"%s",mc->m_custom1_string);
	sscanf(emc->m_custom2_count,"%x",&mc->m_custom2_count);
	sscanf(emc->m_custom2_string,"%s",mc->m_custom2_string);
	sscanf(emc->m_background,"%d",&mc->m_background);
	sscanf(emc->m_sharemode,"%d",&mc->m_sharemode);
	sscanf(emc->m_uniform_file_size_dist,"%d",&mc->m_uniform_file_size_dist);
	sscanf(emc->m_rand_dist_behavior,"%d",&mc->m_rand_dist_behavior);

	if(Op_lat_flag)
	{
	   sscanf(emc->m_read_time,"%lf",&mc->m_read_time);
	   sscanf(emc->m_read_file_time,"%lf",&mc->m_read_file_time);
	   sscanf(emc->m_read_rand_time,"%lf",&mc->m_read_rand_time);
	   sscanf(emc->m_write_time,"%lf",&mc->m_write_time);
	   sscanf(emc->m_write_file_time,"%lf",&mc->m_write_file_time);
	   sscanf(emc->m_mmap_write_time,"%lf",&mc->m_mmap_write_time);
	   sscanf(emc->m_mmap_read_time,"%lf",&mc->m_mmap_read_time);
	   sscanf(emc->m_write_rand_time,"%lf",&mc->m_write_rand_time);
	   sscanf(emc->m_rmw_time,"%lf",&mc->m_rmw_time);
	   sscanf(emc->m_mkdir_time,"%lf",&mc->m_mkdir_time);
	   sscanf(emc->m_unlink_time,"%lf",&mc->m_unlink_time);
	   sscanf(emc->m_create_time,"%lf",&mc->m_create_time);
	   sscanf(emc->m_append_time,"%lf",&mc->m_append_time);
	   sscanf(emc->m_locking_time,"%lf",&mc->m_locking_time);
	   sscanf(emc->m_access_time,"%lf",&mc->m_access_time);
	   sscanf(emc->m_chmod_time,"%lf",&mc->m_chmod_time);
	   sscanf(emc->m_readdir_time,"%lf",&mc->m_readdir_time);
	   sscanf(emc->m_stat_time,"%lf",&mc->m_stat_time);
	   sscanf(emc->m_copyfile_time,"%lf",&mc->m_copyfile_time);
	   sscanf(emc->m_rename_time,"%lf",&mc->m_rename_time);
	   sscanf(emc->m_statfs_time,"%lf",&mc->m_statfs_time);
	   sscanf(emc->m_pathconf_time,"%lf",&mc->m_pathconf_time);
	   sscanf(emc->m_custom1_time,"%lf",&mc->m_custom1_time);
	   sscanf(emc->m_custom2_time,"%lf",&mc->m_custom2_time);
	   for(u=0;u<BUCKETS;u++)
	   {
	     sscanf(&emc->m_bands[u][0],"%lld",&mc->m_bands[u]);
	     if(mdebug)
	        printf("M Band[%d] = %lld\n",u,mc->m_bands[u]);
	   }
	}
	sscanf(emc->m_heartbeat_type,"%x",&mc->m_heartbeat_type);
	sscanf(emc->m_modified_run,"%x",&mc->m_modified_run);
}
void
client_int_to_ext(struct client_command *cc, struct ext_client_command *ecc)
{
	int i,j;
	strcpy(ecc->c_host_name,cc->c_host_name);
	strcpy(ecc->c_pit_hostname,cc->c_pit_hostname);
	strcpy(ecc->c_pit_service,cc->c_pit_service);
	strcpy(ecc->c_client_name,cc->c_client_name);
	strcpy(ecc->c_client_version,cc->c_client_version);
	sprintf(ecc->c_gg_offset,"%d",cc->c_gg_offset);
	sprintf(ecc->c_gg_flag,"%d",cc->c_gg_flag);
	sprintf(ecc->c_client_number,"%d",cc->c_client_number);
	sprintf(ecc->c_child_flag,"%d",cc->c_child_flag);
	sprintf(ecc->c_flush_flag,"%d",cc->c_flush_flag);
	sprintf(ecc->c_sockdebug,"%d",cc->c_sockdebug);
	sprintf(ecc->c_cleanup,"%d",cc->c_cleanup);
	sprintf(ecc->c_dump_files_flag,"%d",cc->c_dump_files_flag);
	sprintf(ecc->c_tracedebug,"%d",cc->c_tracedebug);
	sprintf(ecc->c_ipv6_enable,"%d",cc->c_ipv6_enable);
	sprintf(ecc->c_licensed,"%d",cc->c_licensed);
	sprintf(ecc->c_heartbeat,"%d",cc->c_heartbeat);
	sprintf(ecc->c_fd_caching_limit,"%d",cc->c_fd_caching_limit);
	sprintf(ecc->c_op_lat_flag,"%d",cc->c_op_lat_flag);
	sprintf(ecc->c_command,"%d",cc->c_command);
	sprintf(ecc->c_generation,"%d",cc->c_generation);
	/* transpose OPs */
	for(j=0;j<NUM_WORK_OBJ;j++)
	{
	   strcpy(ecc->ext_worker[j].c_workload_name,cc->client_worker[j].c_workload_name);
	   strcpy(ecc->ext_worker[j].c_fs_type,cc->client_worker[j].c_fs_type);
	   for (i =0; i< NUM_OP_TYPES; i++ )
	   {
	      sprintf(ecc->ext_worker[j].c_op_array[i],"%d",cc->client_worker[j].c_op_array[i]);
	      sprintf(ecc->ext_worker[j].c_op_array_string[i],"%s",cc->client_worker[j].c_op_array_string[i]);
	   }
	   /* transpose xfer mixes */
	   for (i =0; i< MAX_DIST_ELEM; i++ )
   	   {
	      sprintf(ecc->ext_worker[j].c_rdist_array_size_min[i],"%d",cc->client_worker[j].c_rdist_array[i].size_min);
	      sprintf(ecc->ext_worker[j].c_rdist_array_size_max[i],"%d",cc->client_worker[j].c_rdist_array[i].size_max);
	      sprintf(ecc->ext_worker[j].c_rdist_array_percent[i],"%d",cc->client_worker[j].c_rdist_array[i].percent);

	      sprintf(ecc->ext_worker[j].c_wdist_array_size_min[i],"%d",cc->client_worker[j].c_wdist_array[i].size_min);
	      sprintf(ecc->ext_worker[j].c_wdist_array_size_max[i],"%d",cc->client_worker[j].c_wdist_array[i].size_max);
	      sprintf(ecc->ext_worker[j].c_wdist_array_percent[i],"%d",cc->client_worker[j].c_wdist_array[i].percent);
		
	   }
	   sprintf(ecc->ext_worker[j].c_percent_commit,"%d",cc->client_worker[j].c_percent_commit);
	   sprintf(ecc->ext_worker[j].c_percent_direct,"%d",cc->client_worker[j].c_percent_direct);
	   sprintf(ecc->ext_worker[j].c_align,"%d",cc->client_worker[j].c_align);
	   sprintf(ecc->ext_worker[j].c_percent_osync,"%d",cc->client_worker[j].c_percent_osync);
	   sprintf(ecc->ext_worker[j].c_percent_geometric,"%d",cc->client_worker[j].c_percent_geometric);
	   sprintf(ecc->ext_worker[j].c_percent_compress,"%d",cc->client_worker[j].c_percent_compress);
	   sprintf(ecc->ext_worker[j].c_background,"%d",cc->client_worker[j].c_background);
	   sprintf(ecc->ext_worker[j].c_sharemode,"%d",cc->client_worker[j].c_sharemode);
	   sprintf(ecc->ext_worker[j].c_uniform_file_size_dist,"%d",cc->client_worker[j].c_uniform_file_size_dist);
	   sprintf(ecc->ext_worker[j].c_rand_dist_behavior,"%d",cc->client_worker[j].c_rand_dist_behavior);
	}
}
void
client_ext_to_int(struct client_command *cc, struct ext_client_command *ecc)
{
	int i,j;
	strcpy(cc->c_host_name,ecc->c_host_name);
	strcpy(cc->c_pit_hostname,ecc->c_pit_hostname);
	strcpy(cc->c_pit_service,ecc->c_pit_service);
	strcpy(cc->c_client_name,ecc->c_client_name);
	strcpy(cc->c_client_version,ecc->c_client_version);
	sscanf(ecc->c_gg_offset,"%d",&cc->c_gg_offset);
	sscanf(ecc->c_gg_flag,"%d",&cc->c_gg_flag);
	sscanf(ecc->c_client_number,"%d",&cc->c_client_number);
	sscanf(ecc->c_child_flag,"%d",&cc->c_child_flag);
	sscanf(ecc->c_flush_flag,"%d",&cc->c_flush_flag);
	sscanf(ecc->c_sockdebug,"%d",&cc->c_sockdebug);
	sscanf(ecc->c_cleanup,"%d",&cc->c_cleanup);
	sscanf(ecc->c_dump_files_flag,"%d",&cc->c_dump_files_flag);
	sscanf(ecc->c_tracedebug,"%d",&cc->c_tracedebug);
	sscanf(ecc->c_ipv6_enable,"%d",&cc->c_ipv6_enable);
	sscanf(ecc->c_licensed,"%d",&cc->c_licensed);
	sscanf(ecc->c_heartbeat,"%d",&cc->c_heartbeat);
	sscanf(ecc->c_fd_caching_limit,"%d",&cc->c_fd_caching_limit);
	sscanf(ecc->c_op_lat_flag,"%d",&cc->c_op_lat_flag);
	sscanf(ecc->c_command,"%d",&cc->c_command);
	sscanf(ecc->c_generation,"%d",&cc->c_generation);

	for(j=0;j<NUM_WORK_OBJ;j++)
	{
	   strcpy(cc->client_worker[j].c_workload_name,ecc->ext_worker[j].c_workload_name);
	   strcpy(cc->client_worker[j].c_fs_type,ecc->ext_worker[j].c_fs_type);
	   /* Transpose the OPs */
	   for (i =0; i< NUM_OP_TYPES; i++ )
	   {
	      sscanf(ecc->ext_worker[j].c_op_array[i],"%d",&cc->client_worker[j].c_op_array[i]);
	      sscanf(ecc->ext_worker[j].c_op_array_string[i],"%s",cc->client_worker[j].c_op_array_string[i]);
	   }
	   /* Transpose the xfer mix */
	   for (i =0; i< MAX_DIST_ELEM; i++ )
	   {
	      sscanf(ecc->ext_worker[j].c_rdist_array_size_min[i],"%d",&cc->client_worker[j].c_rdist_array[i].size_min);
	      sscanf(ecc->ext_worker[j].c_rdist_array_size_max[i],"%d",&cc->client_worker[j].c_rdist_array[i].size_max);
	      sscanf(ecc->ext_worker[j].c_rdist_array_percent[i],"%d",&cc->client_worker[j].c_rdist_array[i].percent);
	      sscanf(ecc->ext_worker[j].c_wdist_array_size_min[i],"%d",&cc->client_worker[j].c_wdist_array[i].size_min);
	      sscanf(ecc->ext_worker[j].c_wdist_array_size_max[i],"%d",&cc->client_worker[j].c_wdist_array[i].size_max);
	      sscanf(ecc->ext_worker[j].c_wdist_array_percent[i],"%d",&cc->client_worker[j].c_wdist_array[i].percent);
	   }
	   sscanf(ecc->ext_worker[j].c_percent_commit,"%d",&cc->client_worker[j].c_percent_commit);
	   sscanf(ecc->ext_worker[j].c_percent_direct,"%d",&cc->client_worker[j].c_percent_direct);
	   sscanf(ecc->ext_worker[j].c_align,"%d",&cc->client_worker[j].c_align);
	   sscanf(ecc->ext_worker[j].c_percent_osync,"%d",&cc->client_worker[j].c_percent_osync);
	   sscanf(ecc->ext_worker[j].c_percent_geometric,"%d",&cc->client_worker[j].c_percent_geometric);
	   sscanf(ecc->ext_worker[j].c_percent_compress,"%d",&cc->client_worker[j].c_percent_compress);
	   sscanf(ecc->ext_worker[j].c_background,"%d",&cc->client_worker[j].c_background);
	   sscanf(ecc->ext_worker[j].c_sharemode,"%d",&cc->client_worker[j].c_sharemode);
	   sscanf(ecc->ext_worker[j].c_uniform_file_size_dist,"%d",&cc->client_worker[j].c_uniform_file_size_dist);
	   sscanf(ecc->ext_worker[j].c_rand_dist_behavior,"%d",&cc->client_worker[j].c_rand_dist_behavior);
	}
}

/****************************************************************************/
/* END COMM SECTION */
/****************************************************************************/

/*
 * This is a microsecond sleep function. It will permit one 
 * to sleep for units of time that are in the microsecond range..
 * Well.... sort of... Actually there are no existing timers 
 * that actually let one sleep for any duration that is shorter
 * than a time slice. Which on most systems in NOT in microseconds,
 * but actually in 10-ish millisecond range. 
 * All of the nifty new POSIX usleep() and nanosleep() funtions
 * do NOT actually go below a time slice, so they are useless
 * examples of group think ! Thus, nap() below does as good 
 * as one can do, and is portable across more systems.
 */

void
nap( int select_msecs )
{
#if defined(WIN32)
	Sleep(select_msecs); /* Visual studio's Sleep(milliseconds) */
#else
	struct timeval	sleeptime;
	sleeptime.tv_sec = select_msecs / 1000;
	sleeptime.tv_usec = (select_msecs % 1000) * 1000;
	select(0, NULL, NULL , NULL, &sleeptime);
#endif
}

/* 
 * Fill a buffer with a moderately random pattern.
 * In the future one might wish to think about 
 * how dedup is going to be integrated into this tool.
 */
void
generic_fill_buf(long id, long patt, int size, char *buf)
{
	int i;
	long *lbuf;
	lbuf = (long *)buf;
	for(i=0; i < (size/sizeof(long)); i++)
	  *lbuf++=(id^patt) * i;
}

/*
 * Note: The PIT below is a way of getting a timer that is more
 * stable than what happens in a virtual machine, but... it may have
 * scaling issues as the number o threads goes up... this may
 * swarm the PIT server and introduce intrusion.
 */
#define DFLT_SERVICE   "PIT"     /* Default service name.               */
#define INVALID_DESC   -1        /* Invalid file (socket) descriptor.   */
#define MAXBFRSIZE     256       /* Max bfr sz to read remote TOD.      */
#if defined(Windows) || defined(WIN32)
int false = 0;
int true = 1;
#else
typedef enum { false = 0, true } boolean;
#endif

/*
 * The Programmable Interdimensional Timer interface code.
 */


#if defined(_solaris_)
int pit_gettimeofday( struct timeval *tp, long foo,
	char *pit_hostname, char *pit_service)
#else
int pit_gettimeofday( struct timeval *tp, struct timezone *foo,
	char *pit_hostname, char *pit_service)
#endif
{
	int            	sckt;          /* socket descriptor */
	unsigned scopeId = 0;
	double mytime;

	/* See if the interdimensional rift is active */
	
	if(pit_hostname[0] == 0)
	{
#if defined(_solaris_)
		mytime = m_gettimeofday((struct timeval *)tp,(long)foo);
#else
		mytime = m_gettimeofday((struct timeval *)tp,(struct timezone *)foo);
#endif
		if(cdebug)
		{
      		   fprintf( newstderr,"time returned from m_gettimeofday() = %f\n",mytime);
	           fflush(newstdout);
		}
		return (0);
	}

 	if ( ( sckt = openSckt( pit_hostname,
                           pit_service,
                           scopeId ) ) == INVALID_DESC )
   	{
      		fprintf( newstderr,
                  "%s Sorry... a connectionless socket could "
                  "not be set up.\n",get_timestamp(&time_string[0]));
	        fflush(newstdout);
                return -1;
   	}
	/*
   	** Get the remote PIT.
   	*/
   	pit( sckt ,tp );
	close(sckt);
    	return 0;
}

/*
 * Opens a socket for the PIT to use to get the time
 * from a remote time server ( A PIT server ) 
 */
int openSckt( const char   *host,
                     const char   *service,
                     unsigned int  scopeId )
{
   int              sckt;
   /*
    * Initialize the 'hints' structure for getaddrinfo(3).
   */
   memset( &hints, 0, sizeof( hints ) );
   hints.ai_family   = PF_UNSPEC;     /* IPv4 or IPv6 records */

   hints.ai_socktype = SOCK_STREAM;    /* Connection oriented communication.*/
   hints.ai_protocol = IPPROTO_TCP;   /* TCP transport layer protocol only. */
   /*
   ** Look up the host/service information.
   */
   if ( ( aiErr = getaddrinfo( host,
                               service,
                               &hints,
                               &aiHead ) ) != 0 )
   {
         fprintf( newstderr, "(line %d): host - %s, service - %s. ERROR - %s.\n", __LINE__, host, service,
	    gai_strerror( aiErr ) );
	 fflush(newstdout);
         return INVALID_DESC;
   }
   /*
   ** Go through the list and try to open a connection.  Continue until either
   ** a connection is established or the entire list is exhausted.
   */
   for ( ai = aiHead,   sckt = INVALID_DESC;
         ( ai != NULL ) && ( sckt == INVALID_DESC );
         ai = ai->ai_next )
   {
      /*
      ** IPv6 kluge.  Make sure the scope ID is set.
      */
      if ( ai->ai_family == PF_INET6 )
      {
         pSadrIn6 = (sockaddr_in6_t*) ai->ai_addr;
         if ( pSadrIn6->sin6_scope_id == 0 )
         {
            pSadrIn6->sin6_scope_id = scopeId;
         }  /* End IF the scope ID wasn't set. */
      }  /* End IPv6 kluge. */
      /*
      ** Create a socket.
      */
      sckt = socket( ai->ai_family, ai->ai_socktype, ai->ai_protocol );
      if(sckt == -1)
      {
         sckt = INVALID_DESC;
         continue;   /* Try the next address record in the list. */
      }
      /*
      ** Set the target destination for the remote host on this socket.  That
      ** is, this socket only communicates with the specified host.
      */
      if (connect( sckt, ai->ai_addr, ai->ai_addrlen ) )
      {
         (void) close( sckt );   /* Could use system call again here, 
					but why? */
         sckt = INVALID_DESC;
         continue;   /* Try the next address record in the list. */
      }
   }  /* End FOR each address record returned by getaddrinfo(3). */
   /*
   ** Clean up & return.
   */
   freeaddrinfo( aiHead ); /* Keep cached */
   return sckt;
}  /* End openSckt() */


/*
 * Read the PIT, and convert this back into timeval 
 * info, and store it in the timeval structure that was
 * passed in.
 */
void pit( int sckt, struct timeval *tp)
{
   char bfr[ MAXBFRSIZE+1 ];
   int  inBytes;
   unsigned long long value = 0LL;
   /*
   ** Send a datagram to the server to wake it up.  The content isn't
   ** important, but something must be sent to let it know we want the TOD.
   */
   junk=write( sckt, "Author: Don Capps", 17 );
   /*
   ** Read the PIT from the remote host.
   */
   inBytes = read( sckt, bfr, MAXBFRSIZE );
   if(inBytes <= 0 )
   {
	fprintf(newstdout,"PIT read error: Errno:  %s\n",
	   strerror(errno));
	fflush(newstdout);
	inBytes=0;
   }
   bfr[ inBytes ] = '\0';   /* Null-terminate the received string. */
   /* 
    * Convert result to timeval structure format 
    */
   sscanf(bfr,"%llu\n",&value);
   tp->tv_sec = (long)(value / 1000000);
   tp->tv_usec = (long)(value % 1000000);
   if(tp->tv_sec < 0)
   {
	fprintf(newstdout,"%s ****** OOPS ***** tv_sec is negative \n",get_timestamp(&time_string[0]));
	fflush(newstdout);
   }
   if(tp->tv_usec < 0)
   {
	fprintf(newstdout,"%s ****** OOPS ***** tv_usec is negative \n",get_timestamp(&time_string[0]));
	fflush(newstdout);
   }
}  


/*
 * Put the date and time into a string, without the new line.
 */
char *
get_timestamp(char *string)
{
    time_t	run_str;

    run_str = time((time_t *)&dummytime);
    (void) strncpy((char *) string, (char *) ctime((time_t *)&run_str), 24);
    return(string);
}

/*
 * The power monitor external script startup routine.
 * Builds command, starts with popen() so it can read
 * any output from the command, and passes one option
 * to the command.
 */
int
power_mon(int flag, char *options)
{
	char command[MAXECOMMAND];
	char result[MAXECOMMAND];
	FILE *fp;

	/* Zero out buffer */
	memset(command,0,MAXECOMMAND);
	memset(result,0,MAXECOMMAND);
	/* If told to run the command */
	if(flag == SRUN)
	{
	  /* Check for existance and executable */
	  if( (access(power_script,F_OK) == 0) &&
		(access(power_script,X_OK)==0) )
	  {
	    /* Build command */
#if defined(WIN32)
	      sprintf(command,"%s %s\n",power_script,options);
#else
	    /*sprintf(command,"sh -c \'%s %s\'",power_script,options);*/
	      sprintf(command,"%s %s",power_script,options);
#endif
	    /* Run command */
	      fp=(FILE *)popen(command,"r");
	      if(fp == 0)
		return(1);
	    /* Read any returned output 				*/
	    /* The file position could be at EOF before we read the  	*/
	    /* file, this is not an error. 				*/
	      if(feof(fp))
	      {
	          pclose(fp);
	          return(0);
	      }
	      else
	      {
		fread(result,MAXECOMMAND,1,fp);
		if(result[0]!=0)
	          printf("\tPower returned: %s \n",result);
	        pclose(fp);
	        return(0);
	      }
	   }
	   return(1);
	}
	/* Just check to see if the command is present and executable */
	if(flag == SCHECK)
	{
	  if( (access(power_script,F_OK) == 0) &&
		(access(power_script,X_OK)==0) )
	  {
	      return(0);
	  }
	  return(1);
	}
	return(1);
}	

/*
 * The external monitor script startup routine.
 * Builds command, starts with popen() so it can read
 * any output from the command, and passes one option
 * to the command.
 */
int
extern_mon(int flag, char *options)
{
	char command[MAXECOMMAND];
	char result[MAXECOMMAND];
	FILE *fp;

	/* Zero out buffer */
	memset(command,0,MAXECOMMAND);
	memset(result,0,MAXECOMMAND);

	/* If told to run the command */
	if(flag == SRUN)
	{
	  /* Check for existance and executable */
	    if( (access(mon_script,F_OK) == 0) && (access(mon_script,X_OK)==0) )
	    {
	    /* Build command */
#if defined(WIN32)
	      sprintf(command,"%s %s\n",mon_script,options);
#else
	    /*sprintf(command,"sh -c \'%s %s\'",mon_script,options);*/
	      sprintf(command,"%s %s",mon_script,options);
#endif
	    /* Run command */
	      fp=(FILE *)popen(command,"r");
	      if(fp == 0)
		return(1);
	    /* Read any returned output 				*/
	    /* The file position could be at EOF before we read the  	*/
	    /* file, this is not an error. 				*/
	      if(feof(fp))
	      {
	       pclose(fp);
	       return(0);
	      }
	      else
	      {
		  fread(result,MAXECOMMAND,1,fp);
		  if(result[0]!=0)
	            printf("\tExternal monitor returned: %s \n",result);
	          pclose(fp);
	          return(0);
	      }
	    }
	    return(1);
	}
	/* Just check to see if the command is present and executable */
	if(flag == SCHECK)
	{
	  if( (access(mon_script,F_OK) == 0) && (access(mon_script,X_OK)==0) )
	  {
	      return(0);
	  }
	  return(1);
	}
	return(1);
}	

/*
 * The user defined external monitor script startup routine.
 * Builds command, starts with popen() so it can read
 * any output from the command, and passes one option
 * to the command.
 */
int
u_extern_mon(int flag, char *options)
{
	char command[MAXECOMMAND];
	char result[MAXECOMMAND];
	FILE *fp;

	/* Zero out buffer */
	memset(command,0,MAXECOMMAND);
	memset(result,0,MAXECOMMAND);

	/* If told to run the command */
	if(flag == SRUN)
	{
	  /* Check for existance and executable */
	  if( (access(u_external_script_name,F_OK) == 0) && (access(u_external_script_name,X_OK)==0) )
	  {
	    /* Build command */
#if defined(WIN32)
	    sprintf(command,"%s %s %s\n",u_external_script_name,options, 
		u_external_script_args);
#else
	    /*sprintf(command,"sh -c \'%s %s %s\'",u_external_script_name,options,*/
	    sprintf(command,"%s %s %s",u_external_script_name,options,
		u_external_script_args);
#endif
	    /* Run command */
	    fp=(FILE *)popen(command,"r");
	    if(fp == 0)
		return(1);
	    /* Read any returned output 				*/
	    /* The file position could be at EOF before we read the  	*/
	    /* file, this is not an error. 				*/
	    if(feof(fp))
	    {
	       pclose(fp);
	       return(0);
	    }
	    else
	    {
 		fread(result,MAXECOMMAND,1,fp);
		if(result[0]!=0)
	           printf("\tUser defined monitor returned: %s \n",result);
	        pclose(fp);
	        return(0);
	    }
	  }
	  else
		return(1);
	}
	/* Just check to see if the command is present and executable */
	if(flag == SCHECK)
	{
	  if( (access(u_external_script_name,F_OK) == 0) && (access(u_external_script_name,X_OK)==0) )
	  {
	      return(0);
	  }
	  else
	     return(1);
	}
	return(1);
}	

/*
 * The external stat script startup routine.
 * Builds command, starts with popen() so it can read
 * any output from the command, and passes one option
 * to the command.
 */
int
extern_stat(int flag, char *options)
{
	char command[MAXECOMMAND];
	char result[MAXECOMMAND];
	FILE *fp;

	/* Zero out buffer */
	memset(command,0,MAXECOMMAND);
	memset(result,0,MAXECOMMAND);

	/* If told to run the command */
	if(flag == SRUN)
	{
	  /* Check for existance and executable */
	  if( (access(stat_script,F_OK) == 0) && (access(stat_script,X_OK)==0) )
	  {
	    /* Build command */
#if defined(WIN32)
	      sprintf(command,"%s %s\n",stat_script,options);
#else
	    /*sprintf(command,"sh -c \'%s %s\'",stat_script,options);*/
	      sprintf(command,"%s %s",stat_script,options);
#endif
	    /* Run command */
	      fp=(FILE *)popen(command,"r");
	      if(fp == 0)
		return(1);
	      /* Read any returned output 				*/
	      /* The file position could be at EOF before we read the  	*/
	      /* file, this is not an error. 				*/
	      if(feof(fp))
	      {
	         pclose(fp);
	         return(0);
	      }
	      else
	      {
	         fread(result,MAXECOMMAND,1,fp);
		 if(result[0]!=0)
	            printf("\tExternal stat returned: %s \n",result);
	         pclose(fp);
	         return(0);
	      }
	   }
	}
	/* Just check to see if the command is present and executable */
	if(flag == SCHECK)
	{
	  if( (access(stat_script,F_OK) == 0) &&
		(access(stat_script,X_OK)==0) )
	  {
	      return(0);
	  }
	  return(1);
	}
	return(1);
}	

/*
 * Return current time in seconds (double), and also fill in
 * the timeval structure, in case someone cares.
 */
#if defined(_solaris_)
struct timeval Oink;
double m_gettimeofday(struct timeval *tp, long tz)
#else
double m_gettimeofday(struct timeval *tp, struct timezone *tz)
#endif
{
#if defined(_solaris_)
    	if(gettimeofday((struct timeval *)tp, (long) NULL) == -1)
#else
    	if(gettimeofday((struct timeval *)tp, (struct timezone *) NULL) == -1)
#endif
      		perror("gettimeofday");
    	return ((double) (tp->tv_sec)) + (((double) tp->tv_usec/1000000));
}


#define MVFDN

#if defined(MVFDN)

#define NM_MAXNAMLEN 32
/*
 * Creating filenames with a character distribution that
 * approximates English usage. 
 * From the mind of Don Capps. (Mr. Iozone)
 */

/* The maximum number of character in the prefix name. */
#define PRENAME 8
/* The minimum number of character in the prefix name. */
#define MINPRENAME 3
/* The maximum number of character in the suffix name. */
#define POSTNAME 5
/* The minimum number of character in the suffix name. */
#define MINPOSTNAME 0
/* The size of the frequency distribution table. */
#define TABLE_SIZE 10000


int t_size; /* The active size of the distribution table */

/*
 * Frequency of character usage in the English language.
 * Multiplied by 10000 for frequency table initialization.
 */
 /* Concise Oxford Disctionary 9th edition 1995 */
int freq[]={849,207,454,338,1116,181,247,300,754,19,110,548,301,
            665,716,316,19,758,573,695,363,100,129,29,177,27};
/*
 * English character set.
 */
char ch[]="abcdefghijklmnopqrstuvwxyz";

char table[TABLE_SIZE];
int dist_active; /* Flag, to permit only initializing the table once.*/


/*
 * Make a file name that is [3->8].[0->5] compliant, with pre and post dot
 * lengths that are random, and consisting of characters that
 * have a proper frequency distribution.
 *
 * Note: This routine saves the Netmist random number generator seed before
 * and restores it after makeing the name. This is critical as one
 * MUST not disturb the random namespace or it will cause the entire
 * benchmark to get very funky.
 */
void
make_name(char *namebuf, int seed, int type, int filenum)
{
        int x,y,pt,save;
	int skew = 0;
        char pre[NM_MAXNAMLEN]; 	/* Prefix buffer */
        char post[NM_MAXNAMLEN]; 	/* Suffix buffer */
#if defined(WIN32)
	char special[5];

	special[0]=0;
	special[1]=0;
	special[2]=0;
	special[3]=0;
	special[4]=0;
#endif

	if(dist_active==0) 		/* Only once please */
	{
	   init_dist();
	   dist_active=1;
	}
	save=name_gseed(); 		/* Save the seed !!! */
	switch (type)
	{
	   case 0: skew=101; 	/* 0 = Files. Note: Prime */
		break;
	   case 1: skew=997; 	/* 1 = Directories. Note: Prime*/
		break;
	   case 2: skew=2003; 	/* 2 = Symlinks. Note: Prime */
		break;
	   case 3: skew=3001; 	/* 3 = testdir name. Note: Prime */
		break;
	};

	/* 
	 * seed is the Unique file number.
	 * +1 to avoid confusing netmist_srand with a zero.
	 * +skew to move the directories, files and symlinks apart.
	 * +client_id to move the children apart.
 	 */
	name_srand(seed+1+skew+impersonate); 
#if defined(WIN32)
tagain:
#endif
        x=prename_len(); 		/* Length of prefix */
        y=postname_len(); 		/* Length of suffix */
        for(pt=0;pt<x;pt++)   		/* build up name */
        {
           pre[pt]=get_ch1(); 		/* Get character */
        }
        pre[x]=0; /* Put a null at the end so we won't need to memset */

#if defined(WIN32)
        /* Oh Geeze... Have to avoid these magic names in Windows */
	if((x == 3) || (x == 4))
	{
		special[0] = (char)toupper((int)pre[0]);
		special[1] = (char)toupper((int)pre[1]);
		special[2] = (char)toupper((int)pre[2]);
		special[3] = (char)0;
        	if((my_strcasecmp(special,"CON")==0))
			goto tagain;
        	if((my_strcasecmp(special,"PRN")==0))
			goto tagain;
        	if((my_strcasecmp(special,"LPT")==0))
			goto tagain;
        	if((my_strcasecmp(special,"AUX")==0))
			goto tagain;
        	if((my_strcasecmp(special,"COM")==0))
			goto tagain;
        	if((my_strcasecmp(special,"NUL")==0))
			goto tagain;
	}
#endif
        /* Oh Geeze... Have to avoid these magic names in Windows */

        for(pt=0;pt<y;pt++) 		/* build up name */
        {
           post[pt]=get_ch1();  	/* get character */
        }
        post[y]=0; /* Put a null at the end so we won't need to memset */
	if(y==0)
	{
	   sprintf(namebuf,"%s%d",pre,filenum);
	   namebuf[strlen(namebuf)]=0;   /* put a null at the end so we won't need memset */
	}
	else
	{
	   sprintf(namebuf,"%s%d.%s",pre,filenum,post);
	   namebuf[strlen(namebuf)]=0; /* put a null at the end so we won't need memset */
        }
	name_srand(save);  /* Reset seed back to its original value !!! */
}

/*
 * Make a short file name that is [3->8] compliant, with pre 
 * lengths that are random, and consisting of characters that
 * have a proper frequency distribution.
 */
void
make_short_name(char *namebuf, int seed, int type)
{
    int x,pt,save;
	int skew = 0;
    char pre[NM_MAXNAMLEN]; 	/* Prefix buffer */
	if(dist_active==0) 		/* Only once please */
	{
	   init_dist();
	   dist_active=1;
	}
	save=name_gseed(); 		/* Save the seed !!! */
	switch (type)
	{
	   case 0: skew=101; 	/* 0 = Files. Note: Prime */
		break;
	   case 1: skew=997; 	/* 1 = Directories. Note: Prime*/
		break;
	   case 2: skew=2003; 	/* 2 = Symlinks. Note: Prime */
		break;
	   case 3: skew=3001; 	/* 3 = testdir name. Note: Prime */
		break;
	};

	/* 
	 * seed is the Unique file number.
	 * +1 to avoid confusing netmist_srand with a zero.
	 * +skew to move the directories, files and symlinks apart.
	 * +client_id to move the children apart.
 	 */
	name_srand(seed+1+skew+impersonate); 
        x=prename_len(); 		/* Length of prefix */
        for(pt=0;pt<x;pt++)   		/* build up name */
        {
           pre[pt]=get_ch1(); 		/* Get character */
        }
        pre[x]=0; /* Put a null at the end so we won't need to memset */
        sprintf(namebuf,"%s",pre);
	namebuf[x+1]=0;   /* put a null at the end so we won't need memset */
	name_srand(save);  /* Reset seed back to its original value !!! */
}

/*
 * Pick a length for the pre-dot part of the file name.
 */
static int
prename_len(void)
{
        int pre;
        pre=name_rand()%(PRENAME+1); /* Pick a random length */
        if(pre < MINPRENAME) 		/* Enforce minimum */
           pre=MINPRENAME;
        return(pre);
}

/*
 * Pick a length for the post-dot part of the file name.
 */
static int
postname_len(void)
{
        int post;
        post=name_rand()%(POSTNAME+1); /* Pick a random length */
        if(post<MINPOSTNAME)		/* Enforce minimum */
          post=MINPOSTNAME;
        return(post);
}
/*
 * Initialize the distribution table for later use in the
 * frequency distribution mechanism.
 */
static void
init_dist(void)
{
        unsigned int i;
        int j, pos=0;

        for (i=0;i<strlen(ch);i++) /* Letters in the alphabet */
        {
           for(j=0;j<freq[i];j++)  /* Frequency information */
           {
                table[pos]=ch[i];  /* Fill in the table */
                pos++;
           }
        }
        t_size=pos; /* Actual size of the active table plus one. */
}

/*
 * Return a character with the language character frequency distribution.
 */
static char
get_ch1(void)
{
        char val;		/* char to return */
        int where;		/* index into table */
        where=(name_rand()%t_size); /* Random index into generated table */
        val=table[where]; 		/* Pick up the character */
        return(val);		/* Return char */
}

long cur_seed;

void
netmist_srand( int seed )
{
#if defined(WIN32)
	n_spec_srand(seed);
#else
	cur_seed = seed;
	srand(seed);
#endif
}

int
netmist_gseed(void)
{
	return(cur_seed); 
}

int 
netmist_rand(void)
{
	int ret;
#if defined(WIN32)
	cur_seed = n_spec_rand();
#else
	cur_seed = rand();
#endif
	ret=(int) rotateR((unsigned int)cur_seed,4);
	return(ret & 0x7fffffff);
}

unsigned long long 
netmist_llrand(void)
{
	long long ret;
#ifdef MERSENNE
	/* 
	 * Use the Mersenne Twister. This is a much better random number
 	 * generator. The standard LCA (Linear Congruential Algorithm) 
	 * generators stink. The low order bits are not changing frequently.
	 */
	ret=genrand64_int64();
#else
	ret = lrand48();
	ret = ret << 31;
	ret |= lrand48();
#endif
	return(ret);
}


#endif /* MVFDN */

/* Allocate space for file table objects */
void
init_space(void)
{
	int size,i;
	size = sizeof(struct file_object )* client_dirs * client_files;
	for(i=0;i<NUM_OP_TYPES;i++)
	{
	   op_X_array[i] = (struct file_object *) MALLOC(size);
	   if (op_X_array[i])
		   bzero(op_X_array[i], size); /*zero out the memory*/
	}
}

void
dump_file_objects(FILE *out)
{
	int size,i,j;
	struct file_object *file_ptr;
	size = client_dirs * client_files;

	for(i=0;i<NUM_OP_TYPES;i++) /* for all op types */
	{
	   file_ptr = (struct file_object *) op_X_array[i]; 
	   for(j=0;j<size;j++) /* All files in this op type */
	   {
		if(file_ptr && (file_ptr->access_count > 0) && 
		   (file_ptr->filename[0] != 0))
		{
		  fprintf(out,"%8ld %s\n",file_ptr->access_count, 
			file_ptr->filename);
		  fflush(out);
		}
		file_ptr++;
	   }
	}
}

/* Initialize all of the variables that are used to scale the
 * file sizes, directories, and number of files. Must be 
 * called very early.
 */
void
init_vars(void)
{
 	/* Number of directories per proc. Default 10 */
	client_dirs = 10;  
	/* files per directory. Default 100 */
	client_files = 100;

	/* Number of directories that contain files that will be populated */
	/* write, write_rand, read, read_rand, rwm, mmap_read, mmap_write, */
	/* append, copyfile . Default 9 */
	client_files_pop=11; 

	/* Active space on client */
	client_dirs_pop = (client_dirs * client_files * client_files_pop);

	/* Files read by each proc */
	client_files_rpop = (client_dirs * client_files);

	/* Files written by each proc */
	client_files_wpop = (client_dirs * client_files );

	/* Files populated by each proc */
	max_client_dirs_pop  = (client_dirs * client_files * (client_files_pop+1));
}

void
re_init_vars(void)
{
	/* Active space on client */
	client_dirs_pop = (client_dirs * client_files * client_files_pop);

	/* Files read by each proc */
	client_files_rpop = (client_dirs * client_files);

	/* Files written by each proc */
	client_files_wpop = (client_dirs * client_files );

	/* Files populated by each proc */
	max_client_dirs_pop  = (client_dirs * client_files * (client_files_pop+1));
}


/*
 * Buckets: (Based on a Netapp internal consensus)
 *     0       1       2       3      4
 *  <=20us  <=40us  <=60us  <=80us  <=100us
 *
 *    5        6       7       8       9
 * <=200us  <=400us <=600us <=88us  <=1ms
 *
 *   10       11      12      13      14
 * <=2ms    <=4ms   <=6ms   <=8ms   <=10ms
 *
 *   15       16      17      18      19
 * <=12ms   <=14ms  <=16ms  <=18ms  <=20ms
 *
 *   20       21      22      23      24
 * <=20ms   <=40ms  <=60ms  <=80ms  <=100ms
 *
 *   25       26      27      28      29
 * <=200ms  <=400ms <=600ms <=800ms <=1s
 *
 *   30       31      32      33      34
 * <=2s     <=4s    <=6s    <=8s    <=10s
 *
 *   35       36      37      38      39
 * <=20s    <=30s   <=60    <=90s    >90
 */
void
hist_insert(double my_value)
{
        int k;
        long long value;

        /* Convert to micro-seconds */
        value = (long long)(my_value * 1000000);
        for(k=0;k<BUCKETS;k++)
        {
           if(k < (BUCKETS-1))
           {
             if(value <= bucket_val[k])
             {
                buckets[k]++;
                break;
             }
           }
           else
           {
             if(value > bucket_val[k])
             {
                buckets[k]++;
                break;
             }
           }
        }
}
void
dump_hist(void)
{

#ifndef NO_PRINT_LLD
   printf("Band 1: ");
   printf(" 20us:%-7.1lld ",global_buckets[0]);
   printf(" 40us:%-7.1lld ",global_buckets[1]);
   printf(" 60us:%-7.1lld ",global_buckets[2]);
   printf(" 80us:%-7.1lld ",global_buckets[3]);
   printf("100us:%-7.1lld \n",global_buckets[4]);

   printf("Band 2: ");
   printf("200us:%-7.1lld ",global_buckets[5]);
   printf("400us:%-7.1lld ",global_buckets[6]);
   printf("600us:%-7.1lld ",global_buckets[7]);
   printf("800us:%-7.1lld ",global_buckets[8]);
   printf("  1ms:%-7.1lld \n",global_buckets[9]);

   printf("Band 3: ");
   printf("  2ms:%-7.1lld ",global_buckets[10]);
   printf("  4ms:%-7.1lld ",global_buckets[11]);
   printf("  6ms:%-7.1lld ",global_buckets[12]);
   printf("  8ms:%-7.1lld ",global_buckets[13]);
   printf(" 10ms:%-7.1lld \n",global_buckets[14]);

   printf("Band 4: ");
   printf(" 12ms:%-7.1lld ",global_buckets[15]);
   printf(" 14ms:%-7.1lld ",global_buckets[16]);
   printf(" 16ms:%-7.1lld ",global_buckets[17]);
   printf(" 18ms:%-7.1lld ",global_buckets[18]);
   printf(" 20ms:%-7.1lld \n",global_buckets[19]);

   printf("Band 5: ");
   printf(" 40ms:%-7.1lld ",global_buckets[20]);
   printf(" 60ms:%-7.1lld ",global_buckets[21]);
   printf(" 80ms:%-7.1lld ",global_buckets[22]);
   printf("100ms:%-7.1lld \n",buckets[23]);

   printf("Band 6: ");
   printf("200ms:%-7.1lld ",global_buckets[24]);
   printf("400ms:%-7.1lld ",global_buckets[25]);
   printf("600ms:%-7.1lld ",global_buckets[26]);
   printf("800ms:%-7.1lld ",global_buckets[27]);
   printf("   1s:%-7.1lld \n",global_buckets[28]);

   printf("Band 7: ");
   printf("   2s:%-7.1lld ",global_buckets[29]);
   printf("   4s:%-7.1lld ",global_buckets[30]);
   printf("   6s:%-7.1lld ",global_buckets[31]);
   printf("   8s:%-7.1lld ",global_buckets[32]);
   printf("  10s:%-7.1lld \n",global_buckets[33]);

   printf("Band 8: ");
   printf("  20s:%-7.1lld ",global_buckets[34]);
   printf("  40s:%-7.1lld ",global_buckets[35]);
   printf("  60s:%-7.1lld ",global_buckets[36]);
   printf("  80s:%-7.1lld ",global_buckets[37]);
   printf(" 120s:%-7.1lld \n",global_buckets[38]);

   printf("Band 9: ");
   printf("120+s:%-7.1lld \n\n",global_buckets[39]);
#else
   printf("Band 1: ");
   printf(" 20us:%-7.1ld ",global_buckets[0]);
   printf(" 40us:%-7.1ld ",global_buckets[1]);
   printf(" 60us:%-7.1ld ",global_buckets[2]);
   printf(" 80us:%-7.1ld ",global_buckets[3]);
   printf("100us:%-7.1ld \n",global_buckets[4]);

   printf("Band 2: ");
   printf("200us:%-7.1ld ",global_buckets[5]);
   printf("400us:%-7.1ld ",global_buckets[6]);
   printf("600us:%-7.1ld ",global_buckets[7]);
   printf("800us:%-7.1ld ",global_buckets[8]);
   printf("  1ms:%-7.1ld \n",global_buckets[9]);

   printf("Band 3: ");
   printf("  2ms:%-7.1ld ",global_buckets[10]);
   printf("  4ms:%-7.1ld ",global_buckets[11]);
   printf("  6ms:%-7.1ld ",global_buckets[12]);
   printf("  8ms:%-7.1ld ",global_buckets[13]);
   printf(" 10ms:%-7.1ld \n",global_buckets[14]);

   printf("Band 4: ");
   printf(" 12ms:%-7.1ld ",global_buckets[15]);
   printf(" 14ms:%-7.1ld ",global_buckets[16]);
   printf(" 16ms:%-7.1ld ",global_buckets[17]);
   printf(" 18ms:%-7.1ld ",global_buckets[18]);
   printf(" 20ms:%-7.1ld \n",global_buckets[19]);

   printf("Band 5: ");
   printf(" 40ms:%-7.1ld ",global_buckets[20]);
   printf(" 60ms:%-7.1ld ",global_buckets[21]);
   printf(" 80ms:%-7.1ld ",global_buckets[22]);
   printf("100ms:%-7.1ld \n",global_buckets[23]);

   printf("Band 6: ");
   printf("200ms:%-7.1ld ",global_buckets[24]);
   printf("400ms:%-7.1ld ",global_buckets[25]);
   printf("600ms:%-7.1ld ",global_buckets[26]);
   printf("800ms:%-7.1ld ",global_buckets[27]);
   printf("   1s:%-7.1ld \n",global_buckets[28]);

   printf("Band 7: ");
   printf("   2s:%-7.1ld ",global_buckets[29]);
   printf("   4s:%-7.1ld ",global_buckets[30]);
   printf("   6s:%-7.1ld ",global_buckets[31]);
   printf("   8s:%-7.1ld ",global_buckets[32]);
   printf("  10s:%-7.1ld \n",global_buckets[33]);

   printf("Band 8: ");
   printf("  20s:%-7.1ld ",global_buckets[34]);
   printf("  40s:%-7.1ld ",global_buckets[35]);
   printf("  60s:%-7.1ld ",global_buckets[36]);
   printf("  80s:%-7.1ld ",global_buckets[37]);
   printf(" 120s:%-7.1ld \n",global_buckets[38]);

   printf("Band 9: ");
   printf("120+s:%-7.1ld \n",global_buckets[39]);
   fflush(stdout);
#endif
   if(log_flag)
   {
#ifndef NO_PRINT_LLD
      fprintf(log_out,"Band 1: ");
      fprintf(log_out," 20us:%-7.1lld ",global_buckets[0]);
      fprintf(log_out," 40us:%-7.1lld ",global_buckets[1]);
      fprintf(log_out," 60us:%-7.1lld ",global_buckets[2]);
      fprintf(log_out," 80us:%-7.1lld ",global_buckets[3]);
      fprintf(log_out,"100us:%-7.1lld \n",global_buckets[4]);

      fprintf(log_out,"Band 2: ");
      fprintf(log_out,"200us:%-7.1lld ",global_buckets[5]);
      fprintf(log_out,"400us:%-7.1lld ",global_buckets[6]);
      fprintf(log_out,"600us:%-7.1lld ",global_buckets[7]);
      fprintf(log_out,"800us:%-7.1lld ",global_buckets[8]);
      fprintf(log_out,"  1ms:%-7.1lld \n",global_buckets[9]);

      fprintf(log_out,"Band 3: ");
      fprintf(log_out,"  2ms:%-7.1lld ",global_buckets[10]);
      fprintf(log_out,"  4ms:%-7.1lld ",global_buckets[11]);
      fprintf(log_out,"  6ms:%-7.1lld ",global_buckets[12]);
      fprintf(log_out,"  8ms:%-7.1lld ",global_buckets[13]);
      fprintf(log_out," 10ms:%-7.1lld \n",global_buckets[14]);

      fprintf(log_out,"Band 4: ");
      fprintf(log_out," 12ms:%-7.1lld ",global_buckets[15]);
      fprintf(log_out," 14ms:%-7.1lld ",global_buckets[16]);
      fprintf(log_out," 16ms:%-7.1lld ",global_buckets[17]);
      fprintf(log_out," 18ms:%-7.1lld ",global_buckets[18]);
      fprintf(log_out," 20ms:%-7.1lld \n",global_buckets[19]);

      fprintf(log_out,"Band 5: ");
      fprintf(log_out," 40ms:%-7.1lld ",global_buckets[20]);
      fprintf(log_out," 60ms:%-7.1lld ",global_buckets[21]);
      fprintf(log_out," 80ms:%-7.1lld ",global_buckets[22]);
      fprintf(log_out,"100ms:%-7.1lld \n",global_buckets[23]);

      fprintf(log_out,"Band 6: ");
      fprintf(log_out,"200ms:%-7.1lld ",global_buckets[24]);
      fprintf(log_out,"400ms:%-7.1lld ",global_buckets[25]);
      fprintf(log_out,"600ms:%-7.1lld ",global_buckets[26]);
      fprintf(log_out,"800ms:%-7.1lld ",global_buckets[27]);
      fprintf(log_out,"   1s:%-7.1lld \n",global_buckets[28]);

      fprintf(log_out,"Band 7: ");
      fprintf(log_out,"   2s:%-7.1lld ",global_buckets[29]);
      fprintf(log_out,"   4s:%-7.1lld ",global_buckets[30]);
      fprintf(log_out,"   6s:%-7.1lld ",global_buckets[31]);
      fprintf(log_out,"   8s:%-7.1lld ",global_buckets[32]);
      fprintf(log_out,"  10s:%-7.1lld \n",global_buckets[33]);

      fprintf(log_out,"Band 8: ");
      fprintf(log_out,"  20s:%-7.1lld ",global_buckets[34]);
      fprintf(log_out,"  40s:%-7.1lld ",global_buckets[35]);
      fprintf(log_out,"  60s:%-7.1lld ",global_buckets[36]);
      fprintf(log_out,"  80s:%-7.1lld ",global_buckets[37]);
      fprintf(log_out," 120s:%-7.1lld \n",global_buckets[38]);

      fprintf(log_out,"Band 9: ");
      fprintf(log_out,"120+s:%-7.1lld \n",global_buckets[39]);
#else
      fprintf(log_out,"Band 1: ");
      fprintf(log_out," 20us:%-7.1ld ",global_buckets[0]);
      fprintf(log_out," 40us:%-7.1ld ",global_buckets[1]);
      fprintf(log_out," 60us:%-7.1ld ",global_buckets[2]);
      fprintf(log_out," 80us:%-7.1ld ",global_buckets[3]);
      fprintf(log_out,"100us:%-7.1ld \n",global_buckets[4]);

      fprintf(log_out,"Band 2: ");
      fprintf(log_out,"200us:%-7.1ld ",global_buckets[5]);
      fprintf(log_out,"400us:%-7.1ld ",global_buckets[6]);
      fprintf(log_out,"600us:%-7.1ld ",global_buckets[7]);
      fprintf(log_out,"800us:%-7.1ld ",global_buckets[8]);
      fprintf(log_out,"  1ms:%-7.1ld \n",global_buckets[9]);

      fprintf(log_out,"Band 3: ");
      fprintf(log_out,"  2ms:%-7.1ld ",global_buckets[10]);
      fprintf(log_out,"  4ms:%-7.1ld ",global_buckets[11]);
      fprintf(log_out,"  6ms:%-7.1ld ",global_buckets[12]);
      fprintf(log_out,"  8ms:%-7.1ld ",global_buckets[13]);
      fprintf(log_out," 10ms:%-7.1ld \n",global_buckets[14]);

      fprintf(log_out,"Band 4: ");
      fprintf(log_out," 12ms:%-7.1ld ",global_buckets[15]);
      fprintf(log_out," 14ms:%-7.1ld ",global_buckets[16]);
      fprintf(log_out," 16ms:%-7.1ld ",global_buckets[17]);
      fprintf(log_out," 18ms:%-7.1ld ",global_buckets[18]);
      fprintf(log_out," 20ms:%-7.1ld \n",global_buckets[19]);

      fprintf(log_out,"Band 5: ");
      fprintf(log_out," 40ms:%-7.1ld ",global_buckets[20]);
      fprintf(log_out," 60ms:%-7.1ld ",global_buckets[21]);
      fprintf(log_out," 80ms:%-7.1ld ",global_buckets[22]);
      fprintf(log_out,"100ms:%-7.1ld \n",global_buckets[23]);

      fprintf(log_out,"Band 6: ");
      fprintf(log_out,"200ms:%-7.1ld ",global_buckets[24]);
      fprintf(log_out,"400ms:%-7.1ld ",global_buckets[25]);
      fprintf(log_out,"600ms:%-7.1ld ",global_buckets[26]);
      fprintf(log_out,"800ms:%-7.1ld ",global_buckets[27]);
      fprintf(log_out,"   1s:%-7.1ld \n",global_buckets[28]);

      fprintf(log_out,"Band 7: ");
      fprintf(log_out,"   2s:%-7.1ld ",global_buckets[29]);
      fprintf(log_out,"   4s:%-7.1ld ",global_buckets[30]);
      fprintf(log_out,"   6s:%-7.1ld ",global_buckets[31]);
      fprintf(log_out,"   8s:%-7.1ld ",global_buckets[32]);
      fprintf(log_out,"  10s:%-7.1ld \n",global_buckets[33]);

      fprintf(log_out,"Band 8: ");
      fprintf(log_out,"  20s:%-7.1ld ",global_buckets[34]);
      fprintf(log_out,"  40s:%-7.1ld ",global_buckets[35]);
      fprintf(log_out,"  60s:%-7.1ld ",global_buckets[36]);
      fprintf(log_out,"  80s:%-7.1ld ",global_buckets[37]);
      fprintf(log_out," 120s:%-7.1ld \n",global_buckets[38]);

      fprintf(log_out,"Band 9: ");
      fprintf(log_out,"120+s:%-7.1ld \n",global_buckets[39]);
      fflush(log_out);
#endif
   }
}

#define DEDUP_CHUNK 512
void
make_non_dedup(long *buf, int size)
{
   	int i,j;
	int where,start;
	int percent_compress;
	int switch_count;
	long patt,compress_patt;

	patt=netmist_rand() & 0x7fffffff;
	percent_compress = my_mix_table[my_workload].percent_compress;

	switch_count = (DEDUP_CHUNK*percent_compress)/100;

	compress_patt = (long)netmist_rand() & 0xffffffff;

   	for (i = 0; i < size; i += DEDUP_CHUNK) {
		where = i / sizeof(long);

		/* Set first long in 512 byte buffer */
		buf[where] = (long)netmist_rand() & 0xffffffff;
		/* Now create compressibility within the 512 byte */
		start = i + sizeof(long);
		for (j = start; j < start + DEDUP_CHUNK; j += sizeof(long)) {
			where = j/sizeof(long);

			if( (j%DEDUP_CHUNK) <= switch_count )
				buf[where] = (long)compress_patt;
			else
				buf[where] ^= (patt * (j+i));
		}
	}
}

/*
 * Section for the parser. This parser pulls out the configuration 
 * information using a token based scheme.
 */

#define NUM_INPUT_TOKENS 8
#define MAX_INPUT_TOKENLEN 255
#define T_CLIENT_NAME 0
#define T_USER_NAME 1
#define T_PASSWORD 2
#define T_WORK_DIR 3
#define T_EXEC_PATH 4
#define T_WORKLOAD 5
#define T_OPRATE 6
#define T_INSTANCES 7

struct token {
	int optional;     	   /* Set to 1 to ignore if token is not present */
	char tname[MAX_INPUT_TOKENLEN];  /* String to search for */
	char tvalue[MAX_INPUT_TOKENLEN]; /* Resulting value */
} token_objects[NUM_INPUT_TOKENS] = 
	   {{0,"Clientname=",""},
	    {0,"Username=",  ""},
	    {1,"Password=",  ""},
	    {0,"Workdir=",   ""},
	    {0,"Execpath=",  ""},
	    {0,"Workload=",  ""},
	    {1,"Oprate=",    ""},
	    {1,"Instances=",    ""}};

/*
 * Parse an individual token and put result in destination 
 */
int 
parse_token(char *inputlinep, char *searchtokenp, char *destinationp)
{
	char *tokenp,*cp,*cd;
	int count = 0;
	tokenp = (char *)my_strcasestr(inputlinep,searchtokenp);
	if(tokenp)
	{
		cp = tokenp+strlen(searchtokenp);
		cd = destinationp;
		while((*cp != ' ') && (cp != inputlinep+strlen(inputlinep)) && count < MAX_INPUT_TOKENLEN )
		{
		   count++;
		   *cd++=*cp++;
		}
		*cd=0;
		return(0);
	}
	else
	  *destinationp=0;
	return(1);
}

/*
 * Parse all known tokens from the line.
 */
int
parse_all_tokens(char *input )
{
	int i,ret;
	for(i=0;i<NUM_INPUT_TOKENS;i++)
	{
	   ret = parse_token(input,token_objects[i].tname,token_objects[i].tvalue);
	   if(ret && !token_objects[i].optional)
	   {
	      printf("Parsing error while looking for %s\n",token_objects[i].tname);
   	      fflush(stdout);
	      return(1);
	   }
	}
	return(0);
}
	  
/*
 * Validate all tokens are there, and place their values in the
 * client_object table.
 */

int
validate_all_tokens(int client_num)
{
	/* Check for Clientname */
	if(strlen(token_objects[T_CLIENT_NAME].tvalue)==0)
	{
		printf("Clientname is missing\n");
   	        fflush(stdout);
		return(1);
	}
	strcpy(client_obj[client_num].client_name,token_objects[T_CLIENT_NAME].tvalue);

	if(strlen(token_objects[T_USER_NAME].tvalue)==0)
	{
		printf("Username is missing\n");
   	        fflush(stdout);
		return(1);
	}
	strcpy(client_obj[client_num].client_dom_user,token_objects[T_USER_NAME].tvalue);

	if(strlen(token_objects[T_PASSWORD].tvalue) != 0)
	{
	
		strcpy(client_obj[client_num].client_password,token_objects[T_PASSWORD].tvalue);
	}
	else 
		strcpy(client_obj[client_num].client_password,"_"); /* invalid pass */

	if(strlen(token_objects[T_WORK_DIR].tvalue)==0)
	{
		printf("Workdir is missing\n");
   	        fflush(stdout);
		return(1);
	}
	strcpy(client_obj[client_num].client_workdir,token_objects[T_WORK_DIR].tvalue);

	if(strlen(token_objects[T_EXEC_PATH].tvalue)==0)
	{
		printf("Execpath is missing\n");
   	        fflush(stdout);
		return(1);
	}
	strcpy(client_obj[client_num].client_execdir,token_objects[T_EXEC_PATH].tvalue);

	if(strlen(token_objects[T_WORKLOAD].tvalue)==0)
	{
		printf("Workload is missing\n");
   	        fflush(stdout);
		return(1);
	}
	strcpy(client_obj[client_num].client_workload_name,token_objects[T_WORKLOAD].tvalue);

	if(strlen(token_objects[T_OPRATE].tvalue) != 0)
	{
	
		sscanf(token_objects[T_OPRATE].tvalue,"%lf",&client_obj[client_num].op_rate);
	}
	else 
		client_obj[client_num].op_rate=0.0;

	if(strlen(token_objects[T_INSTANCES].tvalue) != 0)
	{
	
		client_obj[client_num].instances=atoi(token_objects[T_INSTANCES].tvalue);
	}
	else
		client_obj[client_num].instances=0;

	return(0);
}

/*
 * There is a check_sum of the workloads. It is printed 
 * out, at the end of a run. If one wishes to have this
 * result submitted for publication, then one acquires a
 * magic key from the publisher. This key is then XOR-d with
 * the check_sum and submitted to the publisher for validation.
 * The magic key, from the publisher, is delivered via a 
 * secure exchange (cypher).
 */
unsigned short
check_sum()
{
	int i,j;
	unsigned short k = 0;
	unsigned char *ptr;
	unsigned int lic_key;
	char my_lic_file[MAXNAME];
	FILE *lic;
	int got = 0;
	
	for(i=0;i<NUM_WORK_OBJ;i++)
	{
	   ptr = (unsigned char *)&my_mix_table[i];
	   for(j=0;j<sizeof(struct mix_table)/sizeof(unsigned char);j++)
	   {
		k+=(unsigned short)*ptr;
		ptr++;
	   }
	}
	/* Look in old /tmp directory first */
#if defined(WIN32)
	sprintf(my_lic_file,"c:\\tmp\\%s",NETMIST_LIC_FILE);
#else
	sprintf(my_lic_file,"/tmp/%s",NETMIST_LIC_FILE);
#endif
	if(access(my_lic_file,F_OK) == 0)
   	{
	   got = 1;
	}
	if(got == 0)
	{
		/* Try the current working directory */
#if defined(WIN32)
	   sprintf(my_lic_file,"%s",NETMIST_LIC_FILE);
#else
	   sprintf(my_lic_file,"%s",NETMIST_LIC_FILE);
#endif
	   if(access(my_lic_file,F_OK) == 0)
	   {
	      got = 1;
	   }
	}
	if(got == 1)
   	{
	   lic = fopen(my_lic_file,"r");
	   if(lic == (FILE *) NULL)
	   {
		if(cdebug)
		   printf("Open of license %s failed\n",my_lic_file);
		return(k); /* Public finger print */
	   }
	}else{

		return(k); /* Public finger print */
	}
	fscanf(lic,"LICENSE KEY %u\n",&lic_key);
	licensed=1;
	k=k^lic_key;
	fclose(lic);
	return(k); /* Licensed finger print */
}

int
license_check(void)
{
	FILE *lic;
	unsigned int lic_key;
	char my_lic_file[MAXNAME];
	int got = 0;

	/* Look in old /tmp directory first */
#if defined(WIN32)
	sprintf(my_lic_file,"c:\\tmp\\%s",NETMIST_LIC_FILE);
#else
	sprintf(my_lic_file,"/tmp/%s",NETMIST_LIC_FILE);
#endif
	if(access(my_lic_file,F_OK) == 0)
   	{
	   got = 1;
	}
	if(got == 0)
	{
		/* Try the current working directory */
#if defined(WIN32)
	   sprintf(my_lic_file,"%s",NETMIST_LIC_FILE);
#else
	   sprintf(my_lic_file,"%s",NETMIST_LIC_FILE);
#endif
	   if(access(my_lic_file,F_OK) == 0)
	   {
	      got = 1;
	   }
	}
	if(got == 1)
	{
	   lic = fopen(my_lic_file,"r");
	   if(lic == (FILE *) NULL)
	   {
		if(cdebug)
		   printf("Open of license %s failed\n",my_lic_file);
		return(0);
	   }
	}else{
		return(0);
	}

	fscanf(lic,"LICENSE KEY %u\n",&lic_key);
	licensed=1;
	fclose(lic);
	return(licensed);
}

#if defined(PDSM_R_REGISTRY)
void
set_pdsm_attributes(void)
{
	/* Copy inbound mix table (workloads) update, into its resting place */
	if(pdsm_control->set_workload)
	{
	   modified_run=1;
	   memcpy(my_mix_table,pdsm_control->work_load_out,sizeof(struct mix_table)* NUM_WORK_OBJ);
	   create_tests(test_ops,num_clients);
	   init_dist_tables();
	   pdsm_control->set_workload = 0;
	}
	if(pdsm_control->set_op_rate_set)
	{
	   modified_run=1;
	   op_rate = client_obj[c_chid].op_rate = pdsm_control->set_op_rate;
	   pdsm_control->set_op_rate_set=0;
	}
}
#endif

#if !defined(WIN32)
#include <sys/mman.h>
#endif

#ifdef HAVE_ANSIC_C
void
generic_mmap_write(struct file_object *fp)
#else
void
generic_mmap_write(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	HANDLE fd, hMap=NULL;
	LARGE_INTEGER largeoffset;
	long long pagesize = 0;
#else
	int fd;
	long long pagesize; 
	int mflags,prot;
#endif
	char *buf=NULL,*pa=NULL, *err_str=NULL;
	double lat = 0.0;
	int trans;
	size_t filebytes;
	int flags = 0;
	int flags2 = 0;
	int use_cached_fd = 0;


	if(cdebug) 
	{
		fprintf(newstdout,"Doing mmap write op\n");
		fflush(newstdout);
	}
	buf = main_buf;
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
	if(fp->file_desc)
	{
	  use_cached_fd = 1;
	  fd = fp->file_desc;
	  if(tracedebug)
	  	fprintf(tracefile,"%d,%lu,%s,mmap_write,Using cached fd,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	}
	else
	{
#if defined(WIN32)
	  fd = CreateFile(fp->filename,
			GENERIC_WRITE | GENERIC_READ,
			FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE,
			NULL,
			OPEN_ALWAYS,	 // create a new file if it does not exist, otherwise open the old file
			FILE_ATTRIBUTE_NORMAL|flags,
			NULL);
	  if (fd == INVALID_HANDLE_VALUE)
	  {
		  c_eval=GetLastError();
		  err_str = win32_strerror(c_eval);
#else
	  fd=I_OPEN(fp->filename,O_CREAT|O_RDWR|flags,0666);
	  if(fd <0)
	  {
		  c_eval=errno;
		  err_str = strerror(c_eval);
#endif
		
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for mmap write op failed. Filename = %s Error: %s Error value: %d\n",
			fp->filename,err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(109);
	  }
	  if( (cached_open_count+10 < max_open) && !(fp->file_desc))
	  {
	     cached_open_count++;
	     fp->file_desc = fd;
	     use_cached_fd = 1;
	  }
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,mmap_write,Opened file,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	total_file_ops +=1;
	work_obj[my_workload].total_file_ops+=1;
#if defined(WIN32)
	largeoffset.QuadPart = fp->mmapwrite_offset;
    	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	I_LSEEK(fd,fp->mmapwrite_offset,SEEK_SET);
#endif
	/* Select write size to use */
	trans = write_dist_array[(netmist_rand() % RW_ARRAY)];

	make_non_dedup((long *)buf,trans);
        pagesize=(long long)getpagesize();
#if !defined(WIN32)
#if defined(_solaris_)
        mflags=MAP_SHARED;
#else
        mflags=MAP_FILE|MAP_SHARED;
#endif
        prot=PROT_WRITE;
#endif

    	filebytes = 
	  ((fp->Original_fsize*(long long)1024) + (pagesize-1LL)) & ~(pagesize-1LL);

#if defined(WIN32)
	hMap = CreateFileMapping(fd,    // use paging file
			NULL,           // default security
			PAGE_READWRITE, // read/write access
			0,             // maximum object size (high-order DWORD)
			0,             // maximum object size (low-order DWORD)
			NULL);         // name of mapping object
	if (hMap == NULL)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
 	    	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
	    	fprintf(newstdout,"Mmap_Write CreateFileMapping on file %s Error: %s Error value: %d\n",fp->filename,err_str,c_eval);
	    	fflush(newstdout);
		c_eflag++;
		CloseHandle(fd);
		R_exit(110);
	}
	if (hMap != NULL)
	{
		pa = (char*)MapViewOfFile( hMap,     // handle to map object
			FILE_MAP_ALL_ACCESS,         // read/write permission
			0,
			0,
			0);
	}
	if (pa == NULL)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else
    	pa = (char *)I_MMAP( ((char *)0),filebytes, prot,
                    mflags, fd, 0);

	if(pa == (char *)-1)
	{
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
		
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
	        fprintf(newstdout,"Mmap Write failed on file %s Error: %s\n",fp->filename,err_str);
	        fflush(newstdout);
		c_eflag++;
	     	if(use_cached_fd)
		{
	           cached_open_count--;
		   fp->file_desc=0;
		}
#if defined(WIN32)
		if (hMap != NULL)
			CloseHandle(hMap);
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(110);
	}
	if(trans < 0)
		trans = 0;
	if((fp->mmapwrite_offset > filebytes) || ((long long)(fp->mmapwrite_offset) < (long long)0))
		fp->mmapwrite_offset = 0;
	if(fp->mmapwrite_offset+trans > filebytes)
		fp->mmapwrite_offset = 0;

	if(fp->mmapwrite_offset+trans > filebytes)
		/*limit the io size, ensure the operation won't go beyond the valid memory size*/
		trans = filebytes - fp->mmapwrite_offset; 

	if(cdebug)
	{
 	  fprintf(newstdout,"Trans %d Size %lld Offset %lld\n",trans,(long long)filebytes,(long long)(fp->mmapwrite_offset));
	  fflush(newstdout);
	}
	memcpy(&pa[(long long)(fp->mmapwrite_offset)],buf,(size_t)trans);
	fp->mmapwrite_offset += trans;
	if((fp->mmapwrite_offset/(long long)1024) >= (fp->Original_fsize))
	   fp->mmapwrite_offset=0;
	fp->op_count++;

	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,mmap_write,Mmap write,%s,,%lld,%d,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,(long long)fp->mmapwrite_offset,trans,
			fp->access_count);
	/* global (across all files in this work object) write count */
	work_obj[my_workload].mmap_write_op_count++;

	total_file_ops += 1;
	work_obj[my_workload].total_file_ops+=1;

	total_write_bytes += (double)(trans);

	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].mmap_write_op_time += lat;
	}
	work_obj[my_workload].mmap_write_op_count++;
#if defined(WIN32)
	UnmapViewOfFile(pa);
	CloseHandle(hMap);
#else
	munmap((void *)pa,filebytes);
#endif
	if(!use_cached_fd)
	{
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		if(tracedebug)
		{
			fprintf(tracefile,"%d,%lu,%s,mmap_write,Closed file,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
		}
	}
	total_file_ops += 1.0;
	work_obj[my_workload].total_file_ops+=1.0;
}	

#ifdef HAVE_ANSIC_C
void
generic_mmap_read(struct file_object *fp)
#else
void
generic_mmap_read(fp)
struct file_object *fp;
#endif
{
#if defined(WIN32)
	HANDLE fd, hMap=NULL;
	LARGE_INTEGER largeoffset;
	long long pagesize = 0;
#else
	int fd;
	long long pagesize;
	int mflags,prot;
#endif
	char *buf, *err_str=NULL;
	char *pa = NULL;
	double lat = 0.0;
	int trans;
	long long filebytes;
	int flags = 0;
	int flags2 = 0;
	int use_cached_fd = 0;


	if(cdebug) 
	{
		fprintf(newstdout,"Doing mmap read op\n");
		fflush(newstdout);
	}
	buf = main_buf;
	if(fp->file_desc)
	{
	  use_cached_fd = 1;
	  fd = fp->file_desc;
	  if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,mmap_read,Using cached fd,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	}
	else
	{
#if defined(WIN32)
	  flags|=FILE_ATTRIBUTE_NORMAL|FILE_FLAG_POSIX_SEMANTICS;
	  fd = CreateFile(fp->filename,
			GENERIC_WRITE | GENERIC_READ,
			FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
			NULL,
			OPEN_ALWAYS,	 // create a new file if it does not exist, otherwise open the old file
			flags,
			NULL);
	  if (fd == INVALID_HANDLE_VALUE)
	  {
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else
	  fd=I_OPEN(fp->filename,O_CREAT|O_RDWR|flags,0666);
	  if(fd <0)
	  {
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
	 
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open for mmap read op failed. Filename = %s Error: %s Error value: %d\n", 
			fp->filename,err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(109);
	  }
	  if( (cached_open_count+10 < max_open) && !(fp->file_desc))
	  {
	     	cached_open_count++;
	     	fp->file_desc = fd;
	     	use_cached_fd = 1;
	  }
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,mmap_read,Opened file,%s,,,,%ld,O_DIRECT=%d O_SYNC=%d\n",
			client_id,trace_seq++,trace_states[tracestate],fp->filename,
			fp->access_count,
			((flags2 & 1)?1:0),
			((flags2 & 2)?1:0));
	total_file_ops +=1;
	work_obj[my_workload].total_file_ops+=1;
#if defined(WIN32)
	largeoffset.QuadPart = fp->mmapread_offset;
    	SetFilePointerEx(fd, largeoffset, NULL, FILE_BEGIN);
#else
	I_LSEEK(fd,fp->mmapread_offset,SEEK_SET);
#endif

	/* Select read size to use */
	trans = read_dist_array[(netmist_rand() % RW_ARRAY)];
    	pagesize=getpagesize();
#if !defined(WIN32)
#if defined(_solaris_)
        mflags=MAP_SHARED;
#else
        mflags=MAP_FILE|MAP_SHARED;
#endif
        prot=PROT_READ;
#endif

    filebytes = ( (fp->Original_fsize*(long long)1024) + (pagesize-1)) & ~(pagesize-1);
#if defined(WIN32)
	hMap = CreateFileMapping(fd,    		// use paging file
		NULL,                   // default security
		PAGE_READONLY,          // read access
		(DWORD)0,     		// maximum object size (high-order DWORD)
		(DWORD)0,     		// maximum object size (low-order DWORD)
		NULL);                  // name of mapping object
	if (hMap == NULL)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
 	    	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
	    	fprintf(newstdout,"Mmap_Read CreateFileMapping on file %s Error: %s Error value: %d\n",fp->filename,err_str,c_eval);
	    	fflush(newstdout);
		c_eflag++;
		CloseHandle(fd);
		R_exit(110);
	}
	if (hMap != NULL)
	{
		if(cdebug)
		{
			fprintf(newstdout,"File size %lld\n",filebytes);
			fflush(newstdout);
		}
		pa = (char*)MapViewOfFile( hMap,   	// handle to map object
				FILE_MAP_READ, 		// read permission
				0,
				0,
				0);
	}
	if (pa == NULL)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else
    pa = (char *)I_MMAP( ((char *)0),filebytes, prot,
                    mflags, fd, 0);

	if(pa == (char *)-1)
	{
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	    	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
	    	fprintf(newstdout,"Mmap_Read MapViewOfFile failed on file %s Error: %s Error value: %d\n",fp->filename,err_str,c_eval);
	    	fflush(newstdout);
		c_eflag++;
	     	if(use_cached_fd)
		{
	           cached_open_count--;
	           fp->file_desc = 0;
		}
#if defined(WIN32)
		if (hMap != NULL)
			CloseHandle(hMap);
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(110);
	}
	if(trans < 0)
		trans =0;
	if((fp->mmapread_offset > filebytes) || ((long long)(fp->mmapread_offset) < (long long)0))
		fp->mmapread_offset =0;
	if(fp->mmapread_offset+trans > filebytes)
		fp->mmapread_offset =0;

	if(fp->mmapread_offset+trans > filebytes)
		/*Limit the io size, ensure the operation won't go beyond the valid memory size*/
		trans = (int)(filebytes - fp->mmapread_offset); 

	if(cdebug)
	{
 	  fprintf(newstdout,"Trans %d Size %lld Offset %lld\n",trans,(long long)filebytes,(long long)(fp->mmapread_offset));
	  fflush(newstdout);
	}
	memcpy(buf,&pa[(long long)(fp->mmapread_offset)],(size_t)trans);
	if(cdebug)
	{
 	  fprintf(newstdout,"memcpy completed\n");
	  fflush(newstdout);
	}
	fp->mmapread_offset += trans;
	if((fp->mmapread_offset/(long long)1024) >= (fp->Original_fsize))
	   fp->mmapread_offset=0;
	fp->op_count++;

	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,mmap_read,Mmap read,%s,,%lld,%d,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,(long long)fp->mmapread_offset,trans,fp->access_count);
	/* global (across all files in this work object) wreadcount */
	work_obj[my_workload].mmap_read_op_count++;

	total_file_ops += 1;
	work_obj[my_workload].total_file_ops+=1;

	total_read_bytes += (double)(trans);

	if(Op_lat_flag)
	{
	   lat = gettime() - my_gen_op_time;
	   hist_insert(lat);
	   work_obj[my_workload].mmap_read_op_time += lat;
	}
	work_obj[my_workload].mmap_read_op_count++;
#if defined(WIN32)
	if (pa != NULL)
	   UnmapViewOfFile(pa);
	if (hMap != NULL)
	{
	   CloseHandle(hMap);
	}
#else
	munmap((void *)pa,filebytes);
#endif
	if(!use_cached_fd)
	{
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		if(tracedebug)
		{
			fprintf(tracefile,"%d,%lu,%s,mmap_read,Closed file,%s,,,,%ld,\n",
				client_id,trace_seq++,trace_states[tracestate],
				fp->filename,fp->access_count);
		}
	}
	total_file_ops += 1.0;
	work_obj[my_workload].total_file_ops+=1.0;
}	

char *
my_strcasestr(char *s, char *find)
{
	char c, sc;
	size_t len;

	if ((c = *find++) != 0) {
		c = tolower((unsigned char)c);
		len = strlen(find);
		do {
			do {
				if ((sc = *s++) == 0)
					return (NULL);
			} while ((char)tolower((unsigned char)sc) != c);
		} while (strncasecmp(s, find, len) != 0);
		s--;
	}
	return ((char *)s);
}

int
my_strcasecmp(char *s1, char *s2)
{
	unsigned char *uns1 = ( unsigned char *)s1;
	unsigned char *uns2 = (unsigned char *)s2;

	while (tolower(*uns1) == tolower(*uns2++))
		if (*uns1++ == '\0')
			return (0);
	return (tolower(*uns1) - tolower(*--uns2));
}

void *
my_malloc(size_t size)
{
	void *tmp;
	tmp = (void *) malloc(size);
	if(tmp == (void *) 0)
	{
		fprintf(newstdout,"Malloc failure\n");
   	   	fflush(newstdout);
		exit(0);
	}
	bzero(tmp,size);
	return(tmp);
}

/* 
 * This function takes a host name and figures out the ip address
 * and sticks this address in the buffer that was provided. This
 * function is IPV6 aware and can return from 4 bytes to 16 bytes 
 * in the address buffer.
 */
int
get_host_address(char *host, unsigned int *buffer, char *name_buf)
{
	struct addrinfo     *server_addr;
	struct addrinfo     ai_hints;
	struct sockaddr_in *sain;
	struct sockaddr_in6 *sain6;
	int error;

    	memset((char *)&ai_hints, '\0', sizeof (ai_hints));
    	/* Get host's address */
    	if (ipv6_enable) {
            ai_hints.ai_family = AF_INET6;
            ai_hints.ai_flags = AI_CANONNAME;
            error = getaddrinfo(host, NULL, &ai_hints, &server_addr);
            if (error != 0) {
                    (void) fprintf(newstdout,
                        "%s could not be located by getaddrinfo\n", host);
	   	    fflush(newstdout);
                    return (1);
            }
    	}
    	else
    	{
            ai_hints.ai_family = AF_INET;
            ai_hints.ai_flags = AI_CANONNAME;
            error = getaddrinfo(host, NULL, &ai_hints, &server_addr);
            if (error != 0) {
		    if(!mflag)
		    {
                       (void) fprintf(newstdout,"%s could not be located by getaddrinfo\n",host);
	   	       fflush(newstdout);
	  	    }
		    else
		    {
                       (void) printf("\t%s could not be located by getaddrinfo\n",host);
	  	    }
                    return (1);
            }
    	}
	if(ipv6_enable)
	{
		sain6 = (struct sockaddr_in6 *)(server_addr->ai_addr);
        	memcpy(buffer,&sain6->sin6_addr,16);
	}
	else
	{
		sain = (struct sockaddr_in *)(server_addr->ai_addr);
        	memcpy(buffer,&sain->sin_addr,4);
	}
	strcpy(name_buf,server_addr->ai_canonname);
	if(!mflag && report_name==0)
	{
              (void) fprintf(newstdout, "Child: Master's hostname as reported by getaddrinfo: %s\n",
		     server_addr->ai_canonname);
	      fflush(newstdout);
	      report_name++;
	}
        return (0);
}	

int
get_host_ip_address(char *host, char *buffer)
{
	struct addrinfo     *server_addr;
	struct addrinfo     ai_hints;
	struct sockaddr_in *sain;
	struct sockaddr_in6 *sain6;
	int error;
	char ibuffer[256];

    	memset((char *)&ai_hints, '\0', sizeof (ai_hints));
    	/* Get host's address */
    	if (ipv6_enable) {
            ai_hints.ai_family = AF_INET6;
            error = getaddrinfo(host, NULL, &ai_hints, &server_addr);
            if (error != 0) {
                    (void) fprintf(stdout,
                        "%s could not be located by getaddrinfo\n", host);
	   	    fflush(stdout);
                    return (1);
            }
    	}
    	else
    	{
            ai_hints.ai_family = AF_INET;
            error = getaddrinfo(host, NULL, &ai_hints, &server_addr);
            if (error != 0) {
                    (void) fprintf(stdout,
                        "%s could not be located by getaddrinfo\n",host);
	   	    fflush(stdout);
                    return (1);
            }
    	}
	if(ipv6_enable)
	{
		sain6 = (struct sockaddr_in6 *)(server_addr->ai_addr);
        	memcpy(ibuffer,&sain6->sin6_addr,16);
	}
	else
	{
		sain = (struct sockaddr_in *)(server_addr->ai_addr);
        	memcpy(ibuffer,&sain->sin_addr,4);
	}
    	if (ipv6_enable) {
	   inet_ntop(AF_INET6, (void*)ibuffer, buffer, sizeof(struct sockaddr_in6));
	}
	else
	{
	   inet_ntop(AF_INET, (void*)ibuffer, buffer, sizeof(struct sockaddr_in));
	}
        return (0);
}
/*
 * Giving a maximum of "val" return a random number that
 * is less than or equal to val that follows a geometric
 * distribution with a lambda of 4.0.
 */

#include <math.h>
int
geometric_rand(int val)
{
    double lambda = 4.0;
    double L;
    unsigned int k = 0;
    unsigned int ra;
    float p = 1.0;
    int good_val;
    L = exp(-lambda);
    do {
        k++;
	good_val = val/3;
	if(good_val == 0)
	   good_val =1;
        ra = netmist_rand() % (good_val); /*  /3 for more dip in curve */
        p *= (float)ra;
    } while( p > L);
    return ((int)((k-1)%val));
}

#define JITTER_PCT 20
int spot_stuff[100];
int pick_init;

/*
 * Given a spot/chunk number, create a 5% jitter.
 * So... 95 % of the time, the same value is returned, but
 * 5% of the time, it will return a different chunk number;
 */
int 
picker(int input, int max)
{
	int i,ret;
	if(!pick_init) /* Init the Xpercent with other possibilities */
	{
	   for(i = 0; i < JITTER_PCT; i++)
		spot_stuff[i]= (netmist_rand() % max);
	}
	for(i=JITTER_PCT;i < 100; i++)
		spot_stuff[i]=input;
	ret = spot_stuff[ (netmist_rand() % 100) ];
	return(ret);
}	

void
client_validate_ops(void)
{
	int ret1,ret;
	int x = 0;
	char buf[1024];
	struct file_object *fp;
	struct test *tp;
	char alt_name[1024];
	int saved;
	int saved_direct_percent;
	int old_max_open;

	/* Turn off file descriptor caching */
	old_max_open = max_open;
	max_open = 1;
	tp=&tests[0];

	saved = tp->op_type; /* Save original value */

	fp= op_X_array[OP_WRITE];
	fp->Original_fsize = fsize;

	/* Let's force this now, so we don't have it blow up later. */
	saved_direct_percent=my_mix_table[my_workload].percent_direct;
	if(my_mix_table[my_workload].percent_direct != 0)
	{
		/*
 		  fprintf(newstdout,"Set Direct \n");
 		  fflush(newstdout);
		*/
	   	my_mix_table[my_workload].percent_direct=100;
	}
	/* Create validate directory */
#if defined(WIN32)
	sprintf(buf,"%s\\CL%d_%s_val",workdir,client_id,
		work_obj[my_workload].work_obj_name);
#else
	sprintf(buf,"%s/CL%d_%s_val",workdir,client_id,
		work_obj[my_workload].work_obj_name);
#endif
	/* Use virtualized, in case this function is workload specific */
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,validate_ops,Make val dir,%s,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],
			buf);
	ret=(*Netmist_Ops[my_workload].init_dir)(buf);
	if((ret < 0) && (errno != EEXIST))
	{
		ret1=errno;
		c_eflag++;
		c_eval = errno;
 		fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Mkdir failed. Error: %s\n",strerror(errno));
		fflush(newstdout);
		R_exit(ret1);
	}
	strcpy(fp->dirname,buf);

	/* Create file to validate op */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Create\n");
		fflush(newstdout);
	}
#if defined(WIN32)
	sprintf(buf,"%s\\CL%d_%s_val\\validate_file",workdir,client_id,
                work_obj[my_workload].work_obj_name);
#else
	sprintf(buf,"%s/CL%d_%s_val/validate_file",workdir,client_id,
                work_obj[my_workload].work_obj_name);
#endif
	strcpy(fp->filename,buf);
	/* Create a file and fill it up with data */
	if(tracedebug) {
		fprintf(tracefile,"%d,%lu,%s,validate_ops,Create validate file,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,fp->access_count);
	}
	(*Netmist_Ops[my_workload].init_file)(buf,0,0,0);

	/* Write */
	tp->op_type=OP_WRITE;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;

	if(cdebug)
	{
		fprintf(newstdout,"validate: Write\n");
		fflush(newstdout);
	}
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].write)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);
	if(cdebug)
	{
		fprintf(newstdout,"validate: Write returns\n");
		fflush(newstdout);
	}

	/* Write file op*/
	tp->op_type=OP_WRITE_FILE;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	if(cdebug)
	{
		fprintf(newstdout,"validate: Write File\n");
		fflush(newstdout);
	}

	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].write_file)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);
	if(cdebug)
	{
		fprintf(newstdout,"validate: Write File returns\n");
		fflush(newstdout);
	}

	/* Write_rand */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Write rand\n");
		fflush(newstdout);
	}
	tp->op_type=OP_WRITE_RAND;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].write_rand)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* Read */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Read\n");
		fflush(newstdout);
	}
	tp->op_type=OP_READ;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].read)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* Read file op */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Read file\n");
		fflush(newstdout);
	}
	tp->op_type=OP_READ_FILE;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].read_file)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* Read rand*/

	if(cdebug)
	{
		fprintf(newstdout,"validate: Read Rand\n");
		fflush(newstdout);
	}
	tp->op_type=OP_READ_RAND;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].read_rand)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* mmap read */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Mmap Read\n");
		fflush(newstdout);
	}
	tp->op_type=OP_MM_READ;
	fp->mmapread_offset=0;
	fp->offset = 0;
        fp->file_desc = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].mmap_read)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	if(cdebug)
	{
		fprintf(newstdout,"validate: Mmap read completed\n");
		fflush(newstdout);
	}

	/* mmap write */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Mmap write\n");
		fflush(newstdout);
	}
	fp->access_count = 0;
	tp->op_type=OP_MM_WRITE;
	fp->offset = 0;
	fp->mmapwrite_offset=0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].mmap_write)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);
	if(cdebug)
	{
		fprintf(newstdout,"validate: Mmap write completed\n");
		fflush(newstdout);
	}

	/* rmw  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: RMW\n");
		fflush(newstdout);
	}
	tp->op_type=OP_RMW;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].rmw)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* Create an mkdir directory */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Mkdir\n");
		fflush(newstdout);
	}
#if defined(WIN32)
	sprintf(buf,"%s\\CL%d_%s_mkdir",workdir,client_id,
		work_obj[my_workload].work_obj_name);
#else
	sprintf(buf,"%s/CL%d_%s_mkdir",workdir,client_id,
		work_obj[my_workload].work_obj_name);
#endif
	/* Use virtualized, in case this function is workload specific */
	ret=(*Netmist_Ops[my_workload].init_dir)(buf);
	if((ret < 0) && (errno != EEXIST))
	{
		ret1=errno;
 		fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Mkdir failed. Error: %s\n",strerror(errno));
		fflush(newstdout);
		c_eflag++;
		c_eval = errno;
		R_exit(ret1);
	}
	strcpy(fp->dirname,buf);
	/* mkdir  */
	tp->op_type=OP_MKDIR;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].mkdir)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* Now remove it */
	ret = (*Netmist_Ops[my_workload].remove_dir)(fp->dirname);
	if(ret < 0)
	{
 		fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"client_validate_ops: remove_dir op failed: %s\n",
			strerror(errno));
		fflush(newstdout);
	}

	/* Restore the fp->dirname */
#if defined(WIN32)
	sprintf(buf,"%s\\CL%d_%s_val",workdir,client_id,
		work_obj[my_workload].work_obj_name);
#else
	sprintf(buf,"%s/CL%d_%s_val",workdir,client_id,
		work_obj[my_workload].work_obj_name);
#endif
	strcpy(fp->dirname,buf);
	/* unlink  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Unlink\n");
		fflush(newstdout);
	}
	tp->op_type=OP_UNLINK;
#if defined(WIN32)
	sprintf(buf,"%s\\CL%d_%s_val\\unlinkfile",workdir,client_id,
		work_obj[my_workload].work_obj_name);
#else
	sprintf(buf,"%s/CL%d_%s_val/unlinkfile",workdir,client_id,
		work_obj[my_workload].work_obj_name);
#endif
	strcpy(fp->filename,buf);
	/* Create an empty file */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Init empty file\n");
		fflush(newstdout);
	}
	(*Netmist_Ops[my_workload].init_empty_file)(fp->filename);
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].unlink)(fp,0);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* The postfix_op routine put it back */
	(*Netmist_Ops[my_workload].remove_file)(fp->filename);

	/* create  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Create\n");
		fflush(newstdout);
	}
	tp->op_type=OP_CREATE;
#if defined(WIN32)
	sprintf(buf,"%s\\CL%d_%s_val\\createfile",workdir,client_id,
		work_obj[my_workload].work_obj_name);
#else
	sprintf(buf,"%s/CL%d_%s_val/createfile",workdir,client_id,
		work_obj[my_workload].work_obj_name);
#endif
	strcpy(fp->filename,buf);
	/* Create an empty file */
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].create)(fp,0);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

#if defined(WIN32)
	sprintf(buf,"%s\\CL%d_%s_val\\validate_file",workdir,client_id,
                work_obj[my_workload].work_obj_name);
#else
	sprintf(buf,"%s/CL%d_%s_val/validate_file",workdir,client_id,
                work_obj[my_workload].work_obj_name);
#endif
	strcpy(fp->filename,buf);

	/* append  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Append\n");
		fflush(newstdout);
	}
	tp->op_type=OP_APPEND;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].append)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* locking  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Lock\n");
		fflush(newstdout);
	}
	tp->op_type=OP_LOCKING;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].locking)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* access  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Access\n");
		fflush(newstdout);
	}
	tp->op_type=OP_ACCESS;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].access)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* stat  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Stat\n");
		fflush(newstdout);
	}
	tp->op_type=OP_STAT;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].stat)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* chmod  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Chmod\n");
		fflush(newstdout);
	}
	tp->op_type=OP_CHMOD;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].chmod)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* readdir  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Rmdir\n");
		fflush(newstdout);
	}
	tp->op_type=OP_READDIR;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].readdir)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* rename  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Rename\n");
		fflush(newstdout);
	}
	tp->op_type=OP_RENAME;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].rename)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* statfs  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Statfs\n");
		fflush(newstdout);
	}
	tp->op_type=OP_STATFS;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].statfs)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* pathconf  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Pathconf\n");
		fflush(newstdout);
	}
	tp->op_type=OP_PATHCONF;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].pathconf)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	/* copyfile  */
	if(cdebug)
	{
		fprintf(newstdout,"validate: Copyfile\n");
		fflush(newstdout);
	}
	tp->op_type=OP_COPYFILE;
	fp->offset = 0;
#if defined(_solaris_) || defined(_macos_)
	fp->directio_on = 0;
#endif
	fp->access_count = 0;
	fp->op_count = 0;
	fp->db_access_count = 0;
	(*Netmist_Ops[my_workload].prefix_op)(x,fp);
	(*Netmist_Ops[my_workload].copyfile)(fp);
	(*Netmist_Ops[my_workload].postfix_op)(x,fp);

	if(tracedebug) {
		fprintf(tracefile,"%d,%lu,%s,validate_ops,Remove validate file,%s,,,,%ld,\n",
			client_id,trace_seq++,trace_states[tracestate],
			fp->filename,fp->access_count);
	}
	(*Netmist_Ops[my_workload].remove_file)(fp->filename);
	strcpy(alt_name,fp->filename);
	strcat(alt_name,"_2");
	(*Netmist_Ops[my_workload].remove_file)(alt_name);
	/* Turn file descriptor caching back on */
	max_open=old_max_open;
#if defined(WIN32)
        sprintf(buf,"%s\\CL%d_%s_val",workdir,client_id,
                work_obj[my_workload].work_obj_name);
#else
        sprintf(buf,"%s/CL%d_%s_val",workdir,client_id,
                work_obj[my_workload].work_obj_name);
#endif
	if(tracedebug) {
		fprintf(tracefile,"%d,%lu,%s,validate_ops,Remove validate dir,%s,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],
			buf);
	}
	ret = (*Netmist_Ops[my_workload].remove_dir)(buf);
	if(ret < 0)
	{
 		fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"client_validate_ops: remove_dir (_val) failed: %s\n",
			strerror(errno));
		fflush(newstdout);
	}
	/* Restore this now */
	my_mix_table[my_workload].percent_direct=saved_direct_percent;

	fprintf(newstdout,"Validation complete\n");
	fflush(newstdout);

	tp->op_type = saved; /* Restore original value */

}

int
my_gethostname(char * buffer, int len)
{
	gethostname(buffer,len);
	return(0);
}

int 
check_dir_w(char *where)
{
#if defined(WIN32)
	if(GetFileAttributes(where) != INVALID_FILE_ATTRIBUTES)
#else
	if(access(where,W_OK) == 0)
#endif
		return(0);
	else
		return(1);
}

/*
 * This creates a file with the path that is provided.
 * It is used in the INIT phase before the actual
 * testing starts, and by postfix() to put files back after an unlink.
 * When this is called during the INIT phase, it can create a new
 * file and fill it, or.. it can truncate it back to the write size,
 * or it can grow it to the right size, or it can just read it
 * so that the client's caches are defeated.
 *
 */
#ifdef HAVE_ANSIC_C
void
generic_init_file(char *filename, int op_type, int which_dir, int filenum)
#else
void
generic_init_file(filename,op_type, which_dir, filenum)
char *filename;
int op_type;
int which_dir, filenum;
#endif
{
	char *buf;
	char *err_str;
	int flags = 0;
	int fringe = 0;
	long long xfer_count  = 0; /* In bytes */
	struct file_object *file_ptr;
	struct file_object *file_array_ptr;
	double pre_time = 0;
	double post_time = 0;
	int lat_delay = 0;
#if defined(WIN32)
	HANDLE fd;
	LARGE_INTEGER largeoffset;
	DWORD ret=0;
	BOOL err=0;
#else
	int fd,ret;
#endif
	long long i;
#ifdef _LARGEFILE64_SOURCE
  /* 
   * Heavy magic happens with the size of the stat and statfs structures
   * about the time that 10.6 came out.
   */
  #if defined(_macos_)
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  > 1050
  	struct stat stbuf;
    #endif
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  < 1050
	struct stat stbuf;
    #endif
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  == 1050
  	struct stat64 stbuf;
    #endif
  #else
	struct stat64 stbuf;
  #endif
#else
  #if defined(WIN32)
	struct _stati64 stbuf;
  #else
	struct stat stbuf;
  #endif
#endif
	file_array_ptr=op_X_array[op_type];
	file_ptr = &file_array_ptr[(which_dir*client_files)+filenum]; 

	buf=main_buf;
#if defined(WIN32)
	flags|=FILE_FLAG_POSIX_SEMANTICS;
#endif
	if(init_phase) /* Just get the length set correctly */
	{
           /* If the file doesn't exist, then skip to the create region. */
#if defined(INIT_TRACE_DEBUG)
	   if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,init_file,Stat,%s,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],filename);
#endif
#if defined(WIN32)
           ret = GetFileAttributes(filename);
           if(ret == INVALID_FILE_ATTRIBUTES)
           {
              goto jump1;
           }
#endif
           ret = I_STAT(filename, &stbuf);
           if(ret != 0 )
           {
              goto jump1;
           }

           /*----------------------------------------------------------------------------
	    * Not right size ? Try to grow it 
	    *----------------------------------------------------------------------------
	    */
           if(((long long)(stbuf.st_size)) < (file_ptr->Original_fsize*1024))
	   {
if(cdebug)
{
fprintf(newstdout,"Grow Was %lld  is %lld\n",(long long)(stbuf.st_size),
	file_ptr->Original_fsize);
fflush(newstdout);
}
#if defined(WIN32)
		fd=CreateFile(filename,
			GENERIC_READ|GENERIC_WRITE, 
			FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
			NULL, 
			OPEN_EXISTING, 
			flags, 
			NULL);
		if(fd == INVALID_HANDLE_VALUE)
		{
			c_eval = GetLastError();
			err_str = win32_strerror(c_eval);
#else
		fd=I_OPEN(filename,O_APPEND|O_RDWR|flags,0666);
		if(fd < 0)
		{
		   	c_eval=errno;
		   	err_str = strerror(c_eval);
#endif
		
 	       	   	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		   	fprintf(newstdout,"Open existing file for write failed. Filename = %s Error: %s Error value: %d\n",
				filename, err_str,c_eval);
		   	fflush(newstdout);
		   	c_eflag++;
		   	R_exit(106);
		}
#if defined(INIT_TRACE_DEBUG)
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,init_file,Open,%s,0666,,,,O_APPEND=1",
				client_id,trace_seq++,trace_states[tracestate],filename);
#endif
#if defined(WIN32)
		/* APPEND ... move file ponter to the EOF*/
		largeoffset.QuadPart = 0;
		SetFilePointerEx(fd, largeoffset, &largeoffset, FILE_END);
#endif
		if(sharing_flag) /* In sharing mode, prevent dual writers */
		{
		  if(try_lock_file_hold(fd)==0)  /* Got lock ? */
		  {
                     I_STAT(filename, &stbuf);
#if defined(INIT_TRACE_DEBUG)
		     if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,init_file,Stat,%s,,,,,\n",client_id,trace_seq++,
				trace_states[tracestate],filename);
#endif
                     if(((long long)(stbuf.st_size)/(long long)1024) == (file_ptr->Original_fsize))
		       goto out;
		  }
		  else
		       goto out; /* Let the other proc do it */
		}
		if((file_ptr->Original_fsize*1024)-(stbuf.st_size) >= max_xfer_size)
		{
		  for(i=((long long)(stbuf.st_size));(i + max_xfer_size) < 
			(file_ptr->Original_fsize*1024); i+=(long long)max_xfer_size)
		  {
		   make_non_dedup((long *)buf,max_xfer_size);
		   xfer_count += max_xfer_size;
  if(cdebug)
  {
  fprintf(newstdout,"Writing chunk, non-fringe. Size %d b. File size %lld bytes\n",
				max_xfer_size,file_ptr->Original_fsize*1024);
  fflush(newstdout);
  }
		   if(init_phase)
		      pre_time = gettime();
#if defined(WIN32)
		   err=WriteFile(fd,buf,max_xfer_size,&ret,NULL);
		   if((FALSE == err) || (ret <= 0))
		   {
			c_eval = GetLastError();
			err_str = win32_strerror(c_eval);
#else
		   ret=write(fd,buf,max_xfer_size);
		   if(ret < 0)
		   {
			c_eval=errno;
			err_str = strerror(c_eval);
#endif
 	        	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		   	fprintf(newstdout,"Write file failed (1) to file %s Error: %s Error value %d Ret %d\n",
				filename,err_str,c_eval,ret);
			fflush(newstdout);
			c_eflag++;
#if defined(WIN32)
			CloseHandle(fd);
#else
			close(fd);
#endif
			R_exit(107);
		   }
#if defined(INIT_TRACE_DEBUG)
		   if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,init_file,Write,%s,,%lld,%d,,\n",client_id,trace_seq++,
				trace_states[tracestate],filename,i,max_xfer_size);
#endif
		   if(init_phase)
		   {
		      post_time = gettime();
		      /*
 		       * Throttle back on I/O rate, if needed. Don't swamp the filer 
		       */
		      if((post_time - pre_time) > MAX_LATENCY)
			   	lat_delay+=1000;
		      if((post_time - pre_time) < MAX_LATENCY)
				lat_delay-=1000;
		      if(lat_delay < 0)
				lat_delay=0;
		      if(lat_delay > 0)
				nap(lat_delay);
		   }
		  }
		}

		if((stbuf.st_size + xfer_count) < (file_ptr->Original_fsize*1024))
		{
		/* Fix up the fringe at the end */
		   fringe = (file_ptr->Original_fsize*1024 - (stbuf.st_size+xfer_count));
  if(cdebug)
  {
  fprintf(newstdout,"Writing fringe of %d bytes. xfer %lld Original %lld st_size %lld\n",fringe,xfer_count,(long long)(file_ptr->Original_fsize), (long long)(stbuf.st_size/1024));
  fflush(newstdout);
  }
#if defined(WIN32)
		   err=WriteFile(fd,buf,fringe,&ret,NULL);
		   if((FALSE == err) || (ret <= 0))
		   {
			c_eval = GetLastError();
			err_str = win32_strerror(c_eval);
#else
		   ret=write(fd,buf,fringe);
		   if(ret < 0)
		   {
			c_eval=errno;
			err_str = strerror(c_eval);
#endif
 	        	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		   	fprintf(newstdout,"Write file failed (2) to file %s Error: %s Error value %d Ret %d\n",
				filename,err_str,c_eval,ret);
			fflush(newstdout);
			c_eflag++;
#if defined(WIN32)
			CloseHandle(fd);
#else
			close(fd);
#endif
			R_exit(107);
		   }
#if defined(INIT_TRACE_DEBUG)
		   if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,init_file,Write,%s,,%lld,%d,,\n",client_id,trace_seq++,
				trace_states[tracestate],filename,
				file_ptr->Original_fsize*1024-fringe,fringe);
#endif
		}
#if defined(WIN32)
		CloseHandle(fd);
#else
	    	close(fd);
#endif
#if defined(INIT_TRACE_DEBUG)
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,init_file,close,%s,,,,,\n",client_id,trace_seq++,
				trace_states[tracestate],filename);
#endif
		purge_file(file_ptr,filename);
	        return;
           }
           /*
	    *----------------------------------------------------------------------------
	    * Case of the file is too big. So we need to then trim it back.
	    *----------------------------------------------------------------------------
	    */
       if(((long long)(stbuf.st_size)) > (file_ptr->Original_fsize * 1024))
	   {
if(cdebug)
{
fprintf(newstdout,"Shrink Was %lld  is %lld\n",(long long)(stbuf.st_size),
	file_ptr->Original_fsize);
fflush(newstdout);
}
	   	I_TRUNC(filename,(file_ptr->Original_fsize*(long long)1024));
#if defined(INIT_TRACE_DEBUG)
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,file_init,Truncate,%s,,,,,size=%lld\n",
				client_id,trace_seq++,trace_states[tracestate],
				filename,(file_ptr->Original_fsize*(long long)1024));
#endif
		/*  Just enough to break the caches. One op type has sufficient 
		 *  files and space allocated to break the client's caches.
		 *  Reading more than this just consumes wall clock with no
		 *  benefit to the results.
	         */
		if( (op_type == OP_READ) || (op_type == OP_READ_FILE))
		{
		   purge_file(file_ptr,filename);
		   return;
		}
		else
		   return;
	   }
	   /* File the file is already the right size */
       if(((long long)(stbuf.st_size)) == (file_ptr->Original_fsize*1024) )
	   {
		/*  Just enough to break the caches. One op type has sufficient 
		 *  files and space allocated to break the client's caches.
		 *  Reading more than this just consumes wall clock with no
		 *  benefit to the results.
	         */
		if( (op_type == OP_READ) || (op_type == OP_READ_FILE))
		{
if(cdebug)
{
fprintf(newstdout,"Purge file\n");
fflush(newstdout);
}
		   purge_file(file_ptr,filename);
		   return;
		}
		else
		{
if(cdebug)
{
fprintf(newstdout,"Skip file\n");
fflush(newstdout);
}
		   return;
		}
	   }
	}

jump1:
        /*
	 *----------------------------------------------------------------------------
	 * Case of the file doesn't exist... So we create and fill here.
	 *----------------------------------------------------------------------------
	 */
if(cdebug)
{
fprintf(newstdout,"New Size %lld \n", file_ptr->Original_fsize);
fflush(newstdout);
}
#if defined(WIN32)
	/*TODO: why use "write through" flag*/
	fd=CreateFile(filename,
		GENERIC_READ | GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		OPEN_ALWAYS, 
		FILE_ATTRIBUTE_NORMAL|flags, 
		NULL);
	if(fd == INVALID_HANDLE_VALUE)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
#else
	fd=I_OPEN(filename,O_CREAT|O_RDWR,0666);
	if(fd < 0)
	{
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
 	    	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Open new file for write failed. Filename = %s Error: %s Error value: %d\n", 
			filename,err_str,c_eval);
		fflush(newstdout);
		c_eflag++;
		R_exit(106);
	}
#if defined(INIT_TRACE_DEBUG)
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,init_file,Open,%s,0666,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],filename);
#endif
	if(sharing_flag) /* In sharing mode, prevent dual writers */
	{
	    if(try_lock_file_hold(fd)==0) /* Got lock ? */
	    {
	      /* Check to see if we have a collision in sharemode */
              I_STAT(filename, &stbuf);
#if defined(INIT_TRACE_DEBUG)
	      if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,init_file,Stat,%s,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],filename);
#endif
              if(((long long)(stbuf.st_size)/(long long)1024) == (file_ptr->Original_fsize))
	    	 goto out;
	    }else
		 goto out;
	}
	total_file_ops += 1;
	work_obj[my_workload].total_file_ops+=1;
	work_obj[my_workload].open_op_count+=1;
	xfer_count = 0;
	if((file_ptr->Original_fsize*1024) >= max_xfer_size)
	{ 
	  for(i=0; (i+max_xfer_size) < (file_ptr->Original_fsize*1024) ; i+=(long long)(max_xfer_size))
	  {
	   	make_non_dedup((long *)buf,max_xfer_size);
		xfer_count += max_xfer_size;
  if(cdebug)
  {
  fprintf(newstdout,"Writing new chunk, non-fringe. Size %d bytes. File size %lld bytes\n",
				max_xfer_size,file_ptr->Original_fsize*1024);
  fflush(newstdout);
  }
#if defined(WIN32)
		err=WriteFile(fd,buf,max_xfer_size,&ret,NULL);
		if((FALSE == err) || (ret <= 0))
		{
			c_eval=GetLastError();
			err_str = win32_strerror(c_eval);
#else
		ret=write(fd,buf,max_xfer_size);
		if(ret < 0)
		{
			c_eval=errno;
			err_str = strerror(c_eval);
#endif
 	        	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		   	fprintf(newstdout,"Write file failed (3) to file %s Error: %s Error value %d Ret %d\n",
				filename,err_str,c_eval,ret);
			fflush(newstdout);
			c_eflag++;
#if defined(WIN32)
			/*explicitly unlock file*/
			UnlockFile(fd, 0, 0, (DWORD)-1, (DWORD)-1);
			CloseHandle(fd);
#else
			close(fd);
#endif
			R_exit(107);
		}
#if defined(INIT_TRACE_DEBUG)
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,init_file,Write,%s,,%lld,%d,,\n",
				client_id,trace_seq++,trace_states[tracestate],filename,
				i,max_xfer_size);
#endif
		total_file_ops += 1;
		work_obj[my_workload].total_file_ops+=1;
		work_obj[my_workload].write_op_count+=1;
	  }
	}
	/* Fix up the fringe at the end */
	if(xfer_count < file_ptr->Original_fsize*1024)
	{
	     fringe = file_ptr->Original_fsize*1024 - xfer_count;
  if(cdebug)
  {
  fprintf(newstdout,"Writing fringe of %d bytes. xfercount %lld\n",fringe,xfer_count);
  fprintf(newstdout,"Original fsize %lld\n", file_ptr->Original_fsize*1024);
  fflush(newstdout);
  }
#if defined(WIN32)
	     err=WriteFile(fd,buf,fringe,&ret,NULL);
	     if((FALSE == err) || (ret <= 0))
	     {
		c_eval = GetLastError();
		err_str = win32_strerror(c_eval);
#else
	     ret=write(fd,buf,fringe);
	     if(ret < 0)
	     {
		c_eval=errno;
		err_str = strerror(c_eval);
#endif
         	fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
	   	fprintf(newstdout,"Write file failed (4) to file %s Error: %s Error value %d Ret %d\n",
			filename,err_str,c_eval,ret);
		fflush(newstdout);
		c_eflag++;
#if defined(WIN32)
		CloseHandle(fd);
#else
		close(fd);
#endif
		R_exit(107);
	     }
#if defined(INIT_TRACE_DEBUG)
	     if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,init_file,Write,%s,,%lld,%d,,\n",
			client_id,trace_seq++,trace_states[tracestate],filename,
			file_ptr->Original_fsize*1024-fringe,fringe);
#endif
	}
/****/
  	if(cdebug)
  	{
        	ret = I_STAT(filename, &stbuf);
        	fprintf(newstdout,"New actual size %lld\n",  (long long)(stbuf.st_size));
		fflush(newstdout);
  	}
/****/
	if(lflag && flush_flag)
	{
	   I_FSYNC(fd); /* Not needed as it is not measured */
#if defined(INIT_TRACE_DEBUG)
	   if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,init_file,Sync,%s,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],filename);
#endif
	}
	total_file_ops += 1;
	work_obj[my_workload].total_file_ops+=1;
	work_obj[my_workload].close_op_count+=1;
out:
#if defined(WIN32)
	/*explicitly unlock file*/
	UnlockFile(fd, 0, 0, (DWORD)-1, (DWORD)-1);
	CloseHandle(fd);
#else
	close(fd);
#endif
#if defined(INIT_TRACE_DEBUG)
	if(tracedebug)
	{
		fprintf(tracefile,"%d,%lu,%s,init_file,Close,%s,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],filename);
	}
#endif
}

/*
 * This creates an empty file with the path that is provided.
 * It is used in the INIT phase before the actual
 * testing starts, and by postfix() to put files back after an unlink.
 */
#ifdef HAVE_ANSIC_C
void
generic_init_empty_file(char *filename)
#else
void
generic_init_empty_file(filename)
char *filename;
#endif
{
	char* err_str = NULL;
	int retry_unlink = 0;
	int flags = 0;
#if !defined(WIN32)
	int fd;
again:
        c_eval = 0;
	fd=I_OPEN(filename,O_CREAT|O_RDWR|flags,0666);
	if(fd < 0)
	{
		c_eval=errno;
		err_str = strerror(c_eval);
#else
	HANDLE fd;
        c_eval = 0;
again:
	flags|=FILE_FLAG_POSIX_SEMANTICS;
	fd=CreateFile(filename,
		GENERIC_READ | GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,  
		NULL, 
		CREATE_NEW, 
		FILE_ATTRIBUTE_NORMAL|flags, 
		NULL);
	if(fd == INVALID_HANDLE_VALUE)
	{
		c_eval=GetLastError();
		err_str = win32_strerror(c_eval);
		if(c_eval == ERROR_FILE_EXISTS) /* If exists, just return */
		{
			return;		
		}
		/* 
		 * Windows unique behavior here !!!! 
		 * If any process anywhere on any node has an open handle
		 * or an open handle in progress, or a delete that is pending,
		 * then this create can fail with ERROR_ACCESS_DENIED.
	 	 *
 		 * In our case, someone else has initialized the dir, so just
		 * skip this dir and move on.
		 */
		if(c_eval ==ERROR_ACCESS_DENIED ) 
		{
			return;		
		}
#endif
#if defined(INIT_TRACE_DEBUG)
		if(tracedebug)
			fprintf(tracefile,"%d,%lu,%s,init_empty_file,Create,%s,0666,,,,\n",
				client_id,trace_seq++,trace_states[tracestate],filename);
#endif
 	        fprintf(newstdout,"%s ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"Write empty file failed %s Error Value %d  Error %s\n", 
			filename, c_eval,err_str);
		fflush(newstdout);
		if(retry_unlink++<10)
		{
		    	fprintf(newstdout,"try again\n");
			fflush(newstdout);
			goto again;
		}
		c_eflag++;
		R_exit(108);
	}
#if !defined(WIN32)
	close(fd);
#else
	CloseHandle(fd);
#endif
}


/*
 * Generic function to check to see if the directory exists,
 * and is usable as a work directory for testing.
 */

#ifdef HAVE_ANSIC_C
int
generic_stat_workdir(char *dirname)
#else
int
generic_stat_workdir(dirname)
char *dirname;
#endif
{
#if !defined(WIN32)
 #ifdef _LARGEFILE64_SOURCE
  /* 
   * Heavy magic happens with the size of the stat and statfs structures
   * about the time that 10.6 came out.
   */
  #if defined(_macos_)
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  > 1050
  	struct stat stbuf;
    #endif
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  < 1050
	struct stat stbuf;
    #endif
    #if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__  == 1050
  	struct stat64 stbuf;
    #endif
  #else
	struct stat64 stbuf;
  #endif
 #else
	struct stat stbuf;
 #endif
#endif
	int ret;
	char* err_str = NULL;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing stat_workdir op\n");
		fflush(newstdout);
	}
#if defined(WIN32)
        ret = GetFileAttributes(workdir);
        if(ret == INVALID_FILE_ATTRIBUTES)
        {
		c_eval = GetLastError();
		err_str = win32_strerror(c_eval);
 	   	fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
           	fprintf(newstdout,"GetFileAttributes for workdir > %s < Error: %s Error Value: %d\n", workdir,err_str,c_eval);
           	fflush(newstdout);
        }else
	   ret = 0;
#else
	ret = I_STAT(workdir,&stbuf);
	if(ret != 0)
	{
	   err_str = strerror(errno);
 	   fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
           fprintf(newstdout,"Can not stat() working directory.(%s) Error: %s\n",
		workdir,err_str);
	   fflush(newstdout);
	}
#endif
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,stat_workdir,Stat,%s,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],workdir);
	return(ret);
}	

/* 
 * Generic_chdir() Virtualized interface so that this functionality
 * can be used on workload dependent filesystems that might have
 * a different type of chdir.
 */
#ifdef HAVE_ANSIC_C
int
generic_chdir(char *dirname)
#else
int
generic_chdir(dirname)
char *dirname;
#endif
{
	int ret;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing generic_chdir\n");
		fflush(newstdout);
	}
	ret = chdir(dirname);
	if(ret < 0)
	{
 	   fprintf(newstdout,"%s\n",get_timestamp(&time_string[0]));
           fprintf(newstdout,"Can not chdir() to working directory.(%s) Errno %d %s\n",
		workdir,errno,strerror(errno));
	   fflush(newstdout);
	}
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,chdir,Chdir,%s,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],dirname);
	return(ret);
}	

/* 
 * Generic_remove_file() Virtualized interface so that this functionality
 * can be used on workload dependent filesystems that might have
 * a different type of unlink.
 */
#ifdef HAVE_ANSIC_C
int
generic_remove_file(char *filename)
#else
int
generic_remove_file(filename)
char *filename;
#endif
{
	int ret;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing generic_remove_file\n");
		fflush(newstdout);
	}
	ret = unlink(filename);
	if(tracedebug)
		fprintf(tracefile,"%d,%lu,%s,remove_file,Unlink,%s,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],filename);
	return(ret);
}	

/* 
 * Generic_remove_dir() Virtualized interface so that this functionality
 * can be used on workload dependent filesystems that might have
 * a different type of rmdir.
 */
#ifdef HAVE_ANSIC_C
int
generic_remove_dir(char *dirname)
#else
int
generic_remove_dir(dirname)
char *dirname;
#endif
{
	int ret;
	if(cdebug) 
	{
		fprintf(newstdout,"Doing generic_remove_dir\n");
		fflush(newstdout);
	}
	ret = rmdir(dirname);
	if(ret < 0)
	{
 		fprintf(newstdout,"%s: ",get_timestamp(&time_string[0]));
		fprintf(newstdout,"generic_remove_dir failed: %s Name %s\n",
			strerror(errno),dirname);
		fflush(newstdout);
	}
	if(tracedebug)
	{
		fprintf(tracefile,"%d,%lu,%s,remove_dir,Rmdir,%s,,,,,\n",
			client_id,trace_seq++,trace_states[tracestate],dirname);
	}
	return(ret);
}	

void generic_setup()
{
	client_validate_ops();
	if( sharing_flag )
	{
	   /* if(client_id == 0)*/
	      client_init_files(0);
	}
	else
	      client_init_files(client_id);
}

void generic_teardown()
{
	if(sharing_flag)
	{
	  if(cleanup_flag )
	     client_remove_files(0);
	}
	else
	{
	   if(cleanup_flag)
	      client_remove_files(client_id);
	}
}

int init_sizes;
long long size_table[136];

/*
 * This creates a bell curve of file sizes. The center of the
 * curve is the natural file size. The file size steps are:
 * (in percentage of the natural file size)
 * 10%, 20%, 40%, 60%, 80%, 100%, 100%, 120%, 140%, 160%, 180%, 190%
 *
 * Curve approximation has steps: 1,3,7,14,20,23,23,20,14,7,3,1
 * There is   1/136 slots has a file size that is  10% of natural size.
 * There are  3/136 slots has a file size that is  20% of natural size.
 * There are  7/136 slots has a file size that is  40% of natural size.
 * There are 14/136 slots has a file size that is  60% of natural size.
 * There are 20/136 slots has a file size that is  80% of natural size.
 * There are 46/136 slots has a file size that is 100% of natural size.
 * There are 20/136 slots has a file size that is 120% of natural size.
 * There are 14/136 slots has a file size that is 140% of natural size.
 * There are  7/136 slots has a file size that is 160% of natural size.
 * There are  3/136 slots has a file size that is 180% of natural size.
 * There are  1/136 slots has a file size that is 190% of natural size.
 */

unsigned long long
get_next_fsize( long long size )
{
	int j,k;
	unsigned long long ret_val;
if(cdebug)
{
fprintf(newstdout,"next size input %lld\n",size);
fflush(newstdout);
}
	if( init_sizes == 0 )
	{
		k = 0;
		size_table[k++] = (size * 10LL)/100.0;
		for(j=0;j<3;j++)
		   size_table[k++] = (size * 20LL)/100LL;
		for(j=0;j<7;j++)
		   size_table[k++] = (size * 40LL)/100LL;
		for(j=0;j<14;j++)
		   size_table[k++] = (size * 60LL)/100LL;
		for(j=0;j<20;j++)
		   size_table[k++] = (size * 80LL)/100LL;
		for(j=0;j<46;j++)
		   size_table[k++] = (size * 100LL)/100LL;
		for(j=0;j<20;j++)
		   size_table[k++] = (size * 120LL)/100LL;
		for(j=0;j<14;j++)
		   size_table[k++] = (size * 140LL)/100LL;
		for(j=0;j<7;j++)
		   size_table[k++] = (size * 160LL)/100LL;
		for(j=0;j<3;j++)
		   size_table[k++] = (size * 180LL)/100LL;
		for(j=0;j<1;j++)
		   size_table[k++] = (size * 190LL)/100LL;
		init_sizes =1;
	}
	ret_val = size_table[((name_rand()>>4)%136)];	
	if(ret_val == 0) /* Set minimum size */
	   ret_val = rsize*1024;
	ret_val = (ret_val + ((long long)(rsize*1024)-1)) & ~((long long)(rsize*1024)-1);
if(cdebug)
{
fprintf(newstdout,"next size %lld\n",ret_val);
fflush(newstdout);
}
	return(ret_val);	
}

void
child_dump_hist(FILE *out)
{

#ifndef NO_PRINT_LLD
   fprintf(out,"Band 1: ");
   fprintf(out," 20us:%-7.1lld ",buckets[0]);
   fprintf(out," 40us:%-7.1lld ",buckets[1]);
   fprintf(out," 60us:%-7.1lld ",buckets[2]);
   fprintf(out," 80us:%-7.1lld ",buckets[3]);
   fprintf(out,"100us:%-7.1lld \n",buckets[4]);

   fprintf(out,"Band 2: ");
   fprintf(out,"200us:%-7.1lld ",buckets[5]);
   fprintf(out,"400us:%-7.1lld ",buckets[6]);
   fprintf(out,"600us:%-7.1lld ",buckets[7]);
   fprintf(out,"800us:%-7.1lld ",buckets[8]);
   fprintf(out,"  1ms:%-7.1lld \n",buckets[9]);

   fprintf(out,"Band 3: ");
   fprintf(out,"  2ms:%-7.1lld ",buckets[10]);
   fprintf(out,"  4ms:%-7.1lld ",buckets[11]);
   fprintf(out,"  6ms:%-7.1lld ",buckets[12]);
   fprintf(out,"  8ms:%-7.1lld ",buckets[13]);
   fprintf(out," 10ms:%-7.1lld \n",buckets[14]);

   fprintf(out,"Band 4: ");
   fprintf(out," 12ms:%-7.1lld ",buckets[15]);
   fprintf(out," 14ms:%-7.1lld ",buckets[16]);
   fprintf(out," 16ms:%-7.1lld ",buckets[17]);
   fprintf(out," 18ms:%-7.1lld ",buckets[18]);
   fprintf(out," 20ms:%-7.1lld \n",buckets[19]);

   fprintf(out,"Band 5: ");
   fprintf(out," 40ms:%-7.1lld ",buckets[20]);
   fprintf(out," 60ms:%-7.1lld ",buckets[21]);
   fprintf(out," 80ms:%-7.1lld ",buckets[22]);
   fprintf(out,"100ms:%-7.1lld \n",buckets[23]);

   fprintf(out,"Band 6: ");
   fprintf(out,"200ms:%-7.1lld ",buckets[24]);
   fprintf(out,"400ms:%-7.1lld ",buckets[25]);
   fprintf(out,"600ms:%-7.1lld ",buckets[26]);
   fprintf(out,"800ms:%-7.1lld ",buckets[27]);
   fprintf(out,"   1s:%-7.1lld \n",buckets[28]);

   fprintf(out,"Band 7: ");
   fprintf(out,"   2s:%-7.1lld ",buckets[29]);
   fprintf(out,"   4s:%-7.1lld ",buckets[30]);
   fprintf(out,"   6s:%-7.1lld ",buckets[31]);
   fprintf(out,"   8s:%-7.1lld ",buckets[32]);
   fprintf(out,"  10s:%-7.1lld \n",buckets[33]);

   fprintf(out,"Band 8: ");
   fprintf(out,"  20s:%-7.1lld ",buckets[34]);
   fprintf(out,"  40s:%-7.1lld ",buckets[35]);
   fprintf(out,"  60s:%-7.1lld ",buckets[36]);
   fprintf(out,"  80s:%-7.1lld ",buckets[37]);
   fprintf(out," 120s:%-7.1lld \n",buckets[38]);

   fprintf(out,"Band 9: ");
   fprintf(out,"120+s:%-7.1lld \n\n",buckets[39]);
#else
   fprintf(out,"Band 1: ");
   fprintf(out," 20us:%-7.1ld ",buckets[0]);
   fprintf(out," 40us:%-7.1ld ",buckets[1]);
   fprintf(out," 60us:%-7.1ld ",buckets[2]);
   fprintf(out," 80us:%-7.1ld ",buckets[3]);
   fprintf(out,"100us:%-7.1ld \n",buckets[4]);

   fprintf(out,"Band 2: ");
   fprintf(out,"200us:%-7.1ld ",buckets[5]);
   fprintf(out,"400us:%-7.1ld ",buckets[6]);
   fprintf(out,"600us:%-7.1ld ",buckets[7]);
   fprintf(out,"800us:%-7.1ld ",buckets[8]);
   fprintf(out,"  1ms:%-7.1ld \n",buckets[9]);

   fprintf(out,"Band 3: ");
   fprintf(out,"  2ms:%-7.1ld ",buckets[10]);
   fprintf(out,"  4ms:%-7.1ld ",buckets[11]);
   fprintf(out,"  6ms:%-7.1ld ",buckets[12]);
   fprintf(out,"  8ms:%-7.1ld ",buckets[13]);
   fprintf(out," 10ms:%-7.1ld \n",buckets[14]);

   fprintf(out,"Band 4: ");
   fprintf(out," 12ms:%-7.1ld ",buckets[15]);
   fprintf(out," 14ms:%-7.1ld ",buckets[16]);
   fprintf(out," 16ms:%-7.1ld ",buckets[17]);
   fprintf(out," 18ms:%-7.1ld ",buckets[18]);
   fprintf(out," 20ms:%-7.1ld \n",buckets[19]);

   fprintf(out,"Band 5: ");
   fprintf(out," 40ms:%-7.1ld ",buckets[20]);
   fprintf(out," 60ms:%-7.1ld ",buckets[21]);
   fprintf(out," 80ms:%-7.1ld ",buckets[22]);
   fprintf(out,"100ms:%-7.1ld \n",buckets[23]);

   fprintf(out,"Band 6: ");
   fprintf(out,"200ms:%-7.1ld ",buckets[24]);
   fprintf(out,"400ms:%-7.1ld ",buckets[25]);
   fprintf(out,"600ms:%-7.1ld ",buckets[26]);
   fprintf(out,"800ms:%-7.1ld ",buckets[27]);
   fprintf(out,"   1s:%-7.1ld \n",buckets[28]);

   fprintf(out,"Band 7: ");
   fprintf(out,"   2s:%-7.1ld ",buckets[29]);
   fprintf(out,"   4s:%-7.1ld ",buckets[30]);
   fprintf(out,"   6s:%-7.1ld ",buckets[31]);
   fprintf(out,"   8s:%-7.1ld ",buckets[32]);
   fprintf(out,"  10s:%-7.1ld \n",buckets[33]);

   fprintf(out,"Band 8: ");
   fprintf(out,"  20s:%-7.1ld ",buckets[34]);
   fprintf(out,"  40s:%-7.1ld ",buckets[35]);
   fprintf(out,"  60s:%-7.1ld ",buckets[36]);
   fprintf(out,"  80s:%-7.1ld ",buckets[37]);
   fprintf(out," 120s:%-7.1ld \n",buckets[38]);

   fprintf(out,"Band 9: ");
   fprintf(out,"120+s:%-7.1ld \n",buckets[39]);
   fflush(out);
#endif
}

/*
 * This is a list of file names for the "locking" directory
 * files, used in the parallelized initialization path for 
 * ShareMode.
 */
struct name_list {
	char name[MAXNAME];
	struct name_list *next;
};

struct name_list *gn_head;

/* 
 * Save the names of the directory locking files to a list. 
 * This list will be processed later, to unlink all of the 
 * directory locking files.
 */
void
push_name(char *name)
{
	struct name_list *buffer;

#if defined(_P1_DEBUG_)
	fprintf(newstdout,"Adding %s to list \n",name);
	fflush(newstdout);
#endif
	buffer = (struct name_list *)malloc(sizeof(struct name_list));
	buffer->next = 0;
	strcpy(buffer->name,name);
	if(gn_head)
	{
		buffer->next = gn_head;
		gn_head = buffer;
	}
	else
	{
		gn_head = buffer;
	}
}

/*
 * This will unlink the list of directory locking files and free
 * the memory that was used to keep track of these.
 */
void
drain_list(void)
{
	struct name_list *buffer,*old;
	buffer = gn_head;
	if(buffer == 0)
	{
		return; /* nothing to do */
/*
		fprintf(newstdout,"Drain_list: Error in malloc\n");
		fflush(newstdout);
		c_eflag = 1;
		c_eval = errno;
		R_exit(errno);
*/
	}
	while(buffer)
	{
#if defined(_P1_DEBUG_)
		fprintf(newstdout,"Unlinking %s from list \n",buffer->name);
		fflush(newstdout);
#endif
		unlink(buffer->name);
		old = buffer;
		buffer = buffer->next;
		gn_head = buffer;
		free(old);
	}
}


/* 
   A C-program for MT19937-64 (2004/9/29 version).
   Coded by Takuji Nishimura and Makoto Matsumoto.

   This is a 64-bit version of Mersenne Twister pseudorandom number
   generator.

   Before using, initialize the state by using init_genrand64(seed)  
   or init_by_array64(init_key, key_length).

   Copyright (C) 2004, Makoto Matsumoto and Takuji Nishimura,
   All rights reserved.                          

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

     2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

     3. The names of its contributors may not be used to endorse or promote 
        products derived from this software without specific prior written 
        permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

   References:
   T. Nishimura, ``Tables of 64-bit Mersenne Twisters''
     ACM Transactions on Modeling and 
     Computer Simulation 10. (2000) 348--357.
   M. Matsumoto and T. Nishimura,
     ``Mersenne Twister: a 623-dimensionally equidistributed
       uniform pseudorandom number generator''
     ACM Transactions on Modeling and 
     Computer Simulation 8. (Jan. 1998) 3--30.

   Any feedback is very welcome.
   http://www.math.hiroshima-u.ac.jp/~m-mat/MT/emt.html
   email: m-mat @ math.sci.hiroshima-u.ac.jp (remove spaces)
*/



#define NN 312
#define MM 156
#define MATRIX_A 0xB5026F5AA96619E9ULL
#define UM 0xFFFFFFFF80000000ULL /* Most significant 33 bits */
#define LM 0x7FFFFFFFULL /* Least significant 31 bits */


/* The array for the state vector */
static unsigned long long mt[NN]; 
/* mti==NN+1 means mt[NN] is not initialized */
static int mti=NN+1; 

/* initializes mt[NN] with a seed */
void init_genrand64(unsigned long long seed)
{
    mt[0] = seed;
    for (mti=1; mti<NN; mti++) 
        mt[mti] =  (6364136223846793005ULL * (mt[mti-1] ^ (mt[mti-1] >> 62)) + mti);
}

/* initialize by an array with array-length */
/* init_key is the array for initializing keys */
/* key_length is its length */
void init_by_array64(unsigned long long init_key[],
		     unsigned long long key_length)
{
    unsigned long long i, j, k;
    init_genrand64(19650218ULL);
    i=1; j=0;
    k = (NN>key_length ? NN : key_length);
    for (; k; k--) {
        mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 62)) * 3935559000370003845ULL))
          + init_key[j] + j; /* non linear */
        i++; j++;
        if (i>=NN) { mt[0] = mt[NN-1]; i=1; }
        if (j>=key_length) j=0;
    }
    for (k=NN-1; k; k--) {
        mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 62)) * 2862933555777941757ULL))
          - i; /* non linear */
        i++;
        if (i>=NN) { mt[0] = mt[NN-1]; i=1; }
    }

    mt[0] = 1ULL << 63; /* MSB is 1; assuring non-zero initial array */ 
}

/* generates a random number on [0, 2^64-1]-interval */
unsigned long long genrand64_int64(void)
{
    int i;
    unsigned long long x;
    static unsigned long long mag01[2]={0ULL, MATRIX_A};

    if (mti >= NN) { /* generate NN words at one time */

        /* if init_genrand64() has not been called, */
        /* a default initial seed is used     */
        if (mti == NN+1) 
            init_genrand64(5489ULL); 

        for (i=0;i<NN-MM;i++) {
            x = (mt[i]&UM)|(mt[i+1]&LM);
            mt[i] = mt[i+MM] ^ (x>>1) ^ mag01[(int)(x&1ULL)];
        }
        for (;i<NN-1;i++) {
            x = (mt[i]&UM)|(mt[i+1]&LM);
            mt[i] = mt[i+(MM-NN)] ^ (x>>1) ^ mag01[(int)(x&1ULL)];
        }
        x = (mt[NN-1]&UM)|(mt[0]&LM);
        mt[NN-1] = mt[MM-1] ^ (x>>1) ^ mag01[(int)(x&1ULL)];

        mti = 0;
    }
  
    x = mt[mti++];

    x ^= (x >> 29) & 0x5555555555555555ULL;
    x ^= (x << 17) & 0x71D67FFFEDA60000ULL;
    x ^= (x << 37) & 0xFFF7EEE000000000ULL;
    x ^= (x >> 43);

    return x;
}

/* generates a random number on [0, 2^63-1]-interval */
long long genrand64_int63(void)
{
    return (long long)(genrand64_int64() >> 1);
}

/* generates a random number on [0,1]-real-interval */
double genrand64_real1(void)
{
    return (genrand64_int64() >> 11) * (1.0/9007199254740991.0);
}

/* generates a random number on [0,1)-real-interval */
double genrand64_real2(void)
{
    return (genrand64_int64() >> 11) * (1.0/9007199254740992.0);
}

/* generates a random number on (0,1)-real-interval */
double genrand64_real3(void)
{
    return ((genrand64_int64() >> 12) + 0.5) * (1.0/4503599627370496.0);
}

/*
 * On Unix boxes, one can delete files, and their parent
 * directoires, even if there is an open file descriptor to
 * one of those files. BUT !!! on Windows, the rmdir will
 * FAIL if there are any open file descriptors, to files that
 * are open, and unlinked.  So.. we MUST close ALL open 
 * file descriptors before we remove the parent directories.
 */
void
close_cached_file_desc(void)
{
	int i,j,k;
	int close_count = 0;
	struct file_object *array,*fp;
#if defined(WIN32)
	HANDLE fd;
#else
	int fd;
#endif
	/*size = sizeof(struct file_object )* client_dirs * client_files;*/
	for(i=0;i<NUM_OP_TYPES;i++)
	{
	   array = (struct file_object *)op_X_array[i];
	   for(j=0;j<client_dirs;j++)
	   {
	      for(k=0;k<client_files;k++)
	      {
	         if (op_X_array[i])
	         {
		   fp=(struct file_object *)&array[(j*client_files)+k];
		   if(fp && (fp->file_desc!=0))
		   {
			fd = fp->file_desc;
			close_count++;
#if defined(WIN32)
		        CloseHandle(fd);
#else
		        close(fd);
#endif
		   }
	         }
	      }
	   }
	}
	if(cdebug)
	{
		fprintf(newstdout,"Closed %d cached file descriptors\n", close_count);
		fflush(newstdout);
	}
}

int generic_impersonate(void* context)
{
	int result = 0;
	if(localflag)
		return result;
#if defined(WIN32)
	if (ERROR_SUCCESS != (result=LogonAndImpersonateUser( wmiUser, wmiPassword, wmiDomain)))
		fprintf(newstdout,"Failed to impersonate as WMI user >%s\\%s<\n", wmiDomain, wmiUser);
#endif
	return result;
}

void* generic_make_impersonate_context(void)
{

	return (void*)NULL;
}

void
master_dump_hist(struct result_object *here, FILE *out)
{

#ifndef NO_PRINT_LLD
   fprintf(out,"Band 1: ");
   fprintf(out," 20us:%-7.1lld ",here->bands[0]);
   fprintf(out," 40us:%-7.1lld ",here->bands[1]);
   fprintf(out," 60us:%-7.1lld ",here->bands[2]);
   fprintf(out," 80us:%-7.1lld ",here->bands[3]);
   fprintf(out,"100us:%-7.1lld \n",here->bands[4]);

   fprintf(out,"Band 2: ");
   fprintf(out,"200us:%-7.1lld ",here->bands[5]);
   fprintf(out,"400us:%-7.1lld ",here->bands[6]);
   fprintf(out,"600us:%-7.1lld ",here->bands[7]);
   fprintf(out,"800us:%-7.1lld ",here->bands[8]);
   fprintf(out,"  1ms:%-7.1lld \n",here->bands[9]);

   fprintf(out,"Band 3: ");
   fprintf(out,"  2ms:%-7.1lld ",here->bands[10]);
   fprintf(out,"  4ms:%-7.1lld ",here->bands[11]);
   fprintf(out,"  6ms:%-7.1lld ",here->bands[12]);
   fprintf(out,"  8ms:%-7.1lld ",here->bands[13]);
   fprintf(out," 10ms:%-7.1lld \n",here->bands[14]);

   fprintf(out,"Band 4: ");
   fprintf(out," 12ms:%-7.1lld ",here->bands[15]);
   fprintf(out," 14ms:%-7.1lld ",here->bands[16]);
   fprintf(out," 16ms:%-7.1lld ",here->bands[17]);
   fprintf(out," 18ms:%-7.1lld ",here->bands[18]);
   fprintf(out," 20ms:%-7.1lld \n",here->bands[19]);

   fprintf(out,"Band 5: ");
   fprintf(out," 40ms:%-7.1lld ",here->bands[20]);
   fprintf(out," 60ms:%-7.1lld ",here->bands[21]);
   fprintf(out," 80ms:%-7.1lld ",here->bands[22]);
   fprintf(out,"100ms:%-7.1lld \n",here->bands[23]);

   fprintf(out,"Band 6: ");
   fprintf(out,"200ms:%-7.1lld ",here->bands[24]);
   fprintf(out,"400ms:%-7.1lld ",here->bands[25]);
   fprintf(out,"600ms:%-7.1lld ",here->bands[26]);
   fprintf(out,"800ms:%-7.1lld ",here->bands[27]);
   fprintf(out,"   1s:%-7.1lld \n",here->bands[28]);

   fprintf(out,"Band 7: ");
   fprintf(out,"   2s:%-7.1lld ",here->bands[29]);
   fprintf(out,"   4s:%-7.1lld ",here->bands[30]);
   fprintf(out,"   6s:%-7.1lld ",here->bands[31]);
   fprintf(out,"   8s:%-7.1lld ",here->bands[32]);
   fprintf(out,"  10s:%-7.1lld \n",here->bands[33]);

   fprintf(out,"Band 8: ");
   fprintf(out,"  20s:%-7.1lld ",here->bands[34]);
   fprintf(out,"  40s:%-7.1lld ",here->bands[35]);
   fprintf(out,"  60s:%-7.1lld ",here->bands[36]);
   fprintf(out,"  80s:%-7.1lld ",here->bands[37]);
   fprintf(out," 120s:%-7.1lld \n",here->bands[38]);

   fprintf(out,"Band 9: ");
   fprintf(out,"120+s:%-7.1lld \n\n",here->bands[39]);
#else
   fprintf(out,"Band 1: ");
   fprintf(out," 20us:%-7.1ld ",here->bands[0]);
   fprintf(out," 40us:%-7.1ld ",here->bands[1]);
   fprintf(out," 60us:%-7.1ld ",here->bands[2]);
   fprintf(out," 80us:%-7.1ld ",here->bands[3]);
   fprintf(out,"100us:%-7.1ld \n",here->bands[4]);

   fprintf(out,"Band 2: ");
   fprintf(out,"200us:%-7.1ld ",here->bands[5]);
   fprintf(out,"400us:%-7.1ld ",here->bands[6]);
   fprintf(out,"600us:%-7.1ld ",here->bands[7]);
   fprintf(out,"800us:%-7.1ld ",here->bands[8]);
   fprintf(out,"  1ms:%-7.1ld \n",here->bands[9]);

   fprintf(out,"Band 3: ");
   fprintf(out,"  2ms:%-7.1ld ",here->bands[10]);
   fprintf(out,"  4ms:%-7.1ld ",here->bands[11]);
   fprintf(out,"  6ms:%-7.1ld ",here->bands[12]);
   fprintf(out,"  8ms:%-7.1ld ",here->bands[13]);
   fprintf(out," 10ms:%-7.1ld \n",here->bands[14]);

   fprintf(out,"Band 4: ");
   fprintf(out," 12ms:%-7.1ld ",here->bands[15]);
   fprintf(out," 14ms:%-7.1ld ",here->bands[16]);
   fprintf(out," 16ms:%-7.1ld ",here->bands[17]);
   fprintf(out," 18ms:%-7.1ld ",here->bands[18]);
   fprintf(out," 20ms:%-7.1ld \n",here->bands[19]);

   fprintf(out,"Band 5: ");
   fprintf(out," 40ms:%-7.1ld ",here->bands[20]);
   fprintf(out," 60ms:%-7.1ld ",here->bands[21]);
   fprintf(out," 80ms:%-7.1ld ",here->bands[22]);
   fprintf(out,"100ms:%-7.1ld \n",here->bands[23]);

   fprintf(out,"Band 6: ");
   fprintf(out,"200ms:%-7.1ld ",here->bands[24]);
   fprintf(out,"400ms:%-7.1ld ",here->bands[25]);
   fprintf(out,"600ms:%-7.1ld ",here->bands[26]);
   fprintf(out,"800ms:%-7.1ld ",here->bands[27]);
   fprintf(out,"   1s:%-7.1ld \n",here->bands[28]);

   fprintf(out,"Band 7: ");
   fprintf(out,"   2s:%-7.1ld ",here->bands[29]);
   fprintf(out,"   4s:%-7.1ld ",here->bands[30]);
   fprintf(out,"   6s:%-7.1ld ",here->bands[31]);
   fprintf(out,"   8s:%-7.1ld ",here->bands[32]);
   fprintf(out,"  10s:%-7.1ld \n",here->bands[33]);

   fprintf(out,"Band 8: ");
   fprintf(out,"  20s:%-7.1ld ",here->bands[34]);
   fprintf(out,"  40s:%-7.1ld ",here->bands[35]);
   fprintf(out,"  60s:%-7.1ld ",here->bands[36]);
   fprintf(out,"  80s:%-7.1ld ",here->bands[37]);
   fprintf(out," 120s:%-7.1ld \n",here->bands[38]);

   fprintf(out,"Band 9: ");
   fprintf(out,"120+s:%-7.1ld \n",here->bands[39]);
   fflush(out);
#endif
}


#if defined(WIN32)
/*
 * Make a local version of rand() and srand() that work on
 * Windows platforms.  We need this becasue the rand() implementation
 * in Visual Studio 'C++' only returns 16 bit values.
 *
 * We looked at rand_s()... DON'T go there. It is expensive (CPU)
 * and ignores srand().  The ability to control the seed IS needed
 * by the rest of Netmist.
 */

int n_seedi = 2231;

/*
 * Seed the random number generator.
 */
void
n_spec_srand(int seed)
{
    n_seedi = seed;
}

/*
 * Returns a random number.
 */
int
n_spec_rand(void)
{
    (void) n_ran();
    return(n_seedi);
}
/*
 * Get the current seed value 
 */
int
n_spec_gseed(void)
{
    return(n_seedi);
}

#define _A_MULTIPLIER  16807L
#define _M_MODULUS     2147483647L /* (2**31)-1 */
#define _Q_QUOTIENT    127773L     /* 2147483647 / 16807 */
#define _R_REMAINDER   2836L       /* 2147483647 % 16807 */

void n_ran(void)
{
    int	lo;
    int	hi;
    int	test;

    hi = n_seedi / _Q_QUOTIENT;
    lo = n_seedi % _Q_QUOTIENT;
    test = _A_MULTIPLIER * lo - _R_REMAINDER * hi;
    if (test > 0) {
	n_seedi = test;
    } else {
	n_seedi = test + _M_MODULUS;
    }
    /* return((float) n_seedi / _M_MODULUS); */
    return;
}
#endif

int name_seedi = 2231;

/*
 * Seed the random number generator.
 */
void
name_srand(int seed)
{
    name_seedi = seed;
}

/*
 * Returns a random number.
 */
int
name_rand(void)
{
    (void) name_ran();
    return(name_seedi);
}
/*
 * Get the current seed value 
 */
int
name_gseed(void)
{
    return(name_seedi);
}

#define _N_A_MULTIPLIER  16807L
#define _N_M_MODULUS     2147483647L /* (2**31)-1 */
#define _N_Q_QUOTIENT    127773L     /* 2147483647 / 16807 */
#define _N_R_REMAINDER   2836L       /* 2147483647 % 16807 */

void name_ran(void)
{
    int	lo;
    int	hi;
    int	test;

    hi = name_seedi / _N_Q_QUOTIENT;
    lo = name_seedi % _N_Q_QUOTIENT;
    test = _N_A_MULTIPLIER * lo - _N_R_REMAINDER * hi;
    if (test > 0) {
	name_seedi = test;
    } else {
	name_seedi = test + _N_M_MODULUS;
    }
}

/*
 * Rotate (circular shift) of 32 bit int. 
 * Used to rotate some low order bits that are typically
 * not having enough variation to be a reasonable set for
 * modulo with low values.
 */
unsigned int 
rotateR(unsigned int value, int shift) 
{
	if ((shift &= 31) == 0)
		return value;
	return (value >> shift) | (value << (32 - shift));
}
